#!/usr/bin/python


import string,re,sys,os
import math
import time
import fixSSEs
from Numeric import *
from LinearAlgebra import *
from copy import deepcopy
from sequence import *
from alignment import *
from nodeList import *

#global defines: platform specific
pdb_dir = "/usr/db/pdb"
scop_pdb_dir = "/usr/db/scop-pdb"
lockprogram = "/usr/local/bin/lock2"
lock_fixSSEs_program = "/usr/local/bin/lock2"
mclprogram = "/usr/local/bin/mcl"

#sets gap penalties for both CBA and LOCK 2
def setMultipliers(opening,extension,lock_opening,lock_extension):
    globals()["gapMultiplier"] = opening
    globals()["gap_extend_penalty"] = extension
    globals()["lock_gapMultiplier"] = lock_opening
    globals()["lock_gap_extend_penalty"] = lock_extension    


#sets LOCK 2 distance threshold
def setDistanceThreshold(distance):
    globals()["distanceThreshold"] = distance

#def dot(array1, array2):
#    return innerproduct(array1, swapaxes(array2, -1, -2))

def readInputFile(structure_list_file):
    #read in list of domains
    domains = []
    infile = open(structure_list_file,"r")
    line = infile.readline()
    while line:
        domain = line[:-1]        
        if(len(domain)<1): line=infile.readline(); continue
        index = domain.rfind("/")
        if index!=-1:
            directory = domain[:index]
            domain = domain[index+1:]
            chain = "all"
        else:
            directory = ""

        #some SCOP domains start with a 'g,' but corresponding PDB file
        #uses the name that starts with a 'd'
        m = re.match("[dg](....\..)",domain)
        if m!=None:
            domain = "d"+m.group(1)
            m = re.match("(.*)(\..)(\..*)$",domain)
            if m:
                domain = m.group(1)+m.group(2)
                extension = m.group(3)
            else:
                extension = ""
        else:
            m = re.match("(.*)(\..*)$",domain)
            if m:
                domain = m.group(1)
                extension = m.group(2)
            else:
                extension = ""
        #If no directory is specified, determine whether domain name is SCOP or PDB identifier
        if directory=="":
            if os.path.exists(domain+extension):
                chain = "all"
                directory = "./"
            else:
                if len(domain)==7: scop = 1
                else: scop = 0
                if scop:
                    if domain[5]=="_": chain = " "
                    elif domain[5]==".": chain="all"
                    else: chain = domain[5].upper()
                else:
                    if len(domain)>4:
                        chain = domain[-1].upper()
                    else:
                        chain = "all"                
            
        domains.append((domain,directory,extension,chain))
        line = infile.readline()
    infile.close()

    return domains

#reads in MCL results. Parameters:
#filename: name of file containing results
#nodes: type nodeList (containing no data)
def readClusters(filename,nodes):
    cluster_membership = {}; duplicates = []

    try:
        infile = open(filename,"r")
    except:
        return None,None,None
    line = infile.readline()

    #get number of clusters and move to first line of cluster definitions
    while line:
        if re.match("dimensions",line):
            dummy,numclusters = re.split("x",line)
            numclusters = string.atoi(numclusters)
            while line:
                if re.match("begin",line): break
                line = infile.readline()
            break
        else:
            line = infile.readline()

    singletons = []
    clusters = []
    for i in range(numclusters):
        duplicates.append({})
        line = infile.readline()[:-1]

        domains_seen = {}
        #collect all lines describing this cluster
        while not re.search("\$$",line):
            line += infile.readline()[:-1]
        fields = map(string.atoi,line.split()[:-1])

        num = len(fields)-1 #exclude cluster number and dollar sign

        clusternum = fields[0]
        clusters.append(zeros([num],Int))
        for j in range(1,len(fields)):
            clusters[i][j-1] = fields[j]
            if not fields[j] in nodes: continue
            node = nodes.getNodeByLabel(fields[j])
            domain_idx = node.domain_idx; resnum = node.resnum

            if not cluster_membership.has_key(domain_idx):
                cluster_membership[domain_idx] = {}
            cluster_membership[domain_idx][resnum]=clusternum
            if not domains_seen.has_key(domain_idx):
                domains_seen[domain_idx] = fields[j]
            else: #have a duplicate
                if not duplicates[i].has_key(domain_idx):
                    duplicates[i][domain_idx] = []
                    duplicates[i][domain_idx].append(domains_seen[domain_idx])
                duplicates[i][domain_idx].append(fields[j])



    return clusters,cluster_membership,duplicates


#Edge counts are used to determine the strength of a node's connection
#to its cluster. When two residues from the same protein appear in the
#same cluster, edge counts are used to determine which one is eliminated
#from the cluster and added as a singleton cluster
def makeEdgeCounts(domains,alignments,cluster_membership,edge_counts):
    for i in range(len(domains)):
        residues1 = domains[i].getResidues()
        domain1 = domains[i].domainname
        for j in range(i+1,len(domains)):
            domain2 = domains[j].domainname
            alignment = alignments.getAlignment(i,j)
            for res in residues1:
                if alignments.alignments[i][j].alignedResidues.has_key(res):
                    res2 = alignments.alignments[i][j].alignedResidues[res]
                else: res2 = -1
                if res2==-1: continue
                if domains[i].node_numbers.has_key(res):
                    cluster1 = cluster_membership[i][res]
                    label1 = domains[i].getNodeNumber(resnum=res)
                else: label1 = None

                if domains[j].node_numbers.has_key(res2):
                    cluster2 = cluster_membership[j][res2]
                    label2 = domains[j].getNodeNumber(resnum=res2)
                else: label2 = None

                if label1 != None:
                    if not edge_counts[cluster1].has_key(label1):
                        edge_counts[cluster1][label1] = 0
                    if label2 != None:
                        if not edge_counts[cluster2].has_key(label2):
                            edge_counts[cluster2][label2] = 0
                if label2 != None:
                    if not edge_counts[cluster2].has_key(label2):
                        edge_counts[cluster2][label2] = 0
                    if label1 != None:
                        if not edge_counts[cluster1].has_key(label1):
                            edge_counts[cluster1][label1] = 0
                if label1 != None:
                    edge_counts[cluster1][label1] += 1
                if label2 != None:
                    edge_counts[cluster2][label2] += 1

    

#determine which residue to keep in cases where two residues
#from the same protein appear in the same cluster. Based on
#edge counts.
def findDuplicateNodes(domains,clusters,edge_counts,duplicates):
    eliminated = {}
    nodes_to_check = {}

    for i in range(len(clusters)):
        for j in range(len(domains)):
            if duplicates[i].has_key(j):
                node_numbers = duplicates[i][j]
                max = maxnode = -1
                for node_num in node_numbers: #ties broken randomly: first seen
                    if edge_counts[i][node_num] > max:
                        max = edge_counts[i][node_num]
                        maxnode = node_num
                nodes_to_check[maxnode] = []
                for node_num in node_numbers:
                    if node_num != maxnode:
                        eliminated[node_num] = 1
                        nodes_to_check[maxnode].append(node_num)
    return eliminated,nodes_to_check


#Calculates the penalty for introducing a gap in the
#middle of a secondary structure element when placing clusters
#into the multiple alignment. Gaps in the middle of the SSE
#are penalized more heavily than gaps at ends. Not to be
#confused with dynamic programming gap penalties.
#Note: sse_penalties was supposed to contain hashed results,
#but it doesn't do anything yet.
def getSSEPenalty(domain,residue1,residue2,sse_penalties):
    index1 = domain.residue_indices[residue1]
    index2 = domain.residue_indices[residue2]

    if residue1==residue2: #shouldn't happen; included for completeness
        return 0
    elif index2 < index1:
        temp=residue1;residue1=residue2;residue2=temp
        temp=index1;index1=index2;index2=temp

    ssenum1 = domain.ssNumber[residue1]
    ssenum2 = domain.ssNumber[residue2]
    if ssenum1!=ssenum2: return 0 #different SSEs
    if ssenum1==-1 or ssenum2==-1: return 0 #one or both in loops    

    sse1_start,sse1_end = domain.sse_ends[ssenum1]


    midpoint_index = domain.sse_midpoints[ssenum1]

    sse_end_index = domain.residue_indices[sse1_end]
        
    sse_start_index = domain.residue_indices[sse1_start]

    temp1=0
    if index1+1 <= midpoint_index:
        if index1+1 != sse_start_index: #shouldn't fail this conditional
            temp1 = index1+1-sse_start_index
    else:
        if index1+1 != sse_end_index:
            temp1 = sse_end_index-index1-1



    if sse_end_index!=sse_start_index:
        return -1*temp1/float(sse_end_index-sse_start_index)
    else:
        return -1*temp1
    
#Prints both a text and html version of the MSTA
def printMSTA(domains,columns,marked_residues,filename):
    filename_text = re.sub(".html$",".txt",filename)
    blockwidth = 10
    blocks_per_line = 6
    columns_per_line=blockwidth*blocks_per_line
    numcols = len(columns)
    strandcolor="#FFFF99"
    helixcolor="#FF3399"
    color = colorend=""

    name_length = 0
    resnum_length = 0
    for domain in domains:
        if len(domain.domainname)> name_length:
            name_length = len(domain.domainname)
        for residue in domain.residue_indices.keys():
            if len(residue[:-1].strip()) > resnum_length:
                resnum_length = len(residue[:-1].strip())
        

    lines2 = {} #html lines
    rulers = {}
    text_rulers = {}
    lines = {} #text lines
    within_sse = dict([(i,0) for i in range(len(domains))])
    for domain_idx in range(len(domains)):
        domain = domains[domain_idx]
        lines[domain_idx] = []
        lines2[domain_idx] = []
        line_idx = 0
        numlines = numcols/(columns_per_line)
        if numcols%columns_per_line != 0: numlines += 1
        ssenum=prev_ssenum=prev_res_idx = -1
        ssetype=""
        for i in range(numlines):
            firstres=lastres=line2=line=""
            ruler = "&nbsp;"*(name_length+resnum_length+2)
            idx = columns_per_line*line_idx+1
            while idx<((line_idx+1)*columns_per_line+1):
                if idx==columns_per_line*line_idx+1 or idx%10==0:
                    temp = "%i" % idx
                    idx += len(temp)
                    ruler += temp
                else:
                    ruler += "&nbsp;"
                    idx += 1
                if idx>=numcols: break
            ruler += "<br>\n"            
            rulers[i] = ruler
            ruler = re.sub("&nbsp;"," ",ruler)
            ruler = re.sub("<br>","",ruler)
            text_rulers[i] = ruler
            for j in range(line_idx*columns_per_line,(line_idx+1)*columns_per_line):
                if j>=numcols: break
                res_idx = columns[j][domain_idx]
                if res_idx != "-1" and res_idx != -1: resnum=domain.residues[res_idx]
                else: resnum=-1
                sse_span_open = ""
                sse_span_close = ""
                if resnum!=-1:
                    if domain.sstype[resnum]!='U':
                        ssenum=domain.ssNumber[resnum]
                        if not within_sse[domain_idx]: #just started sse
                            ssetype = domain.sstype[resnum]
                            if ssetype=="H":
                                sse_span_open = "<span class='helix'>"
                            else:
                                sse_span_open = "<span class='strand'>"
                            within_sse[domain_idx] = 1
                    else:
                        prev_ssenum = -1
                if (res_idx =="-1" or res_idx == -1):
                    line2 += "-"
                    line += "-"
                else:
                    if firstres=="":
                        firstres = domain.residues[res_idx][:-1].strip()
                    resnum = domain.residues[res_idx]
                    aa = domain.aas[res_idx]
                    line += aa
                    if sse_span_open != "":
                        line2 += sse_span_open+aa
                        if domain.isSSEEnd(res_idx)==2:
                            line2 += "</span>"
                            within_sse[domain_idx] = 0
                    else:
                        line2 += aa
                        if domain.isSSEEnd(res_idx)==2:
                            line2 += "</span>"
                            within_sse[domain_idx] = 0
                        
                    if sse_span_close != "":
                        line2 += sse_span_close


                    lastres = domain.residues[res_idx][:-1].strip()
                    prev_res_idx = res_idx
                    prev_ssenum = ssenum

            if within_sse[domain_idx]:
                line2 += "</span>"
                within_sse[domain_idx] = 0

            if firstres != "":
                temp = "%s " % domain.domainname.ljust(name_length)
                temp += "%s " % firstres.rjust(resnum_length)
                line = temp+line
                temp=re.sub(" ","&nbsp;",temp)
                line2 = temp+line2
            else:
                temp = "%s " % domain.domainname.ljust(name_length)
                temp += " "*(resnum_length+1)
                line = temp + line
                temp = re.sub(" ","&nbsp;",temp)
                line2 = temp+line2
            if lastres != "":
                temp =" %s" % lastres.rjust(resnum_length)
                line += temp
                temp = re.sub(" ","&nbsp;",temp)
                line2 += temp


            line2 += "<br>"
            line += "\n"
            lines2[domain_idx].append(line2)
            lines[domain_idx].append(line)
            line_idx +=1

    outfile=open(filename,"w")
    outfile_txt = open(filename_text,"w")
    outfile.write("<HTML><HEAD><style type='text/css'>.alignment {font-family:'Courier'}\n.helix {background-color:#FF3399}\n.strand {background-color:#FFFF99}\n</style>\n</HEAD>\n")
    outfile.write("<BODY>\n")
    outfile.write("<span class='alignment'>\n")
    for i in range(numlines):
        outfile.write("%s" % rulers[i])
        outfile_txt.write("%s" % text_rulers[i])
        for domain_idx in range(len(domains)):
            outfile.write(lines2[domain_idx][i])
            outfile.write("\n")
            outfile_txt.write(lines[domain_idx][i])
#            outfile_txt.write("\n")
        outfile.write("<br>")
        outfile_txt.write("\n")
    outfile.write("</span>\n")
    outfile.write("</BODY></HTML>\n")
    outfile.close()
    outfile_txt.close()

#Places MCL clusters into the MSTA. Since the clusters are unordered
#and MCL doesn't obey sequence order constraints, this is non-trivial.
#Method: Optimize each cluster's placement as it's added to the MSTA,
#but don't worry about best global ordering.
#Quality of a potential position for a cluster is determined primarily
#by the number of residues in the cluster that violate sequence order
#constraints if the cluster is placed in a given position, and then by
#gaps introduced into secondary structure elements.Other parameters
#were added to correct problems as needed, and I wouldn't be surprised
#if some are redundant and/or not independent. But, it seems to work
#for the most part, so tinkering with it has been a low priority
def orderMSAColumns(domains,domain_indices,nodes,clusters,eliminated_nodes,nodes_to_check,marked_residues,sse_penalties,debug=0):
    node_numbers = dict([(x,1) for x in nodes.getLabels()])

    columns = []
    numclusters = len(clusters)
    singletons = []
    uncertain_nodes=[]


    new_column = dict([(x,-1) for x in range(len(domains))])

    for clusternum in range(len(clusters)): #for each cluster
        cluster = clusters[clusternum]
        if len(cluster)==1: #singletons are added separately
            singletons.append(cluster[0])
            continue
        max_in_order = -1; col = -1;
        min_sse_score = -1
        min_sse_separation=min_separation=-1
        min_gaplen=-1
        numcols = len(columns)

        out_of_order = {} #some nodes may be out of order for a given placement
        is_potential_placement=found_placement=0
        found_last_potential_placement=0
        for i in range(-1,numcols): #for each potential placement
            if found_last_potential_placement: break 
            in_order=gap_len=0
            prev_index=next_index=-1
            out_of_order[i+1] = {}
            sse_separation=gaplen=sse_score=separation=0
            domains_in_cluster = {}
            has_node = 0
            gaplen=separation=0

            for node_num in cluster:
                #the next line is probably archaic
                if not node_numbers.has_key(node_num): continue
                node = nodes.nodes[node_num] #holds the domain idx and resnum
                domain_idx = node.domain_idx; resnum = node.resnum
                domain = domains[domain_idx]
                if eliminated_nodes.has_key(node_num): continue #if eliminated from cluster,
                has_node = 1 #probably archaic
                domains_in_cluster[domain_idx] = 1 #domain has a node in this cluster
                resindex = domain.residue_indices[resnum]

                #get the residues (if any) placed in the MSTA before and
                #after this potential column placement (ignoring gaps)
                prev_index=-1; next_index = -1
                prev_resindex = -1; next_resindex = -1
                j=i
                while j>=0 and columns[j][domain_idx]==-1: j-=1
                if j>=0: prev_index=j; prev_resindex = columns[j][domain_idx]
                j=i+1
                while j<numcols and columns[j][domain_idx]==-1:
                    j+=1
                if j<numcols: next_index=j; next_resindex = columns[j][domain_idx]            

                if prev_resindex!=-1 and next_resindex!=-1:
                    if prev_resindex!=-1 and next_resindex!=-1:
                        if resindex>prev_resindex and resindex<next_resindex:
                            in_order += 1
                            is_potential_placement = 1 #at least one residue from
                                                       #cluster can be placed here
                        else: out_of_order[i+1][domain_idx]=node_num
                elif prev_resindex==-1 and next_resindex!=-1:
                    if resindex < next_resindex:
                        in_order += 1
                        is_potential_placement = 1                        
                    else: out_of_order[i+1][domain_idx]=node_num
                elif prev_resindex!=-1 and next_resindex==-1:
                    if resindex>prev_resindex:
                        in_order += 1
                        is_potential_placement = 1                        
                    else: out_of_order[i+1][domain_idx]=node_num
                else:
                    in_order += 1                    


                #Calculate various parameters used to determine the quality of this placement
                #for this cluster. This became very awkward at some point, so overhauling this code
                #may speed things up. I'm not sure.
                temp=temp2=-1

                ssenum=domain.ssNumber[domain.residues[resindex]]

                if ssenum!=-1 and next_resindex!=-1 and domain.ssNumber[domain.residues[next_resindex]]==ssenum:
                    sse_separation += next_index-i-1
                if ssenum!=-1 and prev_resindex!=-1 and domain.ssNumber[domain.residues[prev_resindex]] == ssenum:
                    sse_separation += i-prev_resindex-1
                if next_resindex!=-1: 
                    temp=-1*getSSEPenalty(domain,domain.residues[resindex],domain.residues[next_resindex],sse_penalties)
                if prev_resindex!=-1:
                    temp2=-1*getSSEPenalty(domain,domain.residues[prev_resindex],domain.residues[resindex],sse_penalties)                    
                if temp!=-1 and temp2!=-1:
                    sse_score += (temp+temp2)/2.0
                elif temp!=-1:
                    sse_score += temp
                elif temp2!=-1:
                    sse_score += temp2            


                if next_resindex!=-1 and prev_resindex != -1 and resindex>prev_resindex and resindex<next_resindex:
                    if next_resindex-prev_resindex < next_index+1-prev_index:
                        separation += 1


                if next_resindex!=-1 and resindex<next_resindex:
                    gaplen += next_index-i 
                    if prev_resindex!=-1 and resindex>prev_resindex:
                        gaplen += i-prev_index

            #if this placement obeys sequence order constraints better than any
            #other placement seen so far, continue calculating parameters
            if in_order>=max_in_order:
                for domain2_idx in range(len(domains)):
                    if domains_in_cluster.has_key(domain2_idx): continue

                    prev_index=-1; next_index = -1
                    prev_resindex = -1; next_resindex = -1
                    j=i
                    while j>=0 and columns[j][domain2_idx]==-1: j-=1
                    if j>=0: prev_index=j; prev_resindex = columns[j][domain2_idx]
                    j=i+1
                    while j<numcols and columns[j][domain2_idx]==-1:
                        j+=1
                    if j<numcols: next_index=j; next_resindex = columns[j][domain2_idx]            


                    if prev_resindex!=-1 and next_resindex!=-1:
                        if next_index==prev_index+1:
                            sse_score+=-1*getSSEPenalty(domains[domain2_idx],domains[domain2_idx].residues[prev_resindex],domains[domain2_idx].residues[next_resindex],sse_penalties)
                        gaplen += next_index-prev_index
                        if sse_score > min_sse_score and max_in_order != -1: #can't possibly be best placement
                            break
                        if next_resindex-prev_resindex < next_index+1-prev_index:
                            separation+=1

            #determine if this is the best placement for the cluster seen so far
            if max_in_order==-1: #first placement checked
                max_in_order=in_order;min_sse_score=sse_score;min_separation=separation;col=i+1
                min_gaplen=gaplen
                min_sse_separation=sse_separation
                found_placement=1
            elif max_in_order<in_order:
                max_in_order=in_order;min_sse_score=sse_score;min_separation=separation;col=i+1
                min_gaplen=gaplen
                min_sse_separation=sse_separation
                found_placement=1                
            elif max_in_order==in_order:
                if min_sse_score>sse_score:
                    max_in_order=in_order;min_sse_score=sse_score;min_separation=separation;col=i+1
                    min_gaplen=gaplen
                    min_sse_separation=sse_separation
                    found_placement=1                    
                elif min_sse_score==sse_score and sse_separation<min_sse_separation:
                    max_in_order=in_order;min_sse_score=sse_score;min_separation=separation;col=i+1
                    min_gaplen=gaplen
                    min_sse_separation=sse_separation
                    found_placement=1                    
                elif min_sse_score==sse_score and sse_separation==min_sse_separation and separation<min_separation:
                    max_in_order=in_order;min_sse_score=sse_score;min_separation=separation;col=i+1
                    min_gaplen=gaplen
                    min_sse_separation=sse_separation
                    found_placement=1                    
                elif min_sse_score==sse_score and min_sse_separation==sse_separation and separation==min_separation and gaplen<min_gaplen:
                    max_in_order=in_order;min_sse_score=sse_score;min_separation=separation;col=i+1
                    min_gaplen=gaplen
                    min_sse_separation=sse_separation
                    found_placement=1
            if in_order>0 and found_placement:
                found_last_potential_placement=0
                if in_order==len(domains) and min_sse_score==0 and sse_separation==0 and gaplen==0 and separation==0:
                    found_last_potential_placement = 1                
        if not has_node:continue #archaic
        
            
        #now add nodes to the MSTA that can be placed in column (corresponding to best placement) without
        #violating sequence order constraints
        column_members = {}
        surviving_nodes = {}
        for node_num in cluster:
            if not node_numbers.has_key(node_num): continue #archaic
            if eliminated_nodes.has_key(node_num): continue
            node = nodes.nodes[node_num]
            domain_idx = node.domain_idx; resnum = node.resnum

            domain = domains[domain_idx]

            if not out_of_order[col].has_key(domain_idx):
                column_members[domain_idx] = domains[domain_idx].residue_indices[resnum]
                surviving_nodes[node_num] = 1
            else: #if node was out of order in its domain sequence, add it as a singleton cluster
                numclusters += 1
                singletons.append(node_num)
                marked_residues[domain_idx][resnum]=1
                column_members[domain_idx] = -1

        if len(surviving_nodes.keys())==1: #only one node survived, so treat it as a singleton
            numclusters += 1
            node_num=surviving_nodes.keys()[0]
            singletons.append(node_num)
            continue

        columns = columns[0:col]+[new_column.copy()]+columns[col:] #insert new column        
        for d in range(len(domains)):
            if column_members.has_key(d):
                columns[col][d]=column_members[d]


    #now place singletons, which include nodes that were eliminated from their
    #clusters in cases where two residues from the same protein were in the same cluster
    node_nums = eliminated_nodes.keys()+singletons

    #sort nodes: adding in order helps optimize MSTA
    node_nums.sort() 


    prev_domain=-1
    prev_column=-1
    added = 0
    count = -1
    for node_num in node_nums: #place each node
        count += 1
        try: #probably archaic
            node = nodes.nodes[node_num]
        except:
            continue

        domain_idx = node.domain_idx; resnum = node.resnum
        resindex = domains[domain_idx].residue_indices[resnum]
        domain = domains[domain_idx]
        
        numcols=len(columns)
        col=-1
        min_sse_score=-1
        min_separation=-1
        min_sse_separation=-1
        min_gaplen=-1
        maxcol=-1
        is_potential_placement=found_placement=0
        found_last_potential_placement=0

        #If a singleton from this domain has already been placed in the MSTA,
        #we know that the next residue from that domain must be in a higher
        #numbered column (since nodes were sorted)
        if prev_domain == domain_idx:
            column_start = prev_column
        else:
            column_start=-1
        for i in range(column_start,numcols): #for each possible column placement
            if found_last_potential_placement: break 
            okay_to_place_cluster=0
            prev_index=-1; next_index = -1
            prev_resindex = -1; next_resindex = -1
            j=i
            while j>=0 and columns[j][domain_idx]==-1: j-=1
            if j>=0: prev_index=j; prev_resindex = columns[j][domain_idx]
            j=i+1

            while j<numcols and columns[j][domain_idx]==-1:
                j+=1
            if j<numcols: next_index=j; next_resindex = columns[j][domain_idx]            

            if prev_resindex != -1 and next_resindex != -1: #internal placement
                if prev_resindex!=-1 and next_resindex!=-1:
                    if resindex > prev_resindex and resindex < next_resindex:
                        okay_to_place_cluster=1
                        is_potential_placement = 1
            elif prev_resindex==-1 and next_resindex!=-1: #before first res
                if resindex < next_resindex:
                        okay_to_place_cluster=1
            elif prev_resindex!=-1 and next_resindex==-1: #after last position
                if resindex>prev_resindex:
                        okay_to_place_cluster=1
                        is_potential_placement = 1
            else: #first residue to be placed
                    okay_to_place_cluster=1
                    is_potential_placement = 1


            if not okay_to_place_cluster:
                if found_placement: #Rest of columns will also violate sequence order constraints
                    found_last_potential_placement=1
                prev_domain=domain_idx
                prev_column=col
                continue

            sse_separation=0;gaplen=0;separation=0
            if next_resindex!=-1 and resindex<next_resindex:
                gaplen += next_resindex-resindex
            if prev_resindex!=-1 and resindex>prev_resindex:
                gaplen += resindex-prev_resindex


            sse_score = 0
            temp=temp2=-1

            ssenum = domain.ssNumber[domain.residues[resindex]]
            prev_index2=prev_index;next_index2=next_index

            if ssenum!=-1 and next_index2!=-1 and domain.ssNumber[domain.residues[columns[next_index2][domain_idx]]] == ssenum:
                sse_separation += next_index2-i-1
            if ssenum!=-1 and prev_index2!=-1 and domain.ssNumber[domain.residues[columns[prev_index2][domain_idx]]] == ssenum:
                sse_separation += i-prev_index2-1

            if next_index2!=-1:
                temp = -1*getSSEPenalty(domain,domain.residues[resindex],domain.residues[columns[next_index2][domain_idx]],sse_penalties)
                separation += next_index2-i-1;
            if prev_index2!=-1:
                temp2 = -1*getSSEPenalty(domain,domain.residues[columns[prev_index2][domain_idx]],domain.residues[resindex],sse_penalties)
                separation += i-prev_index2-1;
            if temp!=-1 and temp2!=-1:
                sse_score += (temp+temp2)/2.0
            elif temp!=-1:
                sse_score += temp
            elif temp2!=-1:
                sse_score += temp2


            #See how putting this node here will affect other domains 
            for domain2_idx in range(len(domains)):
                if domain2_idx==domain_idx: continue
                domain2 = domains[domain2_idx]
                prev_index=-1; next_index = -1
                prev_resindex = -1; next_resindex = -1
                j=i
                while j>=0 and columns[j][domain2_idx]==-1: j-=1
                if j>=0: prev_index=j; prev_resindex = columns[j][domain2_idx]
                j=i+1
                while j<numcols and columns[j][domain2_idx]==-1:
                    j+=1
                if j<numcols: next_index=j; next_resindex = columns[j][domain2_idx]            
                

                if prev_resindex!=-1 and next_resindex!=-1:
                    if next_index == prev_index+1:

                        sse_score += -1*getSSEPenalty(domain2,domain2.residues[prev_resindex],domain2.residues[next_resindex],sse_penalties)
                        if sse_score>min_sse_score and min_sse_score != -1: break
                        separation+=1 
                    if next_resindex!=-1 and prev_resindex!=-1 and next_index==prev_index+1:
                        separation+=1
                    gaplen += next_resindex-prev_resindex

            #Check to see if this is the best potential placement seen so far
            if okay_to_place_cluster:
                if min_sse_score==-1:
                    col=i+1;min_sse_score=sse_score;min_separation=separation
                    min_gaplen=gaplen
                    min_sse_separation=sse_separation
                    found_placement=1
                elif min_sse_score>sse_score:
                    col=i+1;min_sse_score=sse_score;min_separation=separation
                    min_gaplen=gaplen
                    min_sse_separation=sse_separation
                    found_placement=1                    
                    if min_sse_score==0 and sse_separation==0 and gaplen==0 and separation==0:
                        found_last_potential_placement = 1
                elif min_sse_score==sse_score and sse_separation<min_sse_separation:
                    col=i+1;min_sse_score=sse_score;min_separation=separation
                    min_gaplen=gaplen
                    min_sse_separation=sse_separation
                    found_placement=1                    
                    if min_sse_score==0 and sse_separation==0 and gaplen==0 and separation==0:
                        found_last_potential_placement = 1

                elif min_sse_score==sse_score and min_sse_separation==sse_separation and separation<min_separation:
                    col=i+1;min_sse_score=sse_score;min_separation=separation
                    min_gaplen=gaplen
                    min_sse_separation=sse_separation
                    found_placement=1                    
                    if min_sse_score==0 and sse_separation==0 and gaplen==0 and separation==0:
                        found_last_potential_placement = 1

                elif min_sse_score==sse_score and min_sse_separation==sse_separation and min_separation==separation and min_gaplen>gaplen:
                    col=i+1;min_sse_score=sse_score;min_separation=separation
                    min_gaplen=gaplen
                    found_placement=1
                    if min_sse_score==0 and sse_separation==0 and gaplen==0 and separation==0:
                        found_last_potential_placement = 1

            else:
                if found_placement:
                    found_last_potential_placement=1



        #place node in MSTA
        if col!=-1:
            columns = columns[0:col]+[new_column.copy()]+columns[col:]        
        else:
            columns = columns[0:col]+[new_column.copy()]+columns[col:]        

        columns[col][domain_idx] = resindex

    return columns



#Simple method to guess how much information there might be
#to help fix a region of the MSTA where columns with many gaps
#occur. If the column is surrounded by highly occupied columns,
#perhaps that's an indication that the region is generally well
#aligned and so can be fixed relatively easily.
def getSurroundingOccupancies(domain_idx,domains,columns,resnum):
    cutoff = 0.7*len(domains)
    domain = domains[domain_idx]
    resindex = domain.residue_indices[resnum]
    col = 0
    while col < len(columns):
        if columns[col][domain_idx] == resindex: break
        col += 1
    if col >= len(columns): return -1 #error
    start = col-1
    while start >= 0:
        num_aligned = 0
        for d in range(len(domains)):
            if columns[start][d]!= -1: num_aligned += 1
        if num_aligned >= cutoff: break #region of low occupancy ends
        start -= 1

    if start < 0: average=0 #error
    else: average = num_aligned
    end = col+1
    while end < len(columns):
        num_aligned = 0
        for d in range(len(domains)):
            if columns[end][d]!=-1: num_aligned += 1
        if num_aligned >= cutoff: break
        end += 1
    if end >= len(columns): return num_aligned/float(2.0*len(domains)) #error

    average += num_aligned
    average /= (float(2*len(domains)))
    return average


#find columns with small number of aligned residues in the midst of
#columns with high numbers of SSE residues
def getInsertionColumns(columns,domains,marked_residues):
    low_cutoff = 0.3
    upper_cutoff = 0.7
    bad_residues = []
    found_start = 0
    i=0

    while i < len(columns):
        num_aligned = 0
        num_sse_residues = 0
        for d in range(len(domains)):
            domain = domains[d]
            if columns[i][d] != -1:
                num_aligned += 1
                if domain.sstype[domain.residues[columns[i][d]]]!='U':
                    num_sse_residues += 1
        percent_aligned = num_aligned/float(len(domains))
        if num_aligned != 0:
            percent_sse_residues = num_sse_residues/float(num_aligned)
        else:
            percent_sse_residues = 0

        if percent_aligned >= upper_cutoff:
            found_start = 1
            first_percent_aligned = percent_aligned
        elif percent_aligned <= low_cutoff:
            if found_start: #high sse column at i-1 followed by small number of aligned residues at
                bad_columns = []
                found_start = 0
                next_good_column = -1
                j = 1
                while j < 10:
                    if 10+i >= len(columns): break
                    num_aligned = num_sse_residues = 0
                    for d in range(len(domains)):
                        if columns[j+i][d] != -1:
                            num_aligned += 1
                            domain = domains[d]
                            if domain.sstype[domain.residues[columns[j+i][d]]] != 'U':
                                num_sse_residues += 1
                    percent_aligned = num_aligned/float(len(domains))
                    if num_aligned != 0:
                        percent_sse_residues = num_sse_residues/float(num_aligned)
                    else:
                        percent_sse_residues = 0
                    if percent_aligned >= upper_cutoff:
                        next_good_column = j+i
                        bad_columns.append(i)
                        for col in bad_columns:
                            for d in range(len(domains)):
                                if columns[col][d] != -1:
                                    domain = domains[d]
                                    resnum = domain.residues[columns[col][d]]
                                    if not marked_residues[d].has_key(resnum):
                                        average = (first_percent_aligned+percent_aligned)/float(2.0)
                                        bad_residues.append(((1-average),d,resnum))
                        break
                    elif percent_aligned <= low_cutoff:
                        bad_columns.append(j+i)
                    j += 1
                i += (j-1)
            
        i += 1
    return bad_residues


#Finds secondary structure elements with gaps that can be marked
#for realignment in step 6.
def getSSEsWithGaps(columns,domains,marked_residues):
    maximum_sse_gap_length = 2
    bad_residues = []
    for d in range(len(domains)):
        domain = domains[d]
        numresidues = domain.num_residues
        col_start = 0
        for s in range(domain.num_SSEs):
            start,end = domain.getSSEEnds(ssenum=s)
            if marked_residues[d].has_key(start):
                continue  #residue from SSE already marked

            start_index = domain.residue_indices[start]
            end_index = domain.residue_indices[end]
                
            for i in range(col_start,len(columns)):
                if columns[i][d] == start_index: col_start=i; break #if this is correct start column for SSE
            col_start = i

            previous_column_with_sse_residue = col_start
            index = start_index+1
            col = col_start+1
            while index <= end_index: #for each residue in the SSE
                if marked_residues[d].has_key(domain.residues[index]):
                    previous_column_with_sse_residue = col
                    break
                while col<len(columns) and columns[col][d] != index: #find the column
                    col += 1
                if col>=len(columns): break #couldn't find residue; perhaps was hetatm

                if col-previous_column_with_sse_residue>maximum_sse_gap_length: #add to marked residue
                    p = getSurroundingOccupancies(d,domains,columns,domain.residues[columns[col][d]])
                    bad_residues.append(((1-p),d,domain.residues[columns[col][d]]))
                    previous_column_with_sse_residue = col
                    break #don't need to add residue in this sse more than once
                previous_column_with_sse_residue = col
                index += 1

                while not domain.residues.has_key(index) and index<numresidues: index += 1
            start_col = previous_column_with_sse_residue
    return bad_residues



#When a loop residue is targeted for realignment in step 6, this
#function returns the loop region to be realigned.
def getLoopRegion(domain_idx,domain,columns,resindex,domains):

    #find column containing the residue targeted for realignment
    for i in range(len(columns)):
        if columns[i][domain_idx]==resindex: start_col = i; break
    resindex_column = start_col
    end_col = start_col

    #find the column corresponding to the beginning of the loop in
    #the targeted domain: might be a gap column
    start_col -= 1
    while start_col >= 0:
        if columns[start_col][domain_idx] == -1:
            start_col -= 1
        elif domain.sstype[domain.residues[columns[start_col][domain_idx]]]=='U':
            start_col -= 1
        else:
            start_col += 1 #found SSE residue: too far
            break
    if start_col<0:
        start_col = 0

    #find the first residue in the loop (not necessarily start_col since gaps could occur)
    i=0
    while start_col+i < len(columns) and columns[start_col+i][domain_idx] == -1: i += 1
    if start_col+i < len(columns): start_resindex = columns[start_col+i][domain_idx]
    else: start_resindex = resindex #?

    #similarly for end column and residue
    end_col += 1
    while end_col < len(columns):
        if columns[end_col][domain_idx]==-1:
            end_col += 1
        elif domain.sstype[domain.residues[columns[end_col][domain_idx]]]=='U':
            end_col += 1
        else:
            end_col -=1
            break

    if end_col >= len(columns):
        end_col = len(columns)-1
    i=0
    while end_col-i >= 0 and columns[end_col-i][domain_idx]==-1: i += 1
    if end_col-i >= 0: end_resindex = columns[end_col-i][domain_idx]
    else: end_resindex = resindex #?
    return (resindex_column,start_resindex,end_resindex,start_col,end_col)


#When a non-loop residue is targeted for realignment in step 6, this function
#is called to find the MSTA region in which the realignment will take place.
def getColumnRegion(columns,domain_idx, start_resindex,end_resindex,domains,alignments,sse_start,sse_end):
    domain = domains[domain_idx]
    startcol=endcol=-1

    #find column containing the first residue in the region targeted for realignment
    for i in range(len(columns)):
        if columns[i][domain_idx] == start_resindex: startcol=i; break

    #find the column containing the last residue in the region targeted for realignment
    resindex_column = startcol
    for i in range(startcol,len(columns)):
        if columns[i][domain_idx] == end_resindex: endcol=i; break

    
    #Find out how far this region could theoretically be extended into loops without
    #picking up additional secondary structure elements
    upper_limit = endcol+1
    while upper_limit<len(columns):
        if columns[upper_limit][domain_idx]!=-1:
            if domain.sstype[domain.residues[columns[upper_limit][domain_idx]]]!='U': break
        upper_limit += 1
    upper_limit -=1
    
    lower_limit = startcol-1
    while lower_limit>=0:
        if columns[lower_limit][domain_idx]!=-1:
            if domain.sstype[domain.residues[columns[lower_limit][domain_idx]]]!='U': break
        lower_limit -= 1
    lower_limit += 1


    #Final region will contain full SSE for all domains with SSEs aligned to
    #target SSE
    new_start_col = startcol
    new_end_col = endcol
    sse_start = domain.residues[sse_start]; sse_end = domain.residues[sse_end]
    for d in range(len(domains)):
        if d==domain_idx: continue
        if alignments.alignments[domain_idx][d].alignedSSEs.has_key(sse_start):
            start = alignments.alignments[domain_idx][d].alignedSSEs[sse_start]
            end = domains[d].sse_ends[domains[d].ssNumber[start]][1]
            start_idx = domains[d].residue_indices[start]
            end_idx = domains[d].residue_indices[end]
            low_col = new_start_col; high_col = new_end_col
            for c in range(lower_limit,upper_limit+1):
                if columns[c][d]>=start_idx and columns[c][d]<=end_idx:
                    if c< low_col: low_col = c
                    elif c>high_col: high_col = c
                if columns[c][d]>end: break
            new_start_col = low_col; new_end_col = high_col
                    
    return resindex_column,new_start_col,new_end_col

#Get domains in a given region that have a residue marked for realignment (step 6)
def getBadDomainsInRegion(columns,domains,marked_residues,colstart,colend):
    bad_domains = []
    for i in range(colstart,colend+1):
        for idx in range(len(domains)):
            if idx in bad_domains: continue
            resindex = columns[i][idx]
            if resindex == -1: continue
            resnum = domains[idx].residues[resindex]
            if marked_residues[idx].has_key(resnum): bad_domains.append(idx)
    return bad_domains


#Determine if at least one domain in the column has an SSE residue that
#is not the first residue in that SSE. If so, then a gap before this position
#would put a gap in an SSE
def columnHasOpenSSE(column,query_idx,domains):
    num_SSE_residues = 0
    for d in range(len(domains)):
        domain = domains[d]
        if d==query_idx: continue
        if column[d]==-1: continue
        if domain.sstype[domain.residues[column[d]]]!='U' and not domain.isSSEEnd(column[d])==1:
            return 1
    return 0


#Gaps in the dynamic programming used in step 6 receive higher penalties
#if they occur in the middle of SSEs. Calculates gap penalty
def gapScore(domain,resnum,openSSE,loop=0,opened_gap=0):
    multiplier=gapMultiplier
    extend_multiplier = gap_extend_penalty
    SSE_open_multiplier = 3.0
    if loop: multiplier*=0.5
    if domain.sstype[resnum]=='U':
        if openSSE:
            if opened_gap: return SSE_open_multiplier*extend_multiplier
            else: return SSE_open_multiplier*multiplier
        else:
            if opened_gap: return extend_multiplier*1.5
            else: return multiplier*1.5

    else:
        if openSSE:
            if opened_gap: return SSE_open_multiplier*extend_multiplier
            else: return SSE_open_multiplier*multiplier
        else:
            if opened_gap: return extend_multiplier*1.5
            else: return multiplier*1.5


#Of all domains in the column, determines the highest gap penalty
def columnGapScore(domain_idx,domains,columns,col,openSSE,loop=0,column_SSE_open=1,gap_open=0):
    if gap_open:
        multiplier=gapMultiplier
    else:
        multiplier= gap_extend_penalty
    if loop: multiplier*=0.5
    SSE_open_multiplier = 3.0
    SSE_open_multiplier=multiplier
    SSE_closed_multiplier=multiplier/2.0

    if loop and openSSE==0 and column_SSE_open==0:
        return 0
   
    if openSSE: SSEmultiplier=3
    else: SSEmultplier=1

    maxpenalty = 0
    num_loops = 0; num_sses = 0
    for d in range(len(domains)):
        if d==domain_idx: continue
        if columns[col][d]==-1: continue
        resindex = columns[col][d]
        domain = domains[d]
        resnum = domain.residues[columns[col][d]]

        if domain.sstype[resnum] != 'U':
            num_sses += 1
            if openSSE:
                newpenalty = 2+(SSEmultiplier-2)*openSSE
            else: newpenalty=1.5 
        else:
            num_loops+=1
            if openSSE:
                newpenalty = 1.5+(SSEmultiplier-1.5)*openSSE
            else:
                newpenalty=1.5 

        if maxpenalty < newpenalty: maxpenalty=newpenalty

    #another special case: no gap penalty when aligning to column that only has loop residues
    if num_sses==0 and num_loops<0.5*len(domains):
        return 0
    return maxpenalty*multiplier


#Determines if a residue can be placed in a given column. For instance,
#can't align a helical residue to a column containing all strands
def isLegalAlignment(query_idx,query_resnum,domains,column,marked_residues):
    has_allowed_residue = 0
    has_allowed_marked_residue = 0
    query_sse_type = domains[query_idx].sstype[query_resnum]
    if query_sse_type == 'U': return 1 #any alignment to loop residue is allowed
    num_domains_in_column = 0
    is_marked = 0
    if marked_residues[query_idx].has_key(query_resnum):         
        return 1
    for d in range(len(domains)):
        if d==query_idx: continue
        if column[d]==-1: continue
        num_domains_in_column += 1
        if marked_residues[d].has_key(domains[d].residues[column[d]]):
            continue
        ssetype = domains[d].sstype[domains[d].residues[column[d]]]
        if ssetype=='U':
            has_allowed_residue = 1
            return 1
        elif ssetype==query_sse_type: return 1
    if num_domains_in_column == 0: return 1 #no other domains, so no clashes between sse types
    return 0


#Get the dynamic programming score for placing a residue in a given column.
#Based on consistency.
def consistencyDPScore(query_idx,query_resindex,domains,column,alignments,marked_residues,debug,loop=0,column_SSE_open=1,allow_marked=0):
    query = domains[query_idx]
    query_resnum = query.residues[query_resindex]
    all_marked = 1
    num_matched = 0
    for d in range(len(domains)):
        if column[d]!=-1:
            if not marked_residues[d].has_key(domains[d].residues[column[d]]):
                all_marked = 0
                break
    if not all_marked:
        if not isLegalAlignment(query_idx,query_resnum,domains,column,marked_residues):
            return -1000
    num_voters = 0
    num_matched = 0
    max_score = 0
    penalty = 0
    negvotes = 0
    query_is_marked = 0
    num_matching_helices = 0; num_matching_strands = 0
    num_loop_to_loop_matches = 0
    if marked_residues[query_idx].has_key(query_resnum):
        query_is_marked = 1
    query_sse_type = query.sstype[query.residues[query_resindex]]
    votes = 0

    sstypes = {}
    num_helices = 0; num_strands = 0
    for i in range(len(domains)):
        if column[i]!=-1:
            sstype = domains[i].sstype[domains[i].residues[column[i]]]
            if sstype=="H": num_helices +=1
            elif sstype=="S": num_strands += 1
            sstypes[i] = sstype

    for i in range(len(domains)):
        if i == query_idx: continue
        
        if column[i] == -1:                             
            #Give a slight penalty to loop residues so that we prefer to align
            #to columns with SSEs (if query resnum is SSE residue).
            if not loop: penalty += -0.1
            elif column_SSE_open: penalty += -0.1
            continue
        if not allow_marked:
            if  marked_residues[i].has_key(domains[i].residues[column[i]]): continue

        matches = 0
        vote = -1

        #See if domains[i] places query_resnum in this column
        if i<=query_idx:
            if alignments.alignments[i][query_idx].alignedResidues.has_key(domains[i].residues[column[i]]):
                vote = alignments.alignments[i][query_idx].alignedResidues[domains[i].residues[column[i]]]
                if vote==query_resnum: matches=1
                else: matches = -1

        else:
            if alignments.alignments[query_idx][i].alignedResidues.has_key(query_resnum):
                vote = alignments.alignments[query_idx][i].alignedResidues[query_resnum]
                if vote==domains[i].residues[column[i]]: matches = 1
                else: matches=-1

        max_score += 1.3
        if matches==1: num_matched += 1
        sse_type = sstypes[i]
        if sse_type=="H" and matches==1: num_matching_helices += 1
        elif sse_type=="S" and matches==1: num_matching_strands += 1

        if sse_type!=query_sse_type: 
            if sse_type=="S":
                if query_sse_type=="H": votes += matches*0; #no sse mismatches allowed
                else:
                    if matches==1:
                        votes += 0.3; #less credit given to loop-SSE alignments
                    else:
                        penalty += -0.3 #small penalty since domains[i] didn't place query_resnum
                                        #in this column
            elif sse_type=="H":
                if query_sse_type == "S": votes += 0
                else:
                    if matches==1:
                        votes += 0.3;
                    else:
                        penalty += -0.3
            else: #sse_type=='U'
                if matches==1:
                    votes += 0.3 #sse_type is loop and query_sse_type is residue
                else:
                    penalty += -0.3
        else: 
            if matches==1:
                if sse_type!='U' or (num_helices>0 or num_strands>0):
                    votes += 1.3
                else:
                    num_loop_to_loop_matches += 1
                    votes += 1.3 #max credit to H-H or S-S match
            else:
                if sse_type!='U' or (num_helices>0 or num_strands>0):
                    penalty += -0.3 
                else:
                    penalty += -0.3

        num_voters += 1

    #If column contains mostly loops (and query resnum is a loop), give higher score
    #than loop-loop matches normally recieve.
    if num_voters>=0.8*(len(domains)-1) and (num_loop_to_loop_matches>=0.5*num_voters):
        votes += num_loop_to_loop_matches*0.7

    #Give bonus if column is more than half occupied with helical or strand matches
    if num_voters>0.5*len(domains):
        if  num_matching_helices>=0.5*num_voters:
            votes += 0.7*num_matching_helices
        elif num_matching_strands>=0.5*num_voters:
            votes += 0.7*num_matching_strands
    if max_score>0:
        score = votes/float(max_score)*20
        score += penalty
    else:
        score = penalty
    return score


#When a residue is left unaligned in step 6 and the pairwise alignments don't place it in
#a column in the region being realigned, this function is called to find the best place to
#insert a new column (will have the residue and gaps for all other structures). Uses same
#method as orderMSAColumns to place the new column.
def placeResidueInColumns(columns,domains,domain_residues,domain_idx,skipped_residues,col_start,col_end,sse_penalties,marked_residues,column_mapping):
    skipped_res_columns = {}
    domain = domains[domain_idx]
    skipped_residues.sort()
    new_column = zeros([len(domains)],Int)-1
    prev_column = -1
    for resindex in skipped_residues:
        numcols = len(columns)
        col=-1
        min_sse_score=-1
        min_separation=-1
        min_sse_separation=-1
        min_gaplen=-1
        maxcol=-1
        is_potential_placement=found_placement=0
        found_last_potential_placement=0
        column_start = prev_column
        for i in range(column_start,numcols):
            if found_last_potential_placement: break
            okay_to_place_cluster=0
            prev_index=-1; next_index = -1
            prev_resindex = -1; next_resindex = -1
            j=i
            while j>=0 and columns[j][domain_idx]==-1: j-=1
            if j>=0: prev_index=j; prev_resindex = columns[j][domain_idx]
            j=i+1

            while j<numcols and columns[j][domain_idx]==-1:
                j+=1
            if j<numcols: next_index=j; next_resindex = columns[j][domain_idx]            
            if prev_resindex != -1 and next_resindex != -1: #internal placement
                if prev_resindex!=-1 and next_resindex!=-1:
                    if resindex > prev_resindex and resindex < next_resindex:
                        okay_to_place_cluster=1
                        is_potential_placement = 1
            elif prev_resindex==-1 and next_resindex!=-1: #before first res
                if resindex < next_resindex:
                        okay_to_place_cluster=1
            elif prev_resindex!=-1 and next_resindex==-1: #after last position
                if resindex>prev_resindex:
                        okay_to_place_cluster=1
                        is_potential_placement = 1
            else: #first residue to be placed
                    okay_to_place_cluster=1
                    is_potential_placement = 1


            if not okay_to_place_cluster:
                if found_placement: found_last_potential_placement=1
                prev_domain=domain_idx
                prev_column=col
                continue
            sse_separation=0;gaplen=0;separation=0
            if next_resindex!=-1 and resindex<next_resindex:
                gaplen += next_resindex-resindex
            if prev_resindex!=-1 and resindex>prev_resindex:
                gaplen += resindex-prev_resindex


            sse_score = 0
            temp=temp2=-1

            ssenum = domain.ssNumber[domain.residues[resindex]]
            prev_index2=prev_index;next_index2=next_index

            if ssenum!=-1 and next_index2!=-1 and domain.ssNumber[domain.residues[columns[next_index2][domain_idx]]] == ssenum:
                sse_separation += next_index2-i-1
            if ssenum!=-1 and prev_index2!=-1 and domain.ssNumber[domain.residues[columns[prev_index2][domain_idx]]] == ssenum:
                sse_separation += i-prev_index2-1

            if next_index2!=-1:
                temp = -1*getSSEPenalty(domain,domain.residues[resindex],domain.residues[columns[next_index2][domain_idx]],sse_penalties)
            if prev_index2!=-1:
                temp2 = -1*getSSEPenalty(domain,domain.residues[columns[prev_index2][domain_idx]],domain.residues[resindex],sse_penalties)
            if temp!=-1 and temp2!=-1:
                sse_score += (temp+temp2)/2.0
            elif temp!=-1:
                sse_score += temp
            elif temp2!=-1:
                sse_score += temp2

            for domain2_idx in range(len(domains)):
                if domain2_idx==domain_idx: continue
                domain2 = domains[domain2_idx]
                prev_index=-1; next_index = -1
                prev_resindex = -1; next_resindex = -1
                j=i
                while j>=0 and columns[j][domain2_idx]==-1: j-=1
                if j>=0: prev_index=j; prev_resindex = columns[j][domain2_idx]
                j=i+1
                while j<numcols and columns[j][domain2_idx]==-1:
                    j+=1
                if j<numcols: next_index=j; next_resindex = columns[j][domain2_idx]            
                

                if prev_resindex!=-1 and next_resindex!=-1:
                    if next_index == prev_index+1:

                        sse_score += -1*getSSEPenalty(domain2,domain2.residues[prev_resindex],domain2.residues[next_resindex],sse_penalties)
                        if sse_score>min_sse_score and min_sse_score != -1: break
                        separation+=1 
                    if next_resindex!=-1 and prev_resindex!=-1 and next_index==prev_index+1:
                        separation+=1
                    gaplen += next_resindex-prev_resindex
            if okay_to_place_cluster:
                if min_sse_score==-1:
                    col=i+1;min_sse_score=sse_score;min_separation=separation
                    min_gaplen=gaplen
                    min_sse_separation=sse_separation
                    found_placement=1
                elif min_sse_score>sse_score:
                    col=i+1;min_sse_score=sse_score;min_separation=separation
                    min_gaplen=gaplen
                    min_sse_separation=sse_separation
                    found_placement=1                    
                    if min_sse_score==0 and sse_separation==0 and gaplen==0 and separation==0:
                        found_last_potential_placement = 1
                elif min_sse_score==sse_score and sse_separation<min_sse_separation:
                    col=i+1;min_sse_score=sse_score;min_separation=separation
                    min_gaplen=gaplen
                    min_sse_separation=sse_separation
                    found_placement=1                    
                    if min_sse_score==0 and sse_separation==0 and gaplen==0 and separation==0:
                        found_last_potential_placement = 1

                elif min_sse_score==sse_score and min_sse_separation==sse_separation and separation<min_separation:
                    col=i+1;min_sse_score=sse_score;min_separation=separation
                    min_gaplen=gaplen
                    min_sse_separation=sse_separation
                    found_placement=1                    
                    if min_sse_score==0 and sse_separation==0 and gaplen==0 and separation==0:
                        found_last_potential_placement = 1

                elif min_sse_score==sse_score and min_sse_separation==sse_separation and min_separation==separation and min_gaplen>gaplen:
                    col=i+1;min_sse_score=sse_score;min_separation=separation
                    min_gaplen=gaplen
                    found_placement=1
                    if min_sse_score==0 and sse_separation==0 and gaplen==0 and separation==0:
                        found_last_potential_placement = 1

            else:
                if found_placement:
                    found_last_potential_placement=1

        if col!=-1:
            columns = columns[0:col]+[new_column.copy()]+columns[col:]        
        else:
            columns = columns[0:col]+[new_column.copy()]+columns[col:]        
        columns[col][domain_idx] = resindex
        skipped_res_columns[resindex] = col-col_start
        new_columns_col = col-col_start
        if col<=col_start:
            col_start += 1; col_end += 1
        elif col>col_start and col<col_end:
            col_end+=1
            for x in range(len(column_mapping)):
                if column_mapping[x]>=new_columns_col:
                    column_mapping[x]+=1
        elif col>=col_end:
            column_mapping.append(-1)
            
    return columns,col_start,col_end,column_mapping



#Integrates the new aligned region in new_columns into the MSTA (columns)
def integrateAlignedRegion(columns,new_columns,domains,domain_residues,domain_idx,bad_resindex,region_start,region_end,col_start,col_end,alignments,aligned,skipped_columns,skipped_residues,marked_residues,debug,sse_penalties,original_columns,loop=0,last_iteration=0):

    #column_mapping keeps track of column in new_columns that each column of columns corresponds to (Do I have that backwards?)
    column_mapping = [x for x in range(len(new_columns))]

    original_new_columns = deepcopy(new_columns)


    numcols = len(new_columns)
    domain = domains[domain_idx]
    sselength = len(domain_residues) #bad name: not all residues are sse residues
    num_changes=0

    #Overwrite domain residues in the region in columns with -1 (gap).
    for j in range(numcols):
        if new_columns[j][domain_idx] >= region_start and new_columns[j][domain_idx] <= region_end:
            columns[j+col_start][domain_idx] = -1
        new_columns[j][domain_idx] = -1

    
    for j in range(numcols):
        if aligned.has_key(j): #if residue was aligned in dynamic programming, initialize new_columns (should be redundant)
            new_columns[j][domain_idx] = aligned[j]
        elif j in skipped_columns: #If it's a skipped column, initialize new_columns with whatever was in columns before
            new_columns[j][domain_idx] = columns[j+col_start][domain_idx]

    num_columns = len(columns)
    num_new_columns = len(new_columns)
    #Place each residue in new_columns, adding extra columns only if necessary
    for i in range(sselength):
        if domain_residues[i] in skipped_residues:
            column = 0
            resindex = domain_residues[i]
            found = 0
            final_col=-1
            col=-1;min_sse_score=-1;min_separation=-1;min_gaplan=-1;maxcol=-1
            prev_index=next_index=prev_resindex=next_resindex=-1
            is_potential_placement=0
            start_column = -1
            end_column = len(new_columns)-1
            found_placement=okay_to_place_cluster=0

            for c in range(num_new_columns): #Check each column to see if residue could be placed here
                if new_columns[c][domain_idx] != -1:
                    if new_columns[c][domain_idx] < resindex:
                        prev_index = c; prev_resindex = new_columns[c][domain_idx]
                        start_column = c;
                    if new_columns[c][domain_idx] > resindex:
                        next_index = c
                        next_resindex = new_columns[c][domain_idx]
                        end_column = c-1
                        column = c; found=1
                        break #found first column in which residue could be placed

            ssenum = domain.ssNumber[domain.residues[resindex]]

            #first try placing in an existing column
            max_score = 0; maxcol = -1

            #Find columns in columns list to which start_column and new_column of new_columns map
            for c in range(len(column_mapping)):
                if column_mapping[c]==start_column:
                    mapped_start_column = c
                    break
            for c in range(len(column_mapping)):
                if column_mapping[c]==end_column:
                    mapped_end_column = c
                    break

                
            #See if any column has a non-zero consistencyDPScore (i.e. see if the pairwise alignments
            #tend to place it anywhere)
            for c in range(start_column,end_column+1):
                if c<0: continue
                debug_score = 0

                #Since last_iteration variable isn't being used, both calls to consistencyDPScore are
                #the same.
                if resindex==bad_resindex and original_columns[resindex]==c: 
                    score = consistencyDPScore(domain_idx,resindex,domains,new_columns[c],alignments,marked_residues,debug_score,loop,columnHasOpenSSE(new_columns[c],domain_idx,domains),last_iteration)
                else:
                    score = consistencyDPScore(domain_idx,resindex,domains,new_columns[c],alignments,marked_residues,debug_score,loop,columnHasOpenSSE(new_columns[c],domain_idx,domains),0)                    
                if score>max_score:
                    if not aligned.has_key(c): #if another residue is not already assigned to this column
                        max_score = score; maxcol = c
                        
            if max_score > 0: #found column with evidence for alignment in pairwise alignments, ignoring issue of ties
                col = final_col = maxcol
                new_columns[col][domain_idx] = resindex

                #need to adjust skipped columns and aligned (Why? Not adding a column. This code is rarely used, so maybe it has a bug and I never noticed. It's also possible that this code compensates for another bug.)
                new_aligned = {}
                for col_placement,r in aligned.items():
                    if col_placement>col: new_aligned[col_placement+1] = r
                    else: new_aligned[col_placement] = r
                aligned = new_aligned
                aligned[col] = resindex
                new_skipped_cols = []
                for col_placement in skipped_columns:
                    if col_placement > col:
                        new_skipped_cols.append(col_placement+1)
                    elif not col_placement == col:
                        new_skipped_cols.append(col_placement)
                skipped_columns = new_skipped_cols
                skipped_residues.remove(domain_residues[i])


    #Look for columns that have all gaps and remove (could happen during realignment)
    gap_columns = []
    numcols = len(new_columns)
    for j in range(numcols):
        all_gaps = 1
        for d in range(len(domains)):
            if new_columns[j][d]!=-1: all_gaps=0;break
        if all_gaps: gap_columns.append(j)
    i=0
    while i<len(gap_columns):
        new_columns = new_columns[:gap_columns[i]-i]+new_columns[gap_columns[i]-i+1:]
        column_mapping = column_mapping[:gap_columns[i]-i]+column_mapping[gap_columns[i]-i+1:] #?
        i += 1
    numcols = len(new_columns)
    for j in range(numcols):
        for c in range(len(column_mapping)):
            if column_mapping[c]==j:
                mapped_col = c
                break
        else: continue
        columns[mapped_col+col_start][domain_idx] = -1


    #Integrate new_columns into columns
    columns = columns[0:col_start]+deepcopy(new_columns)+columns[col_end+1:]
    col_end = col_start+len(new_columns)
    if len(skipped_residues)!=0: #if some of domain's residues have not been placed in new_columns yet
        columns,full_col_start,full_col_end,column_mapping = placeResidueInColumns(columns,domains,domain_residues,domain_idx,skipped_residues,col_start,col_end,sse_penalties,marked_residues,column_mapping)
        new_columns = columns[full_col_start:full_col_end+1]

    numcols = len(new_columns)


    #Update the residue correspondences in the pairwise alignments
    num_moves = 0
    for i in range(numcols):
        for c in range(len(column_mapping)):
            if column_mapping[c]==i:
                mapped_col = c; break
        else:
            continue
        if columns[mapped_col+col_start][domain_idx] == -1: continue
        domain_resnum = domain.residues[columns[mapped_col+col_start][domain_idx]]
        for d in range(len(domains)):
            if d==domain_idx: continue
            if columns[mapped_col+col_start][d] == -1: continue
            new_aligned_resindex = columns[mapped_col+col_start][d]
            new_aligned_residue = domains[d].residues[columns[mapped_col+col_start][d]]
            r = -1
            if domain_idx<=d: #Shouldn't need to check this anymore
                if alignments.alignments[domain_idx][d].alignedResidues.has_key(domain_resnum):
                    r = alignments.alignments[domain_idx][d].alignedResidues[domain_resnum]
                    if r != new_aligned_residue: #includes case where r=-1
                        num_changes += 1
                        alignments.alignments[domain_idx][d].alignedResidues[domain_resnum] = new_aligned_residue
                else:
                    alignments.alignments[domain_idx][d].num_aligned_residues += 1
            else:

                if alignments.alignments[d][domain_idx].alignedResidues.has_key(new_aligned_residue):
                    r = alignments.alignments[d][domain_idx].alignedResidues[new_aligned_residue]
                    if r != domain_resnum: #includes case where r=-1
                        num_changes += 1
                        alignments.alignments[d][domain_idx].alignedResidues[new_aligned_residue] = domain_resnum

                else:
                    alignments.alignments[d][domain_idx].num_aligned_residues += 1

    return columns,num_changes

#Realigns residues from col_start to col_end
def runDynamicProgramming(columns,domains,region_start,region_end,col_start,col_end,domain_idx,bad_resindex,alignments,bad_domains,marked_residues,debug,sse_penalties,loop=0,last_iteration=0):

    new_columns = deepcopy(columns[col_start:col_end+1]) #holds realigned region
    domain_residues = [] #Residues in region for marked domain (in terms of residue indices)
    numcols = len(new_columns)
    original_columns = {}
    for j in range(numcols):
        if columns[j+col_start][domain_idx] != -1:
            domain_residues.append(columns[j+col_start][domain_idx])
            original_columns[columns[j+col_start][domain_idx]] = j

    if len(domain_residues)==0: return columns,-1 #error
    region_start = domain_residues[0]
    region_end = domain_residues[-1]

    sselength = len(domain_residues) #bad name: not all residues are sse residues
    numcols = len(new_columns)
    domain = domains[domain_idx] #the domain marked for realignment
    
    scorearray = []
    pointarray = []
    #initialize scorearray and pointarray
    for i in range(sselength+1): #bad domain's region (SSE)
        scorearray.append([]); pointarray.append([])
        for j in range(numcols+1): #length of newcols: region to which sse will align
            scorearray[i].append(0)
            pointarray[i].append([None,None])

    #initial conditions
    scorearray[0][0] = 0
    pointarray[0][0][0] = -1; pointarray[0][0][1] = -1

    #More initial conditions.
    opened_gap = 0 #Keeps track of whether gap is open
    if domain.sstype[domain.residues[domain_residues[0]]]!='U':
        scorearray[1][0] = scorearray[0][0] - gapScore(domain,domain.residues[domain_residues[0]],0,loop)
        opened_gap = 1
    else: scorearray[1][0] = 0
    pointarray[1][0] = [0,0]


    #More initial conditions
    for i in range(1,sselength):
        if domain.sstype[domain.residues[domain_residues[i]]]!='U':
            gapscore = gapScore(domain,domain.residues[domain_residues[i]],0,loop,opened_gap)
            opened_gap = 1
        else: gapscore = 0
        scorearray[i+1][0] = scorearray[i][0]-gapscore
        pointarray[i+1][0][0] = i; pointarray[i+1][0][1] = 0;


    #More initial conditions
    pointarray[0][1] = [0,0]
    scorearray[0][1] = scorearray[0][0] 
    num_extra_columns = numcols-sselength
    for j in range(1,numcols):
        gapscore = 0
        scorearray[0][j+1] = scorearray[0][j] - gapscore
        pointarray[0][j+1][0] = 0; pointarray[0][j+1][1] = j


    #Run the dynamic programming
    domain_SSE_open=columns_SSE_open=0
    for i in range(1,sselength+1):
        for j in range(1,numcols+1):
            globaldebug=0
            column_SSE_open=0
            column_SSE_open = columnHasOpenSSE(columns[col_start+j-1],domain_idx,domains)
            domain_SSE_open=0
            if domain_residues[i-1]-1 >= 0:
                if domain.sstype[domain.residues[domain_residues[i-1]-1]]!= 'U' and not domain.isSSEEnd(domain_residues[i-1]-1)==1:
                    domain_SSE_open = 1

            if pointarray[i-1][j][0]==i-2 and pointarray[i-1][j][1]==j: #gap at i-1: apply extension penalty                    
                gapscore = gapScore(domain,domain.residues[domain_residues[i-1]],column_SSE_open,loop,1)
            else:
                gapscore = gapScore(domain,domain.residues[domain_residues[i-1]],column_SSE_open,loop,0)

            max = scorearray[i-1][j] - gapscore
            pointarray[i][j][0] = i-1; pointarray[i][j][1] = j;



            if pointarray[i][j-1][0]==i and pointarray[i][j-1][1]==j-2: #gap at j-1: apply extension penalty
                gapscore = columnGapScore(domain_idx,domains,columns,j-1+col_start,domain_SSE_open,loop,column_SSE_open,1)
            else:
                gapscore = columnGapScore(domain_idx,domains,columns,j-1+col_start,domain_SSE_open,loop,column_SSE_open,0)                

            if max < scorearray[i][j-1]-gapscore:
                max = scorearray[i][j-1]-gapscore
                pointarray[i][j] = [i,j-1]


            debug_score = 0
            if domain_residues[i-1]==bad_resindex or (original_columns[domain_residues[i-1]]==j-1):
                score = consistencyDPScore(domain_idx,domain_residues[i-1],domains,new_columns[j-1],alignments,marked_residues,debug_score,loop,column_SSE_open,last_iteration) #never used (archaic)
            else:
                score = consistencyDPScore(domain_idx,domain_residues[i-1],domains,new_columns[j-1],alignments,marked_residues,debug_score,loop,column_SSE_open,0)                
            if max < scorearray[i-1][j-1]+score:
                max = scorearray[i-1][j-1]+score
                pointarray[i][j] = [i-1,j-1]
            scorearray[i][j] = max

    #traceback
    aligned = {}; skipped_columns = []; skipped_residues = []
    row = sselength
    max = -1
    maxcol = numcols

    max = -1; maxcol = numcols
    for col in range(1,numcols+1):
        if scorearray[row][col]>max:
            if scorearray[sselength][col]>max:
                max = scorearray[sselength][col]
                maxcol = col
    i = row; j = maxcol

    while not (i==0 or j==0):
        nexti,nextj = pointarray[i][j]
        if nexti==-1 and nextj==-1 and not (i==0 and j==0): 
            max=-1;maxcol=0
            while max==-1 and i>=0:
                for col in range(1,j):
                    col = j-col
                    if scorearray[i][col]>max: 
                        if scorearray[i][col]==0 and pointarray[i][col]==[-1,-1]: continue
                        max=scorearray[i][col]; maxcol=col                        
                if max==-1:
                    if i!=0:
                        skipped_residues.append(domain_residues[i-1]) #unaligned residue
                    i -= 1
            j = maxcol
            continue
        if nexti==i-1 and nextj==j-1:
            if i-1 >= 0 and j-1 >= 0:
                aligned[j-1] = domain_residues[i-1] #aligned residue
                firstj=j-1;firsti=i-1
        elif nexti==i-1 and nextj==j:
            if i-1 >= 0:
                skipped_residues.append(domain_residues[i-1]) #unaligned residue
                firstj=j;firsti=i-1
        elif nexti==i and nextj==j-1:
            if j-1 >= 0:
                skipped_columns.append(j-1) #domain has gap in this column
                firstj=j-1;firsti=i
        i=nexti; j=nextj;

    for i in range(sselength): #look for domain residues that weren't aligned 
        found = 0
        for col,res in aligned.items():
            if res==domain_residues[i]:
                found=1; break
        if found: continue #if found, domain_residues[i] was aligned in dynamic programming
        if domain_residues[i] not in skipped_residues:
            skipped_residues.append(domain_residues[i])
        
    for j in range(numcols): #look for columns in which domain has gaps
        found = 0
        if aligned.has_key(j): found=1
        if found: continue
        if j not in skipped_columns:
            skipped_columns.append(j)
            

    num_changes=0


    #Integrate the new aligned region into the multiple alignment
    columns,num_changes = integrateAlignedRegion(columns,new_columns,domains,domain_residues,domain_idx,bad_resindex,region_start,region_end,col_start,col_end,alignments,aligned,skipped_columns,skipped_residues,marked_residues,debug,sse_penalties,original_columns,loop,last_iteration)
    return columns,num_changes


#A given secondary structure element or loop region may be marked for
#realignment more than once. This function retains each region only once
#(though the same SSE or loop may be marked for realignment once for each
#structure). 
def filterBadResidues(bad_residues,domains):
    new_bad_residues = []
    by_domain = {}
    
    for i in range(len(domains)):
        by_domain[i] = []
        for j in range(domains[i].num_residues):
            by_domain[i].append((j,2))

    for percent,domain_idx,resnum in bad_residues:
        resindex = domains[domain_idx].residue_indices[resnum]
        by_domain[domain_idx][resindex] = (resindex,percent)


    for i in range(len(domains)):
        for j in range(domains[i].num_residues):
            resnum = domains[i].residues[j]
            new_region = 0

            if j>0 and domains[i].ssNumber[resnum]!= domains[i].ssNumber[domains[i].residues[j-1]]:
                new_region = 1
            elif j==0: new_region = 1
            ssenum = domains[i].ssNumber[resnum]
            if new_region:
                if j!=0 and best_percent != 2:
                    new_bad_residues.append((best_percent,i,domains[i].residues[best_resindex]))

                best_percent = 2

            if by_domain[i][j][1] < best_percent:
                best_resindex=j; best_percent = by_domain[i][j][1]
            if j==domains[i].num_residues-1 and best_percent != 2:
                new_bad_residues.append((best_percent,i,domains[i].residues[best_resindex]))

    return new_bad_residues


#Performs step 6 of CBA: local realignment of regions that are likely to
#contain mistakes.
def fixBadRegions(columns,domains,nodes,clusters,marked_residues,alignments,sse_penalties):
    maxiterations=10

    new_column = zeros([len(domains)],Int)-1
    original_marked_residues = deepcopy(marked_residues)

    done=0
    iterations = 0
    do_last_iteration = 0 #archaic: doesn't do anything
    
    while (iterations < maxiterations or do_last_iteration):
        if do_last_iteration: break #to prevent last_iteration from being used
        if not do_last_iteration:
            num_changes = 0 #number of pairwise registration changes
            bad_residues = [] #residues marked for realignment        
            bad_residues = getSSEsWithGaps(columns,domains,marked_residues)
            bad_residues += getInsertionColumns(columns,domains,marked_residues)

            #collect marked residues in a list for convenience
            for domain_idx in range(len(domains)):
                domain = domains[domain_idx]
                resnums = domain.residues.values()
                for resnum in resnums:
                    if marked_residues[domain_idx].has_key(resnum):
                        avg_percent_aligned = getSurroundingOccupancies(domain_idx,domains,columns,resnum)
                        bad_residues.append(((1-avg_percent_aligned),domain_idx,resnum))

            bad_residues.sort()
            bad_residues = filterBadResidues(bad_residues,domains)
            last_iteration = 0
        else:
            #collect marked residues in a list for convenience
            for domain_idx in range(len(domains)):
                domain = domains[domain_idx]
                resnums = domain.residues.values()
                for resnum in resnums:
                    if original_marked_residues[domain_idx].has_key(resnum):
                        idx = domain.residue_indices[resnum]
                        for c in range(len(columns)):
                            if columns[c][domain_idx]==idx:
                                count = 0
                                for d in range(len(domains)):
                                    if d!=domain_idx and columns[c][d]==-1: count += 1
                                if count==len(domains)-1:

                                    avg_percent_aligned = getSurroundingOccupancies(domain_idx,domains,columns,resnum)
                                    bad_residues.append(((1-avg_percent_aligned),domain_idx,resnum))
                                    break

            bad_residues.sort()
            bad_residues = filterBadResidues(bad_residues,domains)
            last_iteration = 1
        
        bad_domains = []
        count=0
        for p,domain_idx,resnum in bad_residues: #for each marked residue
            numcols = len(columns)
            domain = domains[domain_idx]
            bad_resindex = domain.residue_indices[resnum] #the marked residue

            region_start,region_end = domain.getSSEEnds(resnum=resnum)
            if region_start!=-1: #if is a secondary structure element
                region_start = domain.residue_indices[region_start]
                region_end = domain.residue_indices[region_end]                
                    
                resindex_column,col_start,col_end = getColumnRegion(columns,domain_idx,region_start,region_end,domains,alignments,region_start,region_end)
                first_sse_col = col_start
                last_sse_col = col_end

                #Try to expand regions up to five residues in each direction. Stop
                #if hit SSE residues
                i = 1
                while i <=5 :
                    if col_start-i<0:
                        i-=1
                        break
                    if columns[col_start-i][domain_idx]!=-1:
                        if domain.sstype[domain.residues[columns[col_start-i][domain_idx]]]!='U':
                            break
                    i += 1
                col_start -= (i-1)
                i = 1

                while i<= 5:
                    if col_end+i>=len(columns):
                        i-=1
                        break
                    if columns[col_end+i][domain_idx]!=-1:
                        if domain.sstype[domain.residues[columns[col_end+i][domain_idx]]]!='U':   
                            break
                    i += 1
                col_end += i-1
                if col_start < 0: col_start=0
                if col_end >= numcols: col_end = numcols-1
                loop = 0

            else: #residue marked for realignment is a loop
                resindex_column,region_start,region_end,col_start,col_end = getLoopRegion(domain_idx,domains[domain_idx],columns,bad_resindex,domains)
                loop = 1

            #get other domains that have marked residues in this region
            bad_domains = getBadDomainsInRegion(columns,domains,marked_residues,col_start,col_end)


            changes = 0
            debug_dp = 0

            #Run dynamic programming to realign region
            columns,changes=runDynamicProgramming(columns,domains,region_start,region_end,col_start,col_end,domain_idx,bad_resindex,alignments,bad_domains,original_marked_residues,debug_dp,sse_penalties,loop,last_iteration)

            if changes==-1: continue #no residues in region (error)
            num_changes += changes
            if marked_residues[domain_idx].has_key(resnum): #residue is fixed now
                del marked_residues[domain_idx][resnum]
            count += 1

        if (num_changes == 0 or iterations == maxiterations-1) and not do_last_iteration:
            do_last_iteration = 1
            iterations=maxiterations
            done=1
        elif do_last_iteration:
            do_last_iteration = 0
            break
            done=1

        iterations += 1

    #restore original list of marked residues
    marked_residues = deepcopy(original_marked_residues);
    return columns


#Prints out multiple alignment in MCL format.
def printMCLFormat(domains,nodes,columns,filename):
    numresidues=0
    for i in range(len(domains)):
        numresidues+=domains[i].num_residues

    outfile=open(filename,"w")
    outfile.write("(mclheader\nmcltype matrix\ndimensions %sx%s\n)\n\n" % (numresidues,len(columns)))
    outfile.write("(mclmatrix\nbegin\n")
    for i in range(len(columns)):
        column=columns[i]
        outfile.write("%i\t" % i)
        for j in range(len(domains)):
            if column[j]!= -1:
                outfile.write("%s " % (domains[j].getNodeNumber(index=column[j])))
        outfile.write("$\n")
    outfile.write(")\n");
    outfile.close()
            
#Clusters structures based on the LOCK 2 global alignment score. Uses average linkage.
def clusterStructures(domains,alignments):
    activeClusters = []
    nearestNeighborList = []
    nearestNeighborValueList = []
    clusters = []
    matrix = []
    joins = []
    internal_nodes = []
    
    numElements = len(domains)

    #initialize matrix from scores dictionary
    for i in range(numElements):
        matrix.append([])
        for j in range(numElements):
            matrix[i].append(0)

    for i in range(numElements):
        for j in range(i,numElements):
            
            matrix[i][j] = alignments.getAlignmentScore(i,j)
            matrix[j][i] = matrix[i][j]


    #initialize clusters array
    for i in range(numElements):
        clusters.append({})
        clusters[i]['id'] = i;
        clusters[i]['size']=1;
        clusters[i]['value']=1;
        clusters[i]['leaves'] = [i];

    #each element is considered a cluster initially
    for i in range(numElements):
        activeClusters.append(i)
        nearestNeighborList.append(-1) #initialization
        nearestNeighborValueList.append(-1) #initialization

    numActiveClusters = numElements

    for i in range(numElements):
        maxSimilarity = -10000
        for j in range(numElements):
            if i==j: continue
            if(matrix[i][j] > maxSimilarity):
                maxSimilarity = matrix[i][j]
                nearestNeighborList[i] = j
                nearestNeighborValueList[i] = maxSimilarity
    c = numElements
    #there are (numElements - 1) mergers
    for i in range(numElements-1):
        maxSimilarity = -10000

        #initialize c1
        for j in range(numElements-i):
            if nearestNeighborValueList[activeClusters[j]] > maxSimilarity:
                maxSimilarity = nearestNeighborValueList[activeClusters[j]]
                c1 = activeClusters[j]

        
        #c2 is c1's nearest neighbor
        c2 = nearestNeighborList[c1]

        #update the other active clusters' values from the new cluster
        for j in range(numElements-i):
            if(activeClusters[j] != c1 and activeClusters[j] != c2):
                matrix[activeClusters[j]][c1] = (clusters[c1]['size']*matrix[activeClusters[j]][c1]+clusters[c2]['size']*matrix[activeClusters[j]][c2])/float(clusters[c1]['size']+clusters[c2]['size'])
                matrix[c1][activeClusters[j]] = matrix[activeClusters[j]][c1]
        #merge cluster c1 and cluster c2

        joins.append((c,clusters[c1]['id'],clusters[c2]['id'],clusters[c1]['leaves'][:],clusters[c2]['leaves'][:],1-maxSimilarity,(1-maxSimilarity)-(1-clusters[c1]['value']),(1-maxSimilarity)-(1-clusters[c2]['value'])))
        clusters[c1]['id'] = c; c += 1
        clusters[c1]['size'] += clusters[c2]['size']
        clusters[c1]['value'] = maxSimilarity
        for leaf in clusters[c2]['leaves']:
            clusters[c1]['leaves'].append(leaf)



        #overwrite cluster c2 in the list of active clusters
        found = 0
        for j in range(numActiveClusters-1):
            if(activeClusters[j]==c2):
                del(activeClusters[j])
                break
        numActiveClusters-=1

        #update nearestNeighborList and nearestNeighborValueList for the new cluster
        maxSimilarity = -10000
        for j in range(numActiveClusters):
            if(activeClusters[j] != c1):
                if(matrix[c1][activeClusters[j]] > maxSimilarity):
                    maxSimilarity = matrix[c1][activeClusters[j]]
                    nearestNeighborList[c1] = activeClusters[j]
                    nearestNeighborValueList[c1] = maxSimilarity

                    
        #update nearestNeighborList and nearestNeighborValueList for the other clusters
        for j in range(numActiveClusters):
            if(activeClusters[j] != c1):
                if(nearestNeighborList[activeClusters[j]]==c1 or nearestNeighborList[activeClusters[j]]==c2):
                    maxSimilarity = -1000
                for k in range(numActiveClusters):
                    if(activeClusters[k] != activeClusters[j]):
                        if(matrix[activeClusters[j]][activeClusters[k]] > maxSimilarity):
                            maxSimilarity = matrix[activeClusters[j]][activeClusters[k]]
                            nearestNeighborList[activeClusters[j]] = activeClusters[k]
                            nearestNeighborValueList[activeClusters[j]] = maxSimilarity;

    return joins

#Not used
def findDuplicateSSENodes(domains,clusters,sse_nodes,alignments,duplicates):
    eliminated = {}
    for i in range(len(clusters)):
        for j in range(len(domains)):
            if duplicates[i].has_key(j): #domains[j] has duplicate(s)
                node_nums=duplicates[i][j]
                max=maxnode=-1
                cluster_nodes=clusters[i]
                for node_num in node_nums:
                    node = sse_nodes.getNodeByLabel(node_num)
                    domain_idx = node.domain_idx
                    sse_start = node.resnum
                    average_sse_score = 0
                    for node_num2 in cluster_nodes:
                        node2 = sse_nodes.getNodeByLabel(node_num2)
                        domain2_idx = node2.domain_idx
                        sse_start2 = node2.resnum
                        if domain2_idx==domain_idx: continue
                        if domain_idx <= domain2_idx:
                            score = alignments.alignments[domain_idx][domain2_idx].getSSEScore(sse_start)
                        else:
                            score = alignments.alignments[domain2_idx][domain_idx].getSSEScore(sse_start2)                             
                        if score != 0:
                            if domain_idx<=domain2_idx:
                                if alignments.alignments[domain_idx][domain2_idx].getAlignedSSE(sse_start)==sse_start2:
                                    average_sse_score += score
                            else:
                                if alignments.alignments[domain2_idx][domain_idx].getAlignedSSE(sse_start2)==sse_start:
                                    average_sse_score += score                                
                    if average_sse_score > max:
                        max=average_sse_score; maxnode_num=node_num
                for node_num in node_nums:
                    if node_num != maxnode_num:
                        eliminated[node_num]=1

    return eliminated

#Not used
def findBestNumbering(clusternums,keep_duplicates):
    scorearray = []; pointarray = []
    maxscore = -1
    maxindex = -1 
    for i in range(len(clusternums)):
        scorearray.append(-1); pointarray.append(-1)
        if clusternums[i] != -1 and clusternums[i] != None:
            myscore = 1; mypointer = -1
            for j in range(0,i-1):
                idx = (i-1)-j
                if keep_duplicates:
                    if clusternums[idx] != -1 and clusternums[idx] <= clusternums[i]:
                        if scorearray[idx]+1 > myscore:
                            myscore = scorearray[idx]+1
                            mypointer = idx
                else:
                    if clusternums[idx] != -1 and clusternums[idx] < clusternums[i]:
                        if scorearray[idx]+1 > myscore:
                            myscore = scorearray[idx]+1
                            mypointer = idx
            scorearray[i]=myscore; pointarray[i] = mypointer
            if myscore > maxscore:
                maxscore = myscore
                maxindex = i
    while maxindex != -1:
        scorearray[maxindex] = 0
        maxindex = pointarray[maxindex]
    for i in range(len(clusternums)):
        if scorearray[i] != 0:
            clusternums[i] = None
    return clusternums

#Not used.
def getSSEConservations(domains,weights,clusters,cluster_membership,eliminated_sse_nodes,alignments,sse_nodes,weighted=0):

    #first, build up SSE alignment from clusters: follow format from residue stuff
    #next, take weighted average of scores (need to deal with first score having lower max)
    #then, return conservations so that a rasmol script can be produced that weights MSA and core structure by conservation


    #find the best seed sequence
    max = -1
    numdomains = len(domains)
    alignments_to_seeds = []
    for idx1 in range(numdomains):
        alignments_to_seeds.append([])
        for idx2 in range(numdomains):
            alignments_to_seeds[idx1].append([])

    is_aligned = []
    for domain_idx in range(numdomains):
        avg=0
        nextnum = 0; labels = {}; in_order = 0
        sses = cluster_membership[domain_idx].keys()
        sses.sort()
        num_seed_sses = len(sses)
        is_aligned.append({})
        #give numbers to clusters
        index_to_clusternum = []
        for sse in sses:
            clusternum = cluster_membership[domain_idx][sse]
            labels[clusternum] = nextnum
            index_to_clusternum.append(clusternum)
            nextnum += 1


        #now check other domains against this numbering
        for domain2_idx in range(numdomains):
            if domain_idx == domain2_idx: continue
            sses2 = cluster_membership[domain2_idx].keys()
            sses2.sort()
            num_unlabeled = -1
            clusternums = []; alignment = []
            num_sses = len(sses2)
            eliminated = []
            for i in range(len(sses)):
                found = 0

                for j in range(len(sses2)):
                    clusternum = cluster_membership[domain2_idx][sses2[j]]
                    if clusternum != -1:
                        if clusternum == index_to_clusternum[i]:
                            clusternums.append(labels[clusternum])
                            alignment.append(sses2[j])
                            found = 1
                if found==0:
                    alignment.append(None); num_unlabeled += 1
                num_labeled = len(clusternums)
            clusternums = findBestNumbering(clusternums,0)
            alignment = findBestNumbering(alignment,1)
            alignments_to_seeds[domain2_idx][domain_idx] = alignment[:]

            num_labeled = 0; num_mislabeled = -1
            for i in range(len(clusternums)):
                if clusternums[i] != -1: num_labeled += 1
                else: num_mislabeled += 1
            avg += num_labeled/float(len(sses2)) 
        avg /= float(numdomains)
        if avg > max:
            max = avg
            seed = domain_idx
            
    #now the seed structure is stored in the variable seed, and alignments
    sses = cluster_membership[seed].keys()
    labels = {}
    cluster_order = []
    nextnum = 0
    for sse in sses:
        clusternum = cluster_membership[seed][str(sse)]
        labels[clusternum] = nextnum
        cluster_order.append(clusternum)
        nextnum += 1

    num_columns = nextnum

    print "seed:",seed,domains[seed]
    aligned_sses = []

    for i in range(num_columns):
        aligned_sses.append([])
        aligned_sses[i].append((seed,sses[i],domains[seed].getSSEEnd(resnum=sses[i])))
        for j in range(numdomains):
            if j == seed: continue
            sses2 = cluster_membership[j].keys()
            alignment = alignments_to_seeds[j][seed]
            if alignment[i] != None:
                aligned_sses[i].append((j,alignment[i],domains[j].getSSEEnd(resnum=alignment[i])))
                is_aligned[j][alignment[i]] = 1

                
        
    sse_conservations = {} #domain then cluster index
    for i in range(numdomains):
        sse_conservations[i] = {}
        for j in range(len(clusters)):
            sse_conservations[i][j] = 0
        
    #now find conservations
    cluster_sse_defs = []
    for i in range(len(clusters)): cluster_sse_defs.append([])
    
    for i in range(len(clusters)):
        cluster = clusters[i]
        for idx1 in range(len(cluster)):
            node = sse_nodes.getNodeByLabel(cluster[idx1])
            domain1_idx = node.domain_idx; sse_start1 = node.resnum
            cluster_sse_defs[i].append((domain1_idx,sse_start1,domains[domain1_idx].getSSEEnd(resnum=sse_start1)))
            conservation = 0.0
            total = 0.0
            for idx2 in range(len(cluster)):
                if idx1 == idx2: continue
                node2 = sse_nodes.getNodeByLabel(cluster[idx2])
                domain2_idx = node2.domain_idx; sse_start2 = node2.resnum                
                if domain1_idx==domain2_idx: continue
                if eliminated_sse_nodes.has_key(cluster[idx2]): 
                    continue
                if domain1_idx<=domain2_idx:
                    if not alignments.alignments[domain1_idx][domain2_idx].alignedSSEs.has_key(sse_start1):
                        continue
                else:
                    if not alignments.alignments[domain2_idx][domain1_idx].alignedSSEs.has_key(sse_start2): continue

                

                total += 40.0
                if domain1_idx<=domain2_idx:
                    conservation += alignments.getSSEAlignmentScore(domain1_idx,domain2_idx,sse_start1)
                else:
                    conservation += alignments.getSSEAlignmentScore(domain2_idx,domain1_idx,sse_start2)                    

            if total!=0:
                conservation /= float(total) #normalize to max possible total score
            else:
                conservation = 0
            sse_conservations[domain1_idx][i] = conservation


    cluster_conservations = []; total_weights = []
    for i in range(len(clusters)):
        cluster_conservations.append(0); total_weights.append(0)
        
    for i in range(len(domains)):
        clusters_with_domain = sse_conservations[i].keys()
        average = 0; total_weight = 0
        for cluster_idx in clusters_with_domain:
            if weighted:
                cluster_conservations[cluster_idx] += weights[i]*sse_conservations[i][cluster_idx]
                total_weights[cluster_idx] += weights[i]
            else:
                cluster_conservations[cluster_idx] += sse_conservations[i][cluster_idx]
                total_weights[cluster_idx] += 1

    for i in range(len(clusters)):
        cluster_conservations[i] /= total_weights[i]


    return cluster_conservations,cluster_sse_defs

def getColorByConservation(c):
    if c < 0.5: return (26,1,4)
    elif c < 0.1: return 28,1,5
    elif c < 0.15: return 31,2,5
    elif c < 0.2: return 36,2,6
    elif c < 0.25: return 38,2,6
    elif c < 0.3: return 43,2,7
    elif c < 0.35: return 48,2,8
    elif c < 0.4: return 56,3,9
    elif c < 0.45: return 61,3,10
    elif c < 0.5: return 69,3,11
    elif c < 0.55: return 77,4,12
    elif c < 0.6: return 84,4,14
    elif c < 0.65: return 94,5,15
    elif c < 0.7: return 102,5,16
    elif c < 0.75: return 117,6,19
    elif c < 0.8: return 133,7,21
    elif c < 0.85: return 148,7,24
    elif c < 0.9: return 184,9,30
    else: return 204,10,33

#Not used
def writeRasmolScript(filename,sse_conservations,cluster_sse_defs,domains,chains):
    outfile = open(filename,"w")
    outfile.write("background white\n")
    outfile.write("wireframe off\n")
    outfile.write("backbone 15\n")
    outfile.write("color gray\n")


    for i in range(len(cluster_sse_defs)):
        conservation = sse_conservations[i]
        r,g,b = getColorByConservation(conservation)
        sse_defs = cluster_sse_defs[i]
        for idx in range(len(sse_defs)):
            domain_idx, sse_start,sse_end = sse_defs[idx][0],sse_defs[idx][1],sse_defs[idx][2]

            chain = chains[domain_idx]
            if re.match("[0-9]+[A-Z]$",sse_start): sse_start = sse_start[:-1]
            if re.match("[0-9]+[A-Z]$",sse_end): sse_end = sse_end[:-1]      
            outfile.write("select (%s-%s%s and not hetero)\n" % (sse_start,sse_end,chain))
            outfile.write("color [%i,%i,%i]\n" % (r,g,b))
        outfile.write("\n")
    outfile.close()

def convert1LetterTo3Letter(code):
    if code=="A": return "ALA"
    elif code=="R": return "ARG"
    elif code=="N": return "ASN"
    elif code=="D": return "ASP"
    elif code=="C": return "CYS"
    elif code=="Q": return "GLN"
    elif code=="E": return "GLU"
    elif code=="G": return "GLY"
    elif code=="H": return "HIS"
    elif code=="I": return "ILE"
    elif code=="L": return "LEU"
    elif code=="K": return "LYS"
    elif code=="M": return "MET"
    elif code=="F": return "PHE"
    elif code=="P": return "PRO"
    elif code=="S": return "SER"
    elif code=="T": return "THR"
    elif code=="W": return "TRP"
    elif code=="Y": return "TYR"
    elif code=="V": return "VAL"
    else: return "UNK" #amino acid not identified

#Writes out HELIX and SHEET lines for pdb file
def writeSSEsToPDBFile(domain,chain,outfile):
    #print out sse info
    for i in range(domain.num_SSEs):
        start,end = domain.getSSEEnds(ssenum=i)
        sstype = domain.sstype[start];
        start = start[:-2]; end = end[:-2];
        if sstype=="H":
            line = "HELIX "+" "*13+chain+" "
            line += "%4s      %s %4s\n" % (start,chain,end)
        else:
            line = "SHEET "+" "*15+chain
            line += "%4s      %s%4s\n" % (start,chain,end)
        outfile.write(line)
    



#Prints out domain's coordinates to outfile using the chain identifier passed to the function.
def writePDBChain(domain,coordinates,chain,outfile):
    
    count = 1
    for i in range(domain.num_residues):
        atoms = domain.residue_atomnames[i]
        if len(atoms)>0:
            if coordinates[i][atoms[0]]==None: continue
        for atom in atoms:
            if not (atom=="CA"  or atom=="CB" or atom=="C" or atom=="N" or atom=="O"): continue
##            #next line produces first 22 characters of a pdb atom record line
            line = "ATOM  " 
            line += "%5s " % count
            line += " %-4s%s %s" % (atom,convert1LetterTo3Letter(domain.aas[i]),chain)
            line += "%4s%s" % (domain.residues[i][:-2],domain.residues[i][-2])

            line += "   %8.3f%8.3f%8.3f\n" % (coordinates[i][atom][0],coordinates[i][atom][1],coordinates[i][atom][2])
            outfile.write(line)
            count += 1


#Applies rotation and translation to a domain using the coordinates passed in.
def applyTransformation(domain,rotation,translation,coordinates):
    new_coordinates = []
    atomnames = domain.residue_atomnames
    num_residues = domain.num_residues
    for i in range(num_residues):
        atoms = atomnames[i]
        new_coordinates.append({})
        for atom in atoms:
            new_coordinates[i][atom] = translation + (dot(coordinates[i][atom],rotation))
    return new_coordinates



#Transforms all structures into the same frame of reference (CBA step 2)
def transformStructures(domains,alignments,joins,transformations,inverse_transformations):
    operations_applied = []
    newly_formed_operations = []
    numdomains = len(domains)
    transformed = {}


    #initialize operations_applied array
    for index in range(numdomains):
        operations_applied.append([])
    for i in range(len(domains)):
        transformed[i] = 0
    join_count = 0
    for join in joins:
        join_count += 1
        left_child = join[1]; right_child = join[2]; left_leaves = join[3]; right_leaves = join[4]

        #select query and target for transformation: compare each left leaf to each right leaf to get greatest similarity
        maxSimilarity = -1000
        l1 = -1; l2 = -1
        for leaf1 in left_leaves:
            for leaf2 in right_leaves:
                if leaf1<=leaf2:
                    score = alignments.getAlignmentScore(leaf1,leaf2)
                else:
                    score = alignments.getAlignmentScore(leaf2,leaf1)                    
                if score > maxSimilarity:
                    maxSimilarity = score
                    l1 = leaf1; l2 = leaf2

        if operations_applied[l1]: #if any transformations have been applied to leaf l1, need to undo them
            inverse_rotation = identity(3) #initialization
            inverse_translation = array([0.0,0.0,0.0])
            inverse_translation_component2 = array([0.0,0.0,0.0])
            forward_translation = array([0.0,0.0,0.0])
            forward_rotation = identity(3)
            for i in range(len(operations_applied[l1])):
                op = operations_applied[l1][i]
                forward_rotation = dot(forward_rotation,op['forward'][0])
                forward_translation = dot(forward_translation,op['forward'][0])
                forward_translation += op['forward'][1]


            
            new_operation = {}

            new_operation['forward'] = (forward_rotation.copy(),forward_translation.copy())
            new_operation_inverse = {}
            new_operation_inverse['forward'] = (inverse(forward_rotation).copy(),-1*dot(forward_translation,inverse(forward_rotation)).copy());
            newly_formed_operations.append(new_operation_inverse)
            newly_formed_operations.append(new_operation)
            newly_formed_operations[-2]['inverse'] = newly_formed_operations[-1]
            newly_formed_operations[-1]['inverse'] = newly_formed_operations[-2]


            for leaf in left_leaves: #Apply the inverse of the operations applied to l1 so far to all leaves
                if leaf==l1:
                    operations_applied[leaf].append(newly_formed_operations[-2])
                else:
                    operations_applied[leaf].append(newly_formed_operations[-2])

        if operations_applied[l2]: #Do the same thing for the right leaves
            inverse_rotation = identity(3)
            inverse_translation = array([0.0,0.0,0.0])
            forward_rotation = identity(3)
            forward_translation = array([0.0,0.0,0.0])
            for i in range(len(operations_applied[l2])):
                op = operations_applied[l2][i]
                forward_rotation = dot(forward_rotation,op['forward'][0])
                forward_translation = dot(forward_translation,op['forward'][0])
                forward_translation += op['forward'][1]


            new_operation = {}
            new_operation['forward'] = (forward_rotation.copy(),forward_translation.copy())
            new_operation_inverse = {}
            new_operation_inverse['forward'] = (inverse(forward_rotation)[:],-1*dot(forward_translation,inverse(forward_rotation))[:]);
            newly_formed_operations.append(new_operation_inverse)
            newly_formed_operations.append(new_operation)

            newly_formed_operations[-2]['inverse'] = newly_formed_operations[-1]
            newly_formed_operations[-1]['inverse'] = newly_formed_operations[-2]
            for leaf in right_leaves:
                if leaf==l2:
                    operations_applied[leaf] = [] #inverse would just undo everything
                else:
                    operations_applied[leaf].append(newly_formed_operations[-2])

        #transform one set of leaves onto the other
        if l1 < l2:
            for leaf in right_leaves:
                transformed[leaf] = 1                
                operations_applied[leaf].append(transformations[l2][l1])
        else:
            for leaf in left_leaves:
                transformed[leaf] = 1
                operations_applied[leaf].append(transformations[l1][l2])

        transformed[l1] = 1; transformed[l2] = 1

    #apply transformations
    new_coordinates = {}
    coordinates = []
    for i in range(len(domains)):
        coordinates.append(deepcopy(domains[i].coordinates))

    for i in range(len(operations_applied)):
        leaf = i
        if not transformed[leaf]: continue
        rotation = identity(3); translation = zeros([3],Float)        
        if operations_applied[leaf]:
            for operation in operations_applied[leaf]:
                rotation = dot(rotation,operation['forward'][0])
                translation = dot(translation,operation['forward'][0])
                translation += operation['forward'][1]
            coordinates[leaf] = applyTransformation(domains[leaf],rotation,translation,coordinates[leaf])
            domains[leaf].setReferenceFrame(rotation,translation)


    return coordinates




                            

#Runs the MCL clustering algorithm
def runMCL(domains,alignments,alignmentdir):
    nodes = nodeList();
    label = 0
    labels = [{} for i in range(len(domains))]
    for i in range(len(domains)): #Give a numeric label to each residue of each structure
        domain = domains[i]
        for j in range(domain.num_residues):
            labels[i][domain.residues[j]] = label
            nodes.addNode(label,i,domain.residues[j])
            label += 1
    num_residues = label
            
    try:
        outfile = open("mclmatrix","w")
    except Exception, e:
        printError("Error: Couldn't open matrix file for writing")

    outfile.write("(mclheader\n")
    outfile.write("mcltype matrix\n")
    outfile.write("dimensions %sx%s\n" % (num_residues,num_residues))
    outfile.write(")\n\n")
    outfile.write("(mclmatrix\nbegin\n")
    
    for i in range(len(domains)):
        domain = domains[i]
        for resnum in domain.residue_indices.keys():
            outfile.write("%s" % labels[i][resnum])
            for j in range(len(domains)):
                if i==j: continue
                if alignments.alignments[i][j].alignedResidues.has_key(resnum):
                    resnum2 = alignments.alignments[i][j].alignedResidues[resnum]
                    weight = alignments.alignments[i][j].alignmentscores[resnum]
                    if weight>0:
                        weight/=float(20)
                    else:
                        weight = 0.1
                    outfile.write(" %i:%.3f" % (labels[j][resnum2],weight))
            outfile.write(" $\n")
    outfile.write(")\n")
    outfile.close()
            
    #run mcl
    mclresults_filename = "mclresults"
    try:
        os.system("%s %s -I 2 -o %s" % (mclprogram,"mclmatrix",mclresults_filename))
    except:
        printError("Error: MCL failed.")

    return mclresults_filename,nodes


#Runs steps 3, 4, and 5 of CBA.
def clusterAndFixSSEs(domains,alignments,alignmentdir,domain_names_parsed,scores):
    mcl_results_filename,nodes = runMCL(domains,alignments,alignmentdir)
    fixed_alignmentdir = "fixed-alignments"
    clusters,cluster_membership,duplicates = readClusters(mcl_results_filename,nodes)
    print "CBA: Ran clustering -> step 3 complete."

    
    processID = int(os.getpid())
    while os.path.exists("/tmp/%i" % processID): processID += 1
    dummy = fixSSEs.fixSSEs(domains,alignments,clusters,nodes,alignmentdir,mcl_results_filename,fixed_alignmentdir,lock_fixSSEs_program,processID,lock_gapMultiplier,lock_gap_extend_penalty,distanceThreshold)
    print "CBA: Fixed secondary structure element alignments -> step 4 complete."
    del alignments
    alignments = alignmentSet("fixed-alignments",domain_names_parsed,scores)

    mcl_results_filename,nodes = runMCL(domains,alignments,fixed_alignmentdir)
    clusters,cluster_membership,duplicates = readClusters(mcl_results_filename,nodes)
    print "CBA: Ran clustering -> step 5 complete."
    
    return nodes,clusters,cluster_membership,duplicates


#Gets new residue registration for pairwise alignments after they've been transformed
#into the same frame of reference (CBA step 2)
def runResidueAlignments(domains,alignments,old_alignmentdir):
    alignmentdir = "superimposed-alignments"
    if not os.path.exists(alignmentdir):
        try:
            os.mkdir(alignmentdir)
        except Exception,e:
            printError("Error: Could not make superimposed alignments directory.")
    

    for d1 in range(len(domains)):
        query = domains[d1]
        if query.chain!="all" and query.chain!=" ":
            query_chain_opt = "-qc %s" % query.chain
        else: query_chain_opt = ""
        querypath = os.path.join(query.directory,"%s%s" % (query.domainname,query.extension))
        for d2 in range(d1,len(domains)):
            target = domains[d2]
            if target.chain!="all" and target.chain!=" ":
                target_chain_opt = "-tc %s" % target.chain
            else:
                target_chain_opt = ""
            targetpath = os.path.join(target.directory,"%s%s" % (target.domainname,target.extension))                
            query_rotation_inverse = inverse(query.rotation)
            new_target_translation = dot(target.translation-query.translation,query_rotation_inverse)
            new_target_rotation = dot(target.rotation,query_rotation_inverse)

            matrix_arg = "-transformation"
            for i in range(3):
                matrix_arg += " %.5f %.5f %.5f" % (new_target_rotation[i][0],new_target_rotation[i][1],new_target_rotation[i][2])
            matrix_arg += " %.5f %.5f %.5f" % (new_target_translation[0],new_target_translation[1],new_target_translation[2])
            try:
                os.system("%s -q %s -t %s -o %s %s %s %s -open %s -extend %s -d %s> /dev/null" % (lockprogram,querypath,targetpath,alignmentdir,query_chain_opt,target_chain_opt,matrix_arg,lock_gapMultiplier,lock_gap_extend_penalty,distanceThreshold))
                if os.path.exists("%s/query_target.out" % alignmentdir):
                    os.system("/bin/mv %s/query_target.out %s/%s_%s.out" % (alignmentdir,alignmentdir,query.domainname,target.domainname))
            except:
                print "Error: LOCK 2 run failed"
                sys.exit(-1)


def gzipAlignments(domains,alignmentdir):
    for d1 in range(len(domains)):
        for d2 in range(d1,len(domains)):
            os.system("/bin/gzip -f %s/%s_%s.out" % (alignmentdir,domains[d1].domainname,domains[d2].domainname))




def printUsage():
    print "Usage: CBA.py <path to file containing list of structure files> [-o <gap opening penalty> -e <gap extension penalty> -lo <LOCK 2 gap opening penalty> -le <LOCK 2 gap extension penalty>] -d <LOCK 2 distance threshold>]"


#Gets the LOCK 2 alignment score from an alignment file
def getScoreFromFile(filename):
    if os.path.exists(filename):
        infile = open(filename,"r")
        num1=0;num2=0;found=0
        line=infile.readline()
        while line:
            if re.match("final score",line): found=1; break
            elif re.match("Num Pre-Core Aligned = 0",line): found=0;break
            line=infile.readline()
        if found:
            score = float(line.split()[-1])
            while line:
                if re.match("query:",line): break
                line=infile.readline()
            line=infile.readline()

            while line:
                if re.match("target:",line): break
                num1 += 1
                line=infile.readline()
            line=infile.readline()
            while line:
                if re.match("CPU",line): break
                num2 += 1
                line=infile.readline()
        else:
            infile.close()
            return -1
        infile.close()
        if num1>=num2: #normalization
            return score/float(40*num1-20)
        else:
            return score/float(40*num2-20)
    else: return -1


#Runs the pairwise LOCK 2 alignments (CBA step 1)
def runPairwiseAlignments(domains,alignmentdir):
    '''Runs the LOCK 2 pairwise alignments that are the starting point for CBA. Alignments will
    be placed in the directory specified by the user (the optional second argument to CBA) or
    otherwise in the default directory (alignments). If necessary, the alignment directory will
    be created, but only if its parent directory exists. Alignments are named <query>_<target>.out,
    where <query> and <target> are the names of the aligned domains. Names include chain identifiers,
    if specified.'''
    
    if not os.path.exists(alignmentdir):
        try:
            os.system("mkdir %s" % alignmentdir)
        except:
            print "Alignment directory (%s) does not exist and could not be created (perhaps the parent directory doesn't exist)" % alignmentdir
            sys.exit(-1)

    scores = [[0 for y in domains] for x in domains]

    for d1 in range(len(domains)):
        query = domains[d1]
        if query.chain!="all" and query.chain!=" ":
            query_chain_opt = "-qc %s" % query.chain
        else: query_chain_opt = ""
        querypath = os.path.join(query.directory,"%s%s" % (query.domainname,query.extension))
        print "CBA: running LOCK 2 pairwise alignments for query %s" % (query.domainname)

        for d2 in range(d1,len(domains)):
            target = domains[d2]
            if target.chain!="all" and target.chain!=" ":
                target_chain_opt = "-tc %s" % target.chain
            else: target_chain_opt = ""
            targetpath = os.path.join(target.directory,"%s%s" % (target.domainname,target.extension))                
            try:
                os.system("%s -q %s -t %s -o %s %s %s -open %.3f -extend %.3f -d %.3f > /dev/null" % (lockprogram,querypath,targetpath,alignmentdir,query_chain_opt,target_chain_opt,lock_gapMultiplier,lock_gap_extend_penalty,distanceThreshold))
                if os.path.exists("%s/query_target.out" % alignmentdir):
                    os.system("/bin/mv -f %s/query_target.out %s/%s_%s.out" % (alignmentdir,alignmentdir,query.domainname,target.domainname))
                    score = getScoreFromFile("%s/%s_%s.out" % (alignmentdir,query.domainname,target.domainname))
                    scores[d1][d2] = score
                    scores[d2][d1] = score
                else:
                    scores[d1][d2] = 0; scores[d2][d1] = 0

            except:
                print "Could not run LOCK 2 alignment for %s vs %s . . . exiting" % (query.domainname,target.domainname)
                sys.exit(-1)
 
    return scores

 
#Not used.
def runSSEClustering(domains,alignments,alignmentdir):
    nodes = nodeList()
    try:
        outfile = open(os.path.join(alignmentdir,"mclmatrix"),"w")
    except Exception, e:
        printError("Error: Couldn't open matrix file for writing")

    labels = [{} for i in range(len(domains))]
    label = 0
    for i in range(len(domains)):
        my_sses = domains[i].sse_ends
        for sse_start,sse_end in my_sses:
            labels[i][sse_start] = label
            nodes.addNode(label,i,sse_start)
            label += 1
    total_sses = label

    outfile.write("(mclheader\n")
    outfile.write("mcltype matrix\n")
    outfile.write("dimensions %sx%s\n" % (total_sses,total_sses))
    outfile.write(")\n\n")
    outfile.write("(mclmatrix\nbegin\n")
    for i in range(len(domains)):
        mysses = domains[i].sse_ends
        for k in range(len(mysses)):
            outfile.write("%i " % labels[i][mysses[k][0]])
            for j in range(len(domains)):
                if i==j: continue
                if alignments.alignments[i][j].alignedSSEs.has_key(mysses[k][0]):
                    sse2 = alignments.alignments[i][j].alignedSSEs[mysses[k][0]]
                    outfile.write("%i " % labels[j][sse2])
            outfile.write("$\n")
    outfile.write(")\n")
    outfile.close()

    mclresults_filename = os.path.join(alignmentdir,"sse-mcl-results")

    try:
        os.system("%s %s -I 2 -o %s" % (mcl,os.path.join(alignmentdir,"mclmatrix"),mclresults_filename))
    except:
        print "Error: MCL failed"
        sys.exit(-1)

    clusters,cluster_membership,duplicates = readClusters(mclresults_filename)
    if clusters==None:
        printError("Error: SSE clustering failed")

    eliminated_sse_nodes = findDuplicateSSENodes(domains,clusters,nodes,alignments,duplicates)            
    return clusters,cluster_membership,duplicates,eliminated_sse_nodes


#Prints out a PDB file for the multiple structure superposition. Only works for 36 chains.
#To print a PDB file with more chains, need to use models, but this is inconvenient for rasmol
def writeAlignmentPDBFile(domains,coordinates,pdbfile):
    chains=["A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P","Q","R","S","T","U","V","W","X","Y","Z","0","1","2","3","4","5","6","7","8","9"]

    if(len(domains)>36):
	print "Problem: Can only print 36 chains to PDB file. Will reuse chain identifiers"
    for i in range(len(domains)):
	chain_idx = i%len(chains)
#	if i>=36: break
        pdbfile.write("HEADER    %s: Chain %s\n" % (domains[i].domainname,chains[chain_idx])) 

    for i in range(len(domains)):
	chain_idx = i%len(chains)
#	if i>=36: break
        writeSSEsToPDBFile(domains[i],chains[chain_idx],pdbfile)
    for i in range(len(domains)):
	chain_idx = i%len(chains)
#	if i>=36: break
        writePDBChain(domains[i],coordinates[i],chains[chain_idx],pdbfile)



def parseOptionalArguments(arguments):
    lock_gap_open_penalty = 3;
    lock_gap_extend_penalty = 0.9;
    gap_open_penalty = 6;
    gap_extend_penalty = 0.5;
    lock_distance_threshold = 4.0;
    
    numargs = len(arguments)
    for i in range(len(arguments)):
        if arguments[i] == "-o" or arguments[i]=="-open":
            if i<numargs:
                gap_open_penalty = arguments[i+1]
                try:
                    gap_open_penalty = float(gap_open_penalty)
                except:
                    print "Error: Expected a numerical value for the gap opening penalty (%s)" % gap_open_penalty
                    sys.exit(-1)
            else:
                print "Error: Expected a gap opening penalty"
                sys.exit(-1)
        elif arguments[i]=="-e" or arguments[i]=="-extend":
            if i<numargs:
                gap_extend_penalty = arguments[i+1]
                try:
                    gap_extend_penalty = float(gap_extend_penalty)
                except:
                    print "Error: Expected a numerical value for the gap extension penalty (%s)" % gap_extend_penalty
                    sys.exit(-1)
            else:
                print "Error: Expected a gap extension penalty"
                sys.exit(-1)
        elif arguments[i]=="-lo" or arguments[i]=="-lopen":
            if i<numargs:
                lock_gap_open_penalty = arguments[i+1]
                try:
                    lock_gap_open_penalty = float(lock_gap_open_penalty)
                except:
                    print "Error: Expected a numerical value for LOCK 2's gap opening penalty (%s)" % gap_open_penalty
            else:
                print "Error: Expected a gap opening penalty for LOCK 2"
                sys.exit(-1)
        elif arguments[i]=="-le" or arguments[i]=="-lextend":
            if i<numargs:
                lock_gap_extend_penalty = arguments[i+1]
                try:
                    lock_gap_extend_penalty = float(lock_gap_extend_penalty)
                except:
                    print "Error: Expected a numerical value for LOCK 2's gap extension penalty (%s)" % lock_gap_extend_penalty
                    sys.exit(-1)
            else:
                print "Error: Expected a gap extension penalty for LOCK 2"
                sys.exit(-1)
        elif arguments[i]=="-d":
            if i<numargs:
                lock_distance_threshold = arguments[i+1]
                try:
                    lock_distance_threshold = float(lock_distance_threshold)
                except:
                    print "Error: Expected a numerical value for the LOCK 2 distance threshold (%s)" % lock_distance_threshold
                    sys.exit(-1)
            else:
                print "Error: Expected a distance threshold for LOCK 2"
                sys.exit(-1)
                
    return gap_open_penalty,gap_extend_penalty,lock_gap_open_penalty,lock_gap_extend_penalty,lock_distance_threshold
                    

def main(arguments):
    structure_list_file = arguments[1] #file containing names of domains to be superimposed
    if len(arguments)>2:
        alignmentdir = arguments[2]
    else:
        alignmentdir = "alignments"

    if len(arguments)>3:
        gapMultiplier,gap_extend_penalty,lock_gapMultiplier,lock_gap_extend_penalty,lock_distance_threshold = parseOptionalArguments(arguments[3:])
    else:
        gapMultiplier = 6;
        gap_extend_penalty = 3;
        lock_gapMultiplier = 3;
        lock_gap_extend_penalty = 2.7
        lock_distance_threshold = 4.0
        
    setMultipliers(gapMultiplier,gap_extend_penalty,lock_gapMultiplier,lock_gap_extend_penalty)
    setDistanceThreshold(lock_distance_threshold)

    sequence().setPDBDir(pdb_dir)
    sequence().setSCOPPDBdir(scop_pdb_dir)
    sequence().setAlignmentDir(alignmentdir)

    edge_counts = {}; domains = []; scores = {}
    (domain_names)=readInputFile(structure_list_file)

    domain_indices = {} #give each domain a unique index
    sse_penalties = {}
    domain_names_parsed = [x[0] for x in domain_names]
    #Read in PDB files
    if not os.path.exists("Proteins"):
        os.mkdir("Proteins")
        
    for i in range(len(domain_names)):
        domain_indices[domain_names[i][0]] = i
        os.system("/bin/ln %s/%s%s Proteins/%s%s" % (domain_names[i][1],domain_names[i][0],domain_names[i][2],domain_names[i][0],domain_names[i][2]))
        domains.append(sequence(domain_names[i][0],"Proteins",domain_names[i][2],domain_names[i][3]))
        sse_penalties[domain_names[i][0]] = {}

    marked_residues = {}
    original_marked_residues = {}

    transformations = {}
    inverse_transformations = {}

    #step 1: run pairwise alignments
    scores = runPairwiseAlignments(domains,alignmentdir)

    #Read in pairwise alignments
    alignments = alignmentSet(alignmentdir,domain_names_parsed,scores)
    for domain in domains: #Finish setting up structures
        domain.getSSEs() #This is the best way to get the SSE definitions used by LOCK 2
        domain.makeSSTypeArray()
        domain.getSSEMidpoints()
    
    print "CBA: Done setting up alignments -> step 1 complete."


    for i in range(len(domains)):
        domain1 = domains[i].domainname
        marked_residues[i] = {}
        original_marked_residues[i]  = {}
        transformations[i] = {}; inverse_transformations[i] = {}
        for j in range(0,i):
            transformations[i][j] = {}; inverse_transformations[i][j] = {}


    #Set up transformation pointers. Easy way to find the inverse transformation for an
    #alignment. Will be useful when transforming structures into the same frame of reference.
    for i in range(len(domains)):
        for j in range(i+1,len(domains)):
            rotation,translation = alignments.getTransformation(i,j)
            transformations[j][i]['forward'] = (rotation[:],translation[:])
            inverse_rotation = inverse(rotation) #Couls just getTransformation(j,i)
            inverse_translation = -1*dot(translation,inverse_rotation)
            inverse_transformations[j][i]['forward'] = (inverse_rotation[:],inverse_translation[:])
            transformations[j][i]['inverse'] = inverse_transformations[j][i]
            inverse_transformations[j][i]['inverse'] = transformations[j][i]


    #Use hierarchical clustering to find best way to transform structures into same
    #frame of reference in step 2.
    joins = clusterStructures(domains,alignments)



    #use joins array to transform all structures (step 2)
    new_coordinates = []
    new_coordinates = transformStructures(domains,alignments,joins,transformations,inverse_transformations)

    try:
        pdbfile = open("cba.pdb","w") 
    except Exception,e:
        printError("Error: Could not print superimposed coordinates.")

    writeAlignmentPDBFile(domains,new_coordinates,pdbfile)
    pdbfile.close()

    del new_coordinates #free up some memory

    #redo residue alignment for each pairwise alignment (step 2)
    runResidueAlignments(domains,alignments,alignmentdir)

    print "CBA: Transformed structures -> step 2 complete."
    


    #read in new alignments
    del alignments
    alignments = alignmentSet("superimposed-alignments",domain_names_parsed,scores)  

    #Run steps 3, 4, and 5
    nodes,clusters,cluster_membership,duplicates = clusterAndFixSSEs(domains,alignments,alignmentdir,domain_names_parsed,scores)

        
    #When a cluster has two residues from the same structure, edge_counts determines
    #which residue is more strongly connected to the cluster
    for i in range(len(clusters)):
        edge_counts[i] = {}
        for j in range(len(clusters[i])):
            edge_counts[i][clusters[i][j]] = 0

    makeEdgeCounts(domains,alignments,cluster_membership,edge_counts)


    #eliminated_nodes has a key for each node that was removed because it was a duplicate. 
    eliminated_nodes,nodes_to_check=findDuplicateNodes(domains,clusters,edge_counts,duplicates)
    del edge_counts #don't need it anymore


    for node_num in eliminated_nodes.keys():
        node = nodes.nodes[node_num]
        node.eliminated = 1
        domain = node.domain_idx; resnum = node.resnum
        marked_residues[domain][resnum]=1
        original_marked_residues[domain][resnum] = 1


    #Build the initial MSTA
    columns=orderMSAColumns(domains,domain_indices,nodes,clusters,eliminated_nodes,nodes_to_check,marked_residues,sse_penalties,debug=0)
    print "CBA: Finished setting up current multiple structure alignment. About to begin step 6."
    original_marked = deepcopy(marked_residues)


    #Realign local regions (step 6)
    columns = fixBadRegions(columns,domains,nodes,clusters,marked_residues,alignments,sse_penalties)
    print "CBA: Realigned local regions -> step 6 complete."

    printMSTA(domains,columns,original_marked_residues,"cba.html")
    gzipAlignments(domains,alignmentdir)
    gzipAlignments(domains,"superimposed-alignments")
    gzipAlignments(domains,"fixed-alignments")
    os.remove("mclmatrix")
    os.remove("mclresults")
    

if 1:
    try:
        import psyco
        psyco.bind(orderMSAColumns)
        psyco.bind(fixBadRegions)
        psyco.bind(runDynamicProgramming)
        psyco.bind(integrateAlignedRegion)
    except:
        pass

main(sys.argv)
