#!/usr/bin/python

import string,re,os
import gzip
from Numeric import *
from LinearAlgebra import *



number_re = re.compile("[0-9]+.?")
string_re = re.compile("[A-Z]")

#def dot(array1, array2):
#    return innerproduct(array1, swapaxes(array2, -1, -2))


#Encodes information about an alignment. Note that initializing an alignment
#does not read the alignment file. This is not necessarily intuitive, but is
#done to encourage the use of the alignmentSet class, which encodes all of the
#pairwise alignments among a group of proteins. 
class alignment:
    alignmentdir = "/disk2/LOCK_allvall/alignments"

    def __init__(self,query,target,score=0,alignmentdir=alignmentdir):
        self.__class__.alignmentdir=alignmentdir
        self.query=query
        self.target=target
        self.score=score
        self.num_aligned_residues = 0 #will be set after reading alignmentfile
        alignment.alignmentdir = alignmentdir

        #Rotation and translation transform the target onto the query. Will
        #be initialized when alignment file is read in
        self.rotation = identity(3)
        self.translation = zeros([3],Float)



    def __str__(self):
        return "<alignment of %s to %s>" % (self.query,self.target)

    __repr__ = __str__


    #Initialize the alignment object by reading the alignment file
    def initAlignmentFromFile(self):
        self.readAlignmentFile(alignmentdir=self.__class__.alignmentdir)

        
    def isAligned(self,resnum):
        if self.alignedResidues.has_key(resnum): return 1
        else: return 0

    def alignedResidue(self,resnum):
        try:
            return self.alignedResidues[resnum]
        except KeyError:
            return -1

        
    def setAlignedResidue(self,resnum,aligned_res):
        if not self.alignedResidues.has_key(resnum): self.num_aligned_residues += 1
        self.alignedResidues[resnum] = aligned_res

    def setAlignedSSE(self,sse_start,aligned_sse_start,score=0,first_aligned_SSE=-1):
        if not self.alignedSSEs.has_key(sse_start): self.num_aligned_SSEs += 1
        self.alignedSSEs[sse_start] = aligned_sse_start
        self.sse_scores[sse_start] = score
        if first_aligned_SSE!=-1: 
            if first_aligned_SSE: self.max_sse_scores[sse_start] = 20.0
            else: self.max_sse_scores[sse_start] = 40.0
        else: self.max_sse_scores[sse_start] = 20.0


    def getSSEScore(self,sse_start):
        try:
            return self.sse_scores[sse_start]
        except KeyError:
            return 0

    def getAlignedSSE(self,sse_start):
        try: return self.alignedSSEs[sse_start]
        except KeyError: return -1
        

    #Initializes an alignment of the query vs target from the inverse alignment
    #(target vs. query).
    def initAlignmentFromInverse(self,inverse_alignment):
        aligned_residues = inverse_alignment.alignedResidues.keys()
        self.alignedResidues = {}
        self.alignedSSEs = {}
        self.num_aligned_SSEs = inverse_alignment.num_aligned_SSEs
        self.sse_scores = {}
        self.max_sse_scores = {}
        self.alignmentscores = {}
        for res in aligned_residues:
            self.setAlignedResidue(inverse_alignment.alignedResidue(res),res)
            self.alignmentscores[inverse_alignment.alignedResidue(res)] = inverse_alignment.alignmentscores[res]

        self.num_aligned_residues = len(aligned_residues)
        self.rotation = inverse(inverse_alignment.rotation)
        self.translation = -1*dot(inverse_alignment.translation,self.rotation)
        aligned_sses = inverse_alignment.alignedSSEs.keys()
        count = 0
        for sse_start in aligned_sses:
            if count==0:
                first = 1
            else: first=0
            self.setAlignedSSE(inverse_alignment.alignedSSEs[sse_start],sse_start,inverse_alignment.getSSEScore(sse_start),first_aligned_SSE = first)
            count += 1
                
                
    def readAlignmentFile(self,alignmentdir=""):
        debug = 0
        if debug:
            print "readalignmentfile:",self.query,self.target
        if alignmentdir=="": alignmentdir=self.__class__.alignmentdir
        file = os.path.join(alignmentdir,"%s_%s.out" % (self.query,self.target))
        self.num_aligned_residues = 0
        self.num_aligned_SSEs = 0
        seen = {}
        seen2 = {}
        aligned = {}
        alignmentscores = {}
        gapscores = {}
        transformation_matrix = []
        self.sse_scores = {}
        self.max_sse_scores = {}
        self.alignedSSEs = {}
        num_residues_aligned = 0
        self.finallines = []
        if not os.path.exists(file):
            file += ".gz"
            try:
                infile = gzip.open(file,"rb")
            except:
#                print "Couldn't read LOCK 2 alignment file (%s)" % file
                self.num_aligned_residues = 0
                self.rotation = identity(3)
                self.translation = array([0,0,0])
                self.alignedResidues = {}
                self.num_aligned_SSEs = 0
                self.alignmentscores = {}
                self.gapscores = {}
                return 0 #failure
        else:
            infile = open(file,"r")
        line=infile.readline()
        begin_final_lines = 0
        while line:
            if re.match("Transformation Matrix for query",line):
                query_rotation = zeros([3,3],Float)
                for i in range(3):
                    line=infile.readline()
                    if begin_final_lines:
                        self.finallines.append(line)

                    fields = map(string.atof,line.split()[:-1])
                    for j in range(3):
                        query_rotation[i][j] = fields[j] #learn better way to do this
                line=infile.readline()
                if begin_final_lines:
                    self.finallines.append(line)

                query_translation = array(map(string.atof,line.split()[:-1]))
            elif re.match("Transformation Matrix for target",line):
                target_rotation = zeros([3,3],Float)
                for i in range(3):
                    line=infile.readline()
                    if begin_final_lines:
                        self.finallines.append(line)
                    fields = map(string.atof,line.split()[:-1])
                    for j in range(3):
                        target_rotation[i][j] = fields[j] #learn better way to do this
                line=infile.readline()
                if begin_final_lines:
                    self.finallines.append(line)                
                target_translation = array(map(string.atof,line.split()[:-1]))
            elif re.match("Num Core Aligned",line):
                num_residues_aligned = string.atoi(string.split(line)[-1])
                if num_residues_aligned==0: break
                for i in range(num_residues_aligned):
                    line = infile.readline()
                    fields = re.split("\t",line)
                    resnum1 = fields[0][:-1] #get rid of amino acid
                    resnum1 = resnum1.strip()
                    resnum2 = fields[1][:-1].strip()
                    if debug:
                        print "resnum1=%s,resnum2=%s" % (resnum1,resnum2)
                    aligned[resnum1] = resnum2
                    alignmentscores[resnum1] = float(fields[3])
                    gapscores[resnum1] = float(fields[-1])
                begin_final_lines = 1
            elif re.match("Aligned Secondary Structure Elements",line):
                num_aligned_SSEs = string.atoi(line.split()[-2])
                self.num_aligned_SSEs = num_aligned_SSEs
                for i in range(num_aligned_SSEs):
                    line = infile.readline()
                    if begin_final_lines:
                        self.finallines.append(line)
                    fields = re.split("\t",line)
                    qstart,qend = re.split(",",fields[1])
                    tstart,tend = re.split(",",fields[2])
                    score = float(fields[-2])
                    self.alignedSSEs[qstart] = tstart
                    self.sse_scores[qstart] = score
                    if i==0: self.max_sse_scores[qstart] = 40.0
                    else: self.max_sse_scores[qstart] = 20.0
                if begin_final_lines:
                    line = infile.readline()
                    while line:
                        self.finallines.append(line)
                        line = infile.readline()
                break
            else:
                line=infile.readline()
                if begin_final_lines:
                    self.finallines.append(line)                
        
        #make query stationary by applying inverse transformation to target
        if not num_residues_aligned==0:
            query_rotation_inverse = inverse(query_rotation)
            new_target_translation = dot(target_translation-query_translation,query_rotation_inverse)
            new_target_rotation = dot(target_rotation,query_rotation_inverse)

            #compute inverse in order to transform query onto target
            inverse_target_rotation = inverse(new_target_rotation)
            inverse_target_translation = -1*dot(new_target_translation,inverse_target_rotation)
        else:
            new_target_rotation = identity(3,Float)
            new_target_translation = zeros([3],Float)
        self.num_aligned_residues = num_residues_aligned
        self.rotation = new_target_rotation
        self.translation = new_target_translation
        self.alignedResidues = aligned
        self.alignmentscores = alignmentscores
        self.gapscores = gapscores
        return 1 #success
#        return new_target_rotation,new_target_translation,aligned
        

    def setAlignmentDir(self,dir):
        self.__class__.alignmentdir = dir


#CBA typically accesses alignment information through this class, which encodes
#information about all of the pairwise alignments among a group of structures.
class alignmentSet:
    alignmentdir = "/home/jessicas/fixSSEs/new-alignments-3.3"

    #The scores parameter is a 2D list indexed in the same order
    #as domainnames (which itself is a list of the domains in the
    #alignmentSet object). Thus, scores[1][2] is the LOCK 2 alignment
    #score for domainnames[1] vs domainnames[2].
    def __init__(self,alignment_dir,domainnames,scores=None):
        self.__class__.alignmentdir = alignment_dir
        self.domains = domainnames #a list of domainnames
        self.num_domains = len(domainnames)

        if scores==None: #So python won't complain if scores are not passed in during testing
            scores = [[] for x in range(len(self.domains))]
            for x in range(len(self.domains)):
                scores[x] = [1 for y in range(len(self.domains))]
        self.scores = scores
        self.initializeAlignments()
                    

    def setAlignmentDirectory(self,alignmentdir):
        self.__class__.alignmentdir = alignmentdir
        

    #Initializes the alignments in the alignmentSet by reading alignment files.
    def initializeAlignments(self):
        alignments = []
        domains_to_remove = []
        for i in range(self.num_domains):
            alignments.append([])

            for j in range(self.num_domains):
                alignments[i].append(alignment(self.domains[i],self.domains[j],alignmentdir=self.__class__.alignmentdir,score=self.scores[i][j]))

        for i in range(self.num_domains):
            for j in range(i,self.num_domains):
                alignments[i][j].initAlignmentFromFile()
        for j in range(self.num_domains):
            for i in range(j+1,self.num_domains):
                alignments[i][j].initAlignmentFromInverse(alignments[j][i])
                
        self.alignments = alignments
        return domains_to_remove

    def getAlignment(self,query_idx,target_idx):
        return self.alignments[query_idx][target_idx]

    def isAligned(self,query_idx,target_idx,resnum):
        return self.alignments[query_idx][target_idx].isAligned(resnum)

    def alignedResidue(self,query_idx,target_idx,resnum):
        if self.alignments[query_idx][target_idx].alignedResidues.has_key(resnum):
            return self.alignments[query_idx][target_idx].alignedResidues[resnum]
        else: return -1

    def setAlignedResidue(self,query_idx,target_idx,resnum,aligned_resnum):
        self.alignments[query_idx][target_idx].setAlignedResidue(resnum,aligned_resnum)
        self.alignments[target_idx][query_idx].setAlignedResidue(aligned_resnum,resnum)


    def getAlignmentScore(self,query_idx,target_idx):
        return self.alignments[query_idx][target_idx].score

    def getSSEAlignmentScore(self,query_idx,target_idx,query_sse_start=-1, query_ssenum = -1):
        if self.alignments[query_idx][target_idx].sse_scores.has_key(query_sse_start):
            return self.alignments[query_idx][target_idx].sse_scores[query_sse_start]
        else:
            return 0

    def getTransformation(self,query_idx,target_idx):
        return (self.alignments[query_idx][target_idx].rotation,self.alignments[query_idx][target_idx].translation)





if 1:
    try:
        import psyco
        psyco.full()
    except:
        pass
