#!/usr/bin/python

import os,string,re
import copy
from numarray import *
import gzip
import math
import nodeList

    
zcat_program = "/bin/zcat"
file_program = "/usr/bin/file"
gunzip_program = "/bin/gunzip"


genetic_domain_re = re.compile("g(.*)")

def dot(array1, array2):
    return innerproduct(array1, swapaxes(array2, -1, -2))

compare_re = re.compile("[0-9]+[A-Z]$")
def compareStringNumbers(res1,res2):
    chain1 = res1[-1]
    chain2 = res2[-1]
    res1 = res1[:-2]
    res2 = res2[:-2]

    res1 = int(res1)
    res2 = int(res2)
    if chain1==chain2:
        return cmp(res1,res2)
    elif chain1 < chain2:
        return -1
    else:
        return 1



def convert3LetterTo1Letter(code):
    if code=="ALA": return "A"
    elif code=="ARG": return "R"
    elif code=="ASN": return "N"
    elif code=="ASP": return "D"
    elif code=="CYS": return "C"
    elif code=="GLN": return "Q"
    elif code=="GLU": return "E"
    elif code=="GLY": return "G"
    elif code=="HIS": return "H"
    elif code=="ILE": return "I"
    elif code=="LEU": return "L"
    elif code=="LYS": return "K"
    elif code=="MET": return "M"
    elif code=="PHE": return "F"
    elif code=="PRO": return "P"
    elif code=="SER": return "S"
    elif code=="THR": return "T"
    elif code=="TRP": return "W"
    elif code=="TYR": return "Y"
    elif code=="VAL": return "V"
    else: return "U" #amino acid not identified

#Encodes information about a protein sequence
class sequence:
    scop_pdbdir = "/disk3/scop-pdb" 
    pdbdir = "/pdb_new"
    alignmentdir = "/home/jessicas/MSA/globin-test/alignments-6.9"

    #Initializes the sequence. Recognizes SCOP and PDB identifiers, or a path to
    #a PDB file by the parameters name, directory, extension, and chain. (See readInputFile
    #method in CBA.py.)
    #NOTE: Initialization is a little messy since member variables were added as needed
    #in the development of CBA. Other function calls are necessary to obtain secondary
    #structure definitions (getSSEs, makeSSTypeArray). The function getSSEs depends on the
    #existence of an alignment of the structure to itself to make sure that the SSE definitions
    #used by LOCK 2 are read in (as opposed to DSSP definitions, which LOCK 2 might not have
    #used).
    def __init__(self,name="",directory="",extension="",chain="all"):
        if len(name)==7 and directory=="": scop = 1
        else: scop = 0
        self.domainname = name
        self.chain = chain
        if name != "":
            self.directory = directory
            self.extension = extension
            self.scop = scop
            if scop:
                self.shortname = self.domainname
                self.directory = "%s/%s" % (sequence.scop_pdbdir,name[2:4])
                if self.extension=="":

                    #try to detect extension
                    if not os.path.exists("%s/%s" % (self.directory,self.shortname)):
                        if os.path.exists("%s/%s.ent" % (self.directory,self.shortname)):
                            self.extension = ".ent"
                        elif os.path.exists("%s/%s.pdb" % (self.directory,self.shortname)):
                            self.extension = ".pdb"
                        elif os.path.exists("%s/%s.ent.gz" % (self.directory,self.shortname)):
                            self.extension = ".ent"
                        elif os.path.exists("%s/%s.pdb.gz" % (self.directory,self.shortname)):
                            self.extension = ".pdb"
                        elif os.path.exists("%s/%s.ent.Z" % (self.directory,self.shortname)):
                            self.extension = ".ent"
                        elif os.path.exists("%s/%s.pdb.Z" % (self.directory,self.shortname)):
                            self.extension = ".pdb"
                        
            else:
                if directory=="": #A PDB identifier was specified
                    if len(name)>4:  #Has chain id, as in 1dlw-A
                        self.shortname = name[:-2] #name without chain
                        if self.chain!="all": self.chain = chain.upper()
                    else:
                        self.shortname = name

                    self.directory = "%s/%s" % (sequence.pdbdir,name[1:3])
                    if self.extension=="":
                        #try to detect extension
                        if not os.path.exists("%s/%s" % (self.directory,self.shortname)):
                            if os.path.exists("%s/%s.ent" % (self.directory,self.shortname)):
                                self.extension = ".ent"
                            elif os.path.exists("%s/%s.pdb" % (self.directory,self.shortname)):
                                    self.extension = ".pdb"
                            elif os.path.exists("%s/%s.ent.gz" % (self.directory,self.shortname)):
                                self.extension = ".ent"
                            elif os.path.exists("%s/%s.pdb.gz" % (self.directory,self.shortname)):
                                self.extension = ".pdb"
                            elif os.path.exists("%s/%s.ent.Z" % (self.directory,self.shortname)):
                                self.extension = ".ent"
                            elif os.path.exists("%s/%s.pdb.Z" % (self.directory,self.shortname)):
                                self.extension = ".pdb"
                else: #A PDB file was specified
                    if self.chain != "all":
                        if self.chain!="all": self.chain = chain.upper()	
#                    self.chain = "all"
                    self.shortname = name

            self.readPDBFile() #read the PDB file
            self.num_residues = len(self.residues.keys())
            self.node_numbers = {}
            self.setReferenceFrame(identity(3),zeros([3],Float))
            
    def __str__(self):
        return "<sequence object %s>" % self.domainname

    __repr__ = __str__


    def setReferenceFrame(self,rotation,translation):
        self.translation = translation
        self.rotation = rotation

    def findGlycineBetaCarbon(self,resindex):
        debug=0
        CA = copy.copy(self.coordinates[resindex]["CA"])
        N = copy.copy((self.coordinates[resindex]["N"]))
        C = copy.copy((self.coordinates[resindex]["C"]))
        theta = -2.0*math.pi;
        theta = theta/float(3.0);

        #calculate a unit vector (a,b,c) along the CA-N axis
        a = N[0]-CA[0]; b = N[1]-CA[1]; c = N[2]-CA[2]
        length = math.sqrt(a*a+b*b+c*c)
        #translate atoms such that the alpha carbon is at the origin
        N[0] -= CA[0]; N[1] -= CA[1]; N[2] -= CA[2]
        C[0] -= CA[0]; C[1] -= CA[1]; C[2] -= CA[2]

        #rotate space about the x axis so that the rotation axis lies in the xz plane
        d = math.sqrt(b*b+c*c)
        if d!= 0:
            a2 = C[0]; b2 = C[1]*c/float(d)-C[2]*b/float(d); c2 = C[1]*b/float(d)+C[2]*c/d;
        else:
            a2 = C[0]; b2=C[1]; c2=C[2];

        #rotate space about the y axis so that the rotation axis lies along the z axis
        a3 = a2*d-c2*a; b3=b2; c3=a*a2+d*c2

        #rotation about the z axis by 120 degrees
        a4 = math.cos(theta)*a3+math.sin(theta)*b3;
        b4 = -math.sin(theta)*a3+math.cos(theta)*b3;
        c4 = c3

        #invert rotation about the y axis
        a3 = d*a4+a*c4
        b3=b4
        c3=-a*a4+d*c4

        #invert rotation about x axis
        if d!=0:
            a2=a3; b2=b3*c/float(d)+c3*b/float(d); c2=-b3*b/float(d)+c3*c/float(d)
        else:
            a2=a3; b2=b3; c2=c3

        #invert translation to get final point
        point= array([a2+CA[0], b2+CA[1], c2+CA[2]])
        if debug: print "findglycineCB:",point,a2,a3,a4
        return point
    


    def getNodeNumber(self,resnum=-1,index=-1):
        if resnum==-1: resnum=self.residues[index]
        try:
            return self.node_numbers[resnum]
        except KeyError:
            return -1
    
    def setNodeNumber(self,resnum,node):
        self.node_numbers[resnum] = copy.copy(node.getNumber())

    def isNamed(self,name):
        if self.domainname==name: return 1
        else: return 0

    def setPDBDir(self,dir):
        self.__class__.pdbdir = dir

    def setSCOPPDBdir(self,dir):
        self.__class__.scop_pdbdir = dir

    def setAlignmentDir(self,dir):
        self.__class__.alignmentdir = dir        
        

    #Creates dictionaries of SSE type (H,S, or U) by residue number (sstype) and
    #of SSE number by residue number (ssNumber). 
    def makeSSTypeArray(self):
        debug = 0
        if debug: print "makeSSTypeArray %s" % self.domainname
        sse_types = self.sse_types
        querylen = len(self.residue_indices.keys())
        sstype = {}; ssNumber = {}

        myresidues = self.residues.values()
        myresidues.sort(compareStringNumbers)
        query_res = myresidues[0]

        number = -1
        i = 0
        for s in range(len(self.sse_ends)):
            start,end = self.sse_ends[s]
            resint = self.residue_indices[start]
            number += 1
            testres = self.residue_indices[myresidues[i]]
            if debug:
                print "s=%i, ends=%s,%s, resint=%i,testres=%i" % (s,start,end,resint,testres)

            while testres < resint:
                ssNumber[myresidues[i]] = -1
                sstype[myresidues[i]] = 'U'
                i += 1
                if i<querylen:
                    testres = self.residue_indices[myresidues[i]]
                else:
                    break
            testres = self.residue_indices[myresidues[i]]
            end = self.residue_indices[end]+1
            if debug:
                print "testres=%i, end=%i" % (testres,end)
            while testres < end:
                ssNumber[myresidues[i]] = number
                sstype[myresidues[i]] = sse_types[start]
                i += 1
                if i<querylen:
                    testres = self.residue_indices[myresidues[i]]
                else:
                    break


        for j in range(i,querylen):
            ssNumber[myresidues[j]] = -1
            sstype[myresidues[j]] = 'U'
        if debug:
            for j in range(querylen):
                print j,myresidues[j],ssNumber[myresidues[j]]
        self.sstype = sstype
        self.ssNumber = ssNumber

        
    def getPotentiallyCompressedFile(self,filename,domainname,directory):
        if os.path.exists(filename):
            output = os.popen("%s %s" % (file_program,filename),"r")
            result = output.readline()
            output.close()
            fields = result.split()
            if fields[1]=="gzip":
                os.system("%s -c %s > %s/%s.ent" % (gunzip_program,filename,directory,domainname))
                return (directory,"%s/%s.ent" % (directory,domainname))
            elif (fields[1]=="compressed" or fields[1]=="compress'd"):
                os.system("%s %s > %s/%s.ent" % (zcat_program,filename,directory,domainname))
                return (directory,"%s/%s.ent" % (directory,domainname))
            else:
                return (directory,filename)
        else:
            return -1

                

    def readPDBFile(self):
        atomre = re.compile("^ATOM|HETATM")
        endre = re.compile("^END|ENDMDL|CONECT")
#        hetatomre = re.compile("^HETATM")
        skip_re = re.compile("ANISOU|TER")
        ca = {}
        cb = {}
        C = {}; N = {}
        resnums_seen = {}
        self.residue_atomnames = {}
        self.coordinates = {}
        self.has_insertion_code = {}
        seen_ca={}; seen_cb= {}; seen_c = {}; seen_n = {}        
        domainname = self.domainname
        filename = "%s/%s%s" % (self.directory,self.shortname,self.extension)
        filename = os.path.normpath(filename)
        if not os.path.exists(filename):
            newfilename = filename + ".gz"
            if os.path.exists(newfilename):
                os.system("%s -f %s > %s.ent" % (zcat_program,newfilename,domainname))
                self.directory = "./"
                self.extension = ".ent"
                filename = "%s.ent" % domainname
            else:
                newfilename = filename + ".Z"
                if os.path.exists(newfilename):
                    os.system("%s -f %s > %s.ent" % (zcat_program,newfilename,domainname))
                    self.directory = "./"
                    self.extension = ".ent"
                    filename = "%s.ent" % domainname
                else:
                    print "Couldn't find file for %s: exiting" % domainname
                    sys.exit(-1)
        else:
            #file might still be compressed
            newfilename = self.getPotentiallyCompressedFile(filename,domainname,self.directory)[1]
            if newfilename!=filename: #file was compressed
#                self.directory = "./"
                self.extension = ".ent"
                filename = newfilename
        self.residue_indices = {}; self.residues = {}; self.aas = {}
        index = -1
        old_resnum = ""
        found_alpha_carbon = 0
        pdb = open(filename,"r")
        line=pdb.readline()
        while line:
            if atomre.match(line): break
            line = pdb.readline()
        while line:
            if endre.match(line): break
##            if hetatomre.match(line):
##                line=pdb.readline()
##                continue
            if skip_re.match(line):
                line=pdb.readline()
                continue
            atomname = string.strip(line[12:16])
            if not self.chain=="all" and line[21]!=self.chain: line=pdb.readline(); continue
            x=string.atof(line[30:38]); y=string.atof(line[38:46]); z=string.atof(line[46:54]);
            resnum = string.strip(line[22:26])
            resnum += line[26]
            chain = line[21]
            if chain==" ": chain = "-"
            resnum += chain
            if old_resnum != resnum:
                if resnums_seen.has_key(resnum):
                    line = pdb.readline()
                    continue
                resnums_seen[resnum] = 1
                if found_alpha_carbon or (old_resnum==""):
                    index += 1
                    if old_resnum!="":
                        self.residue_indices[old_resnum] = index-1

                self.coordinates[index] = {}

                self.residues[index] = resnum
                self.aas[index] = convert3LetterTo1Letter(line[17:20])
                self.residue_atomnames[index] = []
                found_alpha_carbon = 0
            self.coordinates[index][atomname] = array([x,y,z])
            self.residue_atomnames[index].append(atomname)
            if atomname=="CA": found_alpha_carbon = 1
            old_resnum = resnum
            line =pdb.readline()
        #add in last residue
        if found_alpha_carbon:
            self.residue_indices[old_resnum] = index
        else:
            del self.residues[index]
            del self.aas[index]
            del self.residue_atomnames[index]
            del self.coordinates[index]
            for r,idx in self.residue_indices.items():
                if idx==index:
                    del self.residue_indices[r]
        for index in self.residues.keys():
            if not self.coordinates[index].has_key("CB"):
                if self.coordinates[index].has_key("CA") and self.coordinates[index].has_key("C") and self.coordinates[index].has_key("N"):
                    self.coordinates[index]["CB"] = self.findGlycineBetaCarbon(index)
                    self.residue_atomnames[index].append("CB")

                


    query_transformation_re=re.compile("Transformation Matrix for query")
    target_transformation_re=re.compile("Transformation Matrix for target")

    #Gets SSE definitions used by LOCK 2 by reading query vs. query alignment file.
    def getSSEs(self):
        file = os.path.join(sequence.alignmentdir,"%s_%s.out" % (self.domainname,self.domainname))
        sse_ends = []
        sse_types = {}
        if not os.path.exists(file):
            file += ".gz"
            try:
                infile = gzip.open(file,"rb")
            except:
                self.sse_ends = []; self.sse_types = {}
                self.num_SSEs = 0
                return -1
#                return sse_ends,sse_types
        else:
            infile = open(file,"r")
        line = infile.readline()
        while line:
            if re.match("SSE aa numbers",line):
                infile.readline(); line=infile.readline()
                while not re.match("target",line):
                    fields = re.split("\t",line[:-1])
                    qstart = fields[1]; qend = fields[2]
                    sse_ends.append((qstart,qend))
                    sse_types[qstart] = fields[0]
                    line=infile.readline()
            else:
                line=infile.readline()
                    
        self.sse_ends = sse_ends; self.sse_types = sse_types
        self.num_SSEs = len(self.sse_ends)
        return 1


    #Finds the midpoint of each secondary structure element
    def getSSEMidpoints(self):
        debug=0
        if debug: print "getssemidpoints",self.sse_ends
        midpoints = {}
        for i in range(len(self.sse_ends)):
            if debug: print "i=%i, ends=" % i,self.sse_ends[i]
            start,end = self.sse_ends[i]
            start_index = self.residue_indices[start]
            end_index = self.residue_indices[end]
            sse_residues = []
            for resindex in range(start_index,end_index+1):
                sse_residues.append(self.residues[resindex])
            midpoints[i] = self.residue_indices[sse_residues[len(sse_residues)/2]]
            
        self.sse_midpoints = midpoints
    

    #Returns the SSE type of the residue (H, S, or U)
    def getSSEType(self,resnum=-1,index=-1):
        if resnum==-1:
            resnum=self.residues[index]
        return self.sstype[resnum]


    #Returns 1 if the residue is the first in a helix or strand, 2 if it's the last,
    # or zero otherwise
    def isSSEEnd(self,index):
        resnum = self.residues[index]
        ssenum = self.ssNumber[resnum]

        if ssenum==-1: return 0
        if index < self.num_residues-1:
            next_resnum = self.residues[index+1]
            if self.ssNumber[next_resnum] != ssenum: return 2
        elif  index==self.num_residues-1:
            return 2
        if index>0:
            prev_resnum = self.residues[index-1]
            if self.ssNumber[prev_resnum] != ssenum: return 1
        elif index==0:
            return 1
        return 0


    #Returns the first and last residue of the SSE, which can be
    #specified either by SSE number or by passing in a residue in
    #the SSE
    def getSSEEnds(self,resnum=-1,ssenum=-1):
        if ssenum==-1:
            ssenum = self.ssNumber[resnum]
            if ssenum==-1: return -1,-1
            else:
                return self.sse_ends[ssenum]
        else:
            return self.sse_ends[ssenum]


    def getSSEStart(self,resnum=-1,ssenum=-1):
        if ssenum==-1:
            ssenum = self.ssNumber[resnum]
        if ssenum==-1: return -1,-1
        else: return self.sse_ends[ssenum][0]
        
    def getSSEEnd(self,resnum=-1,ssenum=-1):
        if ssenum==-1:
            ssenum = self.ssNumber[resnum]
            if ssenum==-1: return -1,-1
            else:
                return self.sse_ends[ssenum][1]
        else:
            if ssenum==-1: return -1,-1
            else: return self.sse_ends[ssenum][1]

        
    def transformCoordinates(self,rotation,translation):
        new_coordinates = []
        atomnames = self.residue_atomnames
        for i in range(self.num_residues):
            atoms = atomnames[i]
            new_coordinates.append({})
            for atom in atoms:
                new_coordinates[i][atom] = array(dot(self.coordinates[i][atom],rotation)+translation)
        return new_coordinates


    #Not used. 
    def getResiduePosition(self,resnum):
        sse_start,sse_end = self.getSSEEnds(resnum)
        if sse_start==-1: return 1 #loop
        sse_start = self.residue_indices[sse_start]
        sse_end = self.residue_indices[sse_end]
        length = sse_end-sse_start+1
        firstquarter = sse_start+length/4
        lastquarter = sse_end-length/4
        while lastquarter-firstquarter+1 < length/2:
            if lastquarter%2==0: lastquarter+=1
            else: lastquarter-=1

        if resindex <= firstquarter: return 2
        elif resindex < lastquarter: return 3
        else: return 2

    def getTransformation(self,query_idx,target_idx):
        return (self.alignments[query_idx][target_idx].rotation,self.alignments[query_idx][target_idx].translation)

    def getResidues(self):
        return self.residue_indices.keys()

if 1:
    try:
        import psyco
        psyco.full()
    except:
        pass
