#!/usr/bin/perl -T

# Copyright 2003 The Board of Trustees of 
# The Leland Stanford Junior University. All Rights Reserved.

# Communication to Jessica Ebert, jessicas@stanford.edu


$SCOPPATH = "/usr/db/scop-pdb";
$PDBPATH = "/usr/db/pdb";
$DSSPPATH = "";
$SCOPDSSPPATH = "/usr/db/scop-dssp";
$DSSPPROGRAM = "/usr/local/bin/dsspcmbi";

$ENV{'PATH'} = "/bin:/usr/bin:/usr/local/bin";

#system commands used by this script:
#you may need to switch to full path names for some shells
$file = "/usr/bin/file";
$cat = "/bin/cat";
$ln = "/bin/ln";
$gunzip = "/bin/gunzip";
#$uncompress = "@UNCOMPRESS@";
$rm = "/bin/rm";
$mkdir = "/bin/mkdir";

open (STDOUT, ">/dev/null") || die "can't open STDOUT: $!\n";
open (STDERR, ">/dev/null") || die "can't open STDERR: $!\n";



select STDOUT; $| = 1;
select STDERR; $| = 1;


$longpdbname = shift(@ARGV);
$newname = shift(@ARGV);

$newname = securityCheckFilename($newname); #problem: will remove certain characters and create a new filename unknown to the program that called this one
if($newname !~ /\w/) {
  exit(-1); #filename was invalid; just exit
}

if (-e $newname) {
    system($rm, $newname);
}


if(length($longpdbname) == 7) { #if using scop id
  $pdbname = $longpdbname;
  $isScopID = 1;
  $pdbname = securityCheck($longpdbname,$isScopID);
}else {
  ($pdbname) = $longpdbname =~ /^(....)/;
  $isScopID = 0;
  $pdbname = securityCheck($longpdbname,$isScopID);
}

$pdbname =~ tr/A-Z/a-z/;

if ((not defined $pdbname) || (not defined $newname)) {
    exit(-1);
}


if($isScopID) {
  ($two_letters) = $pdbname =~ /^..(..)/;
  $myfile = getFile("$SCOPDSSPPATH/$two_letters/$pdbname");
}else {
  ($two_letters) = $pdbname =~/^.(..)./;
  $myfile = getFile("$DSSPPATH/$two_letters/$pdbname");
}



if ($myfile eq "NOFILE") {
 
#  if(!$isScopID) {
#    $two_letter = $pdbname =~ /^[0-9](..)/;
#    $newdsspfile = "$DSSPPATH/$pdbname.dssp";
#  }else {
#    $two_letter = substr($pdbname,2,2);

#    $newdsspfile = "$SCOPDSSPPATH/$two_letter/$pdbname.dssp";

#    if(! -e "$SCOPDSSPPATH/$two_letter") {

#      system("$mkdir", "$SCOPDSSPPATH/$two_letter");

#    }
    
}


$tempfile = findFile($pdbname, $isScopID);

if ($tempfile ne "NOFILE") {
    $tempfile = checkCompressed($tempfile);
    system($DSSPPROGRAM, "-na", $tempfile, $newname, ">", "/dev/null");
    if (-e "/tmp/$$.temp") {
	system($rm, "/tmp/$$.temp");
    }
}else {
  $file = "/tmp/$$\.file";
  system($file,$myfile,"\>",$file);

  open(IN, "/tmp/$$\.file");
  $line = <IN>;
  close(IN);
  system("$rm","-f","/tmp/$$\.file");



  @ftype = split /\s+/, $line;

    if ($ftype[1] eq "gzip") {	
	system($cat, $myfile, "|", $gunzip ,">", $newname);
    }

    elsif (($ftype[1] eq "compressed") || ($ftype[1] eq "compress'd")) {	
	system($cat, $myfile, "|", $gunzip, ">" ,$newname);
    }
    else {
	system($ln, "-s", $myfile, $newname);
    }
}

exit(0);
    
######################################################################

sub findFile() {
  my($filename, $isScopID) = @_;
  my($shortname);

  if(!$isScopID) {
    ($two_letter) = $filename =~ /^[0-9](..)/;
    if($filename =~ /^(.*)-[A-Za-z0-9]$/) {
      $shortname = $1;
    }else {
      $shortname = $filename;
    }
    $mytemp = getFile("$PDBPATH/$two_letter/$shortname.pdb");
    if ($mytemp eq "NOFILE") {
	$mytemp = getFile("$PDBPATH/$shortname.pdb");
    }
    if ($mytemp eq "NOFILE") {
	$mytemp = getFile("$PDBPATH/pdb$shortname.ent");
    }
  }else {

    $two_letter = substr($filename,2,2);
    
    $mytemp = getFile("$SCOPPATH/$two_letter/$filename.ent");
    
    if($mytemp eq "NOFILE") {
      $mytemp = getFile("$SCOPDSSPPATH/$two_letter/$filename.dssp");
    }
  }
    return $mytemp;
}

sub getFile() {

    $mytemp = $_[0];

    if (-e $mytemp) {
	return $mytemp;
    }

    $mytemp = "$_[0]".".gz";
    if (-e $mytemp) {
	return $mytemp;
    }

    $mytemp = "$_[0]".".Z";
    if (-e $mytemp) {
	return $mytemp;
    }

    return "NOFILE";

}

    
sub checkCompressed() {
  my($myfile) = @_;

  open(OLD_STDOUT, ">&STDOUT");
  open(STDOUT,">/tmp/$$\.file");
  select(STDOUT); $| = 1;
  system($file,$myfile);
  close(STDOUT);
  open(STDOUT, ">&OLD_STDOUT");
  
  open(IN, "/tmp/$$\.file");
  $line = <IN>;
  close(IN);
  system("$rm","-f","/tmp/$$\.file");
  @ftype = split /\s+/, $line;
  
  if ($ftype[1] eq "gzip") {	
      open(OLD_STDOUT,">&STDOUT");
      open(STDOUT, ">/tmp/$$.temp.gz");
      select(STDOUT); $|=1;
      system($cat,$myfile);
      close(STDOUT);
      open(STDOUT, ">&OLD_STDOUT");
      system($gunzip, "/tmp/$$.temp.gz");
      
      if (-e "/tmp/$$.temp") {
	  return "/tmp/$$.temp";
      }
      else {
	  return "NOFILE";
      }
  }
  
  if ($ftype[1] eq "compressed" || $ftype[1] eq "compress'd") {	
      open(OLD_STDOUT, ">&STDOUT");
      open(STDOUT,">/tmp/$$.temp.Z");
      select(STDOUT); $| = 1;
      system($cat, $myfile);
      close(STDOUT);
      open(STDOUT, ">&OLD_STDOUT");
      system($gunzip,"/tmp/$$.temp.Z");
      
      if (-e "/tmp/$$.temp") {
	  return "/tmp/$$.temp";
      }
      else {
	  return "NOFILE";
      }
  }
  
  return $_[0];
}

sub securityCheckFilename {
  my($filename) = @_;
#  print "checking $filename\n";
  #Make sure this is a full path that contains only alphanumeric characters and '/'
  if($filename =~ /^([\/A-Za-z.0-9]*)$/) {
    return $1;
  }else {
    return ""; #need to do error checking -- could return empty string
  }
}

#checks user's pdb or scop id (can scop id start with anything other than d or g?)
sub securityCheck {
  my($inputname, $isScop) = @_;
  my($cleanname,$pdbid);
  
  $cleanname = "";
  if($isScop) {
    $inputname =~ /^([dg][0-9][A-Za-z0-9]{3}[\._A-Za-z0-9][_0-9])/;
    $cleanname = $1;
  }else {
    $inputname =~/^([0-9][A-Za-z0-9]{3})(-[A-Za-z0-9])?/;
    $cleanname = $1.$2; #$1 is the pdb id, and $2 is the chain, prefixed by '-'
  }
  return $cleanname; #need to do error checking -- could return empty string
}
