#!/usr/bin/python

import cgi
import string
import re
import sys
import os
import smtplib


cba_program = "/var/www/cgi-bin/CBA/cba.py"
cba_html_dir = "/var/www/html/CBA"
cba_html_subdir = "CBA"
scopdir = "/usr/db/scop-pdb"
pdbdir = "/usr/db/pdb"
customer_service_email_address = "jessicas@stanford.edu"
customer_service_contact = "Jessica Ebert"

#programs needed
gunzip = "/bin/gunzip"

def uploadFile(fileObject,filename):
    #make sure file name is reasonable (for security)
    m = re.match("^([^\*]*)$",filename)
    if m==None:
        return -1 #filename contains illegal characters
    
    try:
        outfile = open("%s/results/%s" % (cba_html_dir,filename),"w")
    except Exception, e:
        return 0
    lines = fileObject.file.readlines()
    outfile.writelines(lines)
    outfile.close()
    return 1


def parseParameters(names,identifiers,chains,files,email,processID,gapOpen,gapExtend,lock_gapOpen,lock_gapExtend,distanceThreshold):
    errors = []
    domains = []
    user_filenames = {}
    num_structures = 0
    if email.value=="":
        errors.append("Please enter an email address. Your address will not be stored, and we will contact you only to inform you that your results are ready.")
    try:
        float(gapOpen)
    except:
        errors.append("Value for the gap opening penalty must be numeric (%s)" % gapOpen)
    try:
        float(gapExtend)
    except:
        errors.append("Value for the gap extension penalty must be numeric (%s)" % gapExtend)
    try:
        float(lock_gapOpen)
    except:
        errors.append("Value for the LOCK 2 gap opening penalty must be numeric (%s)" % lock_gapOpen)
    try:
        float(lock_gapExtend)
    except:
        errors.append("Value for the LOCK 2 gap extension penalty must be numeric (%s)" % lock_gapExtend)
    try:
        float(distanceThreshold)
    except:
        errors.append("Value for the LOCK 2 distance threshold must be numeric (%s)" % distanceThreshold)
    
        
    for i in range(len(names)):
        if identifiers[i].value!="": #specified as pdb or scop identifier
            if names[i].value!="":
                name = names[i].value
            else:
                name = identifiers[i].value
            if chains[i].value=="":
                domains.append((identifiers[i].value,chains[i].value,name,0))
            else:
                domains.append(("%s-%s" % (identifiers[i].value,chains[i].value),chains[i].value,name,0))
            if files[i].filename!="":
                errors.append("Structure %i: Both a file name and a PDB/SCOP identifier were specified. Please leave one of the two fields blank." % (i+1))                
            num_structures += 1
        elif files[i].file:
            if files[i].filename=="": continue
            if names[i].value!="":
                name = names[i].value
            else:
                name = "%i" % (i+1)
            status = uploadFile(files[i], "%s/%s.txt" % (processID,name))
            if not status:
                if status==-1:
                    errors.append("File name (%s) contains illegal characters." % files[i].filename)
                else:
                    errors.append("File %s was not uploaded successfully" % files[i].filename)
            domains.append(("%s/results/%s/%s.txt" % (cba_html_dir,processID,name),chains[i].value,name,1))
            user_filenames[num_structures] = files[i].filename
            if identifiers[i].value!="":
                errors.append("Structure %i: Both a file name and a PDB/SCOP identifier were specified. Please leave one of the two fields blank." % (i+1))
            num_structures += 1
    if num_structures==0:
        errors.append("No valid structures were entered")

    return domains,errors,user_filenames


def reportErrors(errors,domains,identifiers,user_filenames,email):
    print "Content-type: text/html\n"
    
    num_structures = len(domains)
    #open the front page file and print it out with the information the user entered filled in
    infile = open("%s/index.html" % cba_html_dir,"r")
    line=infile.readline()
    while line:
        if re.search("<br><br>",line): #breaks after the CBA logo image
            print line
            print "<font color='red' size=+2><b>Errors occurred:</b></font>\n"
            print "<i><font size=-1>Please note that if you specified files to be uploaded, most browsers will require that you enter them again before resubmitting.</font></i><br>\n"
            print "<ul>\n"
            for error in errors:
                print "<li><font color='red'>%s</font>" % error
            print "</ul><br>\n"

        elif re.search("Enter your email address",line):
            print line
            print "<td><input type=\'text\' name='email' value='%s'></td>" % email.value
            line=infile.readline()
        elif re.search("structure 1",line):
            for i in range(num_structures):
                print "<!-- structure %i -->\n" % (i+1)
                print " <tr>\n  <td>%i&nbsp;</td>" % (i+1)
                if not domains[i][3]:
                    print "  <td align='center'><input type='text' name='identifier' size='7' value='%s'></td>" % identifiers[i].value
                    print "  <td>&nbsp;</td>"
                    print "  <td><input type='file' name='file'></td>"
                else:
                    print "  <td align='center'><input type='text' name='identifier' size='7'></td>"
                    print "  <td>&nbsp;</td>"
                    print "  <td align='center'><input type='file' name='file' value='%s'></td>" % user_filenames[i]

                print "  <td><input type='text' name='chain' size='1' value='%s'></td>" % domains[i][1]
                if domains[i][3]:
                    print "  <td><input type='text' name='name' size='10'></td>"
                else:
                    print "  <td><input type='text' name='name' size='10' value='%s'></td>" % domains[i][2]

                print " </tr>"
                
            for i in range(num_structures,10):
                print '''<!-- structure %i --> 
<tr>
  <td>%i.&nbsp;</td>
  <td align=center><input type="text" name="identifier" size="7"></td>
  <td>&nbsp;</td>
  <td><input type="file" name="file"></td>
  <td><input type="text" name="chain" size="1"></td>
  <td><input type="text" name="name" size="10"></td>

</tr>
''' % (i+1,i+1)
            while line:
                if re.search("colspan=5",line):
                    print "<tr>\n"
                    print line
                    break
                line=infile.readline()
            
        else:
            print line
        line=infile.readline()

            
    infile.close()


def checkForFile(domain,resultssubdir):
    identifier = domain[0]
    if len(identifier)==7: #SCOP domain
        if os.path.exists("%s/%s/%s.ent" % (scopdir,identifier[2:4],identifier)):
            os.system("/bin/cp %s/%s/%s.ent %s/results/%s/%s.ent" % (scopdir,identifier[2:4],identifier,cba_html_dir,resultssubdir,identifier))
            return 1,domain
        else:
            os.system("/bin/cp %s/%s/%s.ent %s/results/%s/%s-%s.ent" % (scopdir,identifier[2:4],identifier,cba_html_dir,resultssubdir,identifier[:7],identifier[8]))            
            return 0,domain
    elif len(identifier)==9:
        if os.path.exists("%s/%s/%s.ent" % (scopdir,identifier[2:4],identifier[0:7])):
            os.system("/bin/cp %s/%s/%s.ent %s/results/%s/%s.ent" % (scopdir,identifier[2:4],identifier[0:7],cba_html_dir,resultssubdir,identifier))
            return 1,domain
        else:
            return 0,domain
    elif len(identifier)==4:
        if os.path.exists("%s/%s/%s.pdb" % (pdbdir,identifier[1:3],identifier)):
            os.system("/bin/cp %s/%s/%s.pdb %s/results/%s/%s.ent.gz" % (pdbdir,identifier[1:3],identifier,cba_html_dir,resultssubdir,identifier))
            os.system("%s -f %s/results/%s/%s.ent" % (gunzip,cba_html_dir,resultssubdir,identifier))         
            return 1,domain
        else: return 0,domain
    elif len(identifier)==5:

        if os.path.exists("%s/%s/%s.pdb" % (pdbdir,identifier[1:3],identifier[:4])):
            os.system("/bin/cp %s/%s/%s.pdb %s/results/%s/%s-%s.ent.gz" % (pdbdir,identifier[1:3],identifier[:4],cba_html_dir,resultssubdir,identifier[:4],identifier[4]))
            os.system("%s -f %s/results/%s/%s-%s.ent" % (gunzip,cba_html_dir,resultssubdir,identifier[:4],identifier[4]))
            domain = ("%s-%s" % (identifier[:4],identifier[4]),identifier[4],domain[2],domain[3])
            return 1,domain
        else: return 0,domain
    elif len(identifier)==6:

        m = re.match("(....)-(.)",identifier)
        if m:
            pdbid = m.group(1)
            chain = m.group(2)
            if os.path.exists("%s/%s/%s.pdb" % (pdbdir,pdbid[1:3],pdbid)):
                os.system("/bin/cp %s/%s/%s.pdb %s/results/%s/%s-%s.ent.gz" % (pdbdir,pdbid[1:3],pdbid[:4],cba_html_dir,resultssubdir,pdbid,chain))
                os.system("%s -f %s/results/%s/%s-%s.ent" % (gunzip,cba_html_dir,resultssubdir,pdbid,chain))
#                domain = ("%s" % (pdbid,chain),domain[1],domain[2],domain[3])
                return 1,domain
            else: return 0,domain
        else:
            return -1,domain
    return -1,domain

def printDomainList(domains,resultssubdir):
    try:
        outfile = open("%s/results/%s/domains.txt" % (cba_html_dir,resultssubdir),"w")
    except Exception, e:
        return -1
    for domain in domains:
        if domain[3]: #an uploaded file
            if domain[1]!="": #chain was specified
                outfile.write("%s %s\n" % (domain[0],domain[1]))
            else:
                outfile.write("%s\n" % domain[0])
        else:
            if len(domain[0])==7: #scop
                outfile.write("%s/results/%s/%s.ent\n" % (cba_html_dir,resultssubdir,domain[0]))
            elif len(domain[0])==9: #scop with chain specified
                outfile.write("%s/results/%s/%s.ent %s\n" % (cba_html_dir,resultssubdir,domain[0],domain[1]))
            elif len(domain[0])==4: #pdb id, no chain
                outfile.write("%s/results/%s/%s.ent\n" % (cba_html_dir,resultssubdir,domain[0]))
            elif len(domain[0])==5: #pdb id with chain; no dash
                if domain[1]!="": #chain was specified
                    outfile.write("%s/results/%s/%s-%s.ent %s\n" % (cba_html_dir,resultssubdir,domain[0][0:4],domain[0][4],domain[1]))
                else:
                    outfile.write("%s/results/%s/%s-%s.ent\n" % (cba_html_dir,resultssubdir,domain[0][0:4],domain[0][4]))                    
            elif len(domain[0])==6: #pdb id with chain appended with dash
                outfile.write("%s/results/%s/%s.ent %s\n" % (cba_html_dir,resultssubdir,domain[0],domain[0][-1]))
    outfile.close()
    return 1
    

def printResultsPage(processID):
    #need to check to see if errors occurred
    try:
        infile = open("%s/results/%s/runlog.txt" % (cba_html_dir,processID),"r")
        outfile = open("%s/results/%s/errors.txt" % (cba_html_dir,processID),"w")
        opened_runlog = 1
    except:
        opened_runlog = 0

    if opened_runlog:
        lines = infile.readlines()
        infile.close()
        
        for line in lines:
            if re.match("Error:",line):
                outfile.write(line)
        outfile.close()
            
    
        
    #prepare results for user
    try:
        outfile = open("%s/results/%s/index.html" % (cba_html_dir,processID),"w")
    except:
        return "<font size=+2 color='red'>Error: Couldn't open results file for printing</font><br>"
        

    try:
        infile = open("%s/results/%s/cba.html" % (cba_html_dir,processID),"r")
    except:
        if not os.path.exists("%s/results/%s/cba.html" % (cba_html_dir,processID)):
            return "<font size=+2 color='red'>Error: CBA run failed for ID %s</font><br>" % processID
        else:
            return "<font size=+2 color='red'>Error: Couldn't read CBA results for ID %s</font><br>" % processID

    lines = infile.readlines()
    infile.close()

    for i in range(len(lines)):
        if re.search("<BODY>",lines[i]):
            lines = lines[i+1:-1]
            break

    outstring =  '''
    <HTML>
      <HEAD>
        <style type="text/css">
        .helix {background-color:#FF3399}
        .strand {background-color:#FFFF99}
        .alignment {font-family:"Courier"}
        </style>
        <title>CBA: Consistency Based Alignment</title>        
      </HEAD>
      <BODY>
       <a href="/CBA/"><img src="/CBA/logo-large.png" border=0></a><br>
       <br>
       Download results files:<br>
       <a href="/%s/results/%s/cba.html">Sequence Alignment (html)</a><br>
       <a href="/%s/results/%s/cba.txt">Sequence Alignment (text)</a><br>
       <a href="/%s/results/%s/cba.pdb">Structure Alignment</a><br>
       <a href="/%s/results/%s/runlog.txt">Log</a>&nbsp;<i><font size=-1>(Potentially useful in case of errors)</font></i><br>
       <br>
       Please send comments, questions, and bug reports to <a href="mailto:%s">%s</a>
       <hr>
       ''' % (cba_html_subdir,processID,cba_html_subdir,processID,cba_html_subdir,processID,cba_html_subdir,processID,customer_service_email_address,customer_service_contact)       


    outfile.write(outstring)
        
            
    outfile.writelines(lines)
    outfile.write("</BODY>\n</HTML>\n");


    outfile.close()

    return ""


def notifyUser(processID,email,errors=None):
    smtpserver = 'localhost'
    recipients = [email]
    sender = customer_service_email_address

    if errors==None:    
        msg = "To: %s\r\nFrom: %s (CBA Server)\r\nSubject: Your CBA results are ready\r\n" % (email,sender)
        msg += "Your CBA results are now available at http://cba.stanford.edu/results/%s\r" % processID
    else:
        msg = "To: %s\r\nFrom: %s (CBA Server)\r\nSubject: CBA encountered an error\r\n" % (email,sender)
        msg += "CBA encountered %i error(s) when running job id %s:\r\n\r\n" % (len(errors),processID)
        for i in range(len(errors)):
            msg += "  (%i) %s\r\n" % (i+1,errors[i])
        msg += "\r\n"
        
    msg += "    Please send any comments, bugs, or questions to mailto:%s.\r" % (customer_service_email_address)
    msg += "."
##    msg = '''To: %s
##    From: %s (CBA Server)
##    Subject: Your CBA results are ready
##    Your CBA results are now available at http://cba.stanford.edu/results/%s.
##    Please send any comments, bugs, or questions to jessicas@stanford.edu.
##    .
##    ''' % (email,sender, processID)


    try:
        session = smtplib.SMTP(smtpserver)
        session.sendmail(sender,recipients,msg)
        session.quit()
    except:
        return "<font size=+2 color='red'>Error: Couldn't send email to user.</font><br>"

    return ""


def main():
    form = cgi.FieldStorage()
    identifiers = form["identifier"]
    names = form["name"]
    chains = form["chain"]
    files = form["file"]
    email = form["email"]
    gapOpen = form["gapOpen"]
    gapExtend = form["gapExtend"]
    lock_gapOpen = form["gapOpenLOCK"]
    lock_gapExtend = form["gapExtendLOCK"]
    distanceThreshold = form["distanceThreshold"]

    #make results directory
    processID = int(os.getpid())
    while os.path.exists("%s/results/%i" % (cba_html_dir,processID)):
        processID += 1
    try:
    	os.mkdir("%s/results/%s" % (cba_html_dir,processID))
    except:
	printError("<font size=+2 color='red'>Can't create results/%s</font><br>" % processID)

    gapOpen = gapOpen.value
    gapExtend = gapExtend.value
    lock_gapOpen = lock_gapOpen.value
    lock_gapExtend = lock_gapExtend.value
    distanceThreshold = distanceThreshold.value
    domains,errors,user_filenames = parseParameters(names,identifiers,chains,files,email,processID,gapOpen,gapExtend,lock_gapOpen,lock_gapExtend,distanceThreshold)



    num_structures = len(domains)

    if len(errors)!=0:
        reportErrors(errors,domains,identifiers,user_filenames,email)
        status = notifyUser(processID,email.value,errors=errors)
#        os.system("/bin/rm -rf %s/results/%s" % (cba_html_dir,processID))
        sys.exit(0)

    #check to make sure PDB/SCOP identifiers are in our local database
    errors = []
    for i in range(len(domains)):
        domain = domains[i]
        if not domain[3]: # if not file
            status,domains[i] = checkForFile(domain,processID)
            if status==0:
                errors.append("Structure %i: We do not have a PDB file for %s in our local database. Please upload the file." % (i+1,domain[0]))
            elif status==-1:
                errors.append("Structure %i: '%s' does not appear to be a valid PDB or SCOP identifier." % (i+1,domain[0]))


    if len(errors)!=0:
        reportErrors(errors,domains,identifiers,user_filenames,email)
        status = notifyUser(processID,email.value,errors=errors)
#        os.system("/bin/rm -rf %s/results/%s" % (cba_html_dir,processID))
        sys.exit(0)


    #print wait page
    print "Content-type: text/html\n"
    print '''
    <HTML>
    <HEAD>
      <title>CBA: Consistency Based Alignment"</title>
    </HEAD>
    <BODY>
       <a href="/CBA/"><img src="/CBA/logo-large.png" border=0></a><br>
       <br>
       Your job has been queued. You will be notified by email when your results are ready to be viewed <a href="/%s/results/%s">here</a>. (Link goes to results directory contents until the run finishes. Refresh occasionally until you see an alignment.)<br>
       <hr>
       Please send comments, questions, and bug reports to <a href="%s">%s</a>
    </BODY>
    </HTML>
    ''' % (cba_html_subdir,processID,customer_service_email_address,customer_service_contact)
    sys.stdout.flush()

    #everything looks good. Run CBA
    status = printDomainList(domains,processID)



    try:
        os.system("%s %s/results/%s/domains.txt %s/results/%s/ %s -o %s -e %s -lo %s -le %s -d %s> %s/results/%s/runlog.txt 2> %s/results/%s/errors.txt" % (cba_program,cba_html_dir,processID,cba_html_dir,processID,processID,gapOpen,gapExtend,lock_gapOpen,lock_gapExtend,distanceThreshold,cba_html_dir,processID,cba_html_dir,processID))
    except:
        notifyUser(processID,email.value,errors=['Could not run CBA.'])
        print "couldn't run cba"
#        os.system("/bin/rm -rf %s/results/%s" % (cba_html_dir,processID))
        sys.exit(-1)

        
    status = printResultsPage(processID)
    if status !="":
        notifyUser(processID,email.value,errors=['Run finished, but could not print results page.'])
#        os.system("/bin/rm -rf %s/results/%s" % (cba_html_dir,processID))

        sys.exit(-1)

    #Our work here is done. Notify the user.
    status = notifyUser(processID,email.value)
    if status!="":
        print "<br> %s" % status
        print "Done. Click <a href='/%s/results/%s'>here</a> for results." % (cba_html_subdir,processID)
        

    
main()
