/* Copyright Amit P. Singh, Douglas L. Brutlag, Stanford University. */

#ifdef DISPLAYON

#include <stdio.h>
#include <stdlib.h>
#include <tcl.h>
#include <tk.h>


extern int isatty _ANSI_ARGS_((int fd));
extern int read _ANSI_ARGS_((int fd, char *buf, size_t size));
extern char *strrchr _ANSI_ARGS_((CONST char *string, int c));

void tellTcl(int, char*, int);
void runTclFile(char*);
void tellTclAnything(char*);
int Tcl_AppInit(Tcl_Interp *interp);


Tcl_Interp *interp;	/* Interpreter for this application. */
Tcl_DString command;	/* Used to assemble lines of terminal input
				 * into Tcl commands. */
Tcl_DString line;	/* Used to read the next line from the
                                 * terminal input. */
int tty;			/* Non-zero means standard input is a
				 * terminal-like device.  Zero means it's
				 * a file. */


extern void setCurrentOne(int);
extern void setCurrentTwo(int);
extern void centerObjects();
extern void structAlignSingle();
extern void structAlignFull();
extern void alignUsingVectorsSingle();
extern void alignUsingVectorsFull();
extern void initAlignUsingVectorsSingle();
extern void initAlignUsingVectorsFull();
extern void sequenceAlign();
extern void quitFunction();
extern void setLocalOrGlobal(int);
extern void setDistThreshold(float);
extern void printOutput();
extern void coreAlignFull();
extern void coreAlignSingle();
extern void setThresholdVar(int);
extern void doAllAlignments();
extern void setDisplayType(int);

int alignStep1(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  initAlignUsingVectorsSingle(NULL,NULL,0);
  alignUsingVectorsFull(NULL,NULL,0);
  return TCL_OK;
}
int alignStep2(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  structAlignFull(NULL,NULL);
  return TCL_OK;
}
int alignStep3(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  coreAlignFull(NULL,NULL);
  structAlignFull(NULL,NULL);
  coreAlignFull(NULL,NULL);
  return TCL_OK;
}


int setCurrentOne_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {

  if ((argc == 2) && (argv[1] != NULL)) {
    setCurrentOne(atoi(argv[1]));
  }  

  return TCL_OK;
}

int setCurrentTwo_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {

  if ((argc == 2) && (argv[1] != NULL)) {
    setCurrentTwo(atoi(argv[1]));
  }  
  return TCL_OK;
}

int centerObjects_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  
  centerObjects();
  return TCL_OK;
}


int structAlignSingle_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  structAlignSingle(NULL,NULL);
  return TCL_OK;
}

int structAlignFull_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  structAlignFull(NULL,NULL,0);
  return TCL_OK;
}

int alignUsingVectorsSingle_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  alignUsingVectorsSingle(NULL,NULL,0);
  return TCL_OK;
}


int alignUsingVectorsFull_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  alignUsingVectorsFull(NULL,NULL,0);
  return TCL_OK;
}

int initAlignUsingVectorsSingle_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  initAlignUsingVectorsSingle(NULL,NULL,0);
  return TCL_OK;
}

int initAlignUsingVectorsFull_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  initAlignUsingVectorsFull(NULL,NULL,0);
  return TCL_OK;
}

int sequenceAlign_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  sequenceAlign();
  return TCL_OK;
}

int quitFunction_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  quitFunction();
  return TCL_OK;
}

int setLocalOrGlobal_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {

  if ((argc == 2) && (argv[1] != NULL)) {
    setLocalOrGlobal(atoi(argv[1]));
  }    
  return TCL_OK;

}

int setDistThreshold_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {

  if ((argc == 2) && (argv[1] != NULL)) {
    setDistThreshold(atof(argv[1]));
  }    
  return TCL_OK;
}

int printOutput_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  printOutput(NULL,NULL,0,0);
  return TCL_OK;
}


int coreAlignFull_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  coreAlignFull(NULL,NULL);
  return TCL_OK;
}

int coreAlignSingle_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  coreAlignSingle(NULL,NULL);
  return TCL_OK;
}

int setDisplayType_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  
  if ((argc == 2) && (argv[1] != NULL)) {
    setDisplayType(atoi(argv[1]));
  }    
  return TCL_OK;
}

int setThresholdVar_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {

  if ((argc == 2) && (argv[1] != NULL)) {
    setThresholdVar(atoi(argv[1]));
  }    
  return TCL_OK;
}

int doAllAlignments_tcl(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  doAllAlignments();
  return TCL_OK;
}


int Tcl_AppInit(Tcl_Interp *interp) {
  char str[200];
  
  if (Tcl_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if (Tk_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }
  Tcl_StaticPackage(interp, "Tk", Tk_Init, (Tcl_PackageInitProc *) NULL); 
  
  Tcl_CreateCommand(interp, "tclAlignStep1", alignStep1, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclAlignStep2", alignStep2, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclAlignStep3", alignStep3, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);

  Tcl_CreateCommand(interp, "tclkeyOne", setCurrentOne_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclkeyTwo", setCurrentTwo_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclkeyFour", centerObjects_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  
  Tcl_CreateCommand(interp, "tclkeya", structAlignSingle_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclkeyA", structAlignFull_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclkeyv", alignUsingVectorsSingle_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclkeyV", alignUsingVectorsFull_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclkeyb", initAlignUsingVectorsSingle_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclkeyB", initAlignUsingVectorsFull_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclkeyS", sequenceAlign_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclkeyQ", quitFunction_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclkeyL", setLocalOrGlobal_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclkeyT", setDistThreshold_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclkeyR", printOutput_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclkeyF", coreAlignFull_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclkeyf", coreAlignSingle_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclkeySPACE", setDisplayType_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tcltoggleT", setThresholdVar_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "tclFullAlign", doAllAlignments_tcl, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
  
  
  Tcl_SetVar(interp, "tcl_rcFileName", "~/.wishrc", TCL_GLOBAL_ONLY);
  
  sprintf (str, "wm withdraw .\n");
  Tcl_Eval (interp, str);
  
  
  return TCL_OK;
}


void tellTclAnything(char *t) {
  Tcl_Eval (interp, t);

  
}

void runTclFile(char *t) {
  int retvalue;
  retvalue = Tcl_EvalFile(interp, t);        


}

void tellTcl(int QorT, char *name, int val) {

  char str[200];

  if (QorT == 0) {

    sprintf (str, "set object1 %d\n", val);
    Tcl_Eval (interp, str);

    sprintf (str, "set queryname %s\n", name);
    Tcl_Eval (interp, str);
  }

  else if (QorT == 1) {
    
    sprintf (str, "set object2 %d\n", val);
    Tcl_Eval (interp, str);

    sprintf (str, "set targetname %s\n", name);
    Tcl_Eval (interp, str);
  }

}




/*
 *----------------------------------------------------------------------
 *
 * My_Tk_Init --
 *
 *	Main program for Wish and most other Tk-based applications.
 *
 * Results:
 *	None. This procedure never returns (it exits the process when
 *	it's done.
 *
 * Side effects:
 *	This procedure initializes the Tk world and then starts
 *	interpreting commands;  almost anything could happen, depending
 *	on the script being interpreted.
 *
 *----------------------------------------------------------------------
 */

int My_Tk_Init(int argc, char **argv, Tcl_AppInitProc* appInitProc)
     /* Number of arguments. */
     /* Array of argument strings. */
     /* Application-specific initialization
      * procedure to call after most
      * initialization but before starting
      * to execute commands. */
{
    char *args, *fileName;
    char buf[20];
    int code;
    size_t length;
    Tcl_Channel inChannel, outChannel, errChannel, chan;


    Tcl_FindExecutable(argv[0]);
    interp = Tcl_CreateInterp();

    /*
     * Parse command-line arguments.  A leading "-file" argument is
     * ignored (a historical relic from the distant past).  If the
     * next argument doesn't start with a "-" then strip it off and
     * use it as the name of a script file to process.
     */

    
    fileName = NULL;
    if (argc > 1) {
      length = strlen(argv[1]);
      if ((length >= 2) && (strncmp(argv[1], "-file", length) == 0)) {
	argc--;
	argv++;
      }
    }
    if ((argc > 1) && (argv[1][0] != '-')) {
      fileName = argv[1];
      argc--;
      argv++;
    }
    

    /*
     * Make command-line arguments available in the Tcl variables "argc"
     * and "argv".
     */

    args = Tcl_Merge(argc-1, argv+1);
    Tcl_SetVar(interp, "argv", args, TCL_GLOBAL_ONLY);
    ckfree(args);
    sprintf(buf, "%d", argc-1);
    Tcl_SetVar(interp, "argc", buf, TCL_GLOBAL_ONLY);
    Tcl_SetVar(interp, "argv0", (fileName != NULL) ? fileName : argv[0],
	    TCL_GLOBAL_ONLY);

    tty = isatty(0);

    Tcl_SetVar(interp, "tcl_interactive",
	    ((fileName == NULL) && tty) ? "1" : "0", TCL_GLOBAL_ONLY);

    /*
     * Invoke application-specific initialization.
     */

    if ((*appInitProc)(interp) != TCL_OK) {
	errChannel = Tcl_GetStdChannel(TCL_STDERR);
	if (errChannel) {
            Tcl_Write(errChannel,
		    "application-specific initialization failed: ", -1);
            Tcl_Write(errChannel, interp->result, -1);
            Tcl_Write(errChannel, "\n", 1);
        }
    }

    /*
     * Invoke the script specified on the command line, if any.
     */

    if (fileName != NULL) {
	code = Tcl_EvalFile(interp, fileName);
	if (code != TCL_OK) {
	    goto error;
	}
	tty = 0;
    }


    outChannel = Tcl_GetStdChannel(TCL_STDOUT);
    if (outChannel) {
      Tcl_Flush(outChannel);
    }
    Tcl_DStringInit(&command);
    Tcl_DStringInit(&line);
    Tcl_ResetResult(interp);
    
    return 1;
    


error:
    /*
     * The following statement guarantees that the errorInfo
     * variable is set properly.
     */

    Tcl_AddErrorInfo(interp, "");
    errChannel = Tcl_GetStdChannel(TCL_STDERR);
    if (errChannel) {
        Tcl_Write(errChannel, Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY),
		-1);
        Tcl_Write(errChannel, "\n", 1);
    }
    Tcl_DeleteInterp(interp);
    Tcl_Exit(1);
    return 1;
}

#endif /* DISPLAYON */		   
