/* Copyright Amit P. Singh, Douglas L. Brutlag, Stanford University. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "defines.h"
#include "ssalign.h"
#include "transform.h"
/*#include "nw.h"*/




#ifdef DISPLAYON
#include <GL/gl.h>
#endif

#define ORIG_DIST_FUNC 0
#define EQUAL_SSE_WEIGHT 1

typedef struct stackElementStruct {

  int Rval, Cval;
  struct stackElementStruct* nextPtr;

} stackElement;

typedef struct stackTypeStruct {
  stackElement* topOfStack;
  int stackSize;
} stackType;

typedef struct resultsStruct {
  float score;           
  int numptrs;
  int Rptr[MAXPTRS], Cptr[MAXPTRS];
  int aligned;

  int bestPointerIndex;

} results;


int colA, rowB;
results* data;
stackType myStack;
float maxScore;
int maxIndex[2];
OBJ *rowObject, *colObject;

extern int runStep1;

typedef	struct	{
		unsigned int dim, max_dim;
		float *ve;
		} VEC;

typedef	struct	{
		unsigned int m, n;
		unsigned int max_m, max_n, max_size;
		float **me,*base;	/* base is base of alloc'd mem */
		} MAT;

typedef struct seqStruct {
  int ss[MAXLENGTH]; /*array of secondary structures: 1 entry per SSE*/
  float vector[MAXLENGTH][3];
  float origin[MAXLENGTH][3];
  float end[MAXLENGTH][3];
  float length[MAXLENGTH];
  float vector_q1[MAXLENGTH][3];
  float vector_center[MAXLENGTH][3];
  float vector_q3[MAXLENGTH][3];
  int alignment[MAXLENGTH];
  char *objSeq;
} seq;

typedef struct alignmentsStruct {
  float *scores;
  int **rowpointers;
  int **columnpointers;
  int *pointers;
  float **dp_scores;
  int numAlignments;
  int index;
} alignments;

typedef struct originScoresStruct{
  float distance;
  float theta;
  float phi;
} *originScores;

originScores **originScoresA;
originScores **originScoresB;

/*some flags*/
int globalFull = 0;
static int all_pairs = 1;
static int chaining = 0;

static seq seqA, seqB;
static seq seqA_orig, seqB_orig;
float orig_query_mat[16];
float orig_target_mat[16];

alignments all_alignments;
int this_transform_oldR;
int this_transform_oldC;
float this_transform_score = 0.0;
float first_alignment_score; /*score when align_ss is called with full==0*/

int num_hashes_tested = 0;
int num_pairs = 0;
/*int globalCalledAddPair;*/

char letA_ss[] = "HS";
char letB_ss[] = "HS";
int numA_ss = 2;
int numB_ss = 2;
float penalty_ss[] = {0, 0, 0, 0};

int bestR, bestC, bestOldR, bestOldC;
float bestFullScore = 0.0;

int secondBestR, secondBestC, secondBestOldR, secondBestOldC;
float secondBestScore = 0.0;

int tempTotalPairs = 0;
float highestPossibleScore = 0.0;
float currentD = 0.0;

float distFuncMax = 12.0;

int local=1;
extern int globalUseSSEWeights;


extern OBJ* currentObjList[10];
extern int numCurrentObjs;
extern OBJ* ObjList[10];
extern int numObjs;
extern char* getFileName(int);

float distanceBetweenResidues(OBJ* obj, int i, int j);
void writeDistances(OBJ*, OBJ*); /* used to get distribution of vector distances within individual proteins*/
void writePair_ss(int, int, OBJ* query, OBJ* target);
void traceBack(int,int,OBJ*,OBJ*);
void printAlignment_ss(stackElement *, int,OBJ *query, OBJ *target);
void computeData_ss(int, int, seq*, seq*);
void computeData_onlyAngles_allPairs(int, int, seq*, seq*);
void computeData_onlyAngles_sequentialPairs(int, int, seq*, seq*);
float computePenalty_ss(int, int, int, int);
float computeScore_ss(int, int, seq*, seq*);
/*void printScores_ss(OBJ*, OBJ*);*/





float computeAngleScore(int, int, int, int, seq*, seq*);
float computeOriginScore(int, int, int, int, seq*, seq*);
void computeAllOriginScores(seq *myseqA, seq* myseqB);
float computeLengthScore(int, int, seq*, seq*);
float testFullAlign(int, int, int, int, seq*, seq*);

void foundHelix(int, int, int, int, int);
void foundSheet(int, int, int, int, int);

int compare_alignments(const void *a, const void *b);
float distance_Point_to_Segment( float* P, float* P1, float* P2);
float SegmentSegmentIntersect(Vector, Vector, Vector, Vector,float);
float distance3D_Segment_to_Segment( Vector P1, Vector P2, Vector P3, Vector P4, float*, float*);
float VectorNorm2(Vector v);
float dot(Vector v, Vector v2);

/*Utility functions*/
void initStackElement(stackElement *s);
void initstackType();
void readStack(stackElement *sArray);
int getStackSize();
void pushStack(stackElement *p);
stackElement* popStack();
void fixMax(int row, int column);

/*setAllPairs: If flag is 1 (default), all pairs will be used in geometric hashing.
 *If flag is 0, either the query or target SSE pair must be sequential in sequence
 *space. This decreases the running time of the algorithm, but may yield a less
 *accurate alignment.
 */
void setAllPairs(int flag) {
  printf("ssalign.c: setting all pairs to %i\n",flag);
  all_pairs = flag;
}


float getTransformationScore(int ordinal) {
  if(all_alignments.numAlignments ==0) {
    return 0;
  }
  if(ordinal >=all_alignments.numAlignments) {
    return 0;
  }
  return(all_alignments.scores[all_alignments.pointers[ordinal]]);
}

float getCurrentAlignmentScore() {

  return(maxScore);
}

float getCurrentTransformationScore() {
    if(all_alignments.numAlignments ==0) {
    return 0;
  }
  return(all_alignments.scores[all_alignments.pointers[all_alignments.index]]);
}
void setAlignment(int ordinal) {
  int index = all_alignments.pointers[ordinal];

  bestOldR=all_alignments.rowpointers[index][0];
  bestR=all_alignments.rowpointers[index][1];
  bestOldC = all_alignments.columnpointers[index][0];
  bestC = all_alignments.columnpointers[index][1];
  all_alignments.index = ordinal;
}

int increaseTransformationIndex() {
  int index;
  int retvalue = 1;

  all_alignments.index++;
  printf("increased index: %i\n",all_alignments.index);
  if(all_alignments.index >=all_alignments.numAlignments) {
    return -1;
  }
  index = all_alignments.pointers[all_alignments.index];
  printf("transformation number %i: ",index);
  bestOldR = all_alignments.rowpointers[index][0];
  bestR = all_alignments.rowpointers[index][1];
  bestOldC = all_alignments.columnpointers[index][0];
  bestC = all_alignments.columnpointers[index][1];
  printf("%i %i %i %i\n",bestOldR,bestR,bestOldC,bestC);
  printf("changed to index %i: rows = %i %i, columns = %i %i, score = %f\n",index,bestOldR,bestR,bestOldC,bestC,all_alignments.scores[index]);
  return 1;
}

int getTransformationIndex() {
  return(all_alignments.index);
}




void foundHelix(int start, int end, int num, int maxnum, int rc) {

  float x1, y1, z1;
  float x2, y2, z2;
  float quarter_lengthx,quarter_lengthy,quarter_lengthz;

  if (end - start == 3) {
    if (end < maxnum) {
      end++;
    }

    if (start > 0) {
      start--;
    }
  }

  else if (end - start == 4) {
    if (end < maxnum) {
      end++;
    }
    else if (start > 0) {
      start--;
    }
  }

  if (rc == 0) {
    
    x1 = ((colObject->points[start][0])*0.7422 + colObject->points[start+1][0] + 
	  colObject->points[start+2][0] + (colObject->points[start+3][0])*0.7422) * 0.26722;
    y1 = ((colObject->points[start][1])*0.7422 + colObject->points[start+1][1] + 
	  colObject->points[start+2][1] + (colObject->points[start+3][1])*0.7422) * 0.26722;
    z1 = ((colObject->points[start][2])*0.7422 + colObject->points[start+1][2] + 
	  colObject->points[start+2][2] + (colObject->points[start+3][2])*0.7422) * 0.26722;
    
    
    x2 = ((colObject->points[end-1][0])*0.7422 + colObject->points[end-2][0] + 
	  colObject->points[end-3][0] + (colObject->points[end-4][0])*0.7422) * 0.26722;
    y2 = ((colObject->points[end-1][1])*0.7422 + colObject->points[end-2][1] + 
	  colObject->points[end-3][1] + (colObject->points[end-4][1])*0.7422) * 0.26722;
    z2 = ((colObject->points[end-1][2])*0.7422 + colObject->points[end-2][2] + 
	  colObject->points[end-3][2] + (colObject->points[end-4][2])*0.7422) * 0.26722;
  
  
    seqA.ss[num] = 0;
    seqA.origin[num][0] = x1;
    seqA.origin[num][1] = y1;
    seqA.origin[num][2] = z1;

    seqA.end[num][0] = x2;
    seqA.end[num][1] = y2;
    seqA.end[num][2] = z2;

    /********
    seqA.vector_center[num][0] = (x1+x2)/2.0;
    seqA.vector_center[num][1] = (y1+y2)/2.0;
    seqA.vector_center[num][2] = (z1+z2)/2.0;

    quarter_lengthx = (x2-x1)/4.0;
    quarter_lengthy = (y2-y1)/4.0;
    quarter_lengthz = (z2-z1)/4.0;

    seqA.vector_q1[num][0] = x1+quarter_lengthx;
    seqA.vector_q1[num][1] = y1+quarter_lengthy;
    seqA.vector_q1[num][2] = z1+quarter_lengthz;

    seqA.vector_q3[num][0] = x2-quarter_lengthx;
    seqA.vector_q3[num][1] = y2-quarter_lengthy;
    seqA.vector_q3[num][2] = z2-quarter_lengthz;
******/    
    seqA.length[num] = sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1));
    
    seqA.vector[num][0] = (x2 - x1)/seqA.length[num];
    seqA.vector[num][1] = (y2 - y1)/seqA.length[num];
    seqA.vector[num][2] = (z2 - z1)/seqA.length[num];


  }

  else {
    x1 = ((rowObject->points[start][0])*0.7422 + rowObject->points[start+1][0] + 
	  rowObject->points[start+2][0] + (rowObject->points[start+3][0])*0.7422) * 0.26722;
    y1 = ((rowObject->points[start][1])*0.7422 + rowObject->points[start+1][1] + 
	  rowObject->points[start+2][1] + (rowObject->points[start+3][1])*0.7422) * 0.26722;
    z1 = ((rowObject->points[start][2])*0.7422 + rowObject->points[start+1][2] + 
	  rowObject->points[start+2][2] + (rowObject->points[start+3][2])*0.7422) * 0.26722;
    
    
    x2 = ((rowObject->points[end-1][0])*0.7422 + rowObject->points[end-2][0] + 
	  rowObject->points[end-3][0] + (rowObject->points[end-4][0])*0.7422) * 0.26722;
    y2 = ((rowObject->points[end-1][1])*0.7422 + rowObject->points[end-2][1] + 
	  rowObject->points[end-3][1] + (rowObject->points[end-4][1])*0.7422) * 0.26722;
    z2 = ((rowObject->points[end-1][2])*0.7422 + rowObject->points[end-2][2] + 
	  rowObject->points[end-3][2] + (rowObject->points[end-4][2])*0.7422) * 0.26722;
  
  
    seqB.ss[num] = 0;
    seqB.origin[num][0] = x1;
    seqB.origin[num][1] = y1;
    seqB.origin[num][2] = z1;

    seqB.end[num][0] = x2;
    seqB.end[num][1] = y2;
    seqB.end[num][2] = z2;
    
    /*******
    seqA.vector_center[num][0] = (x1+x2)/2.0;
    seqA.vector_center[num][1] = (y1+y2)/2.0;
    seqA.vector_center[num][2] = (z1+z2)/2.0;

    quarter_lengthx = (x2-x1)/4.0;
    quarter_lengthy = (y2-y1)/4.0;
    quarter_lengthz = (z2-z1)/4.0;

    seqA.vector_q1[num][0] = x1+quarter_lengthx;
    seqA.vector_q1[num][1] = y1+quarter_lengthy;
    seqA.vector_q1[num][2] = z1+quarter_lengthz;

    seqA.vector_q3[num][0] = x2-quarter_lengthx;
    seqA.vector_q3[num][1] = y2-quarter_lengthy;
    seqA.vector_q3[num][2] = z2-quarter_lengthz;
    ******/
    seqB.length[num] = sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1));
    
    seqB.vector[num][0] = (x2 - x1)/seqB.length[num];
    seqB.vector[num][1] = (y2 - y1)/seqB.length[num];
    seqB.vector[num][2] = (z2 - z1)/seqB.length[num];

  }

}

void foundSheet(int start, int end, int num, int maxnum, int rc) {

  float x1, y1, z1;
  float x2, y2, z2;
  float quarter_lengthx,quarter_lengthy,quarter_lengthz;

  if (rc == 0) {
    x1 = (colObject->points[start][0] + colObject->points[start+1][0]) * 0.5;
    y1 = (colObject->points[start][1] + colObject->points[start+1][1]) * 0.5;
    z1 = (colObject->points[start][2] + colObject->points[start+1][2]) * 0.5;
    
    
    x2 = (colObject->points[end-1][0] + colObject->points[end-2][0]) * 0.5;
    y2 = (colObject->points[end-1][1] + colObject->points[end-2][1]) * 0.5;
    z2 = (colObject->points[end-1][2] + colObject->points[end-2][2]) * 0.5;
    
    
    seqA.ss[num] = 1;
    seqA.origin[num][0] = x1;
    seqA.origin[num][1] = y1;
    seqA.origin[num][2] = z1;
    
    seqA.end[num][0] = x2;
    seqA.end[num][1] = y2;
    seqA.end[num][2] = z2;

    /******
    seqA.vector_center[num][0] = (x1+x2)/2.0;
    seqA.vector_center[num][1] = (y1+y2)/2.0;
    seqA.vector_center[num][2] = (z1+z2)/2.0;

    quarter_lengthx = (x2-x1)/4.0;
    quarter_lengthy = (y2-y1)/4.0;
    quarter_lengthz = (z2-z1)/4.0;

    seqA.vector_q1[num][0] = x1+quarter_lengthx;
    seqA.vector_q1[num][1] = y1+quarter_lengthy;
    seqA.vector_q1[num][2] = z1+quarter_lengthz;

    seqA.vector_q3[num][0] = x2-quarter_lengthx;
    seqA.vector_q3[num][1] = y2-quarter_lengthy;
    seqA.vector_q3[num][2] = z2-quarter_lengthz;
    ********/
    seqA.length[num] = sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1));
    
    seqA.vector[num][0] = (x2 - x1)/seqA.length[num];
    seqA.vector[num][1] = (y2 - y1)/seqA.length[num];
    seqA.vector[num][2] = (z2 - z1)/seqA.length[num];

  }

  else {
    x1 = (rowObject->points[start][0] + rowObject->points[start+1][0]) * 0.5;
    y1 = (rowObject->points[start][1] + rowObject->points[start+1][1]) * 0.5;
    z1 = (rowObject->points[start][2] + rowObject->points[start+1][2]) * 0.5;
    
    
    x2 = (rowObject->points[end-1][0] + rowObject->points[end-2][0]) * 0.5;
    y2 = (rowObject->points[end-1][1] + rowObject->points[end-2][1]) * 0.5;
    z2 = (rowObject->points[end-1][2] + rowObject->points[end-2][2]) * 0.5;
    
    
    seqB.ss[num] = 1;
    seqB.origin[num][0] = x1;
    seqB.origin[num][1] = y1;
    seqB.origin[num][2] = z1;
    
    seqB.end[num][0] = x2;
    seqB.end[num][1] = y2;
    seqB.end[num][2] = z2;

    seqA.vector_center[num][0] = (x1+x2)/2.0;
    seqA.vector_center[num][1] = (y1+y2)/2.0;
    seqA.vector_center[num][2] = (z1+z2)/2.0;

    quarter_lengthx = (x2-x1)/4.0;
    quarter_lengthy = (y2-y1)/4.0;
    quarter_lengthz = (z2-z1)/4.0;

    seqA.vector_q1[num][0] = x1+quarter_lengthx;
    seqA.vector_q1[num][1] = y1+quarter_lengthy;
    seqA.vector_q1[num][2] = z1+quarter_lengthz;

    seqA.vector_q3[num][0] = x2-quarter_lengthx;
    seqA.vector_q3[num][1] = y2-quarter_lengthy;
    seqA.vector_q3[num][2] = z2-quarter_lengthz;

    seqB.length[num] = sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1));
    
    seqB.vector[num][0] = (x2 - x1)/seqB.length[num];
    seqB.vector[num][1] = (y2 - y1)/seqB.length[num];
    seqB.vector[num][2] = (z2 - z1)/seqB.length[num];

  }
}



float distanceBetweenVectors(seq *myseq, int r, int c) {

  float retvalue, dist,dist2;
  float x1, y1, z1, temp;
  float cx1,cy1,cz1,cx2,cy2,cz2;
  float quarter_lengthx1,quarter_lengthy1,quarter_lengthz1,quarter_lengthx2,quarter_lengthy2,quarter_lengthz2; /*one quarter lengths in each coordinate for query and target vectors*/
  float q1x1,q1y1,q1z1,q1x2,q1y2,q1z2; /*first quarter coordinates ("q1") of query and target*/
  float q3x1,q3y1,q3z1,q3x2,q3y2,q3z2; /*third quarter coordinates ("q3") of query and target*/
  float qx1,qy1;
  float tdist;
  int tnum,intersect,i;
  float min,min2;
  float nearest1,nearest2;
  float point[3];
  float mua,mub;
  float Pa[3], Pb[3];
  float origin1[3],end1[3],origin2[3],end2[3];
  float mindists[5];
  float start,end,percent;
  int j;

    
      /*find centers of query and target*/
      cx1 = (myseq->origin[c][0] + myseq->end[c][0])/2.0;
      cy1 = (myseq->origin[c][1] + myseq->end[c][1])/2.0;
      cz1 = (myseq->origin[c][2] + myseq->end[c][2])/2.0;
      cx2 = (myseq->origin[r][0] + myseq->end[r][0])/2.0;
      cy2 = (myseq->origin[r][1] + myseq->end[r][1])/2.0;
      cz2 = (myseq->origin[r][2] + myseq->end[r][2])/2.0;

      
      quarter_lengthx1 = (myseq->end[c][0]-myseq->origin[c][0])/4.0;
      quarter_lengthy1 = (myseq->end[c][1]-myseq->origin[c][1])/4.0;
      quarter_lengthz1 = (myseq->end[c][2]-myseq->origin[c][2])/4.0;
      quarter_lengthx2 = (myseq->end[r][0]-myseq->origin[r][0])/4.0;
      quarter_lengthy2 = (myseq->end[r][1]-myseq->origin[r][1])/4.0;
      quarter_lengthz2 = (myseq->end[r][2]-myseq->origin[r][2])/4.0;
    
      q1x1 = myseq->origin[c][0]+quarter_lengthx1;
      q1y1 = myseq->origin[c][1]+quarter_lengthy1;
      q1z1 = myseq->origin[c][2]+quarter_lengthz1;
      q1x2 = myseq->origin[r][0]+quarter_lengthx2;
      q1y2 = myseq->origin[r][1]+quarter_lengthy2;
      q1z2 = myseq->origin[r][2]+quarter_lengthz2;
      
      q3x1 = myseq->end[c][0]-quarter_lengthx1;
      q3y1 = myseq->end[c][1]-quarter_lengthy1;
      q3z1 = myseq->end[c][2]-quarter_lengthz1;
      q3x2 = myseq->end[r][0]-quarter_lengthx2;
      q3y2 = myseq->end[r][1]-quarter_lengthy2;
      q3z2 = myseq->end[r][2]-quarter_lengthz2;
      

      
    x1 = myseq->origin[c][0] - myseq->origin[r][0];
    y1 = myseq->origin[c][1] - myseq->origin[r][1];
    z1 = myseq->origin[c][2] - myseq->origin[r][2];

    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    mindists[0] = dist;
    dist2 = dist;

    x1= myseq->origin[c][0]-q1x2;
    y1 = myseq->origin[c][1]-q1y2;
    z1 = myseq->origin[c][2]-q1z2;
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[0]) {
      mindists[0] = dist;
    }

    x1 = myseq->origin[c][0] - cx2;
    y1 = myseq->origin[c][1]-cy2;
    z1 = myseq->origin[c][2]-cz2;
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[0]) {
      mindists[0] = dist;
    }

    x1= myseq->origin[c][0]-q3x2;
    y1 = myseq->origin[c][1]-q3y2;
    z1 = myseq->origin[c][2]-q3z2;
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[0]) {
      mindists[0] = dist;
    }

    /*test
    x1 = myseq->origin[c][0]-myseq->end[r][0];
    y1 = myseq->origin[c][1]-myseq->end[r][1];
    z1 = myseq->origin[c][2]-myseq->end[r][2];
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[0]) {
      mindists[0] = dist;
    }*/

    x1 = q1x1-myseq->origin[r][0];
    y1 = q1y1-myseq->origin[r][1];
    z1 = q1z1-myseq->origin[r][2];
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    mindists[1] = dist;

    x1 = q1x1-q1x2;
    y1 = q1y1-q1y2;
    z1 = q1z1-q1z2;
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[1]) {
      mindists[1] = dist;
    }

    x1 = q1x1-cx2;
    y1 = q1y1-cy2;
    z1 = q1z1-cz2;
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[1]) {
      mindists[1] = dist;
    }

    x1 = q1x1-q3x2;
    y1 = q1y1-q3y2;
    z1 = q1z1-q3z2;
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[1]) {
      mindists[1] = dist;
    }

    x1 = q1x1-myseq->end[r][0];
    y1 = q1y1-myseq->end[r][1];
    z1 = q1y1-myseq->end[r][2];
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[1]) {
      mindists[1] = dist;
    }

 


    x1 = cx1 - myseq->origin[r][0];
    y1 = cy1 - myseq->origin[r][1];
    z1 = cz1 - myseq->origin[r][2];
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    mindists[2] = dist;
 

    x1 = cx1-q1x2;
    y1 = cy1-q1y2;
    z1 = cz1-q1z2;
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[2]) {
      mindists[2] = dist;
    }


    x1 = cx1-cx2;
    y1 = cy1-cy2;
    z1 = cz1-cz2;
     dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[2]) {
      mindists[2] = dist;
    }



    x1 = cx1-q3x2;
    y1 = cy1-q3y2;
    z1 = cz1-q3z2;
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[2]) {
      mindists[2] = dist;
    }

    x1 =  cx1 - myseq->end[r][0];
    y1 = cy1 - myseq->end[r][1];
    z1 = cz1 - myseq->end[r][2];
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
     if(dist<mindists[2]) {
      mindists[2] = dist;
    }


    x1 = q3x1-myseq->origin[r][0];
    y1 = q3y1-myseq->origin[r][1];
    z1 = q3z1-myseq->origin[r][2];
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    mindists[3] = dist;


    x1 = q3x1-q1x2;
    y1 = q3y1-q1y2;
    z1 = q3z1-q1z2;
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[3]) {
      mindists[3] = dist;
    }

    x1 = q3x1-cx2;
    y1 = q3y1-cy2;
    z1 = q3z1-cz2;
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[3]) {
      mindists[3] = dist;
    }

    x1 = q3x1-q3x2;
    y1 = q3y1-q3y2;
    z1 = q3z1-q3z2;
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[3]) {
      mindists[3] = dist;
    }

    x1 = q3x1-myseq->end[r][0];
    y1 = q3y1-myseq->end[r][1];
    z1 = q3y1-myseq->end[r][2];
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[3]) {
      mindists[3] = dist;
    }




    x1 = myseq->end[c][0]-q1x2;
    y1 = myseq->end[c][1]-q1y2;
    z1 = myseq->end[c][2]-q1z2;
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    mindists[4] = dist;

    /*test
    x1 = myseq->end[c][0]-myseq->origin[r][0];
    y1 = myseq->end[c][1]-myseq->origin[r][1];
    z1 = myseq->end[c][2]-myseq->origin[r][2];
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[4]) {
      mindists[4] = dist;
    }*/

    x1 = myseq->end[c][0]-cx2;
    y1 = myseq->end[c][1]-cy2;
    z1 = myseq->end[c][2]-cz2;
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[4]) {
      mindists[4] = dist;
    }

    x1 = myseq->end[c][0]-q3x2;
    y1 = myseq->end[c][1]-q3y2;
    z1 = myseq->end[c][2]-q3z2;
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[4]) {
      mindists[4] = dist;
    } 

    x1 = myseq->end[c][0] - myseq->end[r][0];
    y1 = myseq->end[c][1] - myseq->end[r][1];
    z1 = myseq->end[c][2] - myseq->end[r][2];
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[4]) {
      mindists[4] = dist;
    }

    min = mindists[0];
    min2 = 1000;
    for(i=1;i<5;i++) {
      if(mindists[i]<min) {
	min2 = min;
	min = mindists[i];
      }else if(mindists[i]<min2) {
	min2 = mindists[i];
      }
    }

    min = sqrt(min);
    min2 = sqrt(min2);
    min+=min2;
    min/=2.0;
    return(min);
  
}

float distanceBetweenResidues(OBJ* obj, int i, int j) {
  float dist;
  
  dist = (obj->points[i][0]-obj->points[j][0])*(obj->points[i][0]-obj->points[j][0]);
  dist += (obj->points[i][1]-obj->points[j][1])*(obj->points[i][1]-obj->points[j][1]);
  dist += (obj->points[i][2]-obj->points[j][2])*(obj->points[i][2]-obj->points[j][2]);
  return(sqrt(dist));
}

void writeDistances(OBJ* query, OBJ* target){
  FILE *aa, *ab, *bb;
  int i,j,k,l,start1,start2,end1,end2;
  float distance;

  aa=fopen("aa-distances","a");
  ab = fopen("ab-distances","a");
  bb=fopen("bb-distances","a");

  /*do query*/
      /********
      if(seqA.ss[i]==0 && seqA.ss[j]==0) {
	fprintf(aa,"%8.3f\n",distanceBetweenVectors(&seqA,i,j));
      }else if(seqA.ss[i]==1 && seqA.ss[j]==1) {
	fprintf(bb,"%8.3f\n",distanceBetweenVectors(&seqA,i,j));
      }else {
	fprintf(ab,"%8.3f\n",distanceBetweenVectors(&seqA,i,j));
      }
      **********/

  for(i=0;i<colA;i++) {
    for(j=i+1;j<colA;j++) {
      start1 = query->ssArray[2*i];
      start2 = query->ssArray[2*j];
      end1 = query->ssArray[2*i+1];
      end2 = query->ssArray[2*j+1];
      for(k=start1;k<end1;k++) {
	for(l=start2;l<end2;l++) {
	  distance = distanceBetweenResidues(query,k,l);
	  if(seqA.ss[i]==0 && seqA.ss[j]==0) {
	    fprintf(aa,"%.3f ",distanceBetweenResidues(query,k,l));
	  }else if(seqA.ss[i]==1 && seqA.ss[j]==1) {
	    fprintf(bb,"%.3f ",distanceBetweenResidues(query,k,l));
	  }else {
	    fprintf(ab,"%.3f ",distanceBetweenResidues(query,k,l));
	  }
	}
      }
      if(seqA.ss[i]==0 && seqA.ss[j]==0) {
	fprintf(aa,"\n");
      }else if(seqA.ss[i]==1 && seqA.ss[j]==1) {
	fprintf(bb,"\n");
      }else {
	fprintf(ab,"\n");
      }
    }
  }
  

  /*target*/
  /**************
  for(i=0;i<rowB;i++) {
    for(j=i+1;j<rowB;j++) {

      for(k=target->ssArray[2*i];k<target->ssArray[2*i+1];k++) {
	for(l=target->ssArray[2*j];l<target->ssArray[2*j+1];l++) {
	  if(seqA.ss[i]==0 && seqA.ss[j]==0) {
	    fprintf(aa,"%.3f ",distanceBetweenResidues(target,k,l));
	  }else if(seqA.ss[i]==1 && seqA.ss[j]==1) {
	    fprintf(bb,"%.3f ",distanceBetweenResidues(target,k,l));
	  }else {
	    fprintf(ab,"%.3f ",distanceBetweenResidues(target,k,l));
	  }
	}
      }
      if(seqB.ss[i]==0 && seqB.ss[j]==0) {
	fprintf(aa,"\n");
      }else if(seqB.ss[i]==1 && seqB.ss[j]==1) {
	fprintf(bb,"\n");
      }else {
	fprintf(ab,"\n");
      }
    }
  }
  *************/
  fclose(aa);
  fclose(ab);
  fclose(bb);
}

void initvectors_ss(OBJ *query, OBJ *target) {
  int i,j;
  float angle,angle2;
  int done,mergestart;
  float maxangle = 20.0;
  int num_separating = 1;

  maxangle = maxangle*PI/180.0; /*convert to radians*/
  printf("initvectors_ss\n");
  /*merge vectors,changing rowA and colB as necessary*/
  done = 0;
  while(!done) {
    done = 1;
    initsequences_ss(query,target);
    for(i=0;i<colA-1;i++) {
      if(seqA.ss[i] == seqA.ss[i+1]) {
	if((query->ssArray[2*(i+1)]-query->ssArray[2*i+1])<=num_separating) {
	  angle = (float)(acos((double)((seqA.vector[i][0])*(seqA.vector[i+1][0])+
					(seqA.vector[i][1])*(seqA.vector[i+1][1])+
					(seqA.vector[i][2])*(seqA.vector[i+1][2]))));
	  if(angle <=maxangle) { /*merge SSEs*/
	    /*check possibility for merge with next vector*/
	    mergestart = i;
	    if(i<colA-2 && (seqA.ss[i+1] == seqA.ss[i+2])) {
	      if((query->ssArray[2*(i+2)]-query->ssArray[2*(i+1)+1])<=num_separating) {
		angle2 = (float)(acos((double)((seqA.vector[i+1][0])*(seqA.vector[i+2][0])+
					      (seqA.vector[i+1][1])*(seqA.vector[i+2][1])+
					      (seqA.vector[i+1][2])*(seqA.vector[i+2][2]))));
		if(angle2 < angle) {
		  mergestart = i+1;
		  angle = angle2;
		}
	      }
	    }
	    printf("merging query %i (%i to %i) and %i (%i to %i),angle = %f\n",mergestart,query->aaNums[query->ssArray[2*mergestart]],query->aaNums[query->ssArray[2*mergestart+1]-1],mergestart+1,query->aaNums[query->ssArray[2*(mergestart+1)]],query->aaNums[query->ssArray[2*(mergestart+1)+1]-1],angle*180.0/PI);
	    done = 0;
	    for(j=query->ssArray[2*mergestart+1];j<query->ssArray[2*(mergestart+1)];j++) {
	      query->sstype[j] = query->sstype[2*mergestart];
	    }
	    query->ssArray[2*mergestart+1] = query->ssArray[2*(mergestart+1)+1];
	    for(j=2*(mergestart+1);j<2*(colA-1);j++) {
	      query->ssArray[j] = query->ssArray[j+2];
	    }
	    query->ssArrayLength--;
	    if(seqA.ss[mergestart]==0) {
	      query->num_helices--;
	    }else {
	      query->num_strands--;
	    }
	    break;
	  }
	}
	
      }
      
    }
  }
  /*now repeat for target*/
  done = 0;
  while(!done) {
    done =1;
    initsequences_ss(query,target);
    for(i=0;i<rowB-1;i++) {
      if(seqB.ss[i] == seqB.ss[i+1]) {
	if((target->ssArray[2*(i+1)]-target->ssArray[2*i+1])<=num_separating) {
	  angle = (float)(acos((double)((seqB.vector[i][0])*(seqB.vector[i+1][0])+
					(seqB.vector[i][1])*(seqB.vector[i+1][1])+
					(seqB.vector[i][2])*(seqB.vector[i+1][2]))));

	  if(angle <= maxangle) { /*merge SSEs*/

	    /*check possibility for merge with next vector*/
	    mergestart = i;
	    if(i<rowB-2 && (seqB.ss[i+1] == seqB.ss[i+2])) {
	      if((target->ssArray[2*(i+2)]-target->ssArray[2*(i+1)+1])<=num_separating) {
		angle2 = (float)(acos((double)((seqB.vector[i+1][0])*(seqB.vector[i+2][0])+
					      (seqB.vector[i+1][1])*(seqB.vector[i+2][1])+
					      (seqB.vector[i+1][2])*(seqB.vector[i+2][2]))));
		if(angle2 < angle) {
		  mergestart = i+1;
		  angle = angle2;
		}
	      }
	    }
	    printf("merging target %i (%i to %i) and %i (%i to %i), angle =%f\n",mergestart,target->aaNums[target->ssArray[2*mergestart]],target->aaNums[target->ssArray[2*mergestart+1]-1],mergestart+1,target->aaNums[target->ssArray[2*(mergestart+1)]],target->aaNums[target->ssArray[2*(mergestart+1)+1]-1],angle*180.0/PI);
	    done = 0;
	    for(j=target->ssArray[2*mergestart+1];j<target->ssArray[2*(mergestart+1)];j++) {
	      target->sstype[j] = target->sstype[2*mergestart];
	    }
	    target->ssArray[2*mergestart+1] = target->ssArray[2*(mergestart+1)+1];
	    for(j=2*(mergestart+1);j<2*(rowB-1);j++) {
	      target->ssArray[j] = target->ssArray[j+2];
	    }
	    target->ssArrayLength--;
	    if(seqA.ss[mergestart]==0) {
	      target->num_helices--;
	    }else {
	      target->num_strands--;
	    }
	    break;
	  }
	}
	
      }
      
    }
  }

}

void initsequences_ss(OBJ *o1, OBJ *o2) {

  char ch = '0', ch1, ch2;
  char *seq1, *seq2;
  int i;
 
  globalCalledAddPair = 0;

  colObject = o1;
  rowObject = o2;

  seq1 = colObject->sstype;
  seq2 = rowObject->sstype;

  colA = colObject->ssArrayLength;
  rowB = rowObject->ssArrayLength;

  
  o1->num_helices = 0;
  o1->num_strands = 0;
  o2->num_helices = 0;
  o2->num_strands = 0;
  for (i = 0; i < colA; i++) {

    ch = seq1[colObject->ssArray[i*2]];
      
    if (ch == 'H') {
      o1->num_helices++;
      foundHelix(colObject->ssArray[i*2], colObject->ssArray[i*2 + 1], i, colObject->numPts, 0);
    }
    else if (ch == 'S') {
      o1->num_strands++;
      foundSheet(colObject->ssArray[i*2], colObject->ssArray[i*2 + 1], i, colObject->numPts, 0);
    }
        
  }


  for (i = 0; i < rowB; i++) {

    ch = seq2[rowObject->ssArray[i*2]];
      
    if (ch == 'H') {
      o2->num_helices++;
      foundHelix(rowObject->ssArray[i*2], rowObject->ssArray[i*2 + 1], i, rowObject->numPts, 1);
    }
    else if (ch == 'S') {
      o2->num_strands++;
      foundSheet(rowObject->ssArray[i*2], rowObject->ssArray[i*2 + 1], i, rowObject->numPts, 1);
    }
  }

  if(data == NULL) {
    data = (results*) malloc((sizeof(results))*(colA+1)*(rowB+1));
  }


  for (i = 0; i < (colA+1)*(rowB+1); i++) {
    data[i].score = -1000000.0;
    data[i].numptrs = 0;
    data[i].bestPointerIndex = -1;
  }

  maxScore = -1000000.0;
  bestFullScore = -100000.0;
  secondBestScore = -100000.0;
  tempTotalPairs = 0;

  highestPossibleScore = (colA < rowB) ? (40.0 * (float)(colA)) : (40.0 * (float)(rowB));

  /*  currentD = 10.0;*/
  currentD = 10.0;

  bestR = -1;
  bestC = -1;
  bestOldR = -1;
  bestOldC = -1;

  secondBestR = -1;
  secondBestC = -1;
  secondBestOldR = -1;
  secondBestOldC = -1;

  initstackType();
  /*  writeDistances(o1,o2);
  exit(0);*/
}


void writePair_ss(int r, int c,OBJ *query, OBJ *target) {

  int i, myindex, alignSize;
  stackElement *newElement;
  
  newElement = (stackElement*) malloc(sizeof(stackElement));
  initStackElement(newElement);
  newElement->Rval = r;
  newElement->Cval = c;

  pushStack(newElement);
  
  myindex = c + r*colA;
  query->ss_alignment[c][0] = r;
  query->num_SSEs_aligned++;
  query->num_aligned[c]++;
  target->ss_alignment[r][0]=c;
  target->num_SSEs_aligned++;
  target->num_aligned[r]++;
  /*  printf("aligned SSEs: %i %i\n",r,c);*/
  /*  for (i = 0; i < data[myindex].numptrs; i++) {
    writePair_ss(data[myindex].Rptr[i], data[myindex].Cptr[i],query,target);
    }*/
  if(data[myindex].numptrs != 0) {
    writePair_ss(data[myindex].Rptr[data[myindex].bestPointerIndex],
		 data[myindex].Cptr[data[myindex].bestPointerIndex], query, target);
  }

  alignSize = getStackSize();
  
  if ((alignSize > 0) && (data[myindex].numptrs == 0)) {
    newElement = (stackElement*) malloc((sizeof(stackElement))*alignSize);
    for (i = 0; i < alignSize; i++) 
      initStackElement(&(newElement[i]));
    
    readStack(newElement);

    printAlignment_ss(newElement, alignSize,query,target);
    
    free(newElement);
  }

  newElement = popStack();

  free(newElement);

}

GLfloat findNearest_ss(OBJ* target, int targetStart, int window, GLfloat* pVec, int *targetNum, int match_betas) {

  int i, num, start, end;
  GLfloat min = 10000.0;
  GLfloat dist;
  
  for (i = targetStart; i < targetStart + window; i++) {
    
    if(match_betas && target->objSeq[i] != 'G') {
      dist = sqrt((target->cbpoints[i][0] - pVec[0])*(target->cbpoints[i][0] - pVec[0]) +
		  (target->cbpoints[i][1] - pVec[1])*(target->cbpoints[i][1] - pVec[1]) +
		  (target->cbpoints[i][2] - pVec[2])*(target->cbpoints[i][2] - pVec[2]));
    }else {
      dist = sqrt((target->points[i][0] - pVec[0])*(target->points[i][0] - pVec[0]) +
		  (target->points[i][1] - pVec[1])*(target->points[i][1] - pVec[1]) +
		  (target->points[i][2] - pVec[2])*(target->points[i][2] - pVec[2]));
    }
    if (dist < min) {
      min = dist;
      num = i;
    }
  }
  
  *targetNum = num;
  
  return min;
  
}


void printAlignment_ss(stackElement *list, int listSize, OBJ *query, OBJ* target) {

  int i, j, r, c, oldr, oldc, outIndex, tempIndex;
  int order = 0;
  char *out1, *out2;
  int ii, targetstart, targetend, querystart, queryend, difflength, queryPoint, count;
  GLfloat tdist;
  float weight;
  int num_distinct_query_points = 0;
  int num_distinct_target_points = 0;
  int okay_to_match_pairs = 0;
  int prev_query;
  int prev_target;

  /*note: num_distinct_<target or query>_points is not accurate past two distinct
   *points, but this is all that is needed*/

#if SHOWOUTPUT 
  out1 = (char*) malloc((sizeof(char))*(colA + rowB + 20));
  out2 = (char*) malloc((sizeof(char))*(colA + rowB + 20));
  outIndex = 0;
 
  r = list[0].Rval;
  oldr = r - 1;
  c = list[0].Cval;
  oldc = c - 1;

  for (i = 0; i < r - c; i++) {
    out1[outIndex] = '-';
    outIndex++;
  }
  for (i = 0; i < c; i++) {
    out1[outIndex] = letA_ss[seqA.ss[i]];
    outIndex++;
  }
  
  outIndex = 0;
  for (i = 0; i < c - r; i++) {
    out2[outIndex] = '-';
    outIndex++;
  }
  for (i = 0; i < r; i++) {
    out2[outIndex] = letB_ss[seqB.ss[i]];
    outIndex++;
  }

  if (local) {
    out1[outIndex] = '*';
    out2[outIndex] = '*';
    outIndex++;
  }
#endif
  /*everything in this section is aligned (all ss elements), e.g. before '*'*/

    
  for (i = 0; i < listSize; i++) {
    r = list[i].Rval;
    c = list[i].Cval;
#if SHOWOUTPUT 
    for (j = oldr + 1; j < r; j++) {	
      out1[outIndex] = '-';
      out2[outIndex] = letB_ss[seqB.ss[j]];
      outIndex++;
    }
    for (j = oldc + 1; j < c; j++) {	
      out1[outIndex] = letA_ss[seqA.ss[j]];
      out2[outIndex] = '-';
      outIndex++;
    }
    if (seqA.ss[c] == seqB.ss[r]) {
      out1[outIndex] = letA_ss[seqA.ss[c]];
      out2[outIndex] = letB_ss[seqB.ss[r]];
      outIndex++;
    }
    oldr = r;
    oldc = c; 
#endif
    if (seqA.ss[c] == seqB.ss[r]) {

      globalCalledAddPair++;
	
      targetstart = rowObject->ssArray[r*2];
      targetend = rowObject->ssArray[r*2 + 1];

      querystart = colObject->ssArray[c*2];
      queryend = colObject->ssArray[c*2 + 1];

      difflength = (queryend - querystart) - (targetend - targetstart);
      /*      printf("adding SSE pair %i, %i, difflength = %i\n",c,r,difflength);*/
#if EQUAL_SSE_WEIGHT
      /*normalize to shorter SSE -- maybe should do avg. length*/
      if(difflength >= 0) {
	/*	weight = 1.0/((float)(targetend-targetstart+1));*/
	weight = 1.0/((float)(targetend-targetstart));
      }else {
	/*	weight = 1.0/((float)(queryend-querystart+1));*/
	weight = 1.0/((float)(queryend-querystart));
      }

#else 
      weight = 1.0;
#endif
      prev_query = -1;
      prev_target = -1;
      if(target->ssArrayLength < query->ssArrayLength) {
	order = -1;
      }else if((target->ssArrayLength==query->ssArrayLength) && strcmp(query->objSeq,target->objSeq) >0) {
	order = 1;
      }
      /*      order = strcmp(getFileName(0),getFileName(1));*/
      /*      if(difflength==0) {
	printf("order: %i\n",order);
	}*/
      if ((difflength == 0 && order < 0) || difflength > 0) {
	/*	printf("matching target\n");*/
	count = 0;
	for (ii = targetstart; ii < targetend; ii++) {
	  tdist = findNearest_ss(colObject, querystart + count, difflength+1, 
				 rowObject->points[ii], &queryPoint,0);
	  /*	  if(order < 0)*/
	    add_pair(colObject->points[queryPoint], rowObject->points[ii], weight); 
	  /*	  add_pair(colObject->cbpoints[queryPoint], rowObject->cbpoints[ii], weight); */
	    /*	  else
		  add_pair(rowObject->points[ii],colObject->points[queryPoint],weight);*/
	    /*	  printf("adding %i %i, %f, %f\n",queryPoint,ii,tdist,weight);*/
	  globalCalledAddPair++;
	  if(prev_query != queryPoint) {
	    num_distinct_query_points++;
	  }
	  prev_query = queryPoint;
	  count++;
	}

	ii = targetend;
	
	if ((targetend - targetstart == 3) || (queryend - querystart == 3)) {
	  if ((queryPoint+1 < colObject->numPts) && (ii < rowObject->numPts)) {
	    globalCalledAddPair++;

	    add_pair(colObject->points[queryPoint+1], rowObject->points[ii], weight); 
		    /*add_pair(colObject->cbpoints[queryPoint], rowObject->cbpoints[ii], weight); */

	      /*	    printf("adding %i %i, %f, %f\n",queryPoint,ii,tdist,weight);*/
	    num_distinct_query_points++;
	  }
	  else {
	    globalCalledAddPair++;

	    add_pair(colObject->points[queryPoint], rowObject->points[ii-1], weight);
	    /*  add_pair(colObject->cbpoints[queryPoint], rowObject->cbpoints[ii-1], weight); */

	      /*	      	    printf("adding %i %i, %f, %f\n",queryPoint,ii-1,tdist,weight);*/
	  }
	}

	if(num_distinct_query_points >=2) {
	  okay_to_match_pairs = 1;
	}
      }

      else if ((difflength==0 && order>=0) || difflength < 0) {
	/*	printf("matching query\n");*/
	difflength = -1.0 * difflength;
	count = 0;
	prev_query=-1;
	for (ii = querystart; ii < queryend; ii++) {
	  tdist = findNearest_ss(rowObject, targetstart + count, difflength+1, 
				 colObject->points[ii], &queryPoint,0);

	  add_pair(colObject->points[ii], rowObject->points[queryPoint], weight); 
	  /*  add_pair(colObject->cbpoints[ii], rowObject->cbpoints[queryPoint], weight); */


	    /*	  printf("adding %i %i, %f, %f\n",ii,queryPoint,tdist,weight);*/
	  globalCalledAddPair++;
	  if(prev_query == -1) {
	    num_distinct_query_points++;
	  }else if(prev_query != queryPoint) {
	    num_distinct_query_points++;
	  }
	  prev_query = queryPoint;
	  
	  count++;
	}

	ii = queryend;
	
	if ((targetend - targetstart == 3) || (queryend - querystart == 3)) {
	  if ((queryPoint+1 < rowObject->numPts) && (ii < colObject->numPts)) {
	    globalCalledAddPair++;

	     add_pair(colObject->points[ii], rowObject->points[queryPoint+1], weight);
		 /*		 add_pair(colObject->cbpoints[ii], rowObject->cbpoints[queryPoint], weight); */

	      /*	    printf("adding %i %i, %f, %f\n",ii,queryPoint,tdist,weight);*/
	    num_distinct_query_points++;
	  }
	  else {
	    globalCalledAddPair++;

	    add_pair(colObject->points[ii-1], rowObject->points[queryPoint], weight); 
	    /*	      add_pair(colObject->cbpoints[ii-1], rowObject->cbpoints[queryPoint], weight); */

	      /*	    printf("adding %i %i, %f, %f\n",ii-1,queryPoint,tdist,weight);*/
	  }
	}

	if(num_distinct_query_points>=2) {
	  okay_to_match_pairs = 1;
	}
      }
    }
  }
#if SHOWOUTPUT 
  if (local) {
    out1[outIndex] = '*';
    out2[outIndex] = '*';
    outIndex++;
  }

  tempIndex = outIndex;
  for (i = c + 1; i < colA; i++) {
    out1[outIndex] = letA_ss[seqA.ss[i]];
    outIndex++;
  }
  for (i = 0; i < (rowB - r) - (colA - c); i++) {
    out1[outIndex] = '-';
    outIndex++;
  }

  outIndex = tempIndex;
  for (i = r + 1; i < rowB; i++) {
    out2[outIndex] = letB_ss[seqB.ss[i]];
    outIndex++;
  }
  for (i = 0; i < (colA - c) - (rowB - r); i++) {
    out2[outIndex] = '-';
    outIndex++;
  }

  out1[outIndex] = '\0';
  out2[outIndex] = '\0'; 

  printf("%s\n", out1);
  printf("%s\n", out2);

  printf("\n");


  free(out1);
  free(out2);
  if(!okay_to_match_pairs) {
    globalCalledAddPair = 0;
  }
  printf("Secondary structure alignment:\n");
#endif
}

float angleBetweenVectors(int c, int r) {
  float theta,returnvalue;
  /* <a,b> = |a||b|cos(theta), where theta is the angle between a and b*/
  theta = (float)(acos((double)((seqA.vector[c][0])*(seqB.vector[r][0])+
				(seqA.vector[c][1])*(seqB.vector[r][1])+
				(seqA.vector[c][2])*(seqB.vector[r][2]))));
  /*convert from radians to degrees*/
  returnvalue = 180*theta/PI;
  return(returnvalue);
}


/*Score S1*/
float computeAngleScore(int r, int c, int oldr, int oldc, seq *myseqA, seq *myseqB) {
  
  float retvalue, theta, phi;

  if (myseqA->ss[oldc] != myseqB->ss[oldr]) {
    retvalue = 0.0;
  }
  else {
    theta = (float)(acos((double)((myseqA->vector[c][0])*(myseqA->vector[oldc][0]) + 
				  (myseqA->vector[c][1])*(myseqA->vector[oldc][1]) + 
				  (myseqA->vector[c][2])*(myseqA->vector[oldc][2]))));
    
    phi = (float)(acos((double)((myseqB->vector[r][0])*(myseqB->vector[oldr][0]) + 
				(myseqB->vector[r][1])*(myseqB->vector[oldr][1]) + 
				(myseqB->vector[r][2])*(myseqB->vector[oldr][2]))));
    
    
    if (theta > phi) {
      theta = theta - phi;
    }
    else {
      theta = phi - theta;
    }
    
    
    retvalue = (4.0*(theta))/PI;

    retvalue = retvalue*retvalue;
    
    retvalue = (28.0/(1.0 + retvalue)) - 14.0;
    all_alignments.dp_scores[r][0] = retvalue;
  }
  return retvalue;

}

void computeAllOriginScores(seq *myseqA, seq* myseqB) {
  int r,c,oldr,oldc;
  float x1,y1,z1,dist,theta,phi;

  for(oldr=0;oldr<rowB-1;oldr++) {
    for(r=oldr+1;r<rowB;r++) {
      x1 = seqB_orig.origin[r][0]-seqB_orig.end[oldr][0];
      y1 = seqB_orig.origin[r][1]-seqB_orig.end[oldr][1];
      z1 = seqB_orig.origin[r][2]-seqB_orig.end[oldr][2];
      dist = sqrt((x1*x1) + (y1*y1) + (z1*z1)); 
      originScoresB[oldr][r]->distance = dist;
      theta = (acos((double)(seqB_orig.vector[oldr][0]*x1 +
			      seqB_orig.vector[oldr][1]*y1 +
			      seqB_orig.vector[oldr][2]*z1)));
      phi = (acos((double)(seqB_orig.vector[r][0]*x1 +
			    seqB_orig.vector[r][1]*y1 +
			    seqB_orig.vector[r][2]*z1)));
      originScoresB[oldr][r]->theta = theta;
      originScoresB[oldr][r]->phi = phi;
    }
  }

  for(oldc=0;oldc<colA-1;oldc++) {
    for(c=oldr+1;c<colA;c++) {
      x1 = seqA_orig.origin[c][0]-seqA_orig.end[oldc][0];
      y1 = seqA_orig.origin[c][1]-seqA_orig.end[oldc][1];
      z1 = seqA_orig.origin[c][2]-seqA_orig.end[oldc][2];
      dist = sqrt((x1*x1) + (y1*y1) + (z1*z1)); 
      originScoresA[oldc][c]->distance = dist;
      theta = (acos((double)(seqA_orig.vector[oldc][0]*x1 +
			      seqA_orig.vector[oldc][1]*y1 +
			      seqA_orig.vector[oldc][2]*z1)));
      phi = (acos((double)(seqA_orig.vector[c][0]*x1 +
			    seqA_orig.vector[c][1]*y1 +
			    seqA_orig.vector[c][2]*z1)));
      originScoresA[oldc][c]->theta = theta;
      originScoresA[oldc][c]->phi = phi;
    }
  }

  
  
}

float computeOriginScore(int r, int c, int oldr, int oldc, seq *myseqA, seq *myseqB) {
  
  double retvalue, colDist, rowDist, x1, y1, z1, theta, theta2, phi, phi2;

  if (myseqA->ss[oldc] != myseqB->ss[oldr]) {
    retvalue = 0.0;
    return(retvalue);
  }
  if(originScoresA[oldc][c] == NULL) {
    x1 = seqA_orig.origin[c][0]-seqA_orig.end[oldc][0];
    y1 = seqA_orig.origin[c][1]-seqA_orig.end[oldc][1];
    z1 = seqA_orig.origin[c][2]-seqA_orig.end[oldc][2];
    
    colDist = sqrt((x1*x1) + (y1*y1) + (z1*z1));
    /* x1 = myseqA->origin[c][0] - myseqA->end[oldc][0];
       y1 = myseqA->origin[c][1] - myseqA->end[oldc][1];
       z1 = myseqA->origin[c][2] - myseqA->end[oldc][2];*/
    
    
    x1 = x1/colDist;
    y1 = y1/colDist;
    z1 = z1/colDist;
    
    /*    theta = (acos((double)((myseqA->vector[oldc][0])*(x1) + 
	  (myseqA->vector[oldc][1])*(y1) + 
	  (myseqA->vector[oldc][2])*(z1))));*/
    theta = (acos((double)((seqA_orig.vector[oldc][0])*x1 +
			   seqA_orig.vector[oldc][1]*y1 +
			   seqA_orig.vector[oldc][2] *z1)));
    
    
    /*    phi = (acos((double)((myseqA->vector[c][0])*(x1) + 
	  (myseqA->vector[c][1])*(y1) + 
	  (myseqA->vector[c][2])*(z1))));*/
    phi = (acos((double)((seqA_orig.vector[c][0])*x1 +
			 seqA_orig.vector[c][1]*y1 +
			 seqA_orig.vector[c][2] *z1)));
    
    originScoresA[oldc][c] = (originScores)malloc(sizeof(struct originScoresStruct));
    originScoresA[oldc][c]->distance = colDist;
    originScoresA[oldc][c]->phi = phi;
    originScoresA[oldc][c]->theta = theta;
  }else {
    phi = originScoresA[oldc][c]->phi;
    theta = originScoresA[oldc][c]->theta;
    colDist = originScoresA[oldc][c]->distance;
  }
  
    if(originScoresB[oldr][r] == NULL) {
      /*try calculating rowdist from orig sequences*/
      x1 = seqB_orig.origin[r][0]-seqB_orig.end[oldr][0];
      y1 = seqB_orig.origin[r][1]-seqB_orig.end[oldr][1];
      z1 = seqB_orig.origin[r][2]-seqB_orig.end[oldr][2];
      
      rowDist = sqrt((x1*x1) + (y1*y1) + (z1*z1));    
      
      /*    x1 = myseqB->origin[r][0] - myseqB->end[oldr][0];
	    y1 = myseqB->origin[r][1] - myseqB->end[oldr][1];
	    z1 = myseqB->origin[r][2] - myseqB->end[oldr][2];*/
      
      
      x1 = x1/rowDist;
      y1 = y1/rowDist;
      z1 = z1/rowDist;
      
      
      /*    theta2 =(acos((double)((myseqB->vector[oldr][0])*(x1) + 
	    (myseqB->vector[oldr][1])*(y1) + 
	    (myseqB->vector[oldr][2])*(z1))));
	    
	    phi2 = (acos((double)((myseqB->vector[r][0])*(x1) + 
	    (myseqB->vector[r][1])*(y1) + 
	    (myseqB->vector[r][2])*(z1))));*/
      theta2 = (acos((double)(seqB_orig.vector[oldr][0]*x1 +
			      seqB_orig.vector[oldr][1]*y1 +
			      seqB_orig.vector[oldr][2]*z1)));
      phi2 = (acos((double)(seqB_orig.vector[r][0]*x1 +
			    seqB_orig.vector[r][1]*y1 +
			    seqB_orig.vector[r][2]*z1)));
    
      originScoresB[oldr][r] = (originScores)malloc(sizeof(struct originScoresStruct));
      originScoresB[oldr][r]->distance = rowDist;
      originScoresB[oldr][r]->theta = theta2;
      originScoresB[oldr][r]->phi = phi2;
    }else {
      rowDist = originScoresB[oldr][r]->distance;
      theta2 = originScoresB[oldr][r]->theta;
      phi2 = originScoresB[oldr][r]->phi;
    }
      
    if (colDist > rowDist) {
      colDist = colDist - rowDist;
    }
    else {
      colDist = rowDist - colDist;
    }
      
      
    retvalue = colDist/5.0;
    retvalue = retvalue*retvalue;
    all_alignments.dp_scores[r][3] = retvalue;
    /*    all_alignments.dp_scores[r][3] = (float)colDist;*/
    retvalue = (4.0/(1.0 + retvalue)) - 2.0; /*score 4: distance between vectors*/
    /*    if(globalFull)
	  printf("%f ",retvalue);*/
    if (theta > theta2) {
      theta = theta - theta2;
    }
    else {
      theta = theta2 - theta;
    }
    
    
    theta = (4.0*(theta))/PI;
    /*    all_alignments.dp_scores[r][1] = (float)(theta*45.0);*/

    theta = theta*theta;

    retvalue += (4.0/(1.0 + theta)) - 2.0; /*score s2*/
    all_alignments.dp_scores[r][1] = (4.0/(1.0 + theta)) - 2.0;
    /*    if(globalFull)
	  printf("%f ",(4.0/(1.0+theta))-2.0);*/
    if (phi > phi2) {
      phi = phi - phi2;
    }
    else {
      phi = phi2 - phi;
    }
    
    
    phi = (4.0*(phi))/PI;
    /*    all_alignments.dp_scores[r][2] = (float)(phi*45.0);*/
    phi = phi*phi;

    retvalue += (4.0/(1.0 + phi)) - 2.0;    /*score s3*/
    all_alignments.dp_scores[r][2] = (4.0/(1.0 + phi)) - 2.0;
    /*    if(globalFull)
	  printf("%f\n",(4.0/(1.0+phi))-2.0);*/


  return ((float)retvalue);

}

/*score 5: length of vectors*/
float computeLengthScore(int r, int c, seq *myseqA, seq *myseqB) {

  float retvalue, temp;


  if (myseqA->length[c] > myseqB->length[r]) {
    temp = ((myseqA->length[c] - myseqB->length[r])/7.0);
    /*    all_alignments.dp_scores[r][4] = temp*7.0;*/
    temp = temp*temp;
    retvalue = (10.0/(1.0 + temp)) - 5.0;
    all_alignments.dp_scores[r][4] = retvalue;

  }
  else {
    temp = ((myseqB->length[r] - myseqA->length[c])/7.0);
    all_alignments.dp_scores[r][4] = temp*7.0;
    temp = temp*temp;
    retvalue = (10.0/(1.0 + temp)) - 5.0;
    all_alignments.dp_scores[r][4] = retvalue;
  }

  return retvalue;
}

/*score 6: angle between aligned vectors*/
float computeScore_ss(int r, int c, seq *myseqA, seq *myseqB) {

  float retvalue;
  float x1, y1, z1,temp;


  if (myseqA->ss[c] != myseqB->ss[r]) {
    retvalue = -100000.0;
  }
  else {
    
    x1 = (float)((myseqA->vector[c][0])*(myseqB->vector[r][0]) + 
		 (myseqA->vector[c][1])*(myseqB->vector[r][1]) + 
		 (myseqA->vector[c][2])*(myseqB->vector[r][2]));

    if (x1 > 1.0) {
      x1 = 1.0;
    }

    x1 = (float)(acos((double)(x1)));

    temp = (4.0*x1)/PI;

    temp = temp*temp;

    retvalue = (20.0/(1.0 + temp)) - 10.0;
    all_alignments.dp_scores[r][5] = retvalue;
  }

  return retvalue;
}



float computeAvgDistScore_ss(int r, int c, seq *myseqA, seq *myseqB) {

  float retvalue, dist,dist2;
  float x1, y1, z1, temp;
  float cx1,cy1,cz1,cx2,cy2,cz2;
  float quarter_lengthx1,quarter_lengthy1,quarter_lengthz1,quarter_lengthx2,quarter_lengthy2,quarter_lengthz2; /*one quarter lengths in each coordinate for query and target vectors*/
  float q1x1,q1y1,q1z1,q1x2,q1y2,q1z2; /*first quarter coordinates ("q1") of query and target*/
  float q3x1,q3y1,q3z1,q3x2,q3y2,q3z2; /*third quarter coordinates ("q3") of query and target*/
  float qx1,qy1;
  float tdist;
  int tnum,intersect,i;
  float min,min2;
  float nearest1,nearest2;
  float point[3];
  float mua,mub;
  float Pa[3], Pb[3];
  float origin1[3],end1[3],origin2[3],end2[3];
  float mindists[5][5];
  float start,end,percent,dist3,dist4,new_min1,new_min2;
  int j,increased_mub,decreased_mub;
  int minrow, mincol, minrow2, mincol2;
  int calcdists[5][5]; 

  if (myseqA->ss[c] != myseqB->ss[r]) {
    retvalue = -100000.0;
  }
  else {
    for(i=0;i<5;i++) {
      memset(calcdists[i],1,5*sizeof(int));
      for(j=0;j<5;j++) {
	mindists[i][j] = 10000;
      }
    }
      /*find centers of query and target*/
      cx1 = (myseqA->origin[c][0] + myseqA->end[c][0])/2.0;
      cy1 = (myseqA->origin[c][1] + myseqA->end[c][1])/2.0;
      cz1 = (myseqA->origin[c][2] + myseqA->end[c][2])/2.0;
      cx2 = (myseqB->origin[r][0] + myseqB->end[r][0])/2.0;
      cy2 = (myseqB->origin[r][1] + myseqB->end[r][1])/2.0;
      cz2 = (myseqB->origin[r][2] + myseqB->end[r][2])/2.0;

      
      quarter_lengthx1 = (myseqA->end[c][0]-myseqA->origin[c][0])/4.0;
      quarter_lengthy1 = (myseqA->end[c][1]-myseqA->origin[c][1])/4.0;
      quarter_lengthz1 = (myseqA->end[c][2]-myseqA->origin[c][2])/4.0;
      quarter_lengthx2 = (myseqB->end[r][0]-myseqB->origin[r][0])/4.0;
      quarter_lengthy2 = (myseqB->end[r][1]-myseqB->origin[r][1])/4.0;
      quarter_lengthz2 = (myseqB->end[r][2]-myseqB->origin[r][2])/4.0;
    
      q1x1 = myseqA->origin[c][0]+quarter_lengthx1;
      q1y1 = myseqA->origin[c][1]+quarter_lengthy1;
      q1z1 = myseqA->origin[c][2]+quarter_lengthz1;
      q1x2 = myseqB->origin[r][0]+quarter_lengthx2;
      q1y2 = myseqB->origin[r][1]+quarter_lengthy2;
      q1z2 = myseqB->origin[r][2]+quarter_lengthz2;
      
      q3x1 = myseqA->end[c][0]-quarter_lengthx1;
      q3y1 = myseqA->end[c][1]-quarter_lengthy1;
      q3z1 = myseqA->end[c][2]-quarter_lengthz1;
      q3x2 = myseqB->end[r][0]-quarter_lengthx2;
      q3y2 = myseqB->end[r][1]-quarter_lengthy2;
      q3z2 = myseqB->end[r][2]-quarter_lengthz2;
      

      
    x1 = myseqA->origin[c][0] - myseqB->origin[r][0];
    y1 = myseqA->origin[c][1] - myseqB->origin[r][1];
    z1 = myseqA->origin[c][2] - myseqB->origin[r][2];

    /*    dist = sqrt((x1*x1) + (y1*y1) + (z1*z1));*/
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    mindists[0][0] = dist;
    minrow = 0;
    mincol = 0;



    x1= myseqA->origin[c][0]-q1x2;
    y1 = myseqA->origin[c][1]-q1y2;
    z1 = myseqA->origin[c][2]-q1z2;
    /*    dist = sqrt((x1*x1) + (y1*y1) + (z1*z1));*/
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    mindists[0][1] = dist;
    if(dist < mindists[0][mincol]) {
      mincol = 1;
    }else {
      if(dist > mindists[0][0]) {
	for(i=2;i<5;i++) {
	  calcdists[0][i] = 0;
	  mindists[0][i] = 1000;
	}
      }
    }

    
    if(calcdists[0][2]) {
      x1 = myseqA->origin[c][0] - cx2;
      y1 = myseqA->origin[c][1]-cy2;
      z1 = myseqA->origin[c][2]-cz2;
      /*    dist = ((x1*x1) + (y1*y1) + (z1*z1));*/
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[0][2] = dist;
      if(dist < mindists[0][mincol]) {
	mincol = 2;
      }
      if(dist > mindists[0][1]) { 
	  calcdists[0][3] = 0;
	  calcdists[0][4] = 0;
	  mindists[0][3] = 1000;
	  mindists[0][4]  = 1000;
      }
    }

    if(calcdists[0][3]) {
      x1= myseqA->origin[c][0]-q3x2;
      y1 = myseqA->origin[c][1]-q3y2;
      z1 = myseqA->origin[c][2]-q3z2;
      /*    dist = sqrt((x1*x1) + (y1*y1) + (z1*z1));*/
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[0][3] = dist;
      if(dist < mindists[0][mincol]) {
	mincol = 3;
      }
    }

    mindists[0][4] = 1000;

    x1 = q1x1-myseqB->origin[r][0];
    y1 = q1y1-myseqB->origin[r][1];
    z1 = q1z1-myseqB->origin[r][2];
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    mindists[1][0] = dist;
    if(dist < mindists[minrow][mincol]) {
      minrow = 1;
      mincol = 0;
    }
    if(dist > mindists[0][0]) {
      for(i=2;i<5;i++) {
	calcdists[i][0] = 0;
	mindists[i][0] = 1000;
      }
    }

    if(calcdists[1][1]) { /*don't need check b/c [1][1] can't be zero yet?*/
      x1 = q1x1-q1x2;
      y1 = q1y1-q1y2;
      z1 = q1z1-q1z2;
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      if(dist < mindists[minrow][mincol]) {
	minrow = 1;
	mincol = 1;
      }
      mindists[1][1] = dist;
      if(calcdists[0][1]) {
	if(dist > mindists[0][1]) {
	  for(i=2;i<5;i++) {
	    calcdists[i][1] = 0;
	    mindists[i][1] = 1000;
	  }
	}
      }
      if(calcdists[1][0]) {
	if(dist > mindists[1][0]) {
	  for(i=2;i<5;i++) {
	    calcdists[1][i] = 0;
	    mindists[1][i] = 1000;
	  }
	}
      }
    }


    if(calcdists[1][2]) {
      x1 = q1x1-cx2;
      y1 = q1y1-cy2;
      z1 = q1z1-cz2;
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[1][2] = dist;
      if(dist < mindists[minrow][mincol]) {
	minrow = 1;
	mincol = 2;
      } 
      if(calcdists[0][2]) {
	if(dist > mindists[0][2]) {
	  for(i=2;i<5;i++) {
	    calcdists[i][2] = 0;
	    mindists[i][2] = 1000;
	  }
	}
      }
      if(calcdists[1][1]) {
	if(dist > mindists[1][1]) {
	  for(i=3;i<5;i++) {
	    calcdists[1][i] = 0;
	    mindists[1][i] = 1000;
	  }
	}
      }
    }

    if(calcdists[1][3]) {
      x1 = q1x1-q3x2;
      y1 = q1y1-q3y2;
      z1 = q1z1-q3z2;
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      if(dist < mindists[minrow][mincol]) {
	minrow = 1;
	mincol = 3;
      }
      mindists[1][3] = dist;
      if(calcdists[0][3]) {
	if(dist > mindists[0][3]) {
	  for(i=2;i<5;i++) {
	    calcdists[i][3] = 0;
	    mindists[i][3] = 1000;
	  }
	}
      }
      if(calcdists[1][2]) {
	if(dist > mindists[1][2]) {
	  calcdists[1][4] = 0;
	  mindists[1][4] = 1000;
	}
      }
    }
    
    if(calcdists[1][4]) {
      x1 = q1x1-myseqB->end[r][0];
      y1 = q1y1-myseqB->end[r][1];
      z1 = q1z1-myseqB->end[r][2];
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[1][4] = dist;

      if(dist < mindists[minrow][mincol]) {
	minrow = 1;
	mincol = 4;
      }
      if(calcdists[0][4]) {
	if(dist > mindists[0][4]) {
	  for(i=2;i<5;i++) {
	    calcdists[i][4] = 0;
	    mindists[i][4] = 1000;
	  }
	}
      }
    }
 

    if(calcdists[2][0]) {
      x1 = cx1 - myseqB->origin[r][0];
      y1 = cy1 - myseqB->origin[r][1];
      z1 = cz1 - myseqB->origin[r][2];
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[2][0] = dist;
      if(dist < mindists[minrow][mincol]) {
	minrow = 2;
	mincol = 0;
      }
      if(dist > mindists[1][0]) {
	for(i=3;i<5;i++) {
	  calcdists[i][0] = 0;
	  mindists[i][0] = 1000;
	}
      }
    }

    if(calcdists[2][1]) {
      x1 = cx1-q1x2;
      y1 = cy1-q1y2;
      z1 = cz1-q1z2;
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[2][1] = dist;
      if(dist < mindists[minrow][mincol]) {
	minrow = 2;
	mincol = 1;
      }
      if(dist > mindists[1][1]) {
	for(i=3;i<5;i++) {
	  calcdists[i][1] = 0;
	  mindists[i][1] = 1000;
	}
      }
      if(dist > mindists[2][0]) {
	for(i=2;i<5;i++) {
	  calcdists[2][i] = 0;
	  mindists[2][i] = 1000;
	}
      }
    }

    if(calcdists[2][2]) {
      x1 = cx1-cx2;
      y1 = cy1-cy2;
      z1 = cz1-cz2;
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[2][2] = dist;
      if(dist < mindists[minrow][mincol]) {
	minrow = 2;
	mincol = 2;
      }
      dist2 = sqrt(dist)/3.0;
      if(dist > mindists[2][1]) {
	for(i=3;i<5;i++) {
	  calcdists[2][i] = 0;
	  mindists[2][i] = 1000;
	}
      }
      if(dist > mindists[1][2]) {
	for(i=3;i<5;i++) {
	  calcdists[i][2] = 0;
	  mindists[2][i] = 1000;
	}
      }
    }

    if(calcdists[2][3]) {
      x1 = cx1-q3x2;
      y1 = cy1-q3y2;
      z1 = cz1-q3z2;
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[2][3] = dist;
      if(dist < mindists[minrow][mincol]) {
	minrow = 2;
      mincol = 3;
      }
      if(dist > mindists[1][3]) {
	calcdists[3][3] = 0;
	mindists[3][3] = 1000;
	calcdists[4][3] = 0;
	mindists[4][3] = 1000;
      }
      if(dist > mindists[2][2]) {
	calcdists[2][4] = 0;
	mindists[2][4] = 1000;
      }
    }

    if(calcdists[2][4]) {
      x1 =  cx1 - myseqB->end[r][0];
      y1 = cy1 - myseqB->end[r][1];
      z1 = cz1 - myseqB->end[r][2];
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[2][4] = dist;
      if(dist < mindists[minrow][mincol]) {
	minrow = 2;
	mincol = 4;
      }
      if(dist > mindists[1][4]) {
	calcdists[3][4] = 0;
	calcdists[4][4] = 0;
	mindists[3][4] = 1000;
	mindists[4][4]  = 1000;
      }
    }


    if(calcdists[3][0]) {
      x1 = q3x1-myseqB->origin[r][0];
      y1 = q3y1-myseqB->origin[r][1];
      z1 = q3z1-myseqB->origin[r][2];
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[3][0] = dist;
      if(dist < mindists[minrow][mincol]) {
	minrow = 3;
	mincol = 0;
      }
      if(dist > mindists[2][0]) {
	calcdists[4][0] = 0;
	mindists[4][0] = 1000;
      }
    }

    if(calcdists[3][1]) {
      x1 = q3x1-q1x2;
      y1 = q3y1-q1y2;
      z1 = q3z1-q1z2;
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[3][1] = dist;
      if(dist < mindists[minrow][mincol]) {
	minrow = 3;
	mincol = 1;
      }
      if(dist > mindists[2][1]) {
	calcdists[4][1] = 0;
	mindists[4][1] = 1000;
      }
      if(dist > mindists[3][0]) {
	for(i=2;i<5;i++) {
	  calcdists[3][i] = 0;
	  mindists[3][i] = 1000;
	}
      }
    }

    if(calcdists[3][2]) {
      x1 = q3x1-cx2;
      y1 = q3y1-cy2;
      z1 = q3z1-cz2;
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[3][2] = dist;
      if(dist < mindists[minrow][mincol]) {
	minrow = 3;
	mincol = 2;
      }
      if(dist > mindists[2][2]) {
	calcdists[4][2] = 0;
	mindists[4][2] = 1000;
      }
      if(dist > mindists[3][1]) {
	for(i=3;i<5;i++) {
	  calcdists[3][i] = 0;
	  mindists[3][i] = 1000;
	}
      }
    }

    if(calcdists[3][3]) {
      x1 = q3x1-q3x2;
      y1 = q3y1-q3y2;
      z1 = q3z1-q3z2;
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[3][3] = dist;
      if(dist < mindists[minrow][mincol]) {
	minrow = 3;
	mincol = 3;
      }
      if(dist > mindists[2][3]) {
	calcdists[4][3] = 0;
	mindists[4][3] = 1000;
      }
      if(dist > mindists[3][2]) {
	calcdists[3][4] = 0;
	mindists[3][4] = 1000;
      }
    }

    if(calcdists[3][4]) {
      x1 = q3x1-myseqB->end[r][0];
      y1 = q3y1-myseqB->end[r][1];
      z1 = q3z1-myseqB->end[r][2];
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[3][4] = dist;
      if(dist < mindists[minrow][mincol]) {
	minrow = 3;
	mincol = 4;
      }
      if(dist > mindists[2][4]) {
	calcdists[4][4] = 0;
	mindists[4][4] = 1000;
      }
    }

    mindists[4][0] = 10000;

    if(calcdists[4][1]) {
      x1 = myseqA->end[c][0]-q1x2;
      y1 = myseqA->end[c][1]-q1y2;
      z1 = myseqA->end[c][2]-q1z2;
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[4][1] = dist;
      if(dist < mindists[minrow][mincol]) {
	minrow = 4;
	mincol = 1;
      }
      if(dist > mindists[4][0]) {
	for(i=2;i<5;i++) {
	  calcdists[4][i] = 0;
	  mindists[4][i] = 1000;
	}
      }
    }
    /*test
    x1 = myseqA->end[c][0]-myseqB->origin[r][0];
    y1 = myseqA->end[c][1]-myseqB->origin[r][1];
    z1 = myseqA->end[c][2]-myseqB->origin[r][2];
    dist = ((x1*x1) + (y1*y1) + (z1*z1));
    if(dist<mindists[4]) {
      mindists[4] = dist;
    }*/

    if(calcdists[4][2]) {
      x1 = myseqA->end[c][0]-cx2;
      y1 = myseqA->end[c][1]-cy2;
      z1 = myseqA->end[c][2]-cz2;
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[4][2] = dist;
      if(dist < mindists[minrow][mincol]) {
	minrow = 4;
	mincol = 2;
      }
      if(dist > mindists[4][1]) {
	for(i=3;i<5;i++) {
	  calcdists[4][i] = 0;
	  mindists[4][i] = 1000;
	}
      }
    }

    if(calcdists[4][3]) {
      x1 = myseqA->end[c][0]-q3x2;
      y1 = myseqA->end[c][1]-q3y2;
      z1 = myseqA->end[c][2]-q3z2;
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[4][3] = dist;
      if(dist < mindists[minrow][mincol]) {
	minrow = 4;
	mincol = 3;
      }
      if(dist > mindists[4][2]) {
	calcdists[4][4] = 0;
	mindists[4][4] = 1000;
      }
    }

    if(calcdists[4][4]) {
      x1 = myseqA->end[c][0] - myseqB->end[r][0];
      y1 = myseqA->end[c][1] - myseqB->end[r][1];
      z1 = myseqA->end[c][2] - myseqB->end[r][2];
      dist = ((x1*x1) + (y1*y1) + (z1*z1));
      mindists[4][4] = dist;
      if(dist < mindists[minrow][mincol]) {
	minrow = 4;
	mincol = 4;
      }
    }


    min = mindists[minrow][mincol];
    min2 = 1000;
    /*    for(i=1;i<5;i++) {
      if(mindists[i]<min) {
	min2 = min;
	min = mindists[i];
      }else if(mindists[i]<min2) {
	min2 = mindists[i];
      }
      }*/
    min2 = 10000;
    for(i=0;i<5;i++) {
      for(j=0;j<5;j++) {
	if(i!=minrow && j!=mincol) {
	  if(mindists[i][j] < min2) {
	    min2 = mindists[i][j];
	  }
	}
      }
    }

    min = sqrt(min);
    min2 = sqrt(min2);
    /*    if(globalFull && min < 5 && min2 < 5)
	  printf("orig: min, min2: %f %f (%i, %i)\n",min,min2,r,c);*/
    min+=min2;
    min/=2.0;

    /*need to break ties!!*/
    /************************
    if(myseqB->length[r] > myseqA->length[c]) {
      for(i=0;i<3;i++) {
	origin1[i] = myseqA->origin[c][i]+((myseqA->end[c][i]-myseqA->origin[c][i])*0.25);
	origin2[i] = myseqB->origin[r][i]+((myseqB->end[r][i]-myseqB->origin[r][i])*0.25);
      end1[i] = myseqA->end[c][i]+3*((myseqA->end[c][i]-myseqA->origin[c][i])*0.25);
      end2[i] = myseqB->end[r][i]+3*((myseqB->end[r][i]-myseqB->origin[r][i])*0.25);
      }
    }else {
      for(i=0;i<3;i++) {
	origin2[i] = myseqA->origin[c][i]+((myseqA->end[c][i]-myseqA->origin[c][i])*0.25);
	origin1[i] = myseqB->origin[r][i]+((myseqB->end[r][i]-myseqB->origin[r][i])*0.25);
	end2[i] = myseqA->end[c][i]+3*((myseqA->end[c][i]-myseqA->origin[c][i])*0.25);
	end1[i] = myseqB->end[r][i]+3*((myseqB->end[r][i]-myseqB->origin[r][i])*0.25);
      }
    }


    for(i=0;i<3;i++) {
      origin1[i] = myseqA->origin[c][i];
      origin2[i] = myseqB->origin[r][i];
      end1[i] = myseqA->end[c][i];
      end2[i] = myseqB->end[r][i];
    }

    min = distance3D_Segment_to_Segment(origin1,end1,origin2,end2,&mua,&mub);
    dist=1000;
    dist2=1000;

    if(mub<=0.75) {
      for(i=0;i<3;i++) {
	point[i] = origin2[i]+(mub+0.25)*(end2[i]-origin2[i]);
      }
      dist = 1000;
      if(mua <=0.75) {
	dist = 0;
	for(i=0;i<3;i++) {
	  temp = origin1[i]+(mua+0.25)*(end1[i]-origin1[i]);
	  dist += (point[i] - temp) * (point[i] - temp);
	}
      }else { 
	for(i=0;i<3;i++) {
	  point2[i] = origin1[i] + 0.75 * (end1[i] - origin1[i]);
	}
	new_min = distance_Point_to_Segment(point2, origin1, end1);
	for(i=0;i<3;i++) {
	  temp = origin1[i] + 0.75*(end1[i]-origin1[i]);
	}
      }
      dist2 = 10000;
      if(mua >= 0.25) {
	dist2 = 0;
	for(i=0;i<3;i++) {
	  temp = origin1[i]+(mua-0.25)*(end1[i]-origin1[i]);
	  dist2 += (point[i] - temp) * (point[i] - temp);
	}
      }
	
      if(dist2 < dist) {
	dist = sqrt(dist2);
      }else {
	dist = sqrt(dist);
      }

    }else {
      mub = 0.75;
      for(i=0;i<3;i++) {
	point[i] = origin2[i] + mub * (end2[i] - origin2[i]);
      }
      new_min1 = distance_Point_to_Segment(point,origin1,end1);      

      dist = 1000;
      if(mua <=0.75) {
	dist = 0;
	for(i=0;i<3;i++) {
	  temp = origin1[i]+(mua+0.25)*(end1[i]-origin1[i]);
	  dist += (end2[i] - temp) * (end2[i] - temp);
	}
      }
      dist2 = 10000;
      if(mua >= 0.25) {
	dist2 = 0;
	for(i=0;i<3;i++) {
	  temp = origin1[i]+(mua-0.25)*(end1[i]-origin1[i]);
	  dist2 += (end2[i] - temp) * (end2[i] - temp);
	}
      }
      if(dist2 < dist) {
	dist = sqrt(dist2);
      }else {
	dist = sqrt(dist);
      }
      increased_mub = 1;
    }

    if(mub>=0.25) {
      for(i=0;i<3;i++) {
	point[i] = origin2[i]+(mub-0.25)*(end2[i]-origin2[i]);
      }
      dist3 = 10000;
      if(mua <=0.75) {
	dist3 = 0;
	for(i=0;i<3;i++) {
	  temp = origin1[i]+(mua+0.25)*(end1[i]-origin1[i]);
	  dist3 += (point[i] - temp) * (point[i] - temp);
	}
      }
      dist4 = 10000;
      if(mua >= 0.25) {
	dist4 = 0;
	for(i=0;i<3;i++) {
	  temp = origin1[i]+(mua-0.25)*(end1[i]-origin1[i]);
	  dist4 += (point[i] - temp) * (point[i] - temp);
	}
      }
	
      if(dist3 < dist4) {
	dist2 = sqrt(dist3);
      }else {
	dist2 = sqrt(dist4);
      }

    }else {
      mub = 0.25;
      for(i=0;i<3;i++) {
	point[i] = origin2[i] + mub * (end2[i] - origin2[i]);
      }
      new_min2 = distance_Point_to_Segment(point,origin1,end1);      

      
      dist3 = 1000;
      if(mua <=0.75) {
	dist3 = 0;
	for(i=0;i<3;i++) {
	  temp = origin1[i]+(mua+0.25)*(end1[i]-origin1[i]);
	  dist3 += (origin2[i] - temp) * (origin2[i] - temp);
	}
      }
      dist4 = 10000;
      if(mua >= 0.25) {
	dist4 = 0;
	for(i=0;i<3;i++) {
	  temp = origin1[i]+(mua-0.25)*(end1[i]-origin1[i]);
	  dist4 += (origin2[i] - temp) * (origin2[i] - temp);
	}
      }
      if(dist3 < dist4) {
	dist2 = sqrt(dist3);
      }else {
	dist2 = sqrt(dist4);
      }
      decreased_mub = 1;
    }


    if(dist<dist2) {
      if(increased_mub) {
	min = new_min1;
      }
      if(globalFull && dist < 5 && min < 5)
	printf("min, dist: %f %f\n",min,dist);
      min=(min+dist)/2.0;
    }else {
      if(decreased_mub) {
	min = new_min2;
      }
      if(globalFull && dist2 < 5 && min < 5)
	printf("min, dist2: %f %f\n",min,dist2);
      min = (min+dist2)/2.0;
    }
*****************/
   

    /*********************
    start = -1;
    end = -1;
    if(myseqA->length[c] > myseqB->length[r]) {
      for(i=0;i<=10;i++) {
	for(j=0;j<3;j++) {
	  point[j] = myseqB->origin[r][j]+0.1*i*(myseqB->end[r][j]-myseqB->origin[r][j]);
	}
	dist = distance_Point_to_Segment(point,myseqA->origin[c],myseqA->end[c]);
	if(dist<=4.0) {
	  start = i;
	  break;
	}
      }
      for(i=10;i>=0;i--) {
	for(j=0;j<3;j++) {
	  point[j] = myseqB->origin[r][j]+0.1*i*(myseqB->end[r][j]-myseqB->origin[r][j]);
	}
	dist = distance_Point_to_Segment(point,myseqA->origin[c],myseqA->end[c]);
	if(dist<=4.0) {
	  end = i;
	  break;
	}
      }
    }else {
      for(i=0;i<=10;i++) {
	for(j=0;j<3;j++) {
	  point[j] = myseqA->origin[c][j]+0.1*i*(myseqA->end[c][j]-myseqA->origin[c][j]);
	}
	dist = distance_Point_to_Segment(point,myseqB->origin[r],myseqB->end[r]);
	if(dist<=4.0) {
	  start = i;
	  break;
	}
      }
      for(i=10;i>=0;i--) {
	for(j=0;j<3;j++) {
	  point[j] = myseqA->origin[c][j]+0.1*i*(myseqA->end[c][j]-myseqA->origin[c][j]);
	}
	dist = distance_Point_to_Segment(point,myseqB->origin[r],myseqB->end[r]);
	if(dist<=4.0) {
	  end = i;
	  break;
	}
      }


    }
    if(end!=-1 && start!=-1) {
      percent = 10.0*(end-start);
    }else {
      percent = 0;
    }
***********************/      
  

#if ORIG_DIST_FUNC
      retvalue = (10.0/(1.0+(dist2*dist2)/16.0))-5.0;
#else
      retvalue = (10.0/(1.0 + (min*min)/16.0)) - 5.0; /*<-- current version*/
      /*      retvalue = (10.0/(1.0+(25.0*25.0)/(percent*percent)))-5.0;*/
#endif
    /*test version: M=5, do->5
      retvalue = (10.0/(1.0 + (min*min)/25.0)) - 5.0;*/
  }

  all_alignments.dp_scores[r][6] = retvalue;
  
  return retvalue;
}

float dot(Vector P1, Vector P2) {

  return(P1[0]*P2[0]+P1[1]*P2[1]+P1[2]*P2[2]);
}


/*******
  from: http://www.geometryalgorithms.com
 Copyright 2000, softSurfer (www.softsurfer.com)
 This code may be freely used and modified for any purpose
 providing that this copyright notice is included with it.
 SoftSurfer makes no warranty for this code, and cannot be held
 liable for any real or imagined damage resulting from its use.
 Users of this code must verify correctness for their application
 distance3D_Segment_to_Segment():
    Input:  two 3D line segments S1 and S2
    Return: the shortest distance between S1 and S2
*****/
float distance3D_Segment_to_Segment(Vector P1, Vector P2, Vector P3, Vector P4, float *mua, float *mub) {

    Vector   u;
    Vector   v;
    Vector   w;
    Vector dp;
    float a,b,c,d,e,D;
     float    sc, sN, sD;      /* sc = sN / sD, default sD = D >= 0*/
    float    tc, tN, tD;      /* tc = tN / tD, default tD = D >= 0*/
    float SMALL_NUM = 0.0001;
    int i;

    for(i=0;i<3;i++) {
      u[i] = P2[i]-P1[i];
      v[i] = P4[i]-P3[i];
      w[i] = P1[i]-P3[i];
    }
    
    a = dot(u,u); /*always >=0*/
    b = dot(u,v);
    c = dot(v,v); /*always >=0*/
    d = dot(u,w);
    e = dot(v,w);
    D = a*c-b*b;  /*always >=0*/
    sD = D;  /*default: sD = D>=0*/
    tD = D;  /*default: tD = D >=0*/

    /*compute the line parameters of the two closest points*/
    if (D <= SMALL_NUM) { /*the lines are almost parallel*/
        sN = 0.0;
        tN = e;
        tD = c;
    }
    else {                /*get the closest points on the infinite lines*/
        sN = (b*e - c*d);
        tN = (a*e - b*d);
        if (sN < 0) {       /*sc < 0 => the s=0 edge is visible*/
            sN = 0.0;
            tN = e;
            tD = c;
        }
        else if (sN > sD) {  /* sc > 1 => the s=1 edge is visible*/
            sN = sD;
            tN = e + b;
            tD = c;
        }
    }

    if (tN < 0) {           /*tc < 0 => the t=0 edge is visible*/
        tN = 0.0;
        /*recompute sc for this edge*/
        if (-d < 0)
            sN = 0.0;
        else if (-d > a)
            sN = sD;
        else {
            sN = -d;
            sD = a;
        }
    }
    else if (tN > tD) {      /*tc > 1 => the t=1 edge is visible*/
        tN = tD;
        /*recompute sc for this edge*/
        if ((-d + b) < 0)
            sN = 0;
        else if ((-d + b) > a)
            sN = sD;
        else {
            sN = (-d + b);
            sD = a;
        }
    }
    /*finally do the division to get sc and tc*/
    sc = sN / sD;
    tc = tN / tD;
    *mua = sc;
    *mub = tc;

    /* get the difference of the two closest points*/
    for(i=0;i<3;i++) {
      dp[i] = w[i]+(sc*u[i])-(tc*v[i]);  /* = S1(sc) - S2(tc)*/
    }
    /* return the closest distance*/
    return(sqrt(VectorNorm2(dp))); 
}



/****
 distance_Point_to_Segment():
 Input:  a Point P and a Segment S (in any dimension)
    Return: the shortest distance from P to S
******/
float distance_Point_to_Segment( float* P, float* P0, float* P1)
{
  float v[3],w[3],Pb[3];
  double c1,c2,b;
  int i;

  c1=0;
  c2 = 0;
  for(i=0;i<3;i++) {
    v[i] = P1[i]-P0[i];
    w[i] = P[i]-P0[i];
    c1+=w[i]*v[i];
    c2+=v[i]*v[i];
  }

  if ( c1 <= 0 )
    return(sqrt(pow((P0[0]-P[0]),2) + pow((P0[1]-P[1]),2) + pow((P0[2]-P[2]),2)));

  if ( c2 <= c1 )
    return(sqrt(pow((P1[0]-P[0]),2) + pow((P1[1]-P[1]),2) + pow((P1[2]-P[2]),2)));
  
  b = c1 / c2;
  for(i=0;i<3;i++) {
    Pb[i] = P0[i]+b*v[i];
  }
  
  return(sqrt(pow((Pb[0]-P[0]),2) + pow((Pb[1]-P[1]),2) + pow((Pb[2]-P[2]),2)));
}


float MatrixDet(Matrix3 m) {
  float result;

  result  = m[0][0]*(m[1][1]*m[2][2]-m[2][1]*m[1][2]);
  result -= m[0][1]*(m[1][0]*m[2][2]-m[2][0]*m[1][2]);
  result += m[0][2]*(m[1][0]*m[2][1]-m[2][0]*m[1][1]);

  return(result);
}

void CrossProduct(Vector v1, Vector v2, Vector result) {
  result[0] = (v1[1]*v2[2])-(v1[2]*v2[1]);
  result[1] = (v1[2]*v2[0])-(v1[0]*v2[2]);
  result[2] = (v1[0]*v2[1])-(v1[1]*v2[0]);
}

float VectorNorm2(Vector v) {
  float result = 0;
  int i;

  for(i=0;i<3;i++) {
    result+=v[i]*v[i];
  }

  return(result);

}

void Normalize(Vector v) {
  Vector n;
  int i;
  float length=0;

  for(i=0;i<3;i++) {
    length+=v[i]*v[i];
  }
  length = sqrt(length);
  for(i=0;i<3;i++) {
    v[i]/=length;
  }
}

float SegmentSegmentIntersect(Vector o1, Vector end1, Vector o2, Vector end2,float length2) {
  Vector d2, tempVector, tempVector2;
  Vector d1;
  float numer,denom;
  Matrix3 tempmat;
  float s, t,dist,temp;
  int i;




  for(i=0;i<3;i++) {
    d1[i] = end1[i]-o1[i];
    d2[i] = end2[i]-o2[i];
    tempVector[i] = o2[i]-o1[i];
  }

  /*  Normalize(d1);
  Normalize(d2);*/
  CrossProduct(d1,d2,tempVector2);




  for(i=0;i<3;i++) {
    tempmat[0][i] = tempVector[i];
    tempmat[1][i] = d2[i];
    tempmat[2][i] = tempVector2[i];
  }

  /****
  set_col(tempmat,0,tempVector);
  set_col(tempmat,1,d2);
  set_col(tempmat,2,tempVector2);
  ****/
  numer = MatrixDet(tempmat);
  s = numer/VectorNorm2(tempVector2);

  for(i=0;i<3;i++) {
    tempmat[1][i] = d1[i];
  }
  /*  set_col(tempmat,1,d1);*/
  numer = MatrixDet(tempmat);
  t = numer/VectorNorm2(tempVector2);
  
  dist = 0;
  for(i=0;i<3;i++) {
    temp=(o1[i]+s*d1[i]) - (o2[i]+t*d2[i]);
    dist+=temp*temp;
  }
  dist = sqrt(dist);
  /******
  if(s>=0 && s <=1) {
    if(t>=0 && t <= length2) {
      return(dist);
    }else return(10000);
  }else return(10000);
  ******/

  return dist;
}


/*
   Calculate the line segment PaPb that is the shortest route between
   two lines P1P2 and P3P4. Calculate also the values of mua and mub where
      Pa = P1 + mua (P2 - P1)
      Pb = P3 + mub (P4 - P3)
   Return FALSE if no solution exists.
*/
int LineLineIntersect(
   Vector p1,Vector p2,Vector p3,Vector p4,Vector pa,Vector pb,
   double *mua, double *mub)
{
   Vector p13,p43,p21;
   double d1343,d4321,d1321,d4343,d2121;
   double numer,denom;
   double EPS = 0.00001;

   p13[0] = p1[0] - p3[0];
   p13[1] = p1[1] - p3[1];
   p13[2] = p1[2] - p3[2];
   p43[0] = p4[0] - p3[0];
   p43[1] = p4[1] - p3[1];
   p43[2] = p4[2] - p3[2];
   if (abs(p43[0])  < EPS && abs(p43[1])  < EPS && abs(p43[2])  < EPS)
      return(0);
   p21[0] = p2[0] - p1[0];
   p21[1] = p2[1] - p1[1];
   p21[2] = p2[2] - p1[2];
   if (abs(p21[0])  < EPS && abs(p21[1])  < EPS && abs(p21[2])  < EPS)
      return(0);

   d1343 = p13[0] * p43[0] + p13[1] * p43[1] + p13[2] * p43[2];
   d4321 = p43[0] * p21[0] + p43[1] * p21[1] + p43[2] * p21[2];
   d1321 = p13[0] * p21[0] + p13[1] * p21[1] + p13[2] * p21[2];
   d4343 = p43[0] * p43[0] + p43[1] * p43[1] + p43[2] * p43[2];
   d2121 = p21[0] * p21[0] + p21[1] * p21[1] + p21[2] * p21[2];

   denom = d2121 * d4343 - d4321 * d4321;
   if (abs(denom) < EPS)
      return(0);
   numer = d1343 * d4321 - d1321 * d4343;

   *mua = numer / denom;
   *mub = (d1343 + d4321 * (*mua)) / d4343;

   pa[0] = p1[0] + *mua * p21[0];
   pa[1] = p1[1] + *mua * p21[1];
   pa[2] = p1[2] + *mua * p21[2];
   pb[0] = p3[0] + *mub * p43[0];
   pb[1] = p3[1] + *mub * p43[1];
   pb[2] = p3[2] + *mub * p43[2];

   return(1);
}





void computeData_ss(int r, int c, seq *myseqA, seq *myseqB) {
  int myindex, i, tempNum;
  float tempScore, myscore, max;
  float myscore1, myscore2, myconstscore1, myconstscore2, myconstscore3;
  int row_begin=-1, col_begin=-1,start_index;
  float score_correction = 0.0;
  int col,row;
  int found_pointer = 0;
  int iters = 0;

  max = -10000.0;
 

  if (local) 
    max = 0.0;

  myindex = c + r*colA;
  myconstscore1 = computeScore_ss(r, c, myseqA, myseqB);
  myconstscore2 = computeAvgDistScore_ss(r, c, myseqA, myseqB);
  myconstscore3 = computeLengthScore(r, c, myseqA, myseqB); 
  /*   if(globalFull && myconstscore2 > 0) {
       printf("r: %i c: %i constant scores: %8.3f %8.3f %8.3f\n",r,c,myconstscore1, myconstscore2, myconstscore3);
       }*/


  
  if(myconstscore2 > 0) {
    myscore = myconstscore1+myconstscore2+myconstscore3;
  }
  else {
    myscore = -10000;
  }

  
 iters = 0; 
  
 if ((r > 0) && (c > 0) && ((myscore > 0) || (local == 0))) {

   col = c-1-iters;
   row = r-1-iters;
   while(!found_pointer && !(col < 0 && row < 0)) {
     


       for (i = 0; i < r && col >= 0; i++) {
	 myscore1 = computeAngleScore(r,c,i,col, myseqA, myseqB);
	 myscore2 = computeOriginScore(r,c,i,col, myseqA, myseqB);

	 if ((myscore1 + myscore2 > 0) || (local == 0)) {
	   tempScore = data[(col) + i*colA].score + myscore + myscore1 + myscore2;
	   
	 }
	 else {
	   tempScore = -10000.0;
	 }
	 
	 /*check for pointer*/
	 if(tempScore > 0) {
	 /*	 if(tempScore-myscore1-myscore2 > max && data[col+i*colA].score <= 0) {*/

	   /*there's no pointer here, so we shouldn't store this as max score?*/
	   /*this next line is useless because if myscore1 and 2 aren't used,
	    *tempScore is meaningless? If there isn't going to be a pointer,
	    *ignore. The issue is what to do if constant scores are greater
	    *than pointer scores => can't happen because we insist that ptr
	    *scores be greater than zero and constant score >0*/
	   if(tempScore-myscore1-myscore2 > max && data[col+i*colA].score <= 0) {
	     /*	     max = tempScore;	*/
	     /*useless:	     max = tempScore - myscore1 - myscore2;*/
	     /*	     printf("no ptrs col: (%i, %i): best pointer = (%i, %i), index %i; ptr score = %f, max = %f, tempScore = %f\n",r,c,i,col,data[myindex].numptrs,data[col+i*colA].score,max,tempScore);*/
	   }
	   if ((local == 0) || (data[(col) + i*colA].score > 0.0)) {
	     if(tempScore >  max) {
	       max = tempScore;
	       data[myindex].bestPointerIndex = data[myindex].numptrs;
	       /*	       printf("assign col ptr: (%i, %i): best pointer = (%i, %i), index %i, relative score = %f, constscore = %f, max = %f\n",r,c,i,col,data[myindex].bestPointerIndex,myscore1+myscore2,myscore,max);*/
	     }
	     data[myindex].Rptr[data[myindex].numptrs] = i;
	     data[myindex].Cptr[data[myindex].numptrs]= col;
	     data[myindex].numptrs++; 
	     if(data[myindex].numptrs > MAXPTRS) {
	       	       printf("r = %i, c = %i, TOO MANY POINTERS\n",r,c);
	     }
	     found_pointer = 1;
	     
	   }
	   else {
	     /*	  data[myindex].numptrs = 0; */
	     row_begin = i;
	     col_begin = col;
	     score_correction = 0;
	   }	  
	 }
       
       }

    /*    for (i = 0; i < c - 1; i++) {*/


       for (i = 0; i < col && row >=0; i++) {

	 
	 myscore1 = computeAngleScore(r,c,row,i, myseqA, myseqB);
	 myscore2 = computeOriginScore(r,c,row,i, myseqA, myseqB);
	 
	 if ((myscore1 + myscore2 > 0) || (local == 0)) {
	   tempScore = data[i + (row)*colA].score + myscore + myscore1 + myscore2;
	 }
	 else {
	   tempScore = -10000.0;
	 }
	 
	 /*if(tempScore-myscore1-myscore2 > 0) {*/
	 if(tempScore > 0) {
	   /*	if(globalFull)
		printf("%i, %i, %f %f\n",row,i,myscore1,myscore2);*/
	   /*	   if(tempScore-myscore1-myscore2 > max  && data[i+row*colA].score <= 0) {*/
	   if(tempScore-myscore1-myscore2 > max  && data[i+row*colA].score <= 0) {
	     /*useless	     max = tempScore-myscore1-myscore2;*/
	     /*	     max = tempScore;*/
	     /*	     	     printf("no ptrs: row: (%i, %i): best pointer = (%i, %i), index %i; ptr score = %f; max = %f, tempscore = %f\n",r,c,i,row,data[myindex].numptrs,data[i+row*colA].score,max,tempScore);	   */
	   }
	   if ((local == 0) || (data[i + (row)*colA].score > 0.0)) {
	     if(tempScore > max) {
	       max = tempScore;	
	       data[myindex].bestPointerIndex = data[myindex].numptrs;
	       /*	       printf("assign row ptr: (%i, %i): best pointer = (%i, %i), index %i, relative score = %f, constscore = %f, max = %f\n",r,c,row,i,data[myindex].numptrs,myscore1+myscore2,myscore,max);*/
	     }
	     data[myindex].Rptr[data[myindex].numptrs] = row;
	     data[myindex].Cptr[data[myindex].numptrs] = i;
	     data[myindex].numptrs++; 
	     if(data[myindex].numptrs > MAXPTRS) {
	       printf("r = %i, c = %i, TOO MANY POINTERS\n",r,c);
	     }
	     found_pointer = 1; 

	   }
	   else {
	     /* data[myindex].numptrs = 0;*/
	     row_begin = row;
	     col_begin = i;
	     score_correction = 0;
	   }
	 }
       
       }
     iters++;
     row = r - 1-iters;
     col = c -1 -iters;
   }
 }
 else if ((r == 0) || (c == 0)) {
   if (myscore > max) {
     max = myscore;
     score_correction = 0;
   }
 }

 if(!found_pointer && myscore > 0) {
   /*   printf("(%i, %i): storing constant scores  = %f\n",r,c,myscore);*/
   data[myindex].score = myscore;
 }else {
   data[myindex].score = max;	
 }


  
}


float testFullAlign(int newr, int newc, int oldr, int oldc, seq *myseqA, seq *myseqB) {

  seq newseqB;

  Quaternion quat_rot;
  Vector translation;
  Matrix mat;
  float m[16];
  int i,j,r,c;
  float point[3];
  float *firstA,*secA,*firstB,*secB;
  tempTotalPairs++;


  for (i = 0; i < (colA+1)*(rowB+1); i++) {
    data[i].score = -1000000.0;
    data[i].numptrs = 0;
    data[i].bestPointerIndex = -1;
  }
  maxScore = -1000000.0;

  /*    printf("test full align: r = (%i, %i), c = (%i, %i)\n",newr,oldr,newc,oldc);*/
  for(i=0;i<colA;i++) {
    seqA.alignment[i] = -1;
  }
  for(i=0;i<rowB;i++) {
    seqB.alignment[i] = -1;
  }

  initstackType();


  init_pairs();

  add_pair(myseqA->origin[newc], myseqB->origin[newr], 1);
  add_pair(myseqA->end[newc], myseqB->end[newr], 1);
  add_pair(myseqA->origin[oldc], myseqB->origin[oldr], 1);
  add_pair(myseqA->end[oldc], myseqB->end[oldr], 1);

  /*  match_pairs (mat, quat_rot, translation);*/
  match_pairs(mat,translation);

  for (i = 0; i < 4; i++) {
    for (j = 0; j < 4; j++) {
      m[i*4 + j] = mat[i][j];
    }

  }


  for (r = 0; r < rowB; r++) {
    newseqB.ss[r] = myseqB->ss[r];
    newseqB.length[r] = myseqB->length[r];
    applyRottoV(newseqB.vector[r], myseqB->vector[r], m);
    applyMtoV(newseqB.origin[r], myseqB->origin[r], m);
    applyMtoV(newseqB.end[r], myseqB->end[r], m);

  }


  for (r = 0; r < rowB; r++) {
    for (c = 0; c < colA; c++) {
      /*      printf("calling computedata with (%i, %i)\n",r,c);*/
      computeData_ss(r, c, myseqA, &newseqB);
      fixMax(r, c);
      if(data[c+r*colA].score > maxScore) { /*or if maxscore is zero?*/
	seqA.alignment[c] = r;
	seqB.alignment[r]=c;
      }
    }
  }

    return maxScore;
}





/* Function: computeData_onlyAngles_sequentialPairs:
 * ------------------------------------------
 * Given a query vector and a target vector, selects a second vector from each
 * structure and calls testFullAlign if the angle score is high enough. Either
 * the pair of query or the pair of target vectors must be sequential in 
 * sequence space. This reduces the computational complexity of the algorithm,
 * but may not find the best pairs of vectors for geometric hashing.
 */
void computeData_onlyAngles_sequentialPairs(int r, int c, seq *myseqA, seq *myseqB) {
  int i,j;
  float fullScore;
  float oldscore =0;
  int first,second;
  this_transform_score = 0.0;


  j = c-1;
  if ((myseqA->ss[c] == myseqB->ss[r]) && (r > 0) && (c > 0)) {
    for (i = 0; i < r; i++) {
      if(myseqA->ss[j] == myseqB->ss[i]) { /*test version*/
	fullScore = computeAngleScore(r,c,i,j, myseqA, myseqB) + computeOriginScore(r,c,i,j, myseqA, myseqB); 
	num_pairs++;
	if (fullScore > currentD) {
	  fullScore = testFullAlign(r,c,i,j, myseqA, myseqB);
	  num_hashes_tested++;
	  if (fullScore > bestFullScore) {

	    bestFullScore = fullScore;
	    bestR = r;
	    bestC = c;
	    bestOldR = i;
	    bestOldC = j;
	    
	    if (fullScore > 0.6 * highestPossibleScore) {	      
	      currentD = currentD + ((20.0 - currentD)/2.0);	      
	    }
	    if (fullScore > 0.7 * highestPossibleScore) {	      
	      currentD = currentD + ((20.0 - currentD)/2.0);	      
	    }
	    if (fullScore > 0.8 * highestPossibleScore) {	      
	      currentD = currentD + ((20.0 - currentD)/2.0);	      
	    }
	    
	    if(fullScore > this_transform_score) {
	      this_transform_score = fullScore;
	      this_transform_oldR = i;
	      this_transform_oldC = j;
	    }
	  }
	  if(fullScore > this_transform_score) {
	      this_transform_score = fullScore;
	      this_transform_oldR = i;
	      this_transform_oldC = j;
	    }

	}
	
	/*else below supposed to be related to if(fullScore>bestFullScore?*/
	else if (fullScore > secondBestScore) {
	  secondBestR = r;
	  secondBestC = c;
	  secondBestOldR = i;
	  secondBestOldC = j;
	  secondBestScore = fullScore;
	}

      }
      
    }

    
    j=r-1;
    for (i = 0; i < c - 1; i++) {

	if (myseqA->ss[i] == myseqB->ss[j]) {

	  fullScore= computeAngleScore(r,c,j,i,myseqA,myseqB)+computeOriginScore(r,c,j,i,myseqA,myseqB);

	if (fullScore > currentD) {
	  
	  fullScore = testFullAlign(r,c,j,i, myseqA, myseqB);


	  if (fullScore > bestFullScore) {

	    bestFullScore = fullScore;
	    bestR = r;
	    bestC = c;
	    bestOldR = j;
	    bestOldC = i;
	    if (fullScore > 0.6 * highestPossibleScore) {	      
	      currentD = currentD + ((20.0 - currentD)/2.0);	      
	    }
	    if (fullScore > 0.7 * highestPossibleScore) {	      
	      currentD = currentD + ((20.0 - currentD)/2.0);	      
	    }
	    if (fullScore > 0.8 * highestPossibleScore) {	      
	      currentD = currentD + ((20.0 - currentD)/2.0);	      
	    }
	  }
	}
	else if (fullScore > secondBestScore) {
	  secondBestR = r;
	  secondBestC = c;
	  secondBestOldR = j;
	  secondBestOldC = i;
	  secondBestScore = fullScore;
	}
	}
	/*	if(fullScore > this_transform_score) {
	  this_transform_score = fullScore;
	  this_transform_oldR = j;
	  this_transform_oldC = i;
	  }*/
    }

  }
}

/* Function: computeData_onlyAngles_allPairs:
 * ------------------------------------------
 * Given a query vector and a target vector, selects a second vector from each
 * structure and calls testFullAlign if the angle score is high enough. All
 * possible pairs of vectors from the query and target are tested
 */
void computeData_onlyAngles_allPairs(int r, int c, seq *myseqA, seq *myseqB) {
  int i,j;
  float fullScore;
  float oldscore =0;
  int first,second;
  this_transform_score = 0.0;


  if ((myseqA->ss[c] == myseqB->ss[r]) && (r > 0) && (c > 0)) {
    for (i = 0; i < r; i++) {
      for(j=0; j<=c-1; j++) {
	if(myseqA->ss[j] == myseqB->ss[i]) { /*test version*/
	  fullScore = computeAngleScore(r,c,i,j, myseqA, myseqB) + computeOriginScore(r,c,i,j, myseqA, myseqB); 
	  num_pairs++;
	  if (fullScore > currentD) {
	    fullScore = testFullAlign(r,c,i,j, myseqA, myseqB);
	    num_hashes_tested++;
	    if (fullScore > bestFullScore) {
	      
	      bestFullScore = fullScore;
	      bestR = r;
	      bestC = c;
	      bestOldR = i;
	      bestOldC = j;
	      
	      if (fullScore > 0.6 * highestPossibleScore) {	      
		currentD = currentD + ((20.0 - currentD)/2.0);	      
	      }
	      if (fullScore > 0.7 * highestPossibleScore) {	      
		currentD = currentD + ((20.0 - currentD)/2.0);	      
	      }
	      if (fullScore > 0.8 * highestPossibleScore) {	      
		currentD = currentD + ((20.0 - currentD)/2.0);	      
	      }
	      
	      if(fullScore > this_transform_score) {
		this_transform_score = fullScore;
		this_transform_oldR = i;
		this_transform_oldC = j;
	      }
	    }
	    if(fullScore > this_transform_score) {
	      this_transform_score = fullScore;
	      this_transform_oldR = i;
	      this_transform_oldC = j;
	    }
	    
	  }
	  
	  /*else below supposed to be related to if(fullScore>bestFullScore?*/
	  else if (fullScore > secondBestScore) {
	    secondBestR = r;
	    secondBestC = c;
	    secondBestOldR = i;
	    secondBestOldC = j;
	    secondBestScore = fullScore;
	  }
	  
	}
      }
    }
  }
}




float computePenalty_ss(int r, int c, int fromR, int fromC) {

  float retvalue;

  retvalue = 0.0;

  if (r - fromR > 1) {    
    retvalue = penalty_ss[0] + ((float)(r - fromR - 1))*(penalty_ss[1]);
  }
  else if (c - fromC > 1) {
    retvalue = penalty_ss[2] + ((float)(c - fromC - 1))*(penalty_ss[3]);
  }
    
  return retvalue;

}




float align_ss(int full,OBJ *query, OBJ *target) {
  int r, c,i,j;
  float fullScore;
  int myindex,index;
  int found_alignment = 0;
  float point[3];


  if(full) {
    globalFull = 1;
  }else {
    globalFull = 0;
  }

  if(full == 0) {
    distFuncMax = 4.0;
  }else {
    distFuncMax = 4.0;
  }

  /*  printf("align_ss: local = %i\n",local);*/
  if(full == 0 || !runStep1) {
    originScoresA = (originScores**)malloc(sizeof(originScores*)*colA);
    originScoresB = (originScores**)malloc(sizeof(originScores*)*rowB);
    all_alignments.index = 0;
    all_alignments.pointers = (int *)malloc(sizeof(float)*colA*rowB);
    all_alignments.scores = (float*)malloc(sizeof(float)*colA*rowB);
    all_alignments.rowpointers = (int**)malloc(sizeof(int*)*colA*rowB);
    all_alignments.columnpointers = (int**)malloc(sizeof(int*)*colA*rowB);
    for(i=0;i<(colA*rowB);i++) {
      all_alignments.rowpointers[i] = (int*)malloc(sizeof(int)*2);
      all_alignments.columnpointers[i] = (int*)malloc(sizeof(int)*2);

    }
    for(i=0;i<colA;i++) {
      originScoresA[i] = (originScores*)malloc(sizeof(originScores)*colA);
      for(j=0;j<colA;j++) {
	originScoresA[i][j] = NULL;
      }
    }
    for(i=0;i<rowB;i++) {
      originScoresB[i] = (originScores*)malloc(sizeof(originScores)*rowB);
      for(j=0;j<rowB;j++) {
	originScoresB[i][j] = NULL;
      }
    }
    all_alignments.numAlignments = 0;
    for(j=0;j<(rowB*colA);j++) {
      all_alignments.scores[j] = -1000;
    }

    /*allocate space for dynamic programming scores*/
    all_alignments.dp_scores = (float**)malloc(sizeof(float*)*target->ssArrayLength);
    for(i=0;i<target->ssArrayLength;i++) {
      all_alignments.dp_scores[i] = (float*)malloc(sizeof(float)*7);
    } 

    /*copy original, untransformed seqA and seqB to seqA_orig, seqB_orig*/
    memcpy((char*)seqA_orig.ss,(char*)seqA.ss,sizeof(float)*16);
    memcpy((char*)seqB_orig.ss,(char*)seqB.ss,sizeof(float)*16);
    memcpy((char*)seqA_orig.length,(char*)seqA.length,sizeof(float)*16);
    memcpy((char*)seqB_orig.length,(char*)seqB.length,sizeof(float)*16);
    for(i=0;i<MAXLENGTH;i++) {
      memcpy((char*)seqA_orig.origin[i],(char*)seqA.origin[i],sizeof(float)*3);
      memcpy((char*)seqB_orig.origin[i],(char*)seqB.origin[i],sizeof(float)*3);
      memcpy((char*)seqA_orig.end[i],(char*)seqA.end[i],sizeof(float)*3);
      memcpy((char*)seqB_orig.end[i],(char*)seqB.end[i],sizeof(float)*3);
      memcpy((char*)seqA_orig.vector[i],(char*)seqA.vector[i],sizeof(float)*3);
      memcpy((char*)seqB_orig.vector[i],(char*)seqB.vector[i],sizeof(float)*3);
    }

    /*copy original target matrix*/
    memcpy((char*)orig_target_mat,(char*)target->objMat,sizeof(float)*16);
    memcpy((char*)orig_query_mat,(char*)query->objMat,sizeof(float)*16);
  }else {
    found_alignment = 1;
  }

  /*rowB is the number of SSE elements in target (when this function is called for the first
   *time, at least; and rowA is the number of SSE elements in the query*/
    for (r = 0; r < rowB; r++) {
      for (c = 0; c < colA; c++) {
	if (full == 1) {
	  computeData_ss(r, c, &seqA, &seqB);
	  fixMax(r, c);
	}
	else {
	  if(all_pairs) {
	    computeData_onlyAngles_allPairs(r,c, &seqA, &seqB);
	  }else {
	    computeData_onlyAngles_sequentialPairs(r,c, &seqA, &seqB);
	  }

	  myindex = c+r*colA;
	  if(this_transform_score > 0) {
	    all_alignments.scores[all_alignments.numAlignments] = this_transform_score;
	    all_alignments.rowpointers[all_alignments.numAlignments][0] = this_transform_oldR;
	    all_alignments.rowpointers[all_alignments.numAlignments][1] =r;
	    all_alignments.columnpointers[all_alignments.numAlignments][0] = this_transform_oldC;
	    all_alignments.columnpointers[all_alignments.numAlignments][1] = c;
	    all_alignments.pointers[all_alignments.numAlignments] = all_alignments.numAlignments;
	    all_alignments.numAlignments++;
	  }

	}
      }
    }

    if(full==1) {


      /*      if(chaining && maxScore > 0.0) {
	endpoints = NewEndpointList(regions);

	findBestChain(endpoints,regions,alignmentChain,&chainLength);
	maxScore = getChainScore(regions,alignmentChain[0]);
	FreeEndpointList(endpoints);

	for(i=0;i<chainLength-1;i++) {
	  myindex = getColumnStart(regions,alignmentChain[i])+(rowB-1-getRowStart(regions,alignmentChain[i]))*colA;
	  data[myindex].Rptr[0] = rowB-1-getRowEnd(regions,alignmentChain[i+1]);
	  data[myindex].Cptr[0] = getColumnEnd(regions,alignmentChain[i+1]);
	  data[myindex].numptrs = 1;
	}
	maxIndex[0] = rowB-1-getRowEnd(regions,alignmentChain[0]);
	maxIndex[1] = getColumnEnd(regions,alignmentChain[0]);

      }*/

	traceBack(maxIndex[0],maxIndex[1],query,target);
    }


  if (full != 1) {

    qsort(all_alignments.pointers,all_alignments.numAlignments,sizeof(int),compare_alignments);
    printf("number of successful hashes: %i\n",num_hashes_tested);
    printf("number of pairs: %i\n",num_pairs);
    /*    printf("number of transformations attempted: %i\n",all_alignments.numAlignments);
    for(i=0;i<all_alignments.numAlignments;i++) {
      printf("index %i, score %f\n",all_alignments.pointers[i],all_alignments.scores[all_alignments.pointers[i]]);
    }*/
    if(secondBestR != -1) {
      fullScore = testFullAlign(secondBestR,secondBestC,secondBestOldR,secondBestOldC, &seqA, &seqB);
    }else {
      fullScore = bestFullScore;
    }

    if (fullScore > bestFullScore) {
      bestFullScore = fullScore;
      bestR = secondBestR;
      bestC = secondBestC;
      bestOldR = secondBestOldR;
      bestOldC = secondBestOldC;
      
      /*      bestR=1;
      bestOldR = 0;
      bestC =1;
      bestOldC = 0;*/
    }else {
      globalFull=1;
      /*forward
    bestR=1;
    bestOldR = 0;
    bestC =1;
    bestOldC = 0;*/
    
      /*reverse
    bestR = 5;
    bestOldR = 3;
    bestC = 9;
    bestOldC =0;*/
      if(bestR != -1) {
	fullScore = testFullAlign(bestR,bestC,bestOldR,bestOldC,&seqA,&seqB);
      }

      printf("full score: %f, bestfull: %f\n",fullScore,bestFullScore);
    }

    /*forward
    bestR=1;
    bestOldR =0 ;
    bestC =1;
    bestOldC = 0;*/

    /*reverse
    bestR = 5;
    bestOldR = 3;
    bestC = 9;
    bestOldC = 0;*/
    first_alignment_score = fullScore;

    if(fullScore > 0) {
      traceBack(maxIndex[0],maxIndex[1],query,target);
      globalCalledAddPair = 4;
      init_pairs();
      
      add_pair(seqA.origin[bestC], seqB.origin[bestR], 1);
      add_pair(seqA.end[bestC], seqB.end[bestR], 1);
      add_pair(seqA.origin[bestOldC], seqB.origin[bestOldR], 1);
      add_pair(seqA.end[bestOldC], seqB.end[bestOldR], 1);

    }
  }


  return(maxScore);

}


void initTwoPairTransform(OBJ* query, OBJ* target, OBJ* new_target) {
  Matrix mat;
  Vector translation;
  float m[16];
  int i,j;



  initstackType();


  init_pairs();
  memcpy((char*)target->objMat,(char*)orig_target_mat,sizeof(float)*16);
  memcpy((char*)query->objMat,(char*)orig_query_mat,sizeof(float)*16);

  add_pair(seqA_orig.origin[bestC], seqB_orig.origin[bestR], 1);
  add_pair(seqA_orig.end[bestC], seqB_orig.end[bestR], 1);
  add_pair(seqA_orig.origin[bestOldC], seqB_orig.origin[bestOldR], 1);
  add_pair(seqA_orig.end[bestOldC], seqB_orig.end[bestOldR], 1);

  globalCalledAddPair = 4;

  /*  match_pairs (mat, quat_rot, translation);*/
  match_pairs(mat,translation);



  for (i = 0; i < 4; i++) {
    for (j = 0; j < 4; j++) {
      m[i*4 + j] = mat[i][j];
    }
  }
  myMatMult(target->objMat,m);
  target->center[0] = 0;
  target->center[1] = 0;
  target->center[2] = 0;
  applyMtoV(target->center,target->center,m);
  applyMat(new_target, target,target->objMat);
}


void report_ss(int full, OBJ *query, OBJ* target) {
  int i,j;
  float point[3];

  if (full == 1) {
#if SHOWOUTPUT 
    printf("\nMax Score %f\n", maxScore); 
#endif
    query->num_SSEs_aligned = 0;
    target->num_SSEs_aligned = 0;
    for(i=0;i<=colA;i++) {
      for(j=0;j<=rowB;j++) {
	query->ss_alignment[i][j] = -1;
      }
      query->num_aligned[i] = 0;
    }
    for(i=0;i<=rowB;i++) {
      for(j=0;j<=colA;j++) {
	target->ss_alignment[i][j] = -1;
      }
      target->num_aligned[i] = 0;
    }
    if(maxScore > 0) {
      writePair_ss(maxIndex[0], maxIndex[1],query,target);
    }
  }
  else {
    printf("\nMax Full Score %f\n", bestFullScore); 
    printf("Columns: %d %d\tRows: %d %d\n", bestOldC, bestC, bestOldR, bestR);
    printf("Total Pairs = %d\n", tempTotalPairs);

    globalCalledAddPair = 4;
    
    init_pairs();

    add_pair(seqA.origin[bestC], seqB.origin[bestR], 1);
    add_pair(seqA.end[bestC], seqB.end[bestR], 1);
    add_pair(seqA.origin[bestOldC], seqB.origin[bestOldR], 1);
    add_pair(seqA.end[bestOldC], seqB.end[bestOldR], 1);
  }    


}

void printScores_ss(OBJ* query, OBJ* target) {
  
  int i,myindex;
  float prev_score = 0;
  float newscore;

  printf("Columns: %d %d\tRows: %d %d\tMaxIndices: %d %d\n", bestOldC, bestC, bestOldR, bestR,maxIndex[0],maxIndex[1]);
  printf("Scores\n\n");
  /************
  for (i = 0; i < colA*rowB; i++) {
    printf("%f ", data[i].score);
    if ((i+1)%colA == 0) 
      printf("\n");
  }
  for(i=0;i<colA*rowB;i++) {
    printf("%i,%i ",data[i].Rptr[0],data[i].Cptr[0]);
    if((i+1)%colA == 0) {
      printf("\n");
    }
  }
  ***********/
  myindex = maxIndex[1]+maxIndex[0]*colA;
  traceBack(maxIndex[0],maxIndex[1],query,target);
  for(i=0;i<query->ssArrayLength;i++) {
    query->ss_dpscores[i] = 0;
    if(query->ss_alignment[i][0]!=-1) {
      myindex = i+query->ss_alignment[i][0]*colA;
      newscore = data[myindex].score - prev_score;
      /*      printf("%i,%i,%f=> ",data[myindex].Rptr[0],data[myindex].Cptr[0],data[myindex].score);*/
      /*      printf("%8.3f\n",newscore);*/
      query->ss_dpscores[i] = newscore;
      target->ss_dpscores[query->ss_alignment[i][0]] = newscore;
      prev_score = data[myindex].score;
    }
  }
  printf("\n");
}

void traceBack(int r, int c, OBJ *query, OBJ *target) {
  int i, myindex,j;

  

  myindex = c+r*colA;
  /*  printf("traceback: r = %i, c = %i; best pointer idx: %i\n",r,c,data[myindex].bestPointerIndex);
      fflush(stdout);*/
  if((r==maxIndex[0]) && c == maxIndex[1]) { /*first call*/
    query->num_SSEs_aligned = 0;
    target->num_SSEs_aligned = 0;
    for(i=0;i<colA;i++) {
      for(j=0;j<rowB;j++) {
	query->ss_alignment[i][j] = -1;
      }
      query->num_aligned[i] = 0;
    }
    for(i=0;i<rowB;i++) {
      for(j=0;j<colA;j++) {
	target->ss_alignment[i][j] = -1;
      }
      target->num_aligned[i]  = 0;
    }
  }
  /*  printf("adding row %i and column %i\n",r,c);*/
  query->ss_alignment[c][0] = r;
  query->num_SSEs_aligned++;
  query->num_aligned[c]++;
  target->ss_alignment[r][0] = c;
  target->num_SSEs_aligned++;
  target->num_aligned[r]++;
  /*  for(i=0;i<data[myindex].numptrs;i++) {
    traceBack(data[myindex].Rptr[i],data[myindex].Cptr[i],query,target);
    }*/
  if(data[myindex].numptrs != 0) {
    traceBack(data[myindex].Rptr[data[myindex].bestPointerIndex],data[myindex].Cptr[data[myindex].bestPointerIndex],query,target);
  }
}

/*passed to qsort to sort alignment scores in decreasing order*/
int compare_alignments(const void *a, const void *b) {
  int pointer1, pointer2;
  
  pointer1 = *((int *)a);
  pointer2 = *((int *)b);
  if(all_alignments.scores[pointer1] < all_alignments.scores[pointer2]) { 
    return 1;
  }else if(all_alignments.scores[pointer1]>all_alignments.scores[pointer2]) {
    return -1;
  }else {
    return 0;
  }
}


void PrintDPScores(int index) {
  int j;
  float total_SSE_score = 0;

  for(j=0;j<7;j++) {
    total_SSE_score += all_alignments.dp_scores[index][j];
  }
  printf("%8.3f\t",total_SSE_score);
  for(j=0;j<7;j++) {
    printf("%8.3f\t",all_alignments.dp_scores[index][j]);
  }

}

float CalcConstantScore(int r, int c) {
  float retvalue;
  float score1, score2;
  /*  retvalue = computeScore_ss(r,c,&seqA,&seqB)+computeAvgDistScore_ss(r,c,&seqA,&seqB)+computeLengthScore(r,c,&seqA,&seqB);*/

  score1 = computeAvgDistScore_ss(r,c,&seqA,&seqB);
  if(score1<0) {
    retvalue=-1000;
  }else {
    retvalue = computeScore_ss(r,c,&seqA,&seqB)+score1+computeLengthScore(r,c,&seqA,&seqB);
  }


  return(retvalue);
}

void CalcDPScores(OBJ* query, OBJ* target) {
  int i,j,myindex;
  int prev_query = -2;
  int prev_target = -2;
  int query_ssnum;
  float dummy;
  float prev_score = 0,newscore;
  /************
  for(i=0;i<target->ssArrayLength;i++) {
    if(target->ss_alignment[i][0]!=-1) {
      query_ssnum = target->ss_alignment[i][0];
      if((query_ssnum == prev_query+1 || i==prev_target+1)) {
	dummy = computeAngleScore(i,query_ssnum,prev_target,prev_query,&seqA,&seqB);
	dummy = computeOriginScore(i,query_ssnum,prev_target,prev_query,&seqA,&seqB);
      }else {
	all_alignments.dp_scores[i][0] = 0;
	all_alignments.dp_scores[i][1] = 0;
	all_alignments.dp_scores[i][2] = 0;
	all_alignments.dp_scores[i][3] = 0;
      }
      dummy = computeScore_ss(i,query_ssnum,&seqA,&seqB);
      dummy = computeAvgDistScore_ss(i,query_ssnum,&seqA,&seqB);
      dummy = computeLengthScore(i,query_ssnum,&seqA,&seqB);
    }
    prev_query = query_ssnum;
    prev_target = i;
  }
  ************/
  
  myindex = maxIndex[1]+maxIndex[0]*colA;
  /*traceBack(maxIndex[0],maxIndex[1],query,target);*/
  for(i=0;i<query->ssArrayLength;i++) {
    query->ss_dpscores[i] = 0;
    if(query->ss_alignment[i][0]!=-1) {
      myindex = i+query->ss_alignment[i][0]*colA;
      /*      if((prev_query + 1 == i) || (prev_target+1 == query->ss_alignment[i][0])) {
	newscore = data[myindex].score - prev_score;
	if(newscore < 0) {
	  newscore = data[myindex].score;
	}
      }else {
	newscore = data[myindex].score;
	}*/
      newscore = data[myindex].score - prev_score;
      /*      printf("%i,%i,%f=> ",data[myindex].Rptr[0],data[myindex].Cptr[0],data[myindex].score);*/
      /*      printf("%8.3f\n",newscore);*/
      query->ss_dpscores[i] = newscore;
      target->ss_dpscores[query->ss_alignment[i][0]] = newscore;
      prev_score = data[myindex].score;
      prev_query = i;
      prev_target = query->ss_alignment[i][0];
    }
  }
  /*  printf("\n");*/
}


void calcAllDPScores(OBJ* query, OBJ* target, char *alignedfile, char *unalignedfile) {
  FILE *outfile_aligned, *outfile_unaligned;
  int i,j,k,l,h;
  float score;
  
  
  outfile_aligned = fopen(alignedfile,"a");
  outfile_unaligned = fopen(unalignedfile,"a");
  for(i=0;i<query->ssArrayLength;i++) {
    for(k=0;k<target->ssArrayLength;k++) {
      if(seqA.ss[i] == seqB.ss[k]) {
	if(target->ss_alignment[k][0] == i) {
	  fprintf(outfile_aligned,"%7.3f\t",computeScore_ss(k,i,&seqA,&seqB));
	  ;
	  fprintf(outfile_aligned,"%7.3f\t", 	computeAvgDistScore_ss(k,i,&seqA,&seqB));
	  fprintf(outfile_aligned,"%7.3f",	computeLengthScore(k,i,&seqA,&seqB));
	  fprintf(outfile_aligned,"\n");
	}else {
	  fprintf(outfile_unaligned,"%7.3f\t",	computeScore_ss(k,i,&seqA,&seqB));
	  fprintf(outfile_unaligned,"%7.3f\t", 	computeAvgDistScore_ss(k,i,&seqA,&seqB));
	  fprintf(outfile_unaligned,"%7.3f",	computeLengthScore(k,i,&seqA,&seqB));
	  fprintf(outfile_unaligned,"\n");
	}
      }
    }
  }

  fclose(outfile_aligned);
  fclose(outfile_unaligned);
}


void freeAlignments() {
  int i,j;

  if(all_alignments.columnpointers == NULL) {
    return;
  }
  free(data);
  data = NULL;
  free(all_alignments.scores);
  free(all_alignments.pointers);
  for(i=0;i<(rowB*colA);i++) {
    free(all_alignments.rowpointers[i]);
    free(all_alignments.columnpointers[i]);
  }
  for(i=0;i<rowB;i++) {
    free(all_alignments.dp_scores[i]);
    for(j=0;j<rowB;j++) {
      if(originScoresB[i][j]!=NULL) {
	free(originScoresB[i][j]);
      }
    }
    free(originScoresB[i]);
  }
  for(i=0;i<colA;i++) {
    for(j=0;j<colA;j++) {
      if(originScoresA[i][j]!=NULL) {
	free(originScoresA[i][j]);
      }
    }
    free(originScoresA[i]);
  }
  free(originScoresA);
  free(originScoresB);
  free(all_alignments.dp_scores);
  free(all_alignments.rowpointers);
  free(all_alignments.columnpointers);
  all_alignments.columnpointers = NULL;	 
}

/*Some utility functions*/

void initStackElement(stackElement *s) {
  s->Rval = -1;
  s->Cval = -1;
  s->nextPtr = NULL;
}

void initstackType() {
  myStack.stackSize = 0;
  myStack.topOfStack = NULL;
}

void readStack(stackElement *sArray) {
  
  int i;
  stackElement *currentPtr;

  currentPtr = myStack.topOfStack;
  i = 0;

  while (currentPtr != NULL) {
    sArray[i].Rval = currentPtr->Rval;
    sArray[i].Cval = currentPtr->Cval;
    currentPtr = currentPtr->nextPtr;
    i++;
  }

  if (i != myStack.stackSize) 
    printf("Something's amiss in the stack\n");

}

int getStackSize() {

  return myStack.stackSize;

}

void pushStack(stackElement *p) {

  p->nextPtr = myStack.topOfStack;
  myStack.topOfStack = p;

  myStack.stackSize++;
}

stackElement* popStack() {

  stackElement *tempElement;

  tempElement = myStack.topOfStack;
  myStack.topOfStack = tempElement->nextPtr;
  myStack.stackSize--;

  return tempElement;
}


void fixMax(int r, int c) {
    int myindex;

    myindex = c + r*colA;

    if (local){
	if (data[myindex].score > maxScore) {
	    maxScore = data[myindex].score;
	    maxIndex[0] = r;
	    maxIndex[1] = c;
	}
    }
    else if ((r == rowB - 1) || (c == colA - 1)) {		
	if (data[myindex].score > maxScore) {
	    maxScore = data[myindex].score;
	    maxIndex[0] = r;
	    maxIndex[1] = c;
	}
    }       
}


void rotateAboutZAxis(double theta, seq* myseq) {
  double R[3][3];
  double newvec[3];
  int i;
  
  R[0][0] = cos(theta); R[0][1] = -sin(theta); R[0][2] = 0;
  R[1][0] = sin(theta); R[1][1] = cos(theta); R[1][2] = 0;
  R[2][0] = 0; R[2][1] = 0; R[2][2] = 1;

  for(i=0;i<myseq->length;i++) {
    newvec[0]=newvec[1]=newvec[2]=0;
    newvec[0] = myseq->vector[i][0]*R[0][0] + myseq->vector[i][1]*R[1][0] + myseq->vector[i][2]*R[2][0];
    newvec[1] = myseq->vector[i][0]*R[0][1]+myseq->vector[i][1]*R[1][1]+myseq->vector[i][2]*R[2][1];
    newvec[2] = myseq->vector[i][0]*R[0][2]+myseq->vector[i][1]*R[1][2]+myseq->vector[i][2]*R[2][2];
    myseq->vector[i][0]=newvec[0];
    myseq->vector[i][1]=newvec[1];
    myseq->vector[i][2]=newvec[2];

    /*similarly for origin,end,vector_q1,vector_center,vector_q3*/
  }
    
}

void rotateAboutYAxis(double theta, seq* myseq) {
  double R[3][3];
  double newvec[3];
  int i;

  R[0][0]=cos(theta); R[0][1] = 0; R[0][2] = -sin(theta);
  R[1][0]=0;R[1][1]=1;R[1][2]=0;
  R[2][0]=sin(theta);R[2][1]=0;R[2][2]=cos(theta);
}


void rotateAboutXAxis(double theta, seq* myseq) {
  double R[3][3];
  double newvec[3];
  int i;

  R[0][0]=1; R[0][1] = 0; R[0][2] =0;
  R[1][0]=0;R[1][1]=cos(theta);R[1][2]=-sin(theta);
  R[2][0]=cos(theta);R[2][1]=sin(theta);R[2][2]=cos(theta);
}

/*mc plan: move set =
           choose rotation about x,y,or z axis, or translation in x,y, or z direction
           if angle:
              angle += [-0.035,0.035] (somewhere between -2 and 2 degrees; maybe less?)
           if distance:
              distance += [-.2,.2] (move less than 0.2A)

    make metropolis criterion more stringent as move number increases
      p = abs((A-delta(score))/sqrt(move number)), where A is some arbitrary number. CE's is 25% higher than the maximum score for a position. Try something around 10 since we're not expecting major changes
    stop when changes in alignment score are minor


rotation matrices:

rotation about z axis = 

*/ 

