/*Copyright 2003 The Board of Trustees of The Leland Stanford Junior University.
  All Rights Reserved.

  Communication to Jessica Ebert (jessicas@stanford.edu)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "defines.h"
#include "transform.h"
#include "ssalign.h"
#ifdef DISPLAYON
#include <GL/gl.h>
#include <tcl.h>
#include <tk.h>
#include "display.h"
 #endif


#define CHANGE_TRANSFORM 1 
#define PRINT_DP_SCORES 1

int globalgotone = 0;
int globalgottwo = 0;
int mypid;
char* lockdir;
char* scriptdir;
char* dsspprogram;
int useBetaCarbons = 1;
int lock_residues = 0;

extern void usageError(); 
extern char* getFileName(int num);

#ifdef DISPLAYON

extern void setupGLWindow(char*);
extern int My_Tk_Init(int, char**, Tcl_AppInitProc*);
extern void InitPosition(int, int, int, int);
extern void addReshapeFunc(void (*Func)(GLsizei, GLsizei));
extern void addKeyFunc(int key, void (*Func)(void));
extern void addMouseFunc(int mouse, int mode, void (*Func)(AUX_EVENTREC *));
extern void tellTcl(int, char*, int);
extern void runTclFile(char*);
extern void tellTclAnything(char*);

extern Display *mydisplay;
extern Window GLwin;

extern void doGraphicsEvent(ClientData clientdata, XEvent *current);



GLfloat colorTable[2][3] = {
  {0.0, 0.0, 1.0},
  {1.0, 0.0, 0.0}
};

int currentColor = 0;

static GLfloat LdownXold = -1;
static GLfloat LdownYold = -1;
static GLfloat MdownXold = -1;
static GLfloat MdownYold = -1;
static GLfloat RdownXold = -1;
static GLfloat RdownYold = -1;

int jointModelNumber;
GLfloat currentTransformMat[16];
GLfloat jointModelMat[16];

int displayType = 0;

void createObjList(OBJ*);
void createJointList();
void display(void);
void init(void);
void myReshape(GLsizei, GLsizei);
void setBindings(void);
void setCurrentOne(int);
void setCurrentTwo(int);
void centerObjects();
void setLocalOrGlobal(int);
void setDisplayType(int);

void setThresholdVar(int);
void mouseLeftDown(AUX_EVENTREC*);
void mouseLeftUp(AUX_EVENTREC*);
void mouseMiddleDown(AUX_EVENTREC*);
void mouseMiddleUp(AUX_EVENTREC*);
void mouseRightDown(AUX_EVENTREC*);
void mouseRightUp(AUX_EVENTREC*);
void keyUP();
void keyDOWN();

#endif



OBJ* currentObjList[2];
int numCurrentObjs = 0;
OBJ* ObjList[2];

OBJ* currentObjList0[2];
OBJ* ObjList0[2];


extern char username[2][300];
extern char pdbname[2][13];
extern char chain[2]; 
extern FILE* filePtrs[2]; 
extern int newfile[2];
extern char outputDir[300];
extern int runStep1;
extern float rotation[3][3];
extern float translation[3];

int globalStartDisplay = 1;
int globalComputeCore = 1;
int globalUseSSEWeights = 0;
GLfloat distThreshold = 4.0; /*DISTANCE THRESHOLD HERE: used to be 4*/
GLfloat angleWeight = 15.0;
GLfloat distanceWeight = 5.0;
GLfloat gapMultiplier=3.0;
GLfloat gapExtensionMultiplier = 2.7;
int thresholdOFF = 0; /*TURNS THRESHOLD OFF*/
int displayPoints = 0;
FILE *outfile = NULL;

float globalRMSD = 0.0;
int globalNumAligned = 0;
float globalCoreRMSD = 0.0;
int globalCoreNumAligned = 0;
int  globalLastAlignment = 0;  /* 0 for structalign and 1 for refinealign */

float globalRMSD0 = 0.0;
int globalNumAligned0 = 0;
float globalCoreRMSD0 = 0.0;
int globalCoreNumAligned0 = 0;


int globalSearchDatabase = 0;
FILE *globalDatabaseFile = NULL;

float query_points_to_save[MAXLENGTH][3]; /*used to hold coordinates of SSE elements that were re-aligned*/
float target_points_to_save[MAXLENGTH][3];

clock_t start_time;
clock_t end_time;
double runtime;

int minAlignedResidues = 6;


extern OBJ* readFile(int); 
extern void checkopts(int, char**);


extern int local;

void setUseBetaCarbons(int);
void setDistThreshold(float);
void structAlignSingle(OBJ* t0, OBJ* t1);
void structAlignFull(OBJ* t0, OBJ* t1);
void alignUsingVectorsSingle(OBJ* t0, OBJ* t1,int SSE);
void alignUsingVectorsFull(OBJ* t0, OBJ* t1, int SSE);
void initAlignUsingVectorsSingle(OBJ* t0, OBJ* t1, int SSE);
void initAlignUsingVectorsFull(OBJ* t0, OBJ* t1,int SSE);
void sequenceAlign();
void quitFunction();
void printOutput(OBJ*, OBJ*, int, int);
void printStructAlignment(OBJ* query, OBJ* target, int start, int end);
void printAlignmentMatrices();
void loadAlignmentMatrices();
void coreAlignFull(OBJ* t0, OBJ* t1);
void coreAlignSingle(OBJ* t0, OBJ* t1);
void doAllAlignments();
void doSSEAlignments(OBJ* query, OBJ* target);
void lockResidues(OBJ* query, OBJ* target);
float recalculateAlignment(OBJ* query, OBJ* target);
int breakSSELengthTie(OBJ* query, OBJ* target);
void applyMat(OBJ*, OBJ*, GLfloat*);
void applyMtoV(GLfloat*, GLfloat*, GLfloat*);
void applyRottoV(GLfloat*, GLfloat*, GLfloat*);
void printMat(GLfloat*);
GLfloat findNearest(OBJ*, GLfloat*, char, GLfloat*, int*);
GLfloat findNearestSSEResidue(OBJ*, GLfloat*, float *, int start, int end, char, GLfloat*, int*, int match_loops, int match_beta);
GLfloat findNearestUnpaired(OBJ*, int *, GLfloat*, GLfloat*,int, int, char, GLfloat*, int*, int);
float computeRMSD(OBJ*, OBJ*);
float structAlign(OBJ*, OBJ*, int, float);
float alignUsingVectors(OBJ*, OBJ*, int, float, int, int);
float refineAlign(OBJ*, OBJ*, int, float, int);
void findBestSequence(int*, float*, int, int);    
void integrateSSEObject(OBJ* query, OBJ* target);
void freeObj(OBJ*);

int inputObject(int);
void setupOutputFile();
void gzipOutputFile();

void createAlignmentArray(OBJ*, OBJ*);
void createNextPairedResidueArray(OBJ*);
/****************************************************************************************
                                  Matrix operations etc  
****************************************************************************************/

inline double round(double x)
{
 return floor(x + 0.5);
}



void myMatMult(GLfloat *m1, GLfloat*m2) {

  int i;
  GLfloat t[3];

  for (i = 0; i < 4; i++) {
    t[0] = m1[i*4] * m2[0] + m1[i*4 + 1] * m2[4] + m1[i*4 + 2] * m2[8];
    t[1] = m1[i*4] * m2[1] + m1[i*4 + 1] * m2[5] + m1[i*4 + 2] * m2[9];
    t[2] = m1[i*4] * m2[2] + m1[i*4 + 1] * m2[6] + m1[i*4 + 2] * m2[10];

    m1[i*4] = t[0];
    m1[i*4 + 1] = t[1];
    m1[i*4 + 2] = t[2];

  }

  m1[12] = m1[12] + m2[12];
  m1[13] = m1[13] + m2[13];
  m1[14] = m1[14] + m2[14];

}



void applyMat(OBJ* retObj, OBJ* myobj, GLfloat* mat) {

  int i;
  float t[3];


  for (i = 0; i < myobj->numPts; i++) {
    retObj->points[i][0] = (myobj->points[i][0]) * mat[0] + 
                           (myobj->points[i][1]) * mat[4] +
                           (myobj->points[i][2]) * mat[8] + mat[12];
    retObj->points[i][1] = (myobj->points[i][0]) * mat[1] + 
                           (myobj->points[i][1]) * mat[5] +
                           (myobj->points[i][2]) * mat[9] + mat[13];
    retObj->points[i][2] = (myobj->points[i][0]) * mat[2] + 
                           (myobj->points[i][1]) * mat[6] +
                           (myobj->points[i][2]) * mat[10] + mat[14];

    retObj->cbpoints[i][0] = (myobj->cbpoints[i][0]) * mat[0] + 
                           (myobj->cbpoints[i][1]) * mat[4] +
                           (myobj->cbpoints[i][2]) * mat[8] + mat[12];
    retObj->cbpoints[i][1] = (myobj->cbpoints[i][0]) * mat[1] + 
                           (myobj->cbpoints[i][1]) * mat[5] +
                           (myobj->cbpoints[i][2]) * mat[9] + mat[13];
    retObj->cbpoints[i][2] = (myobj->cbpoints[i][0]) * mat[2] + 
                           (myobj->cbpoints[i][1]) * mat[6] +
                           (myobj->cbpoints[i][2]) * mat[10] + mat[14];



  }
}



void applyMtoV(GLfloat *vout, GLfloat *vin, GLfloat *matin) {

  GLfloat t[3];

  t[0] = (vin[0])* matin[0] + (vin[1])* matin[4] + (vin[2])* matin[8] + matin[12];
  t[1] = (vin[0])* matin[1] + (vin[1])* matin[5] + (vin[2])* matin[9] + matin[13];
  t[2] = (vin[0])* matin[2] + (vin[1])* matin[6] + (vin[2])* matin[10] + matin[14];

  vout[0] = t[0];
  vout[1] = t[1];
  vout[2] = t[2];

}


void applyRottoV(GLfloat *vout, GLfloat *vin, GLfloat *matin) {

  GLfloat t[3];

  t[0] = (vin[0])* matin[0] + (vin[1])* matin[4] + (vin[2])* matin[8];
  t[1] = (vin[0])* matin[1] + (vin[1])* matin[5] + (vin[2])* matin[9];
  t[2] = (vin[0])* matin[2] + (vin[1])* matin[6] + (vin[2])* matin[10];

  vout[0] = t[0];
  vout[1] = t[1];
  vout[2] = t[2];

}



float computeRMSD(OBJ* query, OBJ* target) {

  GLfloat tempVec[3];
  GLfloat tdist;
  GLfloat rmsd = 0.0;
  int i, count = 0, tnum;

  for (i = 0; i < target->numPts; i++) {
    tdist = findNearest(query, target->points[i], target->sstype[i], tempVec, &tnum);
    
    if ((tdist <= distThreshold) || (thresholdOFF)) {
      rmsd += tdist*tdist;
      count++;
    }    
  }

  rmsd = sqrt((rmsd)/((GLfloat)(count)));
  globalNumAligned = count;

  return rmsd;
}


float distBetweenVecs(float *vec0, float *vec1) {

  return sqrt((vec1[0] - vec0[0])*(vec1[0] - vec0[0]) +
	      (vec1[1] - vec0[1])*(vec1[1] - vec0[1]) +
	      (vec1[2] - vec0[2])*(vec1[2] - vec0[2]));

}


float distanceBetweenPoints(float *p1, float *p2) {

  return sqrt((p1[0] - p2[0])*(p1[0] - p2[0]) +
	      (p1[1] - p2[1])*(p1[1] - p2[1]) +
	      (p1[2] - p2[2])*(p1[2] - p2[2]));
}
  
/*findNearest: Given a point pVec and its sstype (ss), findNearest returns the nearest target
 *point such that the ss types match, or one of the points is a loop residue. The nearest point
 *is returned in rVec, and the residue number (where the first residue is number 0) is returned
 *in targetNum.
*/
GLfloat findNearest(OBJ* target, GLfloat* pVec, char ss, GLfloat* rVec, int *targetNum) {

  int i, num = 0;
  GLfloat min = 10000.0;
  GLfloat dist;    /*distance between alpha carbons*/

  for (i = 0; i < target->numPts; i++) {
    
    if ((ss == target->sstype[i]) || (ss == 'U') || (target->sstype[i] == 'U')) {
      
      dist = ((target->points[i][0] - pVec[0])*(target->points[i][0] - pVec[0]) +
		  (target->points[i][1] - pVec[1])*(target->points[i][1] - pVec[1]) +
		  (target->points[i][2] - pVec[2])*(target->points[i][2] - pVec[2]));
      if (dist < min) {
	min = dist;
	num = i;
      }
    }
  }

  rVec[0] = target->points[num][0];
  rVec[1] = target->points[num][1];
  rVec[2] = target->points[num][2];

  *targetNum = num;
  min = sqrt(min);
  return min;

}

/*findNearestSSEResidue: Same as findNearest, but if match_loops is 0, loop residues are ignored. If
 *match_loops is 1, loop residues may be matched to the point in pVec if and only if ss=='U' (i.e. if
 *the residue at location pVec is a loop residue. The matching residue is constrained to lie betweein
 *the start and end target residues.
 */
GLfloat findNearestSSEResidue(OBJ* target, GLfloat* pVec, GLfloat* pVecAlpha, int start, int end,char ss, GLfloat* rVec, int *targetNum, int match_loops, int match_betas) {

  int i, j,num = 0;
  GLfloat min = 10000.0;
  GLfloat dist;    /*distance between alpha carbons*/
  GLfloat SSE_min = 10000.0;
  int SSE_num;
  float dotproduct;
  float min_dotproduct;
  float u[3];
  float v[3];
  float len;
  float angle,min_angle=360;
  float epsilon=0.0001;
  float angle_cutoff;

  angle_cutoff = PI/2.0;
  num=-1;
  len=0;
  for(j=0;j<3;j++) {
    u[j]=pVec[j]-pVecAlpha[j];
    len += u[j]*u[j];
  }
  len = sqrt(len);
  for(j=0;j<3;j++) {
    u[j]/=(float)len;
  }
  for (i = start; i <=end; i++) {
    
    if((ss==target->sstype[i]) || !((ss=='H' && target->sstype[i]=='S') || (ss=='S' && target->sstype[i]=='H'))) {
      /*    if ((ss == target->sstype[i]) || ((match_loops)&&(ss=='U' || target->sstype[i]==ss || target->sstype[i] == 'U'))) {*/
      /*      if(match_betas && target->objSeq[i] != 'G') {*/
      if(match_betas) {
	dist = ((target->cbpoints[i][0] - pVec[0])*(target->cbpoints[i][0] - pVec[0]) +
		  (target->cbpoints[i][1] - pVec[1])*(target->cbpoints[i][1] - pVec[1]) +
		  (target->cbpoints[i][2] - pVec[2])*(target->cbpoints[i][2] - pVec[2]));

	len=0;
	for(j=0;j<3;j++) {
	  v[j]=target->cbpoints[i][j]-target->points[i][j];
	  len+=v[j]*v[j];
	}
	len=sqrt(len);
	dotproduct=0;
	for(j=0;j<3;j++) {
	  v[j]/=len;
	  dotproduct += u[j]*v[j];
	} 
	if(dotproduct<= 1+epsilon && dotproduct >= 1-epsilon) {
	  angle=0;
	}else if(dotproduct<=-1+epsilon && dotproduct>=-1+epsilon) {
	  angle=2*PI;
	}else {
	  angle = acos((double)dotproduct);
	}
      }else {
      
	dist = ((target->points[i][0] - pVec[0])*(target->points[i][0] - pVec[0]) +
		    (target->points[i][1] - pVec[1])*(target->points[i][1] - pVec[1]) +
		    (target->points[i][2] - pVec[2])*(target->points[i][2] - pVec[2]));
      }
      /**********8
      if (dist < min) {

	min = dist;
	num = i;
      }
      **********/

      if(dist <= distThreshold*distThreshold && angle < min_angle && angle_cutoff) {
	min_angle = angle;
	min = dist;
	num=i;
      }

    }
  }

  rVec[0] = target->points[num][0];
  rVec[1] = target->points[num][1];
  rVec[2] = target->points[num][2];

  *targetNum = num;
     min = sqrt(min);
  /*  printf("min: %i %f\n",num,min*min);*/

     if(num!=-1) {
       return min;
     }else {
       return -1;
     }

}

/*findNearestUnpaired: Same as findNearest, but the matching residue must not already be paired to a
 *residue. The argument pairNums was added as a convenience so that the target object need not have a 
 *pairNums array associated with it (pairNums is normally target->pairNums). The matching residue is
 *constrained to lie between start and end target residues.
 */
GLfloat findNearestUnpaired(OBJ* target, int *pairNums, GLfloat* pVec, GLfloat* pVecAlpha, int start, int end,char ss, GLfloat* rVec, int *targetNum, int match_betas) {

  int i, j,num = 0;
  GLfloat min = 10000.0;
  GLfloat dist;    /*distance between alpha carbons*/
  float dotproduct,min_dotproduct;
  float u[3],v[3];
  float len;
  float angle,min_angle=360;
  float epsilon=0.0001;
  float angle_cutoff;

  angle_cutoff = PI/2.0;
  /*  match_betas=0;*/
  len=0;
  for(j=0;j<3;j++) {
    u[j]=pVec[j]-pVecAlpha[j];
    len += u[j]*u[j];
  }
  len = sqrt(len);
  for(j=0;j<3;j++) {
    u[j]/=(float)len;
  }
  
  for (i =start; i <=end; i++) {

    if(pairNums[i] == -1) {
      if ((ss == target->sstype[i]) || (ss=='U' || target->sstype[i] == 'U')) {
	
	if(match_betas) {
	  dist = ((target->cbpoints[i][0] - pVec[0])*(target->cbpoints[i][0] - pVec[0]) +
		  (target->cbpoints[i][1] - pVec[1])*(target->cbpoints[i][1] - pVec[1]) +
		  (target->cbpoints[i][2] - pVec[2])*(target->cbpoints[i][2] - pVec[2]));
	
	  len=0;
	  for(j=0;j<3;j++) {
	    v[j]=target->cbpoints[i][j]-target->points[i][j];
	    len+=v[j]*v[j];
	  }
	  len=sqrt(len);
	  dotproduct=0;

	  for(j=0;j<3;j++) {
	    v[j]/=len;
	    dotproduct += u[j]*v[j];
	  } 
	if(dotproduct<= 1+epsilon && dotproduct >= 1-epsilon) {
	  angle=0;
	}else if(dotproduct<=-1+epsilon && dotproduct>=-1+epsilon) {
	  angle=2*PI;
	}else {
	  angle = acos((double)dotproduct);
	}

	}else {
	dist = ((target->points[i][0] - pVecAlpha[0])*(target->points[i][0] - pVecAlpha[0]) +
		    (target->points[i][1] - pVecAlpha[1])*(target->points[i][1] - pVecAlpha[1]) +
		    (target->points[i][2] - pVecAlpha[2])*(target->points[i][2] - pVecAlpha[2]));

	}

	if(!match_betas) {
	  if (dist < min) {
	    min = dist;
	    num = i;
	  }
	}else {
	  if(dist <=distThreshold*distThreshold && angle < min_angle && angle<angle_cutoff) {
	    min_angle=angle;
	    min=dist;
	    num=i;
	  }else if(dist < min && angle==min_angle) {
	    min_angle=angle;
	    min=dist;
	    num=i;	  
	  }
	}

/***********
	if(dist<min) {
	  min=dist;
	  num=i;
	}
      ********/
      }
    }
  }

  rVec[0] = target->points[num][0];
  rVec[1] = target->points[num][1];
  rVec[2] = target->points[num][2];

  *targetNum = num;
  min = sqrt(min);
  return min;

}

/*printMat: Prints the alignment matrix (passed in as argument m)*/
void printMat(GLfloat *m) {
  
  int i;

  for (i = 0; i<16; i++) {
    if (i%4 == 0) 
      fprintf(outfile, "\n");
    fprintf (outfile, "%f  ", m[i]);
  }

  fprintf(outfile, "\n");  
  
}



#ifdef DISPLAYON

/****************************************************************************************
                                  OpenGL stuff   
****************************************************************************************/


void createObjList(OBJ* myobject) {

  int i, j, count, start, end;
  char currentss;

  glNewList(myobject->modelNumber, GL_COMPILE);

  if (displayType == 0) {
    
    glLineWidth(2.0);
    glColor3f(myobject->color[0],myobject->color[1], myobject->color[2]);   

    glBegin(GL_LINE_STRIP);
    for (i = 0; i < myobject->numPts; i++)
      glVertex3fv(&(myobject->points[i][0]));
    glEnd(); 
  }

  else if (displayType == 1) {
    
    glLineWidth(2.0);
    
    for (i = 0; i < myobject->numPts; i++) {
      currentss = myobject->sstype[i];    
      if (currentss == 'H') {
	glColor3f(myobject->color[0],myobject->color[1] + 1.0, myobject->color[2]);
	count = 0;
	while ((i+count < myobject->numPts) && (currentss == myobject->sstype[i+count])){
	  count++;
	}
	
	glBegin(GL_LINE_STRIP);
	
	for (j = i; j < i+count; j++) 
	  glVertex3fv(&(myobject->points[j][0]));
	
	glEnd();
	
	i = i + count - 1;
	
      }
      else if (currentss == 'S') {
	glColor3f(myobject->color[0],myobject->color[1] + 0.5, myobject->color[2]);
	count = 0;
	while ((i+count < myobject->numPts) && (currentss == myobject->sstype[i+count])){
	  count++;
	}
	
	glBegin(GL_LINE_STRIP);
	
	for (j = i; j < i+count; j++) 
	  glVertex3fv(&(myobject->points[j][0]));
	
	glEnd();
	
	i = i + count - 1;
	
      }
      else if (currentss == 'U') {
	glColor3f(myobject->color[0],myobject->color[1], myobject->color[2]);
	count = 0;
	while ((i+count < myobject->numPts) && (currentss == myobject->sstype[i+count])){
	  count++;
	}
	
	start = i;
	end = i + count;
	
	if (start > 0) {
	  start--;
	}
	
	if (end < myobject->numPts - 1) {
	  end++;
	}
	
	glBegin(GL_LINE_STRIP);
	
	for (j = start; j < end; j++) 
	  glVertex3fv(&(myobject->points[j][0]));
	
	glEnd();
	
	
	i = i + count - 1;
	
      }
    }      
    
  }
  
  else {
    if ((displayType == 2) || (displayType == 4)) {
      
      glLineWidth(2.0);
      glColor3f(myobject->color[0],myobject->color[1], myobject->color[2]);

      for (i = 0; i < myobject->numPts; i++) {
	currentss = myobject->sstype[i];    
	if (currentss == 'H') {
	  count = 0;
	  while ((i+count < myobject->numPts) && (currentss == myobject->sstype[i+count])){
	    count++;
	  }
	  
	  glBegin(GL_LINE_STRIP);
	  
	  for (j = i; j < i+count; j++) 
	    glVertex3fv(&(myobject->points[j][0]));
	  
	  glEnd();
	  
	  i = i + count - 1;
	  
	}
      }
    }
      
    if ((displayType == 3) || (displayType == 4)) {
      glLineWidth(2.0);
      glColor3f(myobject->color[0],myobject->color[1], myobject->color[2]);
      
      for (i = 0; i < myobject->numPts; i++) {
	currentss = myobject->sstype[i];    
	if (currentss == 'S') {
	  count = 0;
	  while ((i+count < myobject->numPts) && (currentss == myobject->sstype[i+count])){
	    count++;
	  }
	  
	  glBegin(GL_LINE_STRIP);
	  
	  for (j = i; j < i+count; j++) 
	    glVertex3fv(&(myobject->points[j][0]));
	  
	  glEnd();
	  
	  i = i + count - 1;
	  
	}    
      }
    }

    if (displayType == 5) {
      glLineWidth(2.0);
      glColor3f(myobject->color[0],myobject->color[1], myobject->color[2]);
      for (i = 0; i < myobject->numPts; i++) {
	currentss = myobject->sstype[i];    
	if (currentss == 'U') {
	  count = 0;
	  while ((i+count < myobject->numPts) && (currentss == myobject->sstype[i+count])){
	    count++;
	  }
	  
	  start = i;
	  end = i + count;
	  
	  if (start > 0) {
	    start--;
	  }
	  
	  if (end < myobject->numPts - 1) {
	    end++;
	  }
	  
	  glBegin(GL_LINE_STRIP);
	  
	  for (j = start; j < end; j++) 
	    glVertex3fv(&(myobject->points[j][0]));
	  
	  glEnd();
	  
	  
	  i = i + count - 1;
	  
	}
      }      
    }
  }
  
  glEndList();

}

  
void createJointList() {
  
  int i;
  GLfloat tempvec[3];

  glNewList(jointModelNumber, GL_COMPILE);

  glLineWidth(1.0);
  glColor3f(1.0,1.0,1.0);
  
  glBegin(GL_LINES);

  for (i = 0; i < ObjList[0]->numPts; i++) {
    if (ObjList[0]->pairNums[i] != -1) {
      applyMtoV(tempvec, ObjList[0]->points[i], ObjList[0]->objMat);
      glVertex3fv(tempvec);
      applyMtoV(tempvec, ObjList[1]->points[(ObjList[0]->pairNums[i])], ObjList[1]->objMat);
      glVertex3fv(tempvec);
      
    }
  }

  glEnd();

  glEndList();

  glPushMatrix();
  glLoadIdentity();
  glGetFloatv(GL_MODELVIEW_MATRIX, jointModelMat);
  glPopMatrix();

  displayPoints = 1;

}
    

void display(void) {
  
  int i;

  if (globalStartDisplay) {
    glClearColor(0.0, 0.0, 0.0, 1.0); 
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 
    
    if (globalSearchDatabase == 0) {
      if (globalgotone) {
	glPushMatrix();
	glMultMatrixf(ObjList[0]->objMat);
	glCallList(ObjList[0]->modelNumber);
	glPopMatrix();
      }      
      if (globalgottwo) {
	glPushMatrix();
	glMultMatrixf(ObjList[1]->objMat);
	glCallList(ObjList[1]->modelNumber);
	glPopMatrix();
      }      
      
#if DRAWJOINTS
      if (displayPoints == 1) {
	glPushMatrix();
	glMultMatrixf(jointModelMat);      
	glCallList(jointModelNumber);
	glPopMatrix();
      }
#endif
      
    }
    glFlush();
    glXSwapBuffers(mydisplay, GLwin); 
  }

}

void init(void) {
  
  int i;

  glMatrixMode(GL_MODELVIEW); 
  glShadeModel(GL_FLAT);
  glColor3f(1.0, 0.0, 0.0);

  glLoadIdentity();
  glTranslatef(EYEX, EYEY, EYEZ); 

  glFlush();

}

void myReshape(GLsizei w, GLsizei h) {

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-SIZEX/2.0, SIZEX/2.0, -SIZEY/2.0, SIZEY/2.0, -90.0, 90.0);
  /*  glFrustum(-SIZEX/2.0, SIZEX/2.0, -SIZEY/2.0, SIZEY/2.0, 20.0, 90.0);*/
  glMatrixMode(GL_MODELVIEW); 
  glViewport(0, 0, w, h); 

  display();

}




/****************************************************************************************
                            User interface(tcl-tk) and graphics events
****************************************************************************************/

void setBindings(void) {

  addKeyFunc(AUX_UP, keyUP);      /* increase threshold */
  addKeyFunc(AUX_DOWN, keyDOWN);  /* decrease threshold */

  addMouseFunc(AUX_LEFTBUTTON, AUX_MOUSELOC, mouseLeftDown);
  addMouseFunc(AUX_LEFTBUTTON, AUX_MOUSEUP, mouseLeftUp);
  addMouseFunc(AUX_MIDDLEBUTTON, AUX_MOUSELOC, mouseMiddleDown);
  addMouseFunc(AUX_MIDDLEBUTTON, AUX_MOUSEUP, mouseMiddleUp);
  addMouseFunc(AUX_RIGHTBUTTON, AUX_MOUSELOC, mouseRightDown);
  addMouseFunc(AUX_RIGHTBUTTON, AUX_MOUSEUP, mouseRightUp);
}


void setCurrentOne(int val) {

  if (val == 1) {
      
    if (numCurrentObjs == 0) {     
      currentObjList[0] = ObjList[0];
      numCurrentObjs = 1;
    }
    else if ((numCurrentObjs == 1) && (currentObjList[0] == ObjList[1])) {
      currentObjList[1] = ObjList[0];
      numCurrentObjs = 2;
    }
  }
  
  else if (val == 0) {
    
    if (numCurrentObjs == 1) {
      numCurrentObjs = 0;
    }
    
    else if (numCurrentObjs == 2) {
      currentObjList[0] = ObjList[1];
      numCurrentObjs = 1;
    }
  }
  
}

void setCurrentTwo(int val) {

  if (val == 1) {

    if (numCurrentObjs == 0) {     
      currentObjList[0] = ObjList[1];
      numCurrentObjs = 1;
    }
    else if ((numCurrentObjs == 1) && (currentObjList[0] == ObjList[0])) {
      currentObjList[1] = ObjList[1];
      numCurrentObjs = 2;
    }
  }
  
  else if (val == 0) {
    
    if (numCurrentObjs == 1) {
      numCurrentObjs = 0;
    }
    
    else if (numCurrentObjs == 2) {
      currentObjList[0] = ObjList[0];
      numCurrentObjs = 1;
    }
  }

}


void centerObjects() {

  int i;

  for (i = 0; i < numCurrentObjs; i++) {
    
    glPushMatrix();
    
    glLoadIdentity();
    glTranslatef(-currentObjList[i]->center[0], -currentObjList[i]->center[1],  
		 -currentObjList[i]->center[2]);
    glMultMatrixf(currentObjList[i]->objMat);
    glGetFloatv(GL_MODELVIEW_MATRIX, currentObjList[i]->objMat);

    currentObjList[i]->center[0] = 0.0;
    currentObjList[i]->center[1] = 0.0;
    currentObjList[i]->center[2] = 0.0;
    
    glPopMatrix();
    
  }
  
  displayPoints = 0;
  display();

}



void setLocalOrGlobal(int i) {

  local = i;

}


void setDisplayType(int i) {
  
  displayType = i;
  createObjList(ObjList[0]);
  createObjList(ObjList[1]);
  
  display();
  
}

void setUseBetaCarbons(int flag) {
  useBetaCarbons = flag;
}

void setDistThreshold(float i) {
  
  distThreshold = i;

}

void setThresholdVar(int i) {

  if (i == 1) 
    thresholdOFF = 0;
  else if (i == 0)
    thresholdOFF = 1;

}

void mouseLeftDown(AUX_EVENTREC *mouseEvent) {

  GLfloat Xval, Yval, Xt, Yt;
  GLfloat transValue[3];
  int i;
  int smallestObjLen = 100000;
  int smallestObjNum = -1;
  

  Xt = (float)(mouseEvent->data[AUX_MOUSEX]);
  Yt = (float)(mouseEvent->data[AUX_MOUSEY]);

  if ((LdownXold == -1.0) && (LdownYold == -1.0)) {
    LdownXold = Xt;
    LdownYold = Yt;
    if (numCurrentObjs != 2) 
      displayPoints = 0;
  }
  else {

    Xval = (Xt - LdownXold)*0.5;
    Yval = (Yt - LdownYold)*0.5;

    for (i = 0; i < numCurrentObjs; i++) {
      if (currentObjList[i]->numPts < smallestObjLen) {
	smallestObjLen = currentObjList[i]->numPts;
	smallestObjNum = i;
      }
    }


    if (smallestObjNum != -1) {
      transValue[0] = currentObjList[smallestObjNum]->center[0];
      transValue[1] = currentObjList[smallestObjNum]->center[1];
      transValue[2] = currentObjList[smallestObjNum]->center[2];

    }
    
    glPushMatrix();
    
    glLoadIdentity();
    glTranslatef(transValue[0], transValue[1], transValue[2]);
    glRotatef(Xval, 0.0, 1.0, 0.0);
    glRotatef(Yval, 1.0, 0.0, 0.0);
    glTranslatef(-transValue[0], -transValue[1], -transValue[2]);
    
    glGetFloatv(GL_MODELVIEW_MATRIX, currentTransformMat);
    
    for (i = 0; i < numCurrentObjs; i++) {

      glLoadMatrixf(currentTransformMat);
      glMultMatrixf(currentObjList[i]->objMat);
      glGetFloatv(GL_MODELVIEW_MATRIX, currentObjList[i]->objMat);

      applyMtoV(currentObjList[i]->center, currentObjList[i]->center, currentTransformMat); 


    }
    glPopMatrix();

    LdownXold = Xt;
    LdownYold = Yt;

#if DRAWJOINTS
    if (displayPoints == 1) {
      glPushMatrix();
      glLoadMatrixf(currentTransformMat);
      glMultMatrixf(jointModelMat);
      glGetFloatv(GL_MODELVIEW_MATRIX, jointModelMat);
      glPopMatrix();
    }
#endif

    if (numCurrentObjs > 0)
      display();
  }
}

  
void mouseLeftUp(AUX_EVENTREC *mouseEvent) {

  LdownXold = -1.0; 
  LdownYold = -1.0; 

}


void mouseMiddleDown(AUX_EVENTREC *mouseEvent) {

  OBJ* myobject;

  GLfloat Xval, Yval, Xt, Yt;
  int i;

  Xt = (float)(mouseEvent->data[AUX_MOUSEX]);
  Yt = (float)(mouseEvent->data[AUX_MOUSEY]);

  if ((MdownXold == -1.0) && (MdownYold == -1.0)) {
    MdownXold = Xt;
    MdownYold = Yt;
    if (numCurrentObjs != 2) 
      displayPoints = 0;
  }
  else {
    
    Xval = ((Xt - MdownXold)/500.0)*SIZEX;
    Yval = ((MdownYold - Yt)/500.0)*SIZEY;

    glPushMatrix();
    
    glLoadIdentity();
    glTranslatef(Xval, Yval, 0.0);
    glGetFloatv(GL_MODELVIEW_MATRIX, currentTransformMat);
    
    for (i = 0; i < numCurrentObjs; i++) {
      glLoadMatrixf(currentTransformMat);
      glMultMatrixf(currentObjList[i]->objMat);
      glGetFloatv(GL_MODELVIEW_MATRIX, currentObjList[i]->objMat);


      currentObjList[i]->center[0] += Xval;
      currentObjList[i]->center[1] += Yval;

    }    
    glPopMatrix();

    MdownXold = Xt;
    MdownYold = Yt;

#if DRAWJOINTS
    if (displayPoints == 1) {
      glPushMatrix();
      glLoadMatrixf(currentTransformMat);
      glMultMatrixf(jointModelMat);
      glGetFloatv(GL_MODELVIEW_MATRIX, jointModelMat);
      glPopMatrix();
    }
#endif
    if (numCurrentObjs > 0)
      display();
  }
}
    
  
void mouseMiddleUp(AUX_EVENTREC *mouseEvent) {
  
  MdownXold = -1.0; 
  MdownYold = -1.0; 

}    
  
void mouseRightDown(AUX_EVENTREC *mouseEvent) {

  GLfloat Xval, Yval, Xt, Yt;
  int i;

  Xt = (float)(mouseEvent->data[AUX_MOUSEX]);
  Yt = (float)(mouseEvent->data[AUX_MOUSEY]);

  if ((RdownXold == -1.0) && (RdownYold == -1.0)) {    
    RdownXold = Xt;
    RdownYold = Yt;
    if (numCurrentObjs != 2) 
      displayPoints = 0;
  }
  else {

    Xval = ((Xt - RdownXold)/500.0)*SIZEX;
    Yval = ((RdownYold - Yt)/500.0)*SIZEY;

    glPushMatrix();
    
    glLoadIdentity();
    glTranslatef(0.0, 0.0, Yval);
    glGetFloatv(GL_MODELVIEW_MATRIX, currentTransformMat);
    
    for (i = 0; i < numCurrentObjs; i++) {
      glLoadMatrixf(currentTransformMat);
      glMultMatrixf(currentObjList[i]->objMat);
      glGetFloatv(GL_MODELVIEW_MATRIX, currentObjList[i]->objMat);

      currentObjList[i]->center[2] += Yval;

    }    

    glPopMatrix();

    RdownXold = Xt;
    RdownYold = Yt;


#if DRAWJOINTS
    if (displayPoints == 1) {
      glPushMatrix();
      glLoadMatrixf(currentTransformMat);
      glMultMatrixf(jointModelMat);
      glGetFloatv(GL_MODELVIEW_MATRIX, jointModelMat);
      glPopMatrix();
    }
#endif
    if (numCurrentObjs > 0)
      display();
  }
}
    
  
void mouseRightUp(AUX_EVENTREC *mouseEvent) {
  
  RdownXold = -1.0; 
  RdownYold = -1.0; 

}    

void keyUP() {

  distThreshold = distThreshold + 0.5;

  printf("Distance Threshold = %f\n", distThreshold);

}

void keyDOWN() {
  
  if (distThreshold >= 1.5) {
    distThreshold = distThreshold - 0.5;
  }

  printf("Distance Threshold = %f\n", distThreshold);

}


#endif /* DISPLAYON */

/****************************************************************************************
                                    I/O
****************************************************************************************/

/*freeObj: Frees all memory associated with t*/
void freeObj(OBJ* t) {
  int i;

  free(t->ssArray);
  free(t->sstype);
  free(t->objSeq);
  if(t->ssNumbers !=NULL) {
    free(t->ssNumbers);
  }
  for(i=0;i<=t->ssArrayLength;i++) {
    free(t->ss_alignment[i]);

  }
  free(t->ss_alignment);
  free(t->num_aligned);
  free(t->aaNumbers);
  free(t->ss_dpscores);
}

/*inputObject: Reads in an object and places it in ObjList[num]. If num==0 or num==1 and ObjList[num] is
 *occupied, frees ObjList[num] before reading in new object. Prints information on query and target
 *to the output file (query_target.out).*/ 
int inputObject(int num) {

  char tempname[300];
  int i, imax, count;

  if ((num == 0) && (globalgotone == 1)) {
    freeObj(ObjList[0]);
    free(ObjList[0]);
    globalgotone = 0;
  }

  if ((num == 1) && (globalgottwo == 1)) {
    freeObj(ObjList[1]);
    free(ObjList[1]);
    globalgottwo = 0;
  }

  ObjList[num] = readFile(num);

  if (ObjList[num] != NULL) {
    
    if (newfile[num] == 1) {
      
      if (num == 0)
	printf("\n** Query = %s ", username[0]);
      else 
	printf("\n** Target = %s ", username[1]);
      
      if (chain[num] != ' ') {
	printf("  Chain = %c **\n", chain[num]);
      }
      else {
	printf("**\n");
      }
    }
    else {
      if (num == 0)
	printf("\n** Query = %s **\n", pdbname[0]);
      else 
	printf("\n** Target = %s **\n", pdbname[1]);
    }

    printf("%s\n", ObjList[num]->sstype);
    
    if (num == 0) {
      globalgotone = 1;
      numCurrentObjs = 1;
    }
    else {
      globalgottwo = 1;
      numCurrentObjs = 2;
    }

    currentObjList[num] = ObjList[num];
    displayPoints = 0;

#ifdef DISPLAYON
    if (globalStartDisplay) {
      if (newfile[num] == 1) {
	imax = strlen(username[num]);
	count = 0;
	
	for(i = 0; i < imax; i++) {
	  if (username[num][i] != '/') {	
	    tempname[count] = username[num][i];
	    count++;
	  }
	  else {
	    count = 0;
	  }
	}
	tempname[count] = '\0';	
	tellTcl(num, tempname, 1);    
      }
      else {
	tellTcl(num, pdbname[num], 1);
      }
      display();
    }
#endif    
    return 1;
  }
  else {
    return 0;
  }
}

void gzipOutputFile() {
  char outFileStr[300];
  char temp1[300];
  char syscommand[50];

  fprintf(stderr,"gzipoutputfile\n");
  fflush(stderr);
  strcpy(outFileStr, pdbname[0]);
  strcat(outFileStr, "_");
  strcat(outFileStr, pdbname[1]);
  strcat(outFileStr, ".out");

  sprintf(syscommand,"gzip -f %s",outFileStr);
  system(syscommand);
}

void setupOutputFile() {

  char outFileStr[300];
  char temp1[300];
  char syscommand[50];

  strcpy(outFileStr, pdbname[0]);
  strcat(outFileStr, "_");
  strcat(outFileStr, pdbname[1]);
  strcat(outFileStr, ".out");

  if (outfile != NULL) {
    fclose(outfile);
  }

  if (strcmp(outputDir, "NONE") == 0) {
    outfile = fopen(outFileStr, "w");
  }

  else {
    strcpy(temp1, outputDir);
    strcat(temp1, "/");
    strcat(temp1, outFileStr);    
    outfile = fopen(temp1, "w");

    if (outfile == NULL) {
      printf("Trying to create directory %s\n", outputDir);
      sprintf(syscommand, "mkdir %s", outputDir);
      system(syscommand);
      outfile = fopen(temp1, "w");
    }
  }

  if (outfile == NULL) {
    printf("Sorry.  Could not open output file %s", outFileStr);
  }      

}

void printOutput(OBJ* obj0, OBJ* obj1,int query_alignment_start,int query_alignment_end) {
  int i,j,query_offset,target_offset,num_aligned;
  int target_start,target_end,query_start,query_end;
  int num_SSEs;
  int count;
  float rmsd,percent_aligned,finalscore,maxscore;
  int found_first_aligned_SSE;
  OBJ tempObj0, tempObj1;

  if(obj0 == NULL) {
    obj0 = ObjList[0];
    query_alignment_start = 0;
    query_alignment_end = obj0->numPts-1;
  }
  if(obj1 == NULL) {
    obj1 = ObjList[1];
  }

  /*Make temporary objects to be transformed*/
  tempObj0.objSeq = obj0->objSeq; /*Array*/
  tempObj0.sstype = obj0->sstype; /*Array of SS types*/
  tempObj0.numPts = obj0->numPts; /*Array of points*/
  tempObj0.ssArrayLength = obj0->ssArrayLength;
  tempObj0.ssArray = obj0->ssArray; /*Array of #'s giving beginning and
				     *ends of ss elements*/
  tempObj0.num_SSEs_aligned = obj0->num_SSEs_aligned;
  tempObj0.ss_alignment = obj0->ss_alignment;
  applyMat(&tempObj0, obj0, obj0->objMat);
  
  tempObj1.objSeq = obj1->objSeq;
  tempObj1.sstype = obj1->sstype;
  tempObj1.numPts = obj1->numPts;
  tempObj1.ssArrayLength = obj1->ssArrayLength;
  tempObj1.ssArray = obj1->ssArray;
  tempObj1.num_SSEs_aligned = obj1->num_SSEs_aligned;
  tempObj1.ss_alignment = obj1->ss_alignment;
  applyMat(&tempObj1, obj1, obj1->objMat);

  /*Run dynamic programming algorithm to obtain final SSE alignment and score*/
  initsequences_ss(&tempObj0, &tempObj1);

  if(!lock_residues) {
    finalscore = align_ss(1,obj0,obj1);
    report_ss(1,obj0,obj1);
  }

  /*  end_time = time(&end_time);*/
  end_time = clock();
  /*  runtime = difftime(end_time,start_time);*/
  runtime = (end_time-start_time)/((double)CLOCKS_PER_SEC);

  setupOutputFile();
  printStructAlignment(obj0,obj1,query_alignment_start,query_alignment_end);
  printAlignmentMatrices();
  

  query_offset = obj0->aaNums[0]-1;
  target_offset = obj1->aaNums[0]-1;


  
  fprintf(outfile,"\nAligned Secondary Structure Elements %i %i\n",obj0->num_SSEs_aligned,obj0->num_repeated_SSE_alignments+obj1->num_repeated_SSE_alignments);
  if(!lock_residues) {
    CalcDPScores(obj0,obj1);
  }

  if(obj0->num_SSEs_aligned!=0) {
    found_first_aligned_SSE = 0;
    for(i=0;i<obj0->ssArrayLength;i++) {
      for(j=0;j<obj1->ssArrayLength;j++) {
	if(obj0->ss_alignment[i][j]!=-1) {
	  query_start = obj0->ssArray[i*2];
	  query_end = obj0->ssArray[i*2+1]-1;
	  target_start = obj1->ssArray[obj0->ss_alignment[i][j]*2];
	  target_end = obj1->ssArray[obj0->ss_alignment[i][j]*2+1]-1;
	  maxscore = 40;
	  if(found_first_aligned_SSE == 0) {
	    maxscore = 20;
	    found_first_aligned_SSE = 1;
	  }
	  fprintf(outfile,"%c,%c\t%i%c%c,%i%c%c\t%i%c%c,%i%c%c\t%8.3f\t%8.3f\n",obj0->sstype[query_start],obj1->sstype[target_start],obj0->aaNums[query_start],obj0->insertionCodes[query_start],obj0->chains[query_start],obj0->aaNums[query_end],obj0->insertionCodes[query_end],obj0->chains[query_end],obj1->aaNums[target_start],obj1->insertionCodes[target_start],obj1->chains[target_start],obj1->aaNums[target_end],obj1->insertionCodes[target_end],obj1->chains[target_end],obj0->ss_dpscores[i],maxscore);
	}
      }
    }
  }


#if PRINT_DP_SCORES
  
  if(obj1->num_SSEs_aligned !=0) {
    /*  printf("final score: %f\n",align_ss(1,obj0,obj1));*/
    printf("final score: %f\n",finalscore);
    fprintf(outfile,"\nfinal score: %f\n",finalscore);
    
  }else {
    printf("final score: 0\n");
  }

#endif

  fprintf(outfile,"\nSSE aa numbers\n");
  fprintf(outfile,"query:\n");
  for(i=0;i<obj0->ssArrayLength;i++) {    
    fprintf(outfile,"%c\t%i%c%c\t%i%c%c\n",obj0->sstype[obj0->ssArray[2*i]],obj0->aaNums[obj0->ssArray[2*i]],obj0->insertionCodes[obj0->ssArray[2*i]],obj0->chains[obj0->ssArray[2*i]],obj0->aaNums[obj0->ssArray[2*i+1]-1],obj0->insertionCodes[obj0->ssArray[2*i+1]-1],obj0->chains[obj0->ssArray[2*i+1]-1]);
  }
  fprintf(outfile,"target:\n");
  for(i=0;i<obj1->ssArrayLength;i++) {
    fprintf(outfile,"%c\t%i%c%c\t%i%c%c\n",obj1->sstype[obj1->ssArray[2*i]],obj1->aaNums[obj1->ssArray[2*i]],obj1->insertionCodes[obj1->ssArray[2*i]],obj1->chains[obj1->ssArray[2*i]],obj1->aaNums[obj1->ssArray[2*i+1]-1],obj1->insertionCodes[obj1->ssArray[2*i+1]-1],obj1->chains[obj1->ssArray[2*i+1]-1]);
  }

  fprintf(outfile, "CPU time: %g\n",runtime);

  fclose(outfile);
  outfile = NULL;	    
  
  /*num residues, num SSEs*/
  printf("Query num residues: %i\n",obj0->numPts);
  printf("Target num residues: %i\n",obj1->numPts);

  num_aligned = 0;
  for(i=0;i<obj0->numPts;i++) {
    if(obj0->pairNums[i]!=-1 && obj0->sstype[i]!='U') {
      num_aligned++;
    }
  }
  printf("Query num SSE residues aligned: %i\n",num_aligned);
    num_aligned = 0;
  for(i=0;i<obj1->numPts;i++) {
    if(obj1->pairNums[i]!=-1 && obj1->sstype[i]!='U') {
      num_aligned++;
    }
  }
  printf("Target num SSE residues aligned: %i\n",num_aligned);
  if(obj0->num_helices <=obj1->num_helices) {
    num_SSEs = obj0->num_helices;
  }else {
    num_SSEs = obj1->num_helices;
  }
  if(obj0->num_strands <=obj1->num_strands) {
    num_SSEs+=obj0->num_strands;
  }else {
    num_SSEs+=obj1->num_strands;
  }
  printf("Common SSEs: %i\n",num_SSEs);
  printf("Query num SSEs: %i\n",obj0->ssArrayLength);
  printf("Target num SSEs: %i\n",obj1->ssArrayLength);
  
  /*  gzipOutputFile();*/
}
  

void printPairNums(OBJ* t0, OBJ* t1) {
  /*OBJ *t0, *t1;*/
  int i,j;
  /*  t0 = ObjList[0];
  t1 = ObjList[1];*/
  for(i=0;i<t0->numPts;i++) {
    if(t0->pairNums[i]!=-1) {
      printf("%i,%c,%f  %i,%c,%f\n",t0->aaNums[i],t0->objSeq[i],t0->pairDists[i],t1->aaNums[t0->pairNums[i]],t1->objSeq[t0->pairNums[i]],t1->pairDists[t0->pairNums[i]]);
    }
  }

}

void printStructAlignment(OBJ* query,OBJ* target,int query_alignment_start, int query_alignment_end) {

  int SSEnum,i,j, tpoint, mycount=0, outcount,qpoint;
 
  /*  OBJ *query, *target;*/

  
  OBJ *t0, *t1;
  char qlist[MAXLENGTH];
  char alist[MAXLENGTH];
  char tlist[MAXLENGTH];
  int ilen, tlen, previ, prevt;

  if (outfile != NULL) {

    /*    t0 = ObjList[0];
    t1 = ObjList[1];
    query = ObjList[0];
    target = ObjList[1];*/
    t1 = target;
    t0 = query;

    for (i = 0; i < t0->numPts; i++) {
      if (t0->pairNums[i] != -1) {
	mycount++;
      }
    }

    if (newfile[0] == 1) {
      fprintf(outfile, "Query: %s  ", username[0]);
      if (chain[0] != ' ') {
	fprintf(outfile, "Chain: %c\n", chain[0]);
      }
      else {
	fprintf(outfile, "\n");
      }
    }
    else {
      fprintf(outfile, "Query: %s\n", pdbname[0]);
    }
    
    if (newfile[1] == 1) {
      fprintf(outfile, "Target: %s  ", username[1]);
      if (chain[1] != ' ') {
	fprintf(outfile, "Chain: %c\n", chain[1]);
      }
      else {
	fprintf(outfile, "\n");
      }
    }
    else {
      fprintf(outfile, "Target: %s\n", pdbname[1]);
    }
    
	
    fprintf(outfile, "RMSD = %f  Num Pre-Core Aligned = %d \n", globalRMSD, globalNumAligned);
    if ((globalComputeCore) || (globalLastAlignment == 1))
      fprintf(outfile, "RMSD core = %f  Num Core Aligned = %d\n", globalCoreRMSD, globalCoreNumAligned);
    if ((t0->ssArrayLength == 0) || (t1->ssArrayLength == 0)) {
      fprintf(outfile, "NOSS\n");
    }
    else {
      fprintf(outfile, "SSOK\n");
    }      
	

    if (globalLastAlignment == 1) {

      outcount = 0;
      previ = -1;
      prevt = -1;

      fprintf(outfile, "\n**********************\n\n");

      for (i = 0; i < t0->numPts; i++) { 
	if (t0->pairNums[i] != -1) {
	  ilen = i - previ - 1;
	  tlen = t0->pairNums[i] - prevt - 1;
	  if (ilen > tlen) {	    
	    for (j = 0; j < ilen; j++) {
	      qlist[outcount] = t0->objSeq[previ + j + 1];
	      alist[outcount] = ' ';
	      
	      if (j < tlen/2) {
		tlist[outcount] = t1->objSeq[prevt + j + 1];
	      }
	      else if ((j >= tlen/2) && (j < tlen/2 + (ilen - tlen))) {
		tlist[outcount] = '-';
	      }
	      else {
		tlist[outcount] = t1->objSeq[prevt + j - (ilen - tlen) + 1];
	      }
	      outcount++;
	    }
	  }
	  else if (ilen < tlen) {
	    for (j = 0; j < tlen; j++) {
	      tlist[outcount] = t1->objSeq[prevt + j + 1];
	      alist[outcount] = ' ';
	      
	      if (j < ilen/2) {
		qlist[outcount] = t0->objSeq[previ + j + 1];
	      }
	      else if ((j >= ilen/2) && (j < ilen/2 + (tlen - ilen))) {
		qlist[outcount] = '-';
	      }
	      else {
		qlist[outcount] = t0->objSeq[previ + j - (tlen - ilen) + 1];
	      }
	      outcount++;
	    }
	  }
	  else {
	    for (j = 0; j < ilen; j++) {
	      qlist[outcount] = t0->objSeq[previ + j + 1];
	      alist[outcount] = ' ';
	      tlist[outcount] = t1->objSeq[prevt + j + 1];
	      outcount++;
	    }
	  }

	  qlist[outcount] = t0->objSeq[i];
	  alist[outcount] = '|';
	  tlist[outcount] = t1->objSeq[t0->pairNums[i]];
	  outcount++;

	  previ = i;
	  prevt = t0->pairNums[i];

	}
      }

      ilen = t0->numPts - previ - 1;
      tlen = t1->numPts - prevt - 1;

      j = (ilen > tlen)?ilen:tlen;

      for (i = 1; i <= j; i++) {
	alist[outcount] = ' ';
	if (previ + i < t0->numPts) {
	  qlist[outcount] = t0->objSeq[previ+i];
	}
	else {
	  qlist[outcount] = '-';
	}
	if (prevt + i < t1->numPts) {
	  tlist[outcount] = t1->objSeq[prevt+i];
	}
	else {
	  tlist[outcount] = '-';
	}
	outcount++;
      }

      j = 0;

      while (j < outcount) {
	fprintf(outfile, "Query \t");
	for (i = j; ((i < j + 50) && (i < outcount)); i++) {
	  if (i%10 == 0) 
	    fprintf(outfile, " ");
	  fprintf(outfile, "%c", qlist[i]);
	}
	fprintf(outfile, "\n");
	
	fprintf(outfile, "      \t");
	for (i = j; ((i < j + 50) && (i < outcount)); i++) {
	  if (i%10 == 0) 
	    fprintf(outfile, " ");
	  fprintf(outfile, "%c", alist[i]);
	}
	fprintf(outfile, "\n");
	
	fprintf(outfile, "Target\t");
	for (i = j; ((i < j + 50) && (i < outcount)); i++) {
	  if (i%10 == 0) 
	    fprintf(outfile, " ");
	  fprintf(outfile, "%c", tlist[i]);
	}
	fprintf(outfile, "\n");
	j = j + 50;
	fprintf(outfile, "\n");
      }

      fprintf(outfile, "\nNote that only residues matched with '|' correspond to structurally\naligned residue pairs (see following list).\n");

    }
	

    fprintf(outfile, "\n**********************\n");
    if (globalLastAlignment == 0) {
      fprintf(outfile, "\nNum Pre-Core Aligned = %d\n", mycount);
    }
    else if (globalLastAlignment == 1) {
      fprintf(outfile, "\nNum Core Aligned = %d\n", mycount);
    }

 

    for(i=query_alignment_start;i<=query_alignment_end;i++) {
      if(t0->pairNums[i]!=-1) {
	tpoint = t0->pairNums[i];
	qpoint = i;
	fprintf(outfile,"%4d%c%c %c\t", t0->aaNums[qpoint], t0->insertionCodes[qpoint],t0->chains[qpoint],t0->objSeq[qpoint]);
	fprintf(outfile,"%4d%c%c %c\t%f\t%f\t%f\n", t1->aaNums[tpoint], t1->insertionCodes[tpoint], t1->chains[tpoint],t1->objSeq[tpoint], t0->pairDists[qpoint],t0->alignmentScore[qpoint],t0->gapScore[qpoint]);
	/*	fprintf(outfile,"%4d%c %c\t", qpoint, t0->insertionCodes[qpoint],t0->objSeq[qpoint]);
	fprintf(outfile,"%4d%c %c\t%f\n", tpoint, t1->insertionCodes[tpoint], t1->objSeq[tpoint], t0->pairDists[qpoint]);
	*/
      }
    }

    
    fprintf(outfile, "\n\n**********************\n\n");
    
    printf("Alignment saved\n");    		    
  }

}


void printAlignmentMatrices() {
  
  char *str0, *str1;

  if (outfile != NULL) {

    if (newfile[0] == 1) {
      str0 = username[0];
    }
    else {
      str0 = pdbname[0];
    }
    
    if (newfile[1] == 1) {
      str1 = username[1];
    }
    else {
      str1 = pdbname[1];
    }
    

    fprintf(outfile, "\nTransformation Matrix for query %s:", str0);
    printMat(ObjList[0]->objMat);

    fprintf(outfile, "\nCenter of mass for %s: \n", str0);
    fprintf(outfile, "%f  %f  %f\n", ObjList[0]->center[0], ObjList[0]->center[1], ObjList[0]->center[2]);

    fprintf(outfile, "\nTransformation Matrix for target %s:", str1);
    printMat(ObjList[1]->objMat);

    fprintf(outfile, "\nCenter of mass for %s: \n", str1);
    fprintf(outfile, "%f  %f  %f\n", ObjList[1]->center[0], ObjList[1]->center[1], ObjList[1]->center[2]);

    printf("Matrices saved\n\n");
    
  }


}


void loadAlignmentMatrices() {
  
  FILE *matrixFile;
  char matrixFileName[50];
  GLfloat temp;
  char tempstr[10];
  int i;
  int tempint;

  printf("Enter Matrix File Name > ");

  scanf("%s", matrixFileName);

  if ((matrixFile = fopen(matrixFileName, "r")) == NULL) {
    printf("Cannot open matrix file %s\n", matrixFileName);
  }

  else {  

    fscanf(matrixFile, "%s", tempstr);
    fscanf(matrixFile, "%s", tempstr);
    fscanf(matrixFile, "%f %d", &temp, &tempint);

    fscanf(matrixFile, "%s", tempstr);
    fscanf(matrixFile, "%s", tempstr);
    fscanf(matrixFile, "%s", tempstr);
    fscanf(matrixFile, "%f %d", &temp, &tempint);
    fscanf(matrixFile, "%s", tempstr);
    
    for (i = 0; i < 16; i++) {
      fscanf(matrixFile, "%f", &(ObjList[0]->objMat[i]));
    }
    
    fscanf(matrixFile, "%f %f %f", &(ObjList[0]->center[0]), &(ObjList[0]->center[1]), &(ObjList[0]->center[2]));
    
    for (i = 0; i < 16; i++) {
      fscanf(matrixFile, "%f", &(ObjList[1]->objMat[i]));
    }

    fscanf(matrixFile, "%f %f %f", &(ObjList[1]->center[0]), &(ObjList[1]->center[1]), &(ObjList[1]->center[2]));
        
    displayPoints = 0;
#ifdef DISPLAYON
    if (globalStartDisplay) {
      display();
    }
#endif
    fclose(matrixFile);
    
  }

}

/*breakSSELengthTie: Utility function used to break ties when SSE lengths are the same.
 * Ties are broken arbitrarily but consistently by comparing amino acid sequences*/
int breakSSELengthTie(OBJ* query, OBJ* target) {
  return(strcmp(query->objSeq,target->objSeq));
}
/****************************************************************************************
                                    Alignment Functions
				    Align using vectors (step 1)
****************************************************************************************/

float alignUsingVectors(OBJ* query, OBJ* target, int iterations, float epsilon, int full, int SSE) {

  float best_transform_score;
  int best_transform_index;
  int i,j, converged, iter, tnum, count = 0,order;
  int SSEnum;
  int done=0,result;
  int query_start,query_end,target_start,target_end,target_length,query_length;
  Vector translation;
  Matrix mat;
  GLfloat m[16];
  float oldm[16];
  OBJ tempObj1, tempObj0;
  GLfloat tdist, rmsd, diffrmsd, oldrmsd = 10000.0;
  GLfloat tempVec[3];
  float old_transform_score = -10000;
  float transform_score_epsilon = 2;
  float difftransformscore;
  int change_transform = 1;
  int meetconvergecriteria;
  int have_set_best_transform = 0;

#if !CHANGE_TRANSFORM
  change_transform = 0;
#endif

  


  memset(query->pairNums,-1,query->numPts*sizeof(int));
  memset(target->pairNums,-1,target->numPts*sizeof(int));

  tempObj0.objSeq = query->objSeq; /*Array*/
  tempObj0.sstype = query->sstype; /*Array of SS types*/
  tempObj0.numPts = query->numPts; /*Array of points*/
  tempObj0.ssArrayLength = query->ssArrayLength;
  tempObj0.ssArray = query->ssArray; /*Array of #'s giving beginning and
				     *ends of ss elements*/
  tempObj0.num_SSEs_aligned = query->num_SSEs_aligned;
  tempObj0.ss_alignment = query->ss_alignment;
  applyMat(&tempObj0, query, query->objMat);
  
  tempObj1.objSeq = target->objSeq;
  tempObj1.sstype = target->sstype;
  tempObj1.numPts = target->numPts;
  tempObj1.ssArrayLength = target->ssArrayLength;
  tempObj1.ssArray = target->ssArray;
  tempObj1.num_SSEs_aligned = target->num_SSEs_aligned;
  tempObj1.ss_alignment = target->ss_alignment;
  applyMat(&tempObj1, target, target->objMat);
    

  converged = 0;
  done = 0;
  meetconvergecriteria = 0;
  best_transform_index = 0;
  best_transform_score = -1;
  
  while(!done) {
    meetconvergecriteria = 0;
    for (iter = 0; (iter < iterations); iter++) {

      printf("alignusingvectors: transformation %i,done = %i\n",iter,done);
      init_pairs(); /*initialize transformation*/
      initsequences_ss(&tempObj0, &tempObj1);    /*initialize sequences*/
      align_ss(full,query,target);
      report_ss(full,query,target);
      if (globalCalledAddPair >= 4 && okayToMatchPoints()) {
	match_pairs (mat, translation);
	for (i = 0; i < 4; i++) {
	  for (j = 0; j < 4; j++) {
	    m[i*4 + j] = mat[i][j];
	  }

	}
	
	myMatMult(target->objMat, m);
	applyMtoV(target->center, target->center, m);
	applyMat(&tempObj1, target, target->objMat);
	count = 0;
	rmsd = 0.0;

	for(SSEnum = 0;SSEnum <target->ssArrayLength;SSEnum++) {
	  order = 0;
	  if(target->ss_alignment[SSEnum][0] != -1) {
	    target_start = target->ssArray[SSEnum*2];
	    target_end = target->ssArray[SSEnum*2+1]-1;
	    query_start = query->ssArray[target->ss_alignment[SSEnum][0]*2];
	    query_end = query->ssArray[target->ss_alignment[SSEnum][0]*2+1]-1;
	        if((target_end-target_start)==(query_end-query_start)) {
		  order = breakSSELengthTie(query,target);
	      }
	    if(order < 0 || ((target_end - target_start) < (query_end-query_start))) {
	      for (i = target_start; i <= target_end; i++) {
		if(useBetaCarbons) {
		  tdist = findNearestSSEResidue(&tempObj0,tempObj1.cbpoints[i],tempObj1.points[i],0,tempObj0.numPts-1,tempObj1.sstype[i],tempVec,&tnum,1,useBetaCarbons);
		}else {
		  tdist = findNearestSSEResidue(&tempObj0,tempObj1.points[i],tempObj0.points[i],0,tempObj0.numPts-1,tempObj1.sstype[i],tempVec,&tnum,1,useBetaCarbons);
		}
		if ((tdist <= distThreshold && tdist >= 0) || (thresholdOFF)) {
		  rmsd += tdist*tdist;
		  count++;
		}
		
	      }
	    }else {
	      for (i = query_start; i <= query_end; i++) {
		if(useBetaCarbons) {
		  tdist = findNearestSSEResidue(&tempObj1,tempObj0.cbpoints[i],tempObj0.points[i], 0,tempObj1.numPts-1,tempObj0.sstype[i],tempVec,&tnum,1,useBetaCarbons);
		}else{
		  tdist=findNearestSSEResidue(&tempObj1,tempObj0.points[i],tempObj1.points[i],0,tempObj1.numPts-1,tempObj0.sstype[i],tempVec,&tnum,1,useBetaCarbons);
		}
		if ((tdist <= distThreshold && tdist >=0) || (thresholdOFF)) {
		  rmsd += tdist*tdist;
		  count++;
		}
		
	      }

	    }
	  }
	}


	rmsd = sqrt((rmsd)/((GLfloat)(count)));
	globalNumAligned = count;
	printf("rmsd = %f, num aligned = %i\n",rmsd,globalNumAligned);
	diffrmsd = oldrmsd - rmsd;
	difftransformscore = old_transform_score - getCurrentAlignmentScore();

	if(difftransformscore < 0)
	  difftransformscore *= -1.0;
	
	if (diffrmsd < 0) 
	  diffrmsd *= -1.0;

	if((difftransformscore < transform_score_epsilon) &&(diffrmsd< epsilon)&& (iter < iterations-1)) {

	  done = 1;
	  converged = 1;
	  meetconvergecriteria = 1;
	  printf("align using vectors converged: iter = %i!\n",iter);
	  if(full==1) {
	    if(getCurrentAlignmentScore()<0.9*getCurrentTransformationScore() && change_transform) {
	      if(getCurrentAlignmentScore()> best_transform_score) {
		best_transform_score = getCurrentAlignmentScore();
		best_transform_index = getTransformationIndex();
	      }
	      printf("changing transformations! current score = %f\n",getCurrentAlignmentScore());
	      done = 0;
	      converged = 0;
	      result = increaseTransformationIndex();
	      if(result == -1) {
		if(have_set_best_transform) {
		  done = 1;
		}else if(best_transform_score > getCurrentAlignmentScore()) {
		  setAlignment(best_transform_index);
		  
		  have_set_best_transform = 1;
		  initTwoPairTransform(query,target,&tempObj1);
		  oldrmsd = 10000.0;
		  old_transform_score = -10000;
		  change_transform = 0;
		  done = 0;
		  converged = 0;
		  break;
		  /*   break;*/
		}	
		done = 1;
		converged = 1;
	      }/*ends loop if result==-1*/
	      if(!converged) {
		initTwoPairTransform(query,target,&tempObj1);
		oldrmsd = 10000.0;
		old_transform_score = -10000;
		break; /*need to break out of loop!*/
	      }
	    }else if(query->num_SSEs_aligned < 3 && (query->ssArrayLength > 2 && target->ssArrayLength > 2) && change_transform) { /*corresponding if statement: if(getCurrentAlignmentScore()...*/
	      printf("changing transformations! current score = %f\n",getCurrentAlignmentScore());
	      if(getCurrentAlignmentScore()> best_transform_score) {
		best_transform_score = getCurrentAlignmentScore();
		best_transform_index = getTransformationIndex();
	      }
	      done = 0;
	      converged = 0;
	      result = increaseTransformationIndex();
	      if(result == -1) {
		if(have_set_best_transform) {
		  done = 1;
		}else if(best_transform_score > getCurrentAlignmentScore()) {
		  setAlignment(best_transform_index);
		  have_set_best_transform = 1;
		  initTwoPairTransform(query,target,&tempObj1);
		  oldrmsd = 10000.0;
		  old_transform_score = -10000;
		  change_transform = 0;
		  done = 0;
		  converged = 0;
		  /*test: putting break statement here; used to be commented out*/
		  break;
		  /*break;*/
		}	
		done = 1;
		converged = 1;
	      }
	      if(!converged) {
		initTwoPairTransform(query,target,&tempObj1);
		oldrmsd = 10000.0;
		old_transform_score = -10000;
	      }
	    }
	  }/*matches if(full==1)*/
	  /*  break;*/
	} /*matches if(diffrmsd < epsilon && iter<iterations-1*/
	if(!meetconvergecriteria) {
	  oldrmsd = rmsd;
	  old_transform_score = getCurrentAlignmentScore();
	  if(iter == iterations-1) {
	    if(full == 1) {
	      if((getCurrentAlignmentScore()<0.9*getCurrentTransformationScore() && change_transform)) {
		printf("changing transformations! current score = %f\n",getCurrentAlignmentScore());
		if(getCurrentAlignmentScore()> best_transform_score) {
		  best_transform_score = getCurrentAlignmentScore();
		  best_transform_index = getTransformationIndex();
		}
		done = 0;
		converged = 0;
		result = increaseTransformationIndex();
		if(result == -1) {
		  if(have_set_best_transform) {
		    done = 1;
		  }else if(best_transform_score > getCurrentAlignmentScore()) {
		    setAlignment(best_transform_index);
		    have_set_best_transform = 1;
		    initTwoPairTransform(query,target,&tempObj1);
		    oldrmsd = 10000.0;
		    old_transform_score = -10000;
		    change_transform = 0;
		    done = 0;
		    converged = 0;
		    break;
		    /*break;*/
		  }	
		  done = 1; /*no more transformations to try*/
		  /* break;*/
		} /*matches if(result== -1) -- unable to change transformations*/
		if(!done) {
		  initTwoPairTransform(query,target,&tempObj1);
		  oldrmsd = 10000.0;
		  old_transform_score = -10000;
		  done = 0;
		  /* break;*/
		}
	       }else  { /*matches if(getCurrentAlignmentScore() ...*/
		 done = 1;
		 converged = 0;
		 /*break;*/
	       }
	     }else { /*matchines if(full==1)*/
	       done = 1;
	       converged = 1; /*if full==0, then iterations = 1, and need to surpress
			       *"Did not converge" message*/
	       /*break;*/
	     }
	  } /*matches if(iter==iterations-1*/
      
	}else {/*matches if(!meetconvergecriteria*/
	  break; /*have converged; need to break out of for loop*/
	}
      }else { /*matches if(globalCalledPair > 4)*/
	done = 0;
	converged = 0;
	result = increaseTransformationIndex();
	if(result==-1) {
	 if(have_set_best_transform) {
	    done = 1;
	  }else if(best_transform_score > 0) { /*go to best transformation*/

	    printf("less than four aligned; switching to best transform\n");
	    setAlignment(best_transform_index);
	    have_set_best_transform = 1;
	    initTwoPairTransform(query,target,&tempObj1);
	    oldrmsd = 10000.0;
	    old_transform_score = -10000;
	    change_transform = 0;
	    done = 0;
	    converged = 0;
	    break;
	  }
	  done = 1; /*no more transformations to try*/
	  break; /*break here because iter may not equal iterations-1, and need to break out of for(iter...)
		   loop*/
	} /*unable to change transformations*/
	if(!done) {
	  initTwoPairTransform(query,target,&tempObj1);
	  oldrmsd = 10000.0;
	  old_transform_score = -10000;
	  done = 0;
	}
      }
    }
    
  }
    
    

  
  if(globalNumAligned != 0) {
    
    count = 0;
    rmsd = 0.0;

    for(SSEnum = 0;SSEnum < target->ssArrayLength;SSEnum++) {
      /*    for (i = 0; i < tempObj1.numPts; i++) {*/
      if(target->ss_alignment[SSEnum][0] != -1) {
	target_start = target->ssArray[SSEnum*2];
	target_end = target->ssArray[SSEnum*2+1]-1;
	for(i=target_start;i<=target_end;i++) {
	  if(useBetaCarbons) {
	    tdist = findNearestSSEResidue(&tempObj0, tempObj1.cbpoints[i], tempObj1.points[i],0,tempObj0.numPts-1,tempObj1.sstype[i], tempVec, &tnum,1,useBetaCarbons);
	  }else {
	    tdist = findNearestSSEResidue(&tempObj0, tempObj1.points[i],tempObj1.points[i], 0,tempObj0.numPts-1,tempObj1.sstype[i], tempVec, &tnum,1,useBetaCarbons);
	  }
      
	  if ((tdist <= distThreshold && tdist >=0 ) || (thresholdOFF)) {
	    target->pairNums[i] = tnum;
	    target->pairDists[i] = tdist;


	    if (query->pairNums[tnum] != -1) {
	      if (tdist < query->pairDists[tnum]) {	    
		rmsd -= (query->pairDists[tnum])*(query->pairDists[tnum]);
		query->pairNums[tnum] = i;	    
		query->pairDists[tnum] = tdist;
		rmsd += tdist*tdist;
	      }
	    }
	    else {
	      query->pairNums[tnum] = i;	    
	      query->pairDists[tnum] = tdist;
	      rmsd += tdist*tdist;
	      count++;
	    }	  
	  }
	}
      }else {
	target_start = target->ssArray[2*SSEnum];
	target_end = target->ssArray[2*SSEnum+1]-1;
	for(i=target_start;i<=target_end;i++) {
	  if(useBetaCarbons) {
	    tdist = findNearestSSEResidue(&tempObj0,tempObj1.cbpoints[i],tempObj1.points[i],0,tempObj0.numPts-1,tempObj1.sstype[i],tempVec,&tnum,0,useBetaCarbons);
	  }else {
	    tdist = findNearestSSEResidue(&tempObj0,tempObj1.points[i],tempObj1.points[i],0,tempObj0.numPts-1,tempObj1.sstype[i],tempVec,&tnum,0,useBetaCarbons);
	  }
	  if ((tdist <= distThreshold && tdist >= 0) || (thresholdOFF)) {
	    target->pairNums[i]=tnum;
	    target->pairDists[i] = tdist;
	    if (query->pairNums[tnum] != -1) {
	      if (tdist < query->pairDists[tnum]) {	    
		rmsd -= (query->pairDists[tnum])*(query->pairDists[tnum]);
		query->pairNums[tnum] = i;	    
		query->pairDists[tnum] = tdist;
		rmsd += tdist*tdist;
	      }
	    }else {
	      query->pairNums[tnum] = i;
	      query->pairDists[tnum] = tdist;
	      rmsd += tdist*tdist;
	      count++;
	    }	
	  }
	}
      }
    }
    rmsd = sqrt((rmsd)/((GLfloat)(count)));
    globalNumAligned = count;

  }

  
  /*  if ((converged == 0) && (iterations == MAXITERATIONS)) {*/
  if(!converged) {
    printf("Did not converge\n");
  }

  query->num_helices = tempObj0.num_helices;
  query->num_strands = tempObj0.num_strands;
  target->num_helices = tempObj1.num_helices;
  target->num_strands = tempObj1.num_strands;
  return rmsd;

}




void alignUsingVectorsSingle(OBJ* t0, OBJ* t1, int SSE) {

  /*  OBJ *t0, *t1;*/
  float rmsd;
  int i;

  if(t0 == NULL) {
    t0 = ObjList[0];
  }
  if(t1 == NULL) {
    t1 = ObjList[1];
  }

  if ((t0->ssArrayLength > 0) && (t1->ssArrayLength > 0)) {

    /*    t0 = ObjList[0];
    t1 = ObjList[1];*/
    
    for (i = 0; i < 5; i++) {
      rmsd = alignUsingVectors(t0, t1, 1, 0.0, 1, SSE);
      
      globalRMSD = rmsd;
      
      printf("alignusingvectorssingle: RMSD = %f %d\n", rmsd, globalNumAligned);

    }
#ifdef DISPLAYON
    if (globalStartDisplay) {
#if DRAWJOINTS
      createJointList();
#endif
      display();
    }
#endif
  }

}


void alignUsingVectorsFull(OBJ* t0, OBJ* t1, int SSE) {

  /*  OBJ *t0, *t1;*/
  float rmsd;

 if(t0 == NULL) {
    t0 = ObjList[0];
  }
  if(t1 == NULL) {
    t1 = ObjList[1];
  }

  if ((t0->ssArrayLength > 0) && (t1->ssArrayLength > 0)) {

    /*    t0 = ObjList[0];
    t1 = ObjList[1];*/

    rmsd = alignUsingVectors(t0, t1, MAXITERATIONS, MAXDIFFERENCE, 1,SSE);

    globalRMSD = rmsd;
    
    printf("alignusingvectorsfull: RMSD = %f %d\n", rmsd, globalNumAligned);

#ifdef DISPLAYON
    if (globalStartDisplay) {

#if DRAWJOINTS
      createJointList();
#endif
      display();
    }
#endif
  }


}


void initAlignUsingVectorsSingle(OBJ* t0, OBJ* t1,int SSE) {

  float rmsd;

  if(t0 == NULL) {
    t0 = ObjList[0];
  }
  if(t1 == NULL) {
    t1 = ObjList[1];
  }

  if ((t0->ssArrayLength > 0) && (t1->ssArrayLength > 0)) {

    /*    t0 = ObjList[0];
    t1 = ObjList[1];*/


    rmsd = alignUsingVectors(t0, t1, 1, 0.0, 0,SSE);
    
    globalRMSD = rmsd;
    
    printf("alignusingvectorssingle: RMSD = %f %d\n", rmsd, globalNumAligned);
   
#ifdef DISPLAYON
    if (globalStartDisplay) {   
#if DRAWJOINTS
      createJointList();
#endif
      display();
    }
#endif
  }
}


void initAlignUsingVectorsFull(OBJ* t0, OBJ* t1,int SSE) {

  /*  OBJ *t0, *t1;*/
  float rmsd;

  if(t0 == NULL) {
    t0 = ObjList[0];
  }
  if(t1 == NULL) {
    t1 = ObjList[1];
  }

  if ((t0->ssArrayLength > 0) && (t1->ssArrayLength > 0)) {

    /*    t0 = ObjList[0];
    t1 = ObjList[1];*/
    
    rmsd = alignUsingVectors(t0, t1, MAXITERATIONS, MAXDIFFERENCE, 0,SSE);

    globalRMSD = rmsd;
    
    printf("initalignusingvectorsfull: RMSD = %f %d\n", rmsd, globalNumAligned);
    
#ifdef DISPLAYON
    if (globalStartDisplay) {
#if DRAWJOINTS
      createJointList();
#endif
      display();
    }
#endif
  }

}


/****************************************************************************************
                                    Alignment Functions
				    Minimize RMSD (step 2)
****************************************************************************************/


float structAlign(OBJ* query, OBJ* target, int iterations, float epsilon) {

  OBJ* temp;
  int i, j, iter, converged, count, tnum,order;
  int SSEnum,query_start,query_end,target_start,target_end;
  float target_length,query_length,weight;
  Vector translation;
  Matrix mat;
  GLfloat m[16];
  GLfloat tempVec[3];
  GLfloat tdist;
  GLfloat rmsd = 0.0, oldrmsd = 1000.0, diffrmsd;
  OBJ tempObj1, tempObj0;
  int query_alignment_start,query_alignment_end,target_alignment_start,target_alignment_end;
  int SSE_start,SSE_end;
  int switched = 0;



  /*set protein with more SSEs as target for purposes of symmetry*/
  if(target->ssArrayLength < query->ssArrayLength) {
    temp = query;
    query = target;
    target = temp;
    switched = 1;
    /*    printf("switched: query longer\n");*/
  }else if((target->ssArrayLength==query->ssArrayLength) && strcmp(query->objSeq,target->objSeq) > 0) {
    temp = query;
    query = target;
    target = temp;
    switched = 1;
    /*    printf("switched: same length\n");*/
  }

  tempObj0.numPts = query->numPts;
  tempObj0.sstype = query->sstype;
  tempObj0.objSeq = query->objSeq;
  applyMat(&tempObj0, query, query->objMat);

  tempObj1.numPts = target->numPts;
  tempObj1.sstype = target->sstype;
  tempObj1.objSeq = target->objSeq;
  applyMat(&tempObj1, target, target->objMat);
  converged = 0;


  memset(query->pairNums,-1,query->numPts*sizeof(int));
  memset(target->pairNums,-1,target->numPts*sizeof(int));

  rmsd = 0.0;
  

  init_pairs();
  
  count = 0; /*number of aligned atoms (points)*/
  
  /*if target is longer than query, will match more SSEs in one direction than other
   *--check for "extra" SSEs and attempt to match?, or set protein with more
   *sses as target?*/
  for(SSEnum = 0; SSEnum < target->ssArrayLength;SSEnum++) {
    if(target->ss_alignment[SSEnum][0] != -1) {
      target_start = target->ssArray[2*SSEnum];
      target_end = target->ssArray[2*SSEnum+1]-1;
      target_length = (float)(target_end-target_start+1);
      query_start = query->ssArray[target->ss_alignment[SSEnum][0]*2];
      query_end = query->ssArray[target->ss_alignment[SSEnum][0]*2+1]-1;
      order = 0;

      weight = 1;
      if(target_end - target_start == query_end - query_start) {
	/*	order = strcmp(getFileName(0),getFileName(1));*/
	if(!switched) {
	  order = breakSSELengthTie(query,target);
	}else {
	  order =breakSSELengthTie(target,query);
	}

	if(switched)
	  order*=-1;
	/*	printf("order: %i\n",order);*/
      }
      if((order < 0) || (target_end - target_start > query_end - query_start)) {
	/*	  printf("matching target\n");*/
	for(i=target_start;i<=target_end;i++) {
	  if(useBetaCarbons) {
	    tdist = findNearestSSEResidue(&tempObj0, tempObj1.cbpoints[i], tempObj1.points[i],0,tempObj0.numPts-1,tempObj1.sstype[i], tempVec, &tnum,1,useBetaCarbons);
	  }else {
	    tdist = findNearestSSEResidue(&tempObj0, tempObj1.points[i],tempObj1.points[i], 0,tempObj0.numPts-1,tempObj1.sstype[i], tempVec, &tnum,1,useBetaCarbons);
	  }
	  if ((tdist <= distThreshold && tdist>=0) || (thresholdOFF)) {
	    
	    add_pair(tempVec,tempObj1.cbpoints[i],1);
	    add_pair(tempObj0.points[tnum],tempObj1.points[i],1);
	    rmsd += tdist*tdist;
	    count++;
	  }
	}
      }else {
	/*same as above, but match query to target*/
	query_length = (float)(query_end - query_start +1);
	/*	  printf("matching query\n");*/
	for(i=query_start;i<=query_end;i++) {
	  if(useBetaCarbons) {
	    tdist = findNearestSSEResidue(&tempObj1, tempObj0.cbpoints[i],tempObj0.points[i], 0,tempObj1.numPts-1,tempObj0.sstype[i], tempVec, &tnum,1,useBetaCarbons);

	  }else {
	    tdist = findNearestSSEResidue(&tempObj1, tempObj0.points[i],tempObj1.points[i], 0,tempObj1.numPts-1,tempObj0.sstype[i], tempVec, &tnum,1,useBetaCarbons);
	  }
	  if ((tdist <= distThreshold && tdist>=0) || (thresholdOFF)) {
	    add_pair(tempObj0.cbpoints[i],tempVec,1);
	    add_pair(tempObj0.points[i],tempObj1.points[tnum],1);
	    rmsd += tdist*tdist;
	    count++;
	  }
	}
      }
    }else {
      target_start = target->ssArray[2*SSEnum];
      target_end = target->ssArray[2*SSEnum+1]-1;
      target_length = (float)(target_end-target_start+1);
      /* query_start = query->ssArray[target->ss_alignment[SSEnum][0]*2];
	 query_end = query->ssArray[target->ss_alignment[SSEnum][0]*2+1]-1;*/
      order = 0;
      /*      if(target_end - target_start == query_end - query_start) {
	order = strcmp(getFileName(0),getFileName(1));
	}*/
      /* if(order < 0 || (target_end - target_start > query_end - query_start)) {*/
      if(1) {
	/*	printf("not aligned: matching target\n");*/
	for(i=target_start;i<=target_end;i++) {

	  /*	  if(target->objSeq[i] == 'G') {
	    tdist = findNearestSSEResidue(&tempObj0, tempObj1.points[i], 0,tempObj0.numPts-1,tempObj1.sstype[i], tempVec, &tnum,0,1);
	  }else {
	    tdist = findNearestSSEResidue(&tempObj0,tempObj1.cbpoints[i],0,tempObj0.numPts-1,tempObj1.sstype[i],tempVec,&tnum,0,1);
	  }
	  */
	  if(useBetaCarbons) {
	    tdist = findNearestSSEResidue(&tempObj0,tempObj1.cbpoints[i],tempObj1.points[i],0,tempObj0.numPts-1,tempObj1.sstype[i],tempVec,&tnum,0,useBetaCarbons);
	    /*	    tdist = distanceBetweenPoints(tempObj0.points[tnum],tempObj1.points[i]);*/
	  }else {
	    tdist = findNearestSSEResidue(&tempObj0,tempObj1.points[i],tempObj1.points[i],0,tempObj0.numPts-1,tempObj1.sstype[i],tempVec,&tnum,0,useBetaCarbons);
	  }
	  if ((tdist <= distThreshold && tdist>=0) || (thresholdOFF)) {
	    add_pair(tempVec,tempObj1.cbpoints[i],1);
	    add_pair(tempObj0.points[tnum],tempObj1.points[i],1);

	    rmsd += tdist*tdist;
	    count++;
	  }
	}
      
      }else {
	query_length  = (float)(query_end - query_start + 1);
	for(i=query_start;i<=query_end;i++) {
	  if(useBetaCarbons) {
	    tdist = findNearestSSEResidue(&tempObj1,tempObj0.cbpoints[i],tempObj0.points[i],0,tempObj1.numPts-1,tempObj0.sstype[i],tempVec,&tnum,0,useBetaCarbons);
	  }else {
	    tdist = findNearestSSEResidue(&tempObj1,tempObj0.points[i],tempObj1.points[i],0,tempObj1.numPts-1,tempObj0.sstype[i],tempVec,&tnum,0,useBetaCarbons);
	  }
	  if ((tdist <= distThreshold && tdist >=0) || (thresholdOFF)) {
	    add_pair(tempObj0.cbpoints[i],tempVec,1);
	    add_pair(tempObj0.points[i],tempObj1.points[tnum],1);
	    rmsd += tdist*tdist;
	    count++;
	  }
	}
      }
    }
  }
  for (iter = 0; iter < iterations; iter++) {
    /*    printf("struct align iter %i\n",iter);*/
    if (count >= minAlignedResidues && okayToMatchPoints()) { 

       match_pairs (mat, translation);
      
      for (i = 0; i < 4; i++) {
	for (j = 0; j < 4; j++) {
	  m[i*4 + j] = mat[i][j];
	}
      }
      
      myMatMult(target->objMat, m);
      applyMtoV(target->center, target->center, m);
      
      rmsd = 0.0;
      
      applyMat(&tempObj1, target, target->objMat);
      applyMat(&tempObj0, query, query->objMat);

      init_pairs();
    
      count = 0;

  for(SSEnum = 0; SSEnum < target->ssArrayLength;SSEnum++) {
    if(target->ss_alignment[SSEnum][0] != -1) {
      target_start = target->ssArray[2*SSEnum];
      target_end = target->ssArray[2*SSEnum+1]-1;
      target_length = (float)(target_end-target_start+1);
      query_start = query->ssArray[target->ss_alignment[SSEnum][0]*2];
      query_end = query->ssArray[target->ss_alignment[SSEnum][0]*2+1]-1;
      order = 0;
      if(target_end - target_start == query_end - query_start) {
	if(!switched) {
	  order = breakSSELengthTie(query,target);
	}else {
	  order =breakSSELengthTie(target,query);
	}
	if(switched)
	  order*=-1;
      }
      if((order < 0) || (target_end - target_start > query_end - query_start)) {
	for(i=target_start;i<=target_end;i++) {
	  if(useBetaCarbons) {
	    tdist = findNearestSSEResidue(&tempObj0, tempObj1.cbpoints[i], tempObj1.points[i],0,tempObj0.numPts-1,tempObj1.sstype[i], tempVec, &tnum,1,useBetaCarbons);
	  }else {
	    tdist = findNearestSSEResidue(&tempObj0, tempObj1.points[i],tempObj1.points[i], 0,tempObj0.numPts-1,tempObj1.sstype[i], tempVec, &tnum,1,useBetaCarbons);
	  }
	  if ((tdist <= distThreshold && tdist>=0) || (thresholdOFF)) {
	    add_pair(tempVec,tempObj1.cbpoints[i],1);
	    add_pair(tempObj0.points[tnum],tempObj1.points[i],1);
	    rmsd += tdist*tdist;
	    count++;
	  }
	}
      }else {
	/*same as above, but match query to target*/
	query_length = (float)(query_end - query_start +1);
	for(i=query_start;i<=query_end;i++) {
	  if(useBetaCarbons) {
	    tdist = findNearestSSEResidue(&tempObj1, tempObj0.cbpoints[i],tempObj0.points[i], 0,tempObj1.numPts-1,tempObj0.sstype[i], tempVec, &tnum,1,useBetaCarbons);
	  }else {
	    tdist = findNearestSSEResidue(&tempObj1, tempObj0.points[i],tempObj1.points[i], 0,tempObj1.numPts-1,tempObj0.sstype[i], tempVec, &tnum,1,useBetaCarbons);
	  }
	  if ((tdist <= distThreshold && tdist>=0) || (thresholdOFF)) {
	    add_pair(tempObj0.cbpoints[i],tempVec,1);
	    add_pair(tempObj0.points[i],tempObj1.points[tnum],1);
	    rmsd += tdist*tdist;
	    count++;
	  }
	}
      }
    }else {
      target_start = target->ssArray[2*SSEnum];
      target_end = target->ssArray[2*SSEnum+1]-1;
      target_length = (float)(target_end-target_start+1);
      order = 0;
      if(target_end - target_start == query_end - query_start) {
if(!switched) {
	  order = breakSSELengthTie(query,target);
	}else {
	  order =breakSSELengthTie(target,query);
	}
      }
      if(1) {
	for(i=target_start;i<=target_end;i++) {
	  if(useBetaCarbons) {
	    tdist = findNearestSSEResidue(&tempObj0,tempObj1.cbpoints[i],tempObj1.points[i],0,tempObj0.numPts-1,tempObj1.sstype[i],tempVec,&tnum,0,useBetaCarbons);
	  }else {
	    tdist = findNearestSSEResidue(&tempObj0,tempObj1.points[i],tempObj1.points[i],0,tempObj0.numPts-1,tempObj1.sstype[i],tempVec,&tnum,0,useBetaCarbons);
	  }
	  if ((tdist <= distThreshold && tdist>=0) || (thresholdOFF)) {
	    add_pair(tempVec,tempObj1.cbpoints[i],1);
	    add_pair(tempObj0.points[tnum],tempObj1.points[i],1);
	    rmsd += tdist*tdist;
	    count++;
	  }
	}
      
      }else {
	query_length  = (float)(query_end - query_start + 1);
	for(i=query_start;i<=query_end;i++) {
	  if(useBetaCarbons) {
	    tdist = findNearestSSEResidue(&tempObj1,tempObj0.cbpoints[i],tempObj0.points[i],0,tempObj1.numPts-1,tempObj0.sstype[i],tempVec,&tnum,0,useBetaCarbons);
	  }else{
	    tdist = findNearestSSEResidue(&tempObj1,tempObj0.points[i],tempObj1.points[i],0,tempObj1.numPts-1,tempObj0.sstype[i],tempVec,&tnum,0,useBetaCarbons);
	  }
	  if ((tdist <= distThreshold && tdist>=0) || (thresholdOFF)) {
	    add_pair(tempObj0.cbpoints[i],tempVec,1);
	    add_pair(tempObj0.points[i],tempObj1.points[tnum],1);
	    rmsd += tdist*tdist;
	    count++;
	  }
	}
      }
    }
  }
   
      rmsd = sqrt((rmsd)/((GLfloat)(count)));
      globalNumAligned = count;
      
      diffrmsd = oldrmsd - rmsd;
      
      if (diffrmsd < 0) 
	diffrmsd *= -1.0;
      
      if ((diffrmsd < epsilon) && (iter < iterations - 1))  {
	/*iter = iterations - 1;*/
	converged = 1;
	break;
      }
    
      oldrmsd = rmsd;
    
    }
    else {

      printf("There are less than four atoms correctly aligned\n");      
      globalNumAligned = 0;
      rmsd = 0.0;
      /*iter = iterations - 1;*/
      break;

    }
  }


  if (globalNumAligned != 0) {
    
    count = 0;
    rmsd = 0.0;

    for(SSEnum = 0; SSEnum < target->ssArrayLength;SSEnum++) {
      if(target->ss_alignment[SSEnum][0]!=-1) {
	target_start = target->ssArray[2*SSEnum];
	target_end = target->ssArray[2*SSEnum+1]-1;
	query_start = query->ssArray[target->ss_alignment[SSEnum][0]*2];
	query_end = query->ssArray[target->ss_alignment[SSEnum][0]*2+1]-1;
	order = 0;
	if(target_end - target_start == query_end - query_start) {
	  if(!switched) {
	    order = breakSSELengthTie(query,target);
	  }else {
	    order =breakSSELengthTie(target,query);
	  }

	  if(switched)
	    order*=-1;
	}
	if((order < 0) || (target_end - target_start > query_end - query_start)) {
	  for (i = target_start; i <=target_end; i++) {
	    if(useBetaCarbons) {
	      tdist = findNearestSSEResidue(&tempObj0, tempObj1.cbpoints[i],tempObj1.points[i], 0,tempObj0.numPts-1,tempObj1.sstype[i], tempVec, &tnum,1,useBetaCarbons);
	    }else {
	      tdist = findNearestSSEResidue(&tempObj0, tempObj1.points[i],tempObj1.points[i], 0,tempObj0.numPts-1,tempObj1.sstype[i], tempVec, &tnum,1,useBetaCarbons);
	    }
	    if ((tdist <= distThreshold && tdist>=0) || (thresholdOFF)) {

	      if (query->pairNums[tnum] != -1) {
		if (tdist < query->pairDists[tnum]) {	    
		  rmsd -= (query->pairDists[tnum])*(query->pairDists[tnum]);
		  target->pairNums[query->pairNums[tnum]] = -1;
		  query->pairNums[tnum] = i;	    
		  query->pairDists[tnum] = tdist;
		  target->pairNums[i] = tnum;
		  target->pairDists[i] = tdist;
		  rmsd += tdist*tdist;
		}
	      }
	      else {
		query->pairNums[tnum] = i;	    
		query->pairDists[tnum] = tdist;
		target->pairNums[i] = tnum;
		target->pairDists[i] = tdist;
		rmsd += tdist*tdist;
		count++;
	      }	  
	    }
	  }
	}else {
	  for (i = query_start; i <=query_end; i++) {
	    if(useBetaCarbons) {
	      tdist = findNearestSSEResidue(&tempObj1, tempObj0.cbpoints[i], tempObj0.points[i],0,tempObj1.numPts-1,tempObj0.sstype[i], tempVec, &tnum,1,useBetaCarbons);
	    }else {
	      tdist = findNearestSSEResidue(&tempObj1, tempObj0.points[i],tempObj1.points[i], 0,tempObj1.numPts-1,tempObj0.sstype[i], tempVec, &tnum,1,useBetaCarbons);
	    }
	    if ((tdist <= distThreshold && tdist >=0) || (thresholdOFF)) {
	      if (query->pairNums[i] != -1) {
		if (tdist < query->pairDists[i]) {	    
		  rmsd -= (query->pairDists[i])*(query->pairDists[i]);
		  target->pairNums[query->pairNums[i]] = -1;
		  query->pairNums[i] = tnum;
		  query->pairDists[i] = tdist;
	          target->pairNums[tnum] = i;
		  target->pairDists[tnum] = tdist;
		  rmsd += tdist*tdist;
		}
	      }
	      else {
		target->pairNums[tnum] = i;
		target->pairDists[tnum] = tdist;
		query->pairNums[i] = tnum;
		query->pairDists[i] = tdist;
		rmsd += tdist*tdist;
		count++;
	      }	  
	    }
	  }
	}
      }else {
	target_start = target->ssArray[2*SSEnum];
	target_end = target->ssArray[2*SSEnum+1]-1;
	for(i=target_start;i<=target_end;i++) {
	  if(useBetaCarbons) {
	    tdist = findNearestSSEResidue(&tempObj0,tempObj1.cbpoints[i],tempObj1.points[i],0,tempObj0.numPts-1,tempObj1.sstype[i],tempVec,&tnum,0,useBetaCarbons);
	  }else {
	    tdist = findNearestSSEResidue(&tempObj0,tempObj1.points[i],tempObj1.points[i],0,tempObj0.numPts-1,tempObj1.sstype[i],tempVec,&tnum,0,useBetaCarbons);
	  }
	  if ((tdist <= distThreshold && tdist >=0) || (thresholdOFF)) {
	    if (query->pairNums[tnum] != -1) {
	      if (tdist < query->pairDists[tnum]) {	    
		rmsd -= (query->pairDists[tnum])*(query->pairDists[tnum]);
		target->pairNums[query->pairNums[tnum]] = -1;
		query->pairNums[tnum] = i;	    
		query->pairDists[tnum] = tdist;
		target->pairNums[i] = tnum;
		target->pairDists[i] = tdist;
		rmsd += tdist*tdist;
	      }
	    }else {
	      target->pairNums[i] = tnum;
	      target->pairDists[i] = tdist;
	      query->pairNums[tnum] = i;	    
	      query->pairDists[tnum] = tdist;
	      rmsd += tdist*tdist;
	      count++;
	    }	
	  }
	}
      }
    }
    
    rmsd = sqrt((rmsd)/((GLfloat)(count)));
    globalNumAligned = count;
    
    
  }
  

  /*  if ((converged == 0) && (iterations == MAXITERATIONS)) {*/
  if(!converged) {
    printf("Did not converge\n");
  }



  globalLastAlignment = 0;

  return rmsd;
  
}


void structAlignSingle(OBJ* t0, OBJ* t1) {

  /*  OBJ *t0, *t1;*/
  float rmsd;
  int i;

  /*  t0 = ObjList[0];
  t1 = ObjList[1];*/
  
  for (i = 0; i < 10; i++) {
    rmsd = structAlign(t0, t1, 1, 0.0);
    
    globalRMSD = rmsd;
    
    printf("structalignsingle: RMSD = %f %d\n", rmsd, globalNumAligned);
    
#ifdef DISPLAYON
    if (globalStartDisplay) {
      display();
    }
#endif
  }

#ifdef DISPLAYON
  if (globalStartDisplay) {
#if DRAWJOINTS
    createJointList();
#endif
    display();
  }
#endif

}


void structAlignFull(OBJ* t0, OBJ* t1) {

  /*  OBJ *t0, *t1;*/
  float rmsd;

  /*  t0 = ObjList[0];
  t1 = ObjList[1];*/

  if(t0 == NULL) {
    t0 = ObjList[0];
  }
  if(t1 == NULL) {
    t1 = ObjList[1];
  }

  rmsd = structAlign(t0, t1, MAXITERATIONS, MAXDIFFERENCE);

  globalRMSD = rmsd;
  
  printf("structalignfull: RMSD = %f %d\n", rmsd, globalNumAligned);
  
#ifdef DISPLAYON
  if (globalStartDisplay) { 
#if DRAWJOINTS
    createJointList();
#endif
    display();
  }
#endif
}

/****************************************************************************************
                                    Alignment Functions
				    Core Align (step 3)
****************************************************************************************/


void findBestSequence(int *a, float *dist, int start, int end) {

  int i, j;
  int index,a_start,a_end;
  int myscore, mypointer, maxscore, maxindex;
  int scorearray[MAXLENGTH];
  int pointarray[MAXLENGTH];

  maxscore = -100;
  maxindex = -1;

  /*  for (i = 0; i < end; i++) {*/
  /*  printf("finding best sequence from %i to %i\n",start,end);*/
  for (i = start; i < end; i++) {
    scorearray[i] = -1;
    pointarray[i] = -1;
    
    if (a[i] != -1) {
      myscore = 1;
      mypointer = -1;
      /*      for (j = i-1; j >=0; j--) {  */
      for(j=i-1;j>=start;j--) {
	if ((a[j] != -1) && (a[j] < a[i])) {
	  if (scorearray[j] + 1 > myscore) {	
	    myscore = scorearray[j] + 1;
	    mypointer = j;
	  }
	  else if (scorearray[j] + 1 == myscore) {
	    if (dist[j] < dist[mypointer]) {
	      mypointer = j;
	    }
	  }
	}
      }  
      scorearray[i] = myscore;
      pointarray[i] = mypointer;
      if (myscore > maxscore) {
	maxscore = myscore;
	maxindex = i;
      }
    }

  }

  while (maxindex != -1) {
    scorearray[maxindex] = 0;
    maxindex = pointarray[maxindex];
  }

  /*  for (i = 0; i < end; i++) {*/
  for (i = start; i < end; i++) {
    if (scorearray[i] != 0) {
      a[i] = -1;
    }
  }
  

}

/*calculates aligned pairs between query and target assuming query and target do NOT need
 *to be transformed before calculation*/
float recalculateAlignment(OBJ* query, OBJ* target) {
  int i,len0,len1,tnum;
  float rmsd,myrmsd,tdist;
  int correctArray1[MAXLENGTH];
  float distArray1[MAXLENGTH];
  int posArray0[MAXLENGTH];
  float distArray0[MAXLENGTH];
  GLfloat tempVec[3];
  int totalCount = 0;


  for (i = 0; i < target->numPts; i++) {
    target->pairNums[i] = -1;
  }  
  
  for (i = 0; i < query->numPts; i++) {
    query->pairNums[i] = -1;
  }  
  memset(query->pairNums,-1,query->numPts*sizeof(int));
  memset(target->pairNums,-1,target->numPts*sizeof(int));

  len1 = target->numPts;
  len0 = query->numPts;
  for (i = 0; i < len1; i++) {    
    distArray1[i] = 1000.0;;
    correctArray1[i] = -1;
  }    
  
  for (i = 0; i < len0; i++) {    
    tdist = findNearest(target, query->points[i], query->sstype[i], tempVec, &tnum);
    posArray0[i] = tnum;    
    distArray0[i] = tdist;
  }
  for (i = 0; i < len0; i++) {
    if ((distArray0[i] < distArray1[posArray0[i]]) && ((distArray0[i] < distThreshold) || (thresholdOFF))) {
      correctArray1[posArray0[i]] = i;
      distArray1[posArray0[i]] = distArray0[i];
    }
  }


 findBestSequence(correctArray1, distArray1, 0, len1);

  for(i=0;i<len1;i++) {
    if(correctArray1[i]!=-1) {
      totalCount++;
      myrmsd+=(distArray1[i])*(distArray1[i]);
    }
  }

  rmsd = sqrt((myrmsd)/((GLfloat)(totalCount)));
  if (totalCount > 0) {
    for (i = 0; i < len1; i++) {
      if (correctArray1[i] != -1) {
	target->pairNums[i] = correctArray1[i];
	target->pairDists[i] = distArray1[i];
	
	
	if (query->pairNums[correctArray1[i]] != -1) {
	  if (distArray1[i] < query->pairDists[correctArray1[i]]) {
	    query->pairNums[correctArray1[i]] = i;
	    query->pairDists[correctArray1[i]] = distArray1[i];
	  }
	}
	else {
	  query->pairNums[correctArray1[i]] = i;
	  query->pairDists[correctArray1[i]] = distArray1[i];
	}	  
      }
    }
  }
  
  globalCoreNumAligned = totalCount;

  return rmsd;
}


float refineAlign(OBJ* query, OBJ* target, int iterations, float epsilon, int doss) {

  OBJ tempObj1, tempObj0, *temp;
  GLfloat tempVec[3];
  GLfloat tdist;
  int i, j, tnum,index;
  int start,end;
  float rmsd = 0.0, oldrmsd = 1000.0, diffrmsd;
  int iter, converged = 0;
  int target_start,target_end;
  Vector translation;
  Matrix mat;
  GLfloat m[16];
  int SSEnum,query_start,query_end,switched = 0;
  int SSEstart,SSEend,query_alignment_start,query_alignment_end,target_alignment_start,target_alignment_end;

  float distArray1[MAXLENGTH];
  float distArray0[MAXLENGTH];
  int correctArray1[MAXLENGTH];
  int correctArray0[MAXLENGTH];
  int tempArray[MAXLENGTH];
  int len1;
  int order;
  int posArray1[MAXLENGTH],posArray0[MAXLENGTH];
  float correctDistArray0[MAXLENGTH],correctDistArray1[MAXLENGTH];
  int numcorrect = 0;
  int len0;

  int totalCount = 0;
  float myrmsd = 0.0;

 /*set protein with more SSEs as target for purposes of symmetry*/

  if(target->ssArrayLength < query->ssArrayLength) {
    temp = query;
    query = target;
    target = temp;
    switched = 1;

  }else if((target->ssArrayLength==query->ssArrayLength) && strcmp(query->objSeq,target->objSeq) > 0) {
    temp = query;
    query = target;
    target = temp;
    switched = 1;

  }

  /*  printf("aligned residues at start of refinealign:\n");
  for(i=0;i<query->numPts;i++) {
    if(query->pairNums[i] != -1) {
      printf("%i %i\n",query->aaNums[i],target->aaNums[query->pairNums[i]]);
    }
  }
  printf("\n");
  */
  memset(query->pairNums,-1,query->numPts*sizeof(int));
  memset(target->pairNums,-1,target->numPts*sizeof(int));

  tempObj1.numPts = target->numPts;
  tempObj0.numPts = query->numPts;

  tempObj1.sstype = target->sstype;
  tempObj0.sstype = query->sstype;
  
  tempObj1.objSeq = target->objSeq;
  tempObj0.objSeq = query->objSeq;

  applyMat(&tempObj0, query, query->objMat);
  applyMat(&tempObj1, target, target->objMat);
  
  len1 = target->numPts;
  len0 = query->numPts;

  init_pairs();
  
  numcorrect = 0;
  
  totalCount = 0;
  myrmsd = 0.0;
  
  applyMat(&tempObj1, target, target->objMat);    
  

  for(i=0;i<len0;i++) {
    distArray0[i] = 1000.0;	 
    correctDistArray0[i] = 1000.0;
    posArray0[i] = -1;
    correctArray0[i] = -1;
  }
  for(i=0;i<len1;i++) {
    correctDistArray1[i] = 1000.0;
    distArray1[i] = 1000.0;
    correctArray1[i] = -1;
    posArray1[i] = -1;
  }


  for(SSEnum=0;SSEnum<target->ssArrayLength;SSEnum++) {
    if(target->ss_alignment[SSEnum][0]!=-1) {

      target_start = target->ssArray[2*SSEnum];
      target_end = target->ssArray[2*SSEnum+1]-1;
      query_start = query->ssArray[2*target->ss_alignment[SSEnum][0]];
      query_end = query->ssArray[2*target->ss_alignment[SSEnum][0]+1]-1;
      order = 0;
      if(target_end - target_start == query_end-query_start) {
	if(!switched) {
	  order = breakSSELengthTie(query,target);
	}else {
	  order =breakSSELengthTie(target,query);
	}
	if(switched)
	  order *= -1;
      }      
      if((order < 0) || (target_end - target_start  > query_end - query_start)) {

	for(i=target_start;i<=target_end;i++) {
	  if(useBetaCarbons) {
	    tdist = findNearestSSEResidue(&tempObj0, tempObj1.cbpoints[i],tempObj1.points[i], 0,query->numPts-1,target->sstype[i], tempVec, &tnum,1,useBetaCarbons);
	  }else {
	    tdist = findNearestSSEResidue(&tempObj0, tempObj1.points[i],tempObj1.points[i], 0,query->numPts-1,target->sstype[i], tempVec, &tnum,1,useBetaCarbons);
	  }

	  if(posArray0[tnum] != -1) { /*is query residue already aligned?*/
	    if(tdist >= 0 && tdist < distArray0[tnum]) { /*if this distance is better*/
	      if(posArray1[i]!= -1) { /*if target residue is already aligned*/
		if(tdist < distArray1[i]) { /*if this distance is better*/
		  posArray1[posArray0[tnum]] = -1;
		  posArray0[posArray1[i]] = -1;
		  posArray1[i] = tnum;    
		  posArray0[tnum] = i;
		  distArray1[i] = tdist;
		  distArray0[tnum] = tdist;
		}
	      }else {/*query residue already aligned; target not*/
		posArray1[posArray0[tnum]] = -1;
		posArray1[i] = tnum;    
		posArray0[tnum] = i;
		distArray1[i] = tdist;
		distArray0[tnum] = tdist;
	      }
	    }
	  }else if(posArray1[i]!=-1) { /*query residue not aligned, target is*/
	    if(tdist >=0 && tdist < distArray1[i]) {
	      posArray0[posArray1[i]] = -1;
	      posArray1[i] = tnum;    
	      posArray0[tnum] = i;
	      distArray1[i] = tdist;
	      distArray0[tnum] = tdist;
	    }
	  }else { /*neither residue previously aligned*/
	      posArray1[i] = tnum;    
	      posArray0[tnum] = i;
	      distArray1[i] = tdist;
	      distArray0[tnum] = tdist;
	    }
	  
	}
      }else {

	for(i=query_start;i<=query_end;i++) {
	  if(useBetaCarbons) {
	    tdist = findNearestSSEResidue(&tempObj1, tempObj0.cbpoints[i],tempObj0.points[i],0,target->numPts-1,query->sstype[i],tempVec, &tnum,1,useBetaCarbons);
	  }else {
	    tdist = findNearestSSEResidue(&tempObj1, tempObj0.points[i],tempObj1.points[i],0,target->numPts-1,query->sstype[i],tempVec, &tnum,1,useBetaCarbons);
	  }
	  if(posArray0[i]!=-1) {
	    if(tdist >= 0 && tdist < distArray0[i]) {
	      if(posArray1[tnum]!=-1) { /*query and target previously aligned*/
		if(tdist < distArray1[tnum]) {
		  posArray1[posArray0[i]] = -1;
		  posArray0[posArray1[tnum]] = -1;
		  posArray1[tnum] = i;
		  posArray0[i] = tnum;
		  distArray1[tnum] = tdist;
		  distArray0[i] = tdist;
		}
	      }else { /*query previously aligned, target not*/
		posArray1[posArray0[i]] = -1;
		posArray1[tnum] = i;
		posArray0[i] = tnum;
		distArray1[tnum] = tdist;
		distArray0[i] = tdist;
	      }
	    }
	  }else if(posArray1[tnum]!=-1) { /*query not aligned, target is*/
	    if(tdist >=0 && tdist < distArray1[tnum]) {
	      posArray0[posArray1[tnum]] = -1;
	      posArray1[tnum] = i;
	      posArray0[i] = tnum;
	      distArray1[tnum] = tdist;
	      distArray0[i] = tdist;
	    }

	  }else { /*neither previously aligned*/
	    posArray1[tnum] = i;
	    posArray0[i] = tnum;
	    distArray1[tnum] = tdist;
	    distArray0[i] = tdist;
	  }
	}
      }
    }else {
      /*if not aligned, always match target SSE*/
      target_start =target->ssArray[SSEnum*2];
      target_end = target->ssArray[SSEnum*2+1]-1;

      for(i=target_start;i<=target_end;i++) {
	if(useBetaCarbons) {
	  tdist = findNearestSSEResidue(&tempObj0, tempObj1.cbpoints[i], tempObj1.points[i],0,query->numPts-1,target->sstype[i], tempVec, &tnum,0,useBetaCarbons);
	}else{
	  tdist = findNearestSSEResidue(&tempObj0, tempObj1.points[i],tempObj1.points[i], 0,query->numPts-1,target->sstype[i], tempVec, &tnum,0,useBetaCarbons);
	}
	if(posArray0[tnum] != -1) {
	  if(tdist >=0 && tdist < distArray0[tnum]) {
	    
	    if(posArray1[i]!= -1) { /*if target residue is already aligned*/
	      if(tdist < distArray1[i]) { /*if this distance is better*/
		  posArray1[posArray0[tnum]] = -1;
		  posArray0[posArray1[i]] = -1;
		  posArray1[i] = tnum;    
		  posArray0[tnum] = i;
		  distArray1[i] = tdist;
		  distArray0[tnum] = tdist;
		}
	      }else {/*query residue already aligned; target not*/
		posArray1[posArray0[tnum]] = -1;
		posArray1[i] = tnum;    
		posArray0[tnum] = i;
		distArray1[i] = tdist;
		distArray0[tnum] = tdist;
	      }
	    }
	}else if(posArray1[i]!=-1) { /*query residue not aligned, target is*/
	  if(tdist>=0 && tdist < distArray1[i]) {
	    posArray0[posArray1[i]] = -1;
	    posArray1[i] = tnum;    
	    posArray0[tnum] = i;
	    distArray1[i] = tdist;
	    distArray0[tnum] = tdist;
	  }
	}else { /*neither residue previously aligned*/
	  posArray1[i] = tnum;    
	  posArray0[tnum] = i;
	  distArray1[i] = tdist;
	  distArray0[tnum] = tdist;
	}
	
      }
    }
  }
              
  for(SSEnum=0;SSEnum<target->ssArrayLength;SSEnum++) {
    target_start = target->ssArray[SSEnum*2];
    target_end = target->ssArray[SSEnum*2+1]-1;
    if(target->ss_alignment[SSEnum][0] !=-1) {
      query_start = query->ssArray[2*target->ss_alignment[SSEnum][0]];
      query_end = query->ssArray[2*target->ss_alignment[SSEnum][0]+1]-1;
      order = 0;
      if(target_end - target_start == query_end-query_start) {
	if(!switched) {
	  order = breakSSELengthTie(query,target);
	}else {
	  order =breakSSELengthTie(target,query);
	}
	if(switched)
	  order *= -1;
      }
    
      if(order < 0 || (target_end - target_start > query_end - query_start)) {
	for(i=target_start;i<=target_end;i++) {
	  if(posArray1[i] != -1) {
	  if ((distArray1[i] < correctDistArray0[posArray1[i]]) && ((distArray1[i] < distThreshold) || (thresholdOFF))) {
	    correctArray0[posArray1[i]] = i;
	    correctArray1[i] = posArray1[i];
	    correctDistArray0[posArray1[i]] = distArray1[i];
	    correctDistArray1[i] = distArray1[i];
	  }
	}
	}
      }else {
	for(i=query_start;i<=query_end;i++) {
	  if(distArray0[i] < correctDistArray1[posArray0[i]] && ((distArray0[i] < distThreshold) || thresholdOFF)) {
	    correctArray0[i] = posArray0[i];
	    correctArray1[posArray0[i]] = i;
	    correctDistArray0[i] = distArray0[i];
	    correctDistArray1[posArray0[i]] = distArray0[i];
	  }
	}
      }
    }else {
      /*if not aligned, always match target SSE*/
      target_start = target->ssArray[SSEnum*2];
      target_end = target->ssArray[SSEnum*2+1]-1;
      for(i=target_start;i<=target_end;i++) {
	if(posArray1[i] != -1) {
	if ((distArray1[i] < correctDistArray0[posArray1[i]]) && ((distArray1[i] < distThreshold) || (thresholdOFF))) {
	  /*	      correctArray1[posArray0[i]] = i;*/
	  correctArray0[posArray1[i]] = i;
	  correctArray1[i] = posArray1[i];
	  correctDistArray0[posArray1[i]] = distArray1[i];
	  correctDistArray1[i] = distArray1[i];
	  /*	      distArray1[posArray0[i]] = distArray0[i];*/
	}
	}
      }
    }
  }  


    /*orig:    for(i=0;i<query->ssArrayLength;i++) {*/
  for(i=0;i<target->ssArrayLength;i++) {
    target_start = target->ssArray[i*2];
    target_end = target->ssArray[i*2+1]-1;
    if(target->ss_alignment[i][0] !=-1) {
      query_start = query->ssArray[2*target->ss_alignment[i][0]];
      query_end = query->ssArray[2*target->ss_alignment[i][0]+1]-1;
      order = 0;
      if(target_end - target_start == query_end-query_start) {
	/*order = strcmp(getFileName(0),getFileName(1));*/
	/*order= strcmp(query->objSeq,target->objSeq);*/
if(!switched) {
	  order = breakSSELengthTie(query,target);
	}else {
	  order =breakSSELengthTie(target,query);
	}
	if(switched)
	  order *= -1;
      }
      if(order < 0 || (target_end - target_start > query_end - query_start)) {
	for(j=target_start;j<=target_end;j++) {
	  tempArray[j-target_start] = correctArray1[j];
	}
	findBestSequence(correctArray1,correctDistArray1,target_start,target_end+1);
	for(j=target_start;j<=target_end;j++) {
	  if(correctArray1[j]!=-1) {
	    correctArray0[correctArray1[j]] = j;
	  }
	}
	for(j=target_start;j<=target_end;j++) {
	  if(correctArray1[j] != tempArray[j-target_start]) {
	    correctArray0[tempArray[j-target_start]] = correctArray1[j];
	  }
	}
      }else {

	for(j=query_start;j<=query_end;j++) {
	  tempArray[j-query_start] = correctArray0[j];
	}
	findBestSequence(correctArray0,correctDistArray0,query_start,query_end+1);
	for(j=query_start;j<=query_end;j++) {
	  if(correctArray0[j]!=-1) {
	    correctArray1[correctArray0[j]] = j;
	  }
	}
	for(j=query_start;j<=query_end;j++) {
	  if(correctArray0[j] != tempArray[j-query_start]) {
	    correctArray1[tempArray[j-query_start]] = correctArray0[j];
	  }
	}
      }
    }else {

      for(j=target_start;j<=target_end;j++) {
	tempArray[j-target_start] = correctArray1[j];
      }
      findBestSequence(correctArray1,correctDistArray1,target_start,target_end+1);
      for(j=target_start;j<=target_end;j++) {
	if(correctArray1[j]!=-1) {
	  correctArray0[correctArray1[j]] = j;
	}
      }
      for(j=target_start;j<=target_end;j++) {
	if(correctArray1[j] != tempArray[j-target_start]) {
	  correctArray0[tempArray[j-target_start]] = correctArray1[j];
	}
      }
    }
  }
    


  findBestSequence(correctArray1,distArray1,0,len1);
  for(j=0;j<len0;j++) {
    correctArray0[j]=-1;
  }
  for(j=0;j<len1;j++) {
    if(correctArray1[j]!=-1) {
      correctArray0[correctArray1[j]]=j;
    }
  }


  for (i = 0; i < len1; i++) {
    if (correctArray1[i] != -1) {
      add_pair(tempObj0.cbpoints[correctArray1[i]], tempObj1.cbpoints[i], 1);

      totalCount++;
      myrmsd += (correctDistArray1[i])*(correctDistArray1[i]);

    }
  }

  /*still need to make modifications here*/
  for (iter = 0; iter < iterations; iter++) {

    if (totalCount >= minAlignedResidues && okayToMatchPoints()) {
      
      match_pairs (mat, translation);
    
      for (i = 0; i < 4; i++) {
	for (j = 0; j < 4; j++) {
	  m[i*4 + j] = mat[i][j];
	}
      }
      
          
      myMatMult(target->objMat, m);
      applyMtoV(target->center, target->center, m);      


      init_pairs();
      
      numcorrect = 0;
      
      totalCount = 0;
      myrmsd = 0.0;
      
      applyMat(&tempObj1, target, target->objMat);    
      

      for(i=0;i<len0;i++) {
	distArray0[i] = 1000.0;
	correctDistArray0[i] = 1000.0;
	posArray0[i] = -1;
	correctArray0[i] = -1;
      }
      for(i=0;i<len1;i++) {
	correctDistArray1[i] = 1000.0;
	distArray1[i] = 1000.0;
	correctArray1[i] = -1;
	posArray1[i] = -1;
      }

      for(SSEnum=0;SSEnum<target->ssArrayLength;SSEnum++) {
	if(target->ss_alignment[SSEnum][0]!=-1) {
	  target_start = target->ssArray[2*SSEnum];
	  target_end = target->ssArray[2*SSEnum+1]-1;
	  query_start = query->ssArray[2*target->ss_alignment[SSEnum][0]];
	  query_end = query->ssArray[2*target->ss_alignment[SSEnum][0]+1]-1;
	  order = 0;
	  if(target_end - target_start == query_end-query_start) {
	    if(!switched) {
	      order = breakSSELengthTie(query,target);
	    }else {
	      order =breakSSELengthTie(target,query);
	    }
	    if(switched)
	      order *= -1;
	  }
	  if((order < 0) || (target_end - target_start  > query_end - query_start)) {
	 
	    for(i=target_start;i<=target_end;i++) {
	      if(useBetaCarbons) {
		tdist = findNearestSSEResidue(&tempObj0, tempObj1.cbpoints[i],tempObj1.points[i], 0,query->numPts-1,target->sstype[i], tempVec, &tnum,1,useBetaCarbons);
		/*		tdist = distanceBetweenPoints(tempObj0.points[tnum],tempObj1.points[i]);*/
	      }else {
		tdist = findNearestSSEResidue(&tempObj0, tempObj1.points[i],tempObj1.points[i], 0,query->numPts-1,target->sstype[i], tempVec, &tnum,1,useBetaCarbons);
	      }
	      if(posArray0[tnum] != -1) { /*is query residue already aligned?*/
		if(tdist >=0 && tdist < distArray0[tnum]) { /*if this distance is better*/
		  if(posArray1[i]!= -1){ /*if target residue is already aligned*/
		    if(tdist < distArray1[i]) {
		      posArray1[posArray0[tnum]] = -1;
		      posArray0[posArray1[i]] = -1;
		      posArray1[i] = tnum;    
		      posArray0[tnum] = i;
		      distArray1[i] = tdist;
		      distArray0[tnum] = tdist;
		    }
		  }else {/*query residue already aligned; target not*/
		    posArray1[posArray0[tnum]] = -1;
		    posArray1[i] = tnum;    
		    posArray0[tnum] = i;
		    distArray1[i] = tdist;
		    distArray0[tnum] = tdist;
		  }
		}
	      }else if(posArray1[i]!=-1) { /*query residue not aligned, target is*/
		if(tdist >=0 && tdist < distArray1[i]) {
		  posArray0[posArray1[i]] = -1;
		  posArray1[i] = tnum;    
		  posArray0[tnum] = i;
		  distArray1[i] = tdist;
		  distArray0[tnum] = tdist;
		}
	      }else { /*neither residue previously aligned*/
		  posArray1[i] = tnum;    
		  posArray0[tnum] = i;
		  distArray1[i] = tdist;
		  distArray0[tnum] = tdist;
	      }	      
	    }
	  }else {
	    for(i=query_start;i<=query_end;i++) {
	      if(useBetaCarbons) {
		tdist = findNearestSSEResidue(&tempObj1, tempObj0.cbpoints[i],tempObj0.points[i],0,target->numPts-1,query->sstype[i],tempVec, &tnum,1,useBetaCarbons);
		/*		tdist = distanceBetweenPoints(tempObj0.points[i],tempObj1.points[tnum]);*/
	      }else {
		tdist = findNearestSSEResidue(&tempObj1, tempObj0.points[i],tempObj1.points[i],0,target->numPts-1,query->sstype[i],tempVec, &tnum,1,useBetaCarbons);
	      }
	      if(posArray0[i]!=-1) {
		if(tdist >= 0 && tdist < distArray0[i]) {
		  if(posArray1[tnum]!=-1) { /*query and target previously aligned*/
		    if(tdist < distArray1[tnum]) {
		      posArray1[posArray0[i]] = -1;
		      posArray0[posArray1[tnum]] = -1;
		      posArray1[tnum] = i;
		      posArray0[i] = tnum;
		      distArray1[tnum] = tdist;
		      distArray0[i] = tdist;
		    }
		  }else { /*query previously aligned, target not*/
		    posArray1[posArray0[i]] = -1;
		    posArray1[tnum] = i;
		    posArray0[i] = tnum;
		    distArray1[tnum] = tdist;
		    distArray0[i] = tdist;
		  }
		}
	      }else if(posArray1[tnum]!=-1) { /*query not aligned, target is*/
		if(tdist >=0 && tdist < distArray1[tnum]) {
		  posArray0[posArray1[tnum]] = -1;
		  posArray1[tnum] = i;
		  posArray0[i] = tnum;
		  distArray1[tnum] = tdist;
		  distArray0[i] = tdist;
		}
	      }else { /*neither previously aligned*/
		posArray1[tnum] = i;
		posArray0[i] = tnum;
		distArray1[tnum] = tdist;
		distArray0[i] = tdist;
	      }
	    }
	  }
	
	}else {
	  target_start =target->ssArray[SSEnum*2];
	  target_end = target->ssArray[SSEnum*2+1]-1;
	  for(i=target_start;i<=target_end;i++) {
	    if(useBetaCarbons) {
	      tdist = findNearestSSEResidue(&tempObj0, tempObj1.cbpoints[i],tempObj1.points[i], 0,query->numPts-1,target->sstype[i], tempVec, &tnum,0,useBetaCarbons);
	    }else{
	      tdist = findNearestSSEResidue(&tempObj0, tempObj1.points[i],tempObj1.points[i], 0,query->numPts-1,target->sstype[i], tempVec, &tnum,0,useBetaCarbons);
	    }
	    if(posArray0[tnum]!= -1) {
	      if(tdist >= 0 && tdist < distArray0[tnum]) {
		if(posArray1[i]!= -1) { /*if target residue is already aligned*/
		  if(tdist < distArray1[i]) { /*if this distance is better*/
		    posArray1[posArray0[tnum]] = -1;
		    posArray0[posArray1[i]] = -1;
		    posArray1[i] = tnum;    
		    posArray0[tnum] = i;
		    distArray1[i] = tdist;
		    distArray0[tnum] = tdist;
		  }
		}else {/*query residue already aligned; target not*/
		  posArray1[posArray0[tnum]] = -1;
		  posArray1[i] = tnum;    
		  posArray0[tnum] = i;
		  distArray1[i] = tdist;
		  distArray0[tnum] = tdist;
		}
	      }
	    }else if(posArray1[i]!=-1) { /*query residue not aligned, target is*/
	      if(tdist >=0 && tdist < distArray1[i]) {
		posArray0[posArray1[i]] = -1;
		posArray1[i] = tnum;    
		posArray0[tnum] = i;
		distArray1[i] = tdist;
		distArray0[tnum] = tdist;
	      } /*remove bracket to put bug back in*/
	    }else { /*neither residue previously aligned*/
	      posArray1[i] = tnum;    
	      posArray0[tnum] = i;
	      distArray1[i] = tdist;
	      distArray0[tnum] = tdist;
	      }
	  }
	}
      }
      /* }*//*extra bracket to put bug back in*/
  
      for(SSEnum=0;SSEnum<target->ssArrayLength;SSEnum++) {
	target_start = target->ssArray[SSEnum*2];
	target_end = target->ssArray[SSEnum*2+1]-1;
	if(target->ss_alignment[SSEnum][0] !=-1) {
	  query_start = query->ssArray[2*target->ss_alignment[SSEnum][0]];
	  query_end = query->ssArray[2*target->ss_alignment[SSEnum][0]+1]-1;
	  order = 0;
	  if(target_end - target_start == query_end-query_start) {
	    /*order = strcmp(getFileName(0),getFileName(1));*/
	    /*	    order = strcmp(query->objSeq,target->objSeq);*/
	    if(!switched) {
	      order = breakSSELengthTie(query,target);
	    }else {
	      order =breakSSELengthTie(target,query);
	    }
	    if(switched)
	      order *= -1;
	  }
	  
	  if(order < 0 || (target_end - target_start > query_end - query_start)) {
	    for(i=target_start;i<=target_end;i++) {
	      if(posArray1[i] != -1) {
	      if ((distArray1[i] < correctDistArray0[posArray1[i]]) && ((distArray1[i] < distThreshold) || (thresholdOFF))) {
		/*	      correctArray1[posArray0[i]] = i;*/
		correctArray0[posArray1[i]] = i;
		correctArray1[i] = posArray1[i];
		correctDistArray0[posArray1[i]] = distArray1[i];
		correctDistArray1[i] = distArray1[i];
		/*	      distArray1[posArray0[i]] = distArray0[i];*/
	      }
	    }
	    }
	  }else {
	    for(i=query_start;i<=query_end;i++) {
	      if(distArray0[i] < correctDistArray1[posArray0[i]] && ((distArray0[i] < distThreshold) || thresholdOFF)) {
		correctArray0[i] = posArray0[i];
		correctArray1[posArray0[i]] = i;
		correctDistArray0[i] = distArray1[posArray0[i]];
		correctDistArray1[posArray0[i]] = distArray1[posArray0[i]];
	      }
	    }
	  }
	}else {
	  /*if not aligned, always match target SSE*/
	  target_start = target->ssArray[SSEnum*2];
	  target_end = target->ssArray[SSEnum*2+1]-1;
	  for(i=target_start;i<=target_end;i++) {
	    if(posArray1[i] != -1) {
	    if ((distArray1[i] < correctDistArray0[posArray1[i]]) && ((distArray1[i] < distThreshold) || (thresholdOFF))) {
	      /*	      correctArray1[posArray0[i]] = i;*/
	      correctArray0[posArray1[i]] = i;
	      correctArray1[i] = posArray1[i];
	      correctDistArray0[posArray1[i]] = distArray1[i];
	      correctDistArray1[i] = distArray1[i];
	      /*	      distArray1[posArray0[i]] = distArray0[i];*/
	    }
	    }
	  }
	}
      }
    
      /*orig:      for(i=0;i<query->ssArrayLength;i++) {*/
      for(i=0;i<target->ssArrayLength;i++) {
	target_start = target->ssArray[i*2];
	target_end = target->ssArray[i*2+1]-1;
	if(target->ss_alignment[i][0] !=-1) {
	  query_start = query->ssArray[2*target->ss_alignment[i][0]];
	  query_end = query->ssArray[2*target->ss_alignment[i][0]+1]-1;
	  order = 0;
	  if(target_end - target_start == query_end-query_start) {
	    /*order = strcmp(getFileName(0),getFileName(1));*/
	    /*		order = strcmp(query->objSeq, target->objSeq);*/
	    if(!switched) {
	      order = breakSSELengthTie(query,target);
	    }else {
	      order =breakSSELengthTie(target,query);
	    }
	    if(switched)
	      order *= -1;
	  }
	  if(order < 0 || (target_end - target_start > query_end - query_start)) {
	  for(j=target_start;j<=target_end;j++) {
	    tempArray[j-target_start] = correctArray1[j];
	  }
	    findBestSequence(correctArray1,correctDistArray1,target_start,target_end+1);
	    for(j=target_start;j<=target_end;j++) {
	      if(correctArray1[j]!=-1) {
		correctArray0[correctArray1[j]] = j;
	      }
	    }
	    /****************
	    for(j=query_start;j<=query_end;j++) {
	      if(correctArray0[j]!=-1) {
		if(correctArray1[correctArray0[j]]!=j) {
		  correctArray0[j] = -1;
		}
	      }
	    }
	    ******************/
	    for(j=target_start;j<=target_end;j++) {
	      if(correctArray1[j] != tempArray[j-target_start]) {
		correctArray0[tempArray[j-target_start]] = correctArray1[j];
	      }
	    }
	  }else {
	    for(j=query_start;j<=query_end;j++) {
	      tempArray[j-query_start] = correctArray0[j];
	    }
	    findBestSequence(correctArray0,correctDistArray0,query_start,query_end+1);
	    for(j=query_start;j<=query_end;j++) {
	      if(correctArray0[j]!=-1) {
		correctArray1[correctArray0[j]] = j;
	      }
	    }
	    /*****************
	    for(j=target_start;j<=target_end;j++) {
	      if(correctArray1[j]!=-1) {
		if(correctArray0[correctArray1[j]]!=j) {
		  correctArray1[j] = -1;
		}
	      }
	    }
	    ****************/
	    for(j=query_start;j<=query_end;j++) {
	      if(correctArray0[j] != tempArray[j-query_start]) {
		correctArray1[tempArray[j-query_start]] = correctArray0[j];
	      }
	    }
	    
	  }
	}else {
	  for(j=target_start;j<=target_end;j++) {
	    tempArray[j-target_start] = correctArray1[j];
	  }
	  findBestSequence(correctArray1,correctDistArray1,target_start,target_end+1);
	  for(j=target_start;j<=target_end;j++) {
	    if(correctArray1[j]!=-1) {
	      correctArray0[correctArray1[j]] = j;
	    }
	  }
	  /*******************
	  for(j=query_start;j<=query_end;j++) {
	    if(correctArray0[j]!=-1) {
	      if(correctArray1[correctArray0[j]]!=j) {
		correctArray0[j] = -1;
	      }
	    }
	  }
	  ******************/
	  for(j=target_start;j<=target_end;j++) {
	    if(correctArray1[j] != tempArray[j-target_start]) {
	      correctArray0[tempArray[j-target_start]] = correctArray1[j];
	    }
	  }
	}
      }

    

  /***
  for(i=0;i<len1;i++) {
    correctArray1[i] = -1;
    distArray1[i] = 1000.0;
  }
  for(i=0;i<len0;i++) {
    if(correctArray0[i]!=-1) {
      correctArray1[correctArray0[i]] = i;
      distArray1[correctArray0[i]] = correctDistArray0[i];
    }
  }
  ******/


      
      findBestSequence(correctArray1, distArray1, 0, len1);
  for(j=0;j<len0;j++) {
    correctArray0[j]=-1;
  }
  for(j=0;j<len1;j++) {
    if(correctArray1[j]!=-1) {
      correctArray0[correctArray1[j]]=j;
    }
  }


  /*      for (i = 0; i < len1; i++) {*/
      for(i=0;i<len1;i++) {
	if (correctArray1[i] != -1) {


	  /*	  if(!(query->objSeq[correctArray1[i]] == 'G' || target->objSeq[i] == 'G')) {
	    add_pair(tempObj0.points[correctArray1[i]], tempObj1.points[i], 1);
	  }else {
	    add_pair(tempObj0.points[correctArray1[i]], tempObj1.points[i],1);
	    }*/
	  add_pair(tempObj0.cbpoints[correctArray1[i]], tempObj1.cbpoints[i],1);
	  totalCount++;
	  myrmsd += (correctDistArray1[i])*(correctDistArray1[i]);
	}
      }

      rmsd = sqrt((myrmsd)/((GLfloat)(totalCount)));
      diffrmsd = oldrmsd - rmsd;
      
      if (diffrmsd < 0) 
	diffrmsd *= -1.0;
      
      if ((diffrmsd < epsilon) && (iter < iterations - 1))  {
	/*	iter = iterations - 1;*/
	converged = 1;
	break;
      }
      
      oldrmsd = rmsd;

    }
    else {

      printf("There are less than four core atoms correctly aligned\n");      
      totalCount = 0;
      rmsd = 0.0;
      /*      iter = iterations - 1;*/
      break;

    }

  }

  /*  if ((converged == 0) && (iterations == MAXITERATIONS)) {*/
  if(!converged) {
    printf("Refinealign did not converge\n");
  }

  if (totalCount > 0) {
    for (i = 0; i < len0; i++) {
      if (correctArray0[i] != -1) {
	query->pairNums[i] = correctArray0[i];
	query->pairDists[i] = correctDistArray0[i];


	if (target->pairNums[correctArray0[i]] != -1) {
	  if (correctDistArray0[i] < target->pairDists[correctArray0[i]]) {
	    target->pairNums[correctArray0[i]] = i;
	    target->pairDists[correctArray0[i]] = correctDistArray0[i];
	  }else {
	    /*take out pair: incorrect*/
	    query->pairNums[i] = -1;
	  }
	}
	else {
	  target->pairNums[correctArray0[i]] = i;
	  target->pairDists[correctArray0[i]] = correctDistArray0[i];
	}	  
      }
    }
  }



  globalCoreNumAligned = totalCount;
  
  globalLastAlignment = 1;
  
  return rmsd;

}


float integrateSSEAlignment(OBJ* query, OBJ* target){

  OBJ tempObj1, tempObj0;
  GLfloat tempVec[3];
  GLfloat tdist;
  int i, j, tnum;
  float rmsd = 0.0, oldrmsd = 1000.0, diffrmsd;
  int iter, converged = 0;

  Vector translation;
  Matrix mat;
  GLfloat m[16];
  
  float distArray1[MAXLENGTH];

  int correctArray1[MAXLENGTH];
  int len1;

  int posArray0[MAXLENGTH];
  float distArray0[MAXLENGTH];
  int numcorrect = 0;
  int len0;

  int totalCount = 0;
  float myrmsd = 0.0;

  for (i = 0; i < target->numPts; i++) {
    target->pairNums[i] = -1;
  }  

  for (i = 0; i < query->numPts; i++) {
    query->pairNums[i] = -1;
  }  
  
  tempObj1.numPts = target->numPts;
  tempObj0.numPts = query->numPts;

  tempObj1.sstype = target->sstype;
  tempObj0.sstype = query->sstype;
  
  
  len1 = target->numPts;
  len0 = query->numPts;

  init_pairs();
  
  numcorrect = 0;
  
  totalCount = 0;
  myrmsd = 0.0;
  
  
  for (i = 0; i < len1; i++) {    
    distArray1[i] = 1000.0;;
    correctArray1[i] = -1;
  }    
    
  for (i = 0; i < len0; i++) {    
    tdist = findNearest(&tempObj1, tempObj0.points[i], query->sstype[i], tempVec, &tnum);
    posArray0[i] = tnum;    
    distArray0[i] = tdist;
  }
              

  for (i = 0; i < len0; i++) {
    if ((distArray0[i] < distArray1[posArray0[i]]) && ((distArray0[i] < distThreshold) || (thresholdOFF))) {
      correctArray1[posArray0[i]] = i;
      distArray1[posArray0[i]] = distArray0[i];
    }
  }


  findBestSequence(correctArray1, distArray1, 0, len1);


  for (i = 0; i < len1; i++) {
    if (correctArray1[i] != -1) {
      /*      if(!(query->objSeq[correctArray1[i]] == 'G' || target->objSeq[i] == 'G')) {
	add_pair(tempObj0.points[correctArray1[i]], tempObj1.points[i], 1);
      }else {
	add_pair(tempObj0.cbpoints[correctArray1[i]], tempObj1.cbpoints[i],1);
	}*/
      add_pair(tempObj0.cbpoints[correctArray1[i]], tempObj1.cbpoints[i], 1);
      totalCount++;
      myrmsd += (distArray1[i])*(distArray1[i]);
    }
  }


  if (totalCount > 0) {
    for (i = 0; i < len1; i++) {
      if (correctArray1[i] != -1) {
	target->pairNums[i] = correctArray1[i];
	target->pairDists[i] = distArray1[i];


	if (query->pairNums[correctArray1[i]] != -1) {
	  if (distArray1[i] < query->pairDists[correctArray1[i]]) {
	    query->pairNums[correctArray1[i]] = i;
	    query->pairDists[correctArray1[i]] = distArray1[i];
	  }
	}
	else {
	  query->pairNums[correctArray1[i]] = i;
	  query->pairDists[correctArray1[i]] = distArray1[i];
	}	  
      }
    }
  }

  globalCoreNumAligned = totalCount;

  globalLastAlignment = 1;

  return rmsd;

}

void coreAlignSingle(OBJ* t0, OBJ* t1) {
  /*  OBJ *t0, *t1;*/
  float rmsd;
  int i;

  /*  t0 = ObjList[0];
  t1 = ObjList[1];*/
  
  for (i = 0; i < 5; i++) {
    rmsd = refineAlign(t0, t1, 1, 0.0, 1);
    
    globalCoreRMSD = rmsd;
    
    printf("Core RMSD = %f %d\n", globalCoreRMSD, globalCoreNumAligned);
    
#ifdef DISPLAYON
    if (globalStartDisplay) {
      display();
    }
#endif
  }

#ifdef DISPLAYON
  if (globalStartDisplay) {
#if DRAWJOINTS
    createJointList();
#endif
    display();
  }
#endif  
}


void coreAlignFull(OBJ* t0, OBJ* t1) {


  float rmsd;

 if(t0 == NULL) {
    t0 = ObjList[0];
  }
  if(t1 == NULL) {
    t1 = ObjList[1];
  }

  
  rmsd = refineAlign(t0, t1, MAXITERATIONS, MAXDIFFERENCE, 1);

  
  globalCoreRMSD = rmsd;
  
  printf("Core RMSD = %f %d\n", globalCoreRMSD, globalCoreNumAligned);

#ifdef DISPLAYON
  if (globalStartDisplay) {    
#if DRAWJOINTS
    createJointList();
#endif
    display();
  }
#endif
}





void createNextPairedResidueArray(OBJ* protein) {
  int i,prev,j;
  int next;
  int found_end;

  protein->nextPairedResidue = (int*)malloc(sizeof(int)*(globalCoreNumAligned+1));
  found_end=1;
  for(i=0;i<protein->numPts;i++) {
    if(protein->pairNums[i]!=-1) {
      next = i;
      found_end=0;
      break;
    }
  }
  
  for(i=0;i<=protein->numPts && !found_end;i++) {
    protein->nextPairedResidue[i] = next;
    found_end=1;
    for(j=next+1;j<protein->numPts;j++) {
      if(protein->pairNums[j]!=-1) {
	next=j;
	found_end=0;
	break;
      }
    }
  }
  protein->nextPairedResidue[globalCoreNumAligned] = -1;
  return;
  prev = -1;
  protein->nextPairedResidue = (int*)malloc(sizeof(int)*(protein->numPts));
  for(i=0;i<protein->numPts;i++) {
    protein->nextPairedResidue[i] = -1;
    if(protein->pairNums[i]!=-1) {
      if(prev!=-1) {
	protein->nextPairedResidue[prev] = i;
      }
      prev = i;
    }
  }
}



void checkLoopRegions(OBJ* query, OBJ* target, int query_alignment_start, int query_alignment_end) {

  int tnum,tnum2;
  int numaligned;
  int prev_query,last_query,prev_target,last_target;
  int prev_index;
  int i,j,done;
  GLfloat tdist;
  float rmsd;
  OBJ tempObj0, tempObj1;
  GLfloat tempVec[3];

  numaligned = globalCoreNumAligned;
  rmsd = globalCoreRMSD*globalCoreRMSD;
  rmsd*=globalCoreNumAligned;

  tempObj1.sstype = target->sstype;
  tempObj1.numPts = target->numPts;

  tempObj0.numPts = query->numPts;
  tempObj0.sstype = query->sstype;


  applyMat(&tempObj0, query, query->objMat);
  applyMat(&tempObj1,target,target->objMat);


  i=0;
  done=0;
  last_target=0;
  last_query=0;
  while(!done) {
    prev_target=last_target;
    last_target=target->nextPairedResidue[i];
    if(last_target == -1) {
      break;
    }
    prev_query = last_query;
    last_query = query->nextPairedResidue[i];
    if(last_query == -1) {
      break;
    }
    for(j=prev_target+1;j<last_target;j++) {
      tdist = findNearestUnpaired(&tempObj0, query->pairNums,tempObj1.cbpoints[j],tempObj1.points[j], prev_query+1,last_query-1,target->sstype[j], tempVec, &tnum,1);
      if((tdist<=distThreshold && tdist >= 0)||thresholdOFF) {
	tdist = findNearestUnpaired(&tempObj1, target->pairNums,tempObj0.cbpoints[tnum],tempObj0.points[tnum],prev_target,last_target,query->sstype[tnum], tempVec, &tnum2,1);
	if(tnum2 == j) {
	  query->pairNums[tnum]=j;
	  query->pairDists[tnum]=tdist;
	  target->pairNums[j]=tnum;
	  target->pairDists[j]=tdist;
	  rmsd += tdist*tdist;
	  numaligned++;
	}
      }
    }
    i++;
  }

  findBestSequence(target->pairNums,target->pairDists,0,target->numPts);
  for(i=0;i<query->numPts;i++) {
    query->pairNums[i]=-1;
  }
  numaligned=0;
  rmsd=0;
  for(i=0;i<target->numPts;i++) {
    if(target->pairNums[i]!=-1) {
      query->pairNums[target->pairNums[i]] = i;
      numaligned++;
      rmsd += target->pairDists[i]*target->pairDists[i];
    }
  }
  globalCoreNumAligned = numaligned;
  rmsd = sqrt(rmsd/((GLfloat)globalCoreNumAligned));
  globalCoreRMSD = rmsd;


  return;
  
  prev_target = -1;
  for(i=0;i<target->numPts;i++) {
    if(target->pairNums[i]!=-1) {
      prev_target = i;
      prev_index=i;
      break;
    }
  }
  if(prev_target == -1) {
    return; /*no aligned residues*/
  }
  /*do beginning of chains -- may be able to integrate this into loop*/
  /*******
  for(i=0;i<target->numPts;i++) {
    if(target->pairNums[i]!=-1) {
      last_target = i;
      break;
    }
  }

  for(i=0;i<query->numPts;i++) {
  if(query->pairNums[i]!=-1) {
      last_query = i;
      break;
    }
  }
  last_query--;
  last_target--;
  

 for(j=0;j<=last_target;j++) {
	if(target->sstype[j]!='U') {
	  continue;
	}
	tdist = findNearestUnpaired(&tempObj0, query->pairNums,tempObj1.points[j], 0,last_query,target->sstype[j], tempVec, &tnum);
	if((tdist<=distThreshold) || thresholdOFF) {
	  tdist = findNearestUnpaired(&tempObj1, target->pairNums,tempObj0.points[tnum],0,last_target,query->sstype[i], tempVec, &tnum2);
	  if(tnum2==j) {
	    query->pairNums[tnum] = j;
	    query->pairDists[tnum] = tdist;
	    target->pairNums[j] = tnum;
	    target->pairDists[j] = tdist;
	    rmsd += tdist*tdist;
	    globalCoreNumAligned++;
	  }
	}
 }

 prev_target=last_target+1;

 ******/

  done = 0;
  
  while(!done) {
    prev_target = prev_index+1;
    last_target=target->nextPairedResidue[prev_index]-1;
    /*    last_target = target->nextPairedResidue[prev_target]-1;*/
    if(last_target<0) {
      done = 1;
    }else {
      /*
      prev_query = target->pairNums[prev_target-1]+1;
            prev_target++;
      last_query = query->nextPairedResidue[prev_query-1]-1;
      */
      prev_query = target->pairNums[prev_target-1]+1;
      last_query = query->nextPairedResidue[prev_query-1]-1;
      for(j=prev_target;j<=last_target;j++) {
	if(target->sstype[j]!='U') {
	  continue;
	}
	tdist = findNearestUnpaired(&tempObj0, query->pairNums,tempObj1.cbpoints[j],tempObj1.points[j], prev_query,last_query,target->sstype[j], tempVec, &tnum,1);
	if((tdist<=distThreshold && tdist >= 0) || thresholdOFF) {
	  /*fixed next line 09/12/01: query->sstype[tnum] instead of sstype[i]*/
	  /*	  tdist = findNearestUnpaired(&tempObj1, target->pairNums,tempObj0.points[tnum],prev_target,last_target,query->sstype[i], tempVec, &tnum2);*/
	  tdist = findNearestUnpaired(&tempObj1, target->pairNums,tempObj0.cbpoints[tnum],tempObj0.points[tnum],prev_target,last_target,query->sstype[tnum], tempVec, &tnum2,1);
	  if(tnum2==j) {
		query->pairNums[tnum] = j;
		query->pairDists[tnum] = tdist;
		target->pairNums[j] = tnum;
		target->pairDists[j] = tdist;
		rmsd += tdist*tdist;
		globalCoreNumAligned++;
	  }
	}
      }
      prev_target=last_target+1;

      prev_index++;
      while(target->pairNums[prev_index]==-1 && prev_index < target->numPts) {
	prev_index++;
      }
    }
  }

  /*do end of chains -- can probably integrate into loop above*/
  /******
  for(i=query->numPts-1;i>=0;i--) {
    if(query->pairNums[i]!=-1) {
      prev_query = i-1;
      break;
    }
  }


  last_target = target->numPts-1;
  last_query = query->numPts-1;

  for(j=prev_target+1;j<=last_target;j++) {
	if(target->sstype[j]!='U') {
	  continue;
	}
	tdist = findNearestUnpaired(&tempObj0, query->pairNums,tempObj1.points[j], prev_query,last_query,target->sstype[j], tempVec, &tnum);
	if((tdist<=distThreshold) || thresholdOFF) {
	  tdist = findNearestUnpaired(&tempObj1, target->pairNums,tempObj0.points[tnum],prev_query,last_target,query->sstype[i], tempVec, &tnum2);
	  if(tnum2==j) {
	    query->pairNums[tnum] = j;
	    query->pairDists[tnum] = tdist;
	    target->pairNums[j] = tnum;
	    target->pairNums[j] = tdist;
	    rmsd += tdist*tdist;
	    globalCoreNumAligned++;
	  }
	}
	}
	******/


  rmsd= sqrt((rmsd)/((GLfloat)(globalCoreNumAligned)));
  globalCoreRMSD = rmsd;
}
  
void quitFunction() {

  exit(0);

}

void lockResidues(OBJ* query, OBJ* target) {
  int i,j;
  Matrix mat;
  Vector translation;
  GLfloat m[16];
  int num_unique_pairs=0;
  char query_residues[MAXLENGTH][5];
  char target_residues[MAXLENGTH][5];
  char temp[5];
  int num_query_residues_seen=0;
  int num_target_residues_seen=0;
  int unique_query_residues=0;
  int unique_target_residues=0;
  int unique;

  /*switch query and target?*/

  /*initialize transformation matrix*/
  for(i=0;i<16;i++) {
    query->objMat[i] = 0; target->objMat[i] = 0;
  }
  query->objMat[0] = query->objMat[5] = query->objMat[10] = 1;
  target->objMat[0] = target->objMat[5] = target->objMat[10] = 1;


  /*do the transformation*/
  init_pairs();
  for(i=0;i<query->num_lock_positions;i++) {
    if(query->lock_positions[i] != -1 && target->lock_positions[i] != -1) {
      printf("locking position %i to %i\n",query->lock_positions[i],target->lock_positions[i]);
      /*      unique = 1;
      sprintf(temp,"%i",query->lock_positions[i]);
      for(j=0;j<unique_query_residues;j++) {

	if(strcmp(query_residues[j],temp)==0) {
	  unique=0;
	  printf("NOT unique\n");
	  break;
	}
      }
      if(unique) {
	unique_query_residues+=1;
	strcpy(query_residues[num_query_residues_seen],temp);
      }
      num_query_residues_seen++;

      unique = 1;
      sprintf(temp,"%i",target->lock_positions[i]);
      for(j=0;j<unique_target_residues;j++) {
	if(strcmp(target_residues[j],temp)==0) {
	  unique=0;
	  printf("NOT unique\n");
	  break;
	}
      }
      if(unique) {
	unique_target_residues+=1;
	strcpy(target_residues[num_target_residues_seen],temp);
      }
      num_target_residues_seen++;
      */
      if(!(query->objSeq[query->lock_positions[i]] == 'G' || target->objSeq[target->lock_positions[i]]=='G')) {
	add_pair(query->cbpoints[query->lock_positions[i]],target->cbpoints[target->lock_positions[i]],1);
      }else {
	add_pair(query->points[query->lock_positions[i]],target->points[target->lock_positions[i]],1);
      }
    }
  }


  /*  if(okayToMatchPoints()) {*/
  /*
  if(unique_query_residues>1 && unique_target_residues>1) {
    match_pairs(mat,translation);
  }else {
    printOutput(query,target,0,-1);
    exit(0);
  }
  */
  if(okayToMatchPoints()) {
    match_pairs(mat,translation);
  }else {
    printOutput(query,target,0,-1);
    exit(0);
  }
  
    /*save the transformation matrix*/
    for(i=0;i<4;i++) {
      for(j=0;j<4;j++) {
	/*      query->objMat[i*4+j] = mat[i][j];*/
	m[i*4+j] = mat[i][j];
      }
    }
    myMatMult(target->objMat,m);
    applyMtoV(target->center,target->center,m);
    /*  }*/
}


int getResiduePosition(OBJ* obj, int index) {
  int firstquarter,lastquarter;
  int i;
  int start,end,length;
  float position=1;

  for(i=0;i<obj->ssArrayLength;i++) {
    if(index >=obj->ssArray[i*2] && index <= obj->ssArray[i*2+1]-1) {
      start = obj->ssArray[i*2]; end = obj->ssArray[i*2+1]-1;
      length = end-start+1;
      firstquarter = start + length/4;
      lastquarter = end-length/4;
      while(lastquarter-firstquarter+1 < length/2) {
	if(lastquarter%2==0) lastquarter++;
	else firstquarter--;
      }
      if(index <= firstquarter) {
	return(2);
      }else if(index < lastquarter) {
	return(3);
      }else {
	return(2);
      }
    }
  }
  return(1); /*loop*/
}

float gapScore(OBJ* obj, int index,int SSE_open, int gap_open) {
  int firstquarter,lastquarter;
  /*  float multiplier = 4.0;*/
  float multiplier = gapMultiplier;
  float penalty1 = -1;
  int i;
  int start,end,length;
  float position, q_position, t_position;
  int prev_query, prev_target;

  if(gap_open) {
    multiplier = gapExtensionMultiplier;
  }

  position = getResiduePosition(obj,index);
  if(position!=1) { /*if not loop residue*/
    if(SSE_open) {
      return (3*multiplier);
    }else {
      return(multiplier);
      /*      return(2*multiplier);*/
    }
  }else{ /*if loop residue*/
    if(SSE_open) {
      return(3*multiplier);
    }else {
      return(multiplier);
    }
  }


}



float alignmentScore(OBJ* query, OBJ* target, int qpoint,  int tpoint, int ***pointarray){
  /*  float dist_weight = 5.0;
  float angle_weight = 15.0;
  */
  float dist_weight = distanceWeight;
  float angle_weight = angleWeight;
  float w1,w2=0;
  int querySSE,targetSSE;
  float tdist,dist_score;
  float dotproduct;
  float angle,anglescore;
  float angle_cutoff = PI/2.0;
  int i,j;
  float u[3], v[3];
  float len1,len2;
  float epsilon = 0.0001; /*for floating point equality*/
  float penalty = 0;
  float weight;
  float angle_diff;
  int qpoint2, tpoint2;
  int error;
  float alpha1,alpha2;
  float distance_cutoff;
  float score;

  if((query->sstype[qpoint]=='U' && target->sstype[tpoint]!='U') || (target->sstype[tpoint]=='U' && query->sstype[qpoint]!='U')) {
    /*    angle_cutoff = PI/4.0;*/
    angle_cutoff = PI/2.0;
    distance_cutoff = distThreshold;
    /*    distance_cutoff = distThreshold*1.5;*/
  }else {
    angle_cutoff = PI/2.0;
    distance_cutoff=distThreshold*1.5;
  }

  tdist = distanceBetweenPoints(query->cbpoints[qpoint],target->cbpoints[tpoint]);
  if(!((query->sstype[qpoint]==target->sstype[tpoint]) || !((query->sstype[qpoint]=='H' && target->sstype[tpoint]=='S') || (query->sstype[qpoint]=='S' && target->sstype[tpoint]=='H')))) {
    return(-10000);
  }

  if(tdist>distance_cutoff) {
    return -10000;
    }

  
  len1=len2=0;
  dotproduct=0;
  for(j=0;j<3;j++) {
    u[j] = query->cbpoints[qpoint][j]-query->points[qpoint][j];
    len1 += u[j]*u[j];
    v[j] = target->cbpoints[tpoint][j]-target->points[tpoint][j];
    len2 += v[j]*v[j];
  }
  len1 = sqrt(len1); len2 = sqrt(len2); 
  for(j=0;j<3;j++) {
    u[j] /= len1;
    v[j] /= len2;
    dotproduct += u[j]*v[j];
  }


  if(dotproduct<= 1+epsilon && dotproduct >= 1-epsilon) {
    angle=0;
  }else if(dotproduct<=-1+epsilon && dotproduct>=-1+epsilon) {
    angle=2*PI;
  }else {
    angle = acos((double)dotproduct);
  }



  if(angle <= 0.7854) {
    dist_score = (10*2)/(1+(tdist*tdist/(distThreshold*distThreshold)))-10;
    anglescore = (10*2)/(1+angle*angle/(angle_cutoff*angle_cutoff))-10;
  }else {
    dist_score = (5*2)/(1+(tdist*tdist/(distThreshold*distThreshold)))-5;
    anglescore = (15*2)/(1+angle*angle/(angle_cutoff*angle_cutoff))-15;
  }


   score = dist_score + anglescore;
   score = round(score*1000)/1000.0;
   return(score);
}

void previousQueryResidue(int ***pointarray, int istart, int jstart, int *qpoint) {
  int i,j,nexti,nextj;
  *qpoint=-1;
  i=pointarray[istart][jstart][0]; j=pointarray[istart][jstart][1];
  while(i!=-1 && j!=-1) {
    nexti=pointarray[i][j][0]; nextj=pointarray[i][j][1];
    if(nextj==j-1) {
      *qpoint = j-1;
      return;
    }
    i=nexti; j=nextj;
  }
}

void previousTargetResidue(int ***pointarray,int istart, int jstart, int *tpoint) {
  int i,j,nexti,nextj;
  *tpoint=-1;
  i=pointarray[istart][jstart][0]; j=pointarray[istart][jstart][1];
  while(i!=-1 && j!=-1) {
    nexti=pointarray[i][j][0]; nextj=pointarray[i][j][1];
    if(nexti==i-1) {
      *tpoint = i-1;
      return;
    }
    i=nexti; j=nextj;
  }
}

void partialTraceback(int ***pointarray, int istart, int jstart, int *tpoint, int *qpoint) {
  int i,j,nexti,nextj;

  *tpoint=-1; *qpoint=-1;
  i=pointarray[istart][jstart][0]; j=pointarray[istart][jstart][1];
  while(i!=-1 && j!=-1) {
    nexti=pointarray[i][j][0]; nextj=pointarray[i][j][1];
    if(nexti==i-1 && nextj==j-1) {
      *qpoint = j-1; *tpoint = i-1;
      return;
    }
    i=nexti; j=nextj;
  }
}

void runDynamicProgramming(OBJ* query, OBJ* target, float **scorearray, int ***pointarray) {
  int i,j,k,iters;
  int nexti,nextj;
  int maxpointer[2];
  int index;
  float myscore, maxscore,max;
  int mypointer,maxindex;
  float gapscore,score;
  float gapweight;
  float rmsd,oldrmsd,diffrmsd,epsilon=MAXDIFFERENCE;
  int count=0;
  float gap_extension_multiplier = gapExtensionMultiplier;
  /*  float gap_extension_multiplier=0.3;
  float sse_end_extension_multiplier=0.7;
  float sse_middle_extension_multiplier=1.0;
  float loop_extension_multiplier=0.1;*/
  int position;
  int query_SSE_open=0;
  int target_SSE_open=0;
  int SSE_open = 0;
  int prev_query,prev_target;

  rmsd = 0.0;
  oldrmsd = 1000;

  

  maxscore=-100; maxindex=-1;

  /*initial conditions*/
  scorearray[0][0] = 0;
  pointarray[0][0][0] = -1; pointarray[0][0][1] = -1;


  /*gap penalty for first unaligned residue before aligned region*/
  if(target->sstype[0]=='U') {
    scorearray[1][0] = scorearray[0][0] - gapScore(target,0,0,0);
  }else {
    scorearray[1][0] = scorearray[0][0] - gapScore(target,0,0,0);
  }
  pointarray[1][0][0] = 0; pointarray[1][0][1] = 0;

  /*gap penalty for subsequent unaligned residue before aligned region*/
  
  for(i=2;i<=target->numPts;i++) {
    gapscore = round((gapScore(target,i-1,0,1))*1000)/1000.0;
    scorearray[i][0] = scorearray[i-1][0]-gapscore;
    pointarray[i][0][0] = i-1; pointarray[i][0][1]=0;
  }

  /*gap penalty for first unaligned residue before aligned region*/
  scorearray[0][1] = scorearray[0][0] - gapScore(query,0,0,0);

  pointarray[0][1][0]=0; pointarray[0][1][1]=0;
  /*gap penalty for subsequent unaligned residue before aligned region*/  
  for(j=2;j<=query->numPts;j++) {
    gapscore = round((gapScore(query,j-1,0,1))*1000)/1000.0;
    scorearray[0][j] = scorearray[0][j-1]-gapscore;
    pointarray[0][j][0] = 0;      pointarray[0][j][1] = j-1;
  }

  for(i=1;i<=target->numPts;i++) {
    for(j=1;j<=query->numPts;j++) {

      /*do up arrow*/

      /*get last query and target residues to have been written down in alignment (whether aligned or not)*/
      previousQueryResidue(pointarray,i-1,j, &prev_query);


      /*given prev_query, see if an SSE is open in the query*/
      query_SSE_open=0;
      if(prev_query != -1 && query->objSeq[prev_query]!='U') {
	/*prev_query is an SSE residue, so see if it was the end of the SSE*/
	for(k=0;k<query->ssArrayLength;k++) {
	  if(prev_query >= query->ssArray[k*2] && prev_query <= query->ssArray[k*2+1]-1) {
	    if(prev_query != query->ssArray[k*2]) {
	      query_SSE_open=1;
	      break;
	    }
	    break;
	  }
	}
      }
      previousTargetResidue(pointarray,i-1,j, &prev_target);
      /*given prev_target, see if an SSE is open in the target*/
      target_SSE_open=0;
      if(prev_target!=-1 && target->objSeq[prev_target]!='U') {
	for(k=0;k<target->ssArrayLength;k++) {
	  if(prev_target>=target->ssArray[k*2] && prev_target <= target->ssArray[k*2+1]-1) {
	    if(prev_target != target->ssArray[k*2]) {
	      target_SSE_open=1;
	      break;
	    }
	    break;
	  }
	}
      }
      SSE_open = 0;
      if(query_SSE_open || target_SSE_open) SSE_open = 1;
      /*continuing with up arrow score . . .*/


      if(pointarray[i-1][j][0]==i-2 && pointarray[i-1][j][1]==j) { /*if gap is already open*/
	gapscore = gapScore(target,i-1,query_SSE_open,1);
      }else {
	gapscore = gapScore(target,i-1,query_SSE_open,0);	
      }
      gapscore = round(gapscore*1000)/1000.0;
      max = scorearray[i-1][j] - gapscore;
      pointarray[i][j][0] = i-1; pointarray[i][j][1]=j;


      /*do left arrow*/

      /*continuing with left arrow . . .*/

      if(pointarray[i][j-1][0]==i && pointarray[i][j-1][1]==j-2) {
	gapscore = gapScore(query,j-1,target_SSE_open,1);
      }else {
	gapscore = gapScore(query,j-1,target_SSE_open,0);
      }

      gapscore = round(gapscore*1000)/1000.0;
      if(max < scorearray[i][j-1]-gapscore) {
	max = scorearray[i][j-1]-gapscore;
	pointarray[i][j][0] = i; pointarray[i][j][1] = j-1;
      }
      score = alignmentScore(query,target,j-1,i-1,pointarray);
      if(max < scorearray[i-1][j-1]+ score) {
	max = scorearray[i-1][j-1]+score;
	pointarray[i][j][0] = i-1; pointarray[i][j][1] = j-1;
      }
      scorearray[i][j] = max;

    }
  }

}

float alignResidues(OBJ* query, OBJ* target) {
  OBJ *temp;
  int switched=0;
  int i,j,iter;
  int nexti,nextj;
  int maxpointer[2];
  int index;
  int weight;
  float myscore, maxscore,max;
  int mypointer,maxindex;
  float **scorearray;
  int ***pointarray;
  OBJ tempObj1, tempObj0;
  float rmsd,oldrmsd,diffrmsd,epsilon=0.05;
  int count=0;
  int converged;
  Matrix mat;
  float m[16];
  Vector translation;
  int iterations =  MAXITERATIONS;
  float tdist;

  printf("multipliers: %f %f %f\n",gapMultiplier,gapExtensionMultiplier,distThreshold);
  oldrmsd = 1000.0;

   if(target->ssArrayLength < query->ssArrayLength) {
    temp = query;
    query = target;
    target = temp;
    switched = 1;

  }else if((target->ssArrayLength==query->ssArrayLength) && strcmp(query->objSeq,target->objSeq) > 0) {
    temp = query;
    query = target;
    target = temp;
    switched = 1;

   }

  

  memset(query->pairNums,-1,query->numPts*sizeof(int));
  memset(target->pairNums,-1,target->numPts*sizeof(int));

  scorearray = (float **)malloc(sizeof(float *)*(target->numPts+1));
  pointarray = (int ***)malloc(sizeof(int **)*(target->numPts+1));
  for(i=0;i<=target->numPts;i++) {
    scorearray[i] = (float *)malloc(sizeof(float)*(query->numPts+1));
    pointarray[i] = (int **)malloc(sizeof(int *)*(query->numPts+1));
    for(j=0;j<=query->numPts;j++) {
      pointarray[i][j] = (int *)malloc(sizeof(int)*2);
    }
  }



  memcpy(tempObj1.aaNums,target->aaNums,target->numPts*sizeof(int));
  memcpy(tempObj0.aaNums,query->aaNums,query->numPts*sizeof(int));


  tempObj1.numPts = target->numPts;
  tempObj0.numPts = query->numPts;

  tempObj1.sstype = target->sstype;
  tempObj0.sstype = query->sstype;
  
  tempObj1.objSeq = target->objSeq;
  tempObj0.objSeq = query->objSeq;

  tempObj0.ssArrayLength = query->ssArrayLength;
  tempObj1.ssArrayLength = target->ssArrayLength;
  tempObj0.ssArray = query->ssArray;
  tempObj1.ssArray = target->ssArray;

  applyMat(&tempObj0, query, query->objMat);
  applyMat(&tempObj1, target, target->objMat);

  /*  runDynamicProgramming(&tempObj0,&tempObj1,scorearray,pointarray);*/

  for(iter=0;iter<iterations;iter++) {

    runDynamicProgramming(&tempObj0,&tempObj1,scorearray,pointarray);
    memset(query->pairNums,-1,query->numPts*sizeof(int));
    memset(target->pairNums,-1,target->numPts*sizeof(int));
    init_pairs();
    i = target->numPts;
    j = query->numPts;

    while(i!=-1 && j!=-1) {
      nexti = pointarray[i][j][0]; nextj = pointarray[i][j][1];
      if(nexti==i-1 && nextj==j-1 && i-1>=0 && j-1>=0) {
	query->pairNums[j-1] = i-1; target->pairNums[i-1] = j-1; /*check this*/
	query->pairDists[j-1] = distanceBetweenPoints(tempObj0.cbpoints[j-1],tempObj1.cbpoints[query->pairNums[j-1]]);
	target->pairDists[i-1] = query->pairDists[j-1];
	query->alignmentScore[j-1] = alignmentScore(&tempObj0,&tempObj1,j-1,i-1,pointarray);
	target->alignmentScore[i-1] = query->alignmentScore[j-1];
	query->gapScore[j-1] = -1; /*haven't calculated this correctly yet*/
	target->gapScore[i-1] = -1;
	weight=1;
	/*	weight = getResiduePosition(query,j-1);*/

	add_pair(tempObj0.cbpoints[j-1],tempObj1.cbpoints[i-1],weight);
	add_pair(tempObj0.points[j-1],tempObj1.points[i-1],weight);
      }
      i=nexti; j=nextj;
    }


    if(okayToMatchPoints()) {
      match_pairs(mat,translation);
      for(i=0;i<4;i++) {
	for(j=0;j<4;j++) {
	  m[i*4+j] = mat[i][j];
	}
      }
      myMatMult(target->objMat,m);
      applyMtoV(target->center,target->center,m);
      applyMat(&tempObj0, query, query->objMat);
      applyMat(&tempObj1, target, target->objMat);
    }

    rmsd = 0;
    count=0;
    for(i=0;i<query->numPts;i++) {
      if(query->pairNums[i]!=-1) {
	query->pairDists[i] = distanceBetweenPoints(tempObj0.cbpoints[i],tempObj1.cbpoints[query->pairNums[i]]);
	target->pairDists[query->pairNums[i]] = query->pairDists[i];
	count++;
	rmsd += query->pairDists[i]*query->pairDists[i];
      }
    }
    rmsd = sqrt(rmsd/((float)count));
    printf("alignresidues: iter=%i, rmsd=%.3f, numaligned=%i\n",iter,rmsd,count);
    diffrmsd = oldrmsd - rmsd;
    if(diffrmsd < 0) {
      diffrmsd *= -1.0;
    }
    if(diffrmsd < epsilon && (iter < iterations-1)) {
      converged = 1;
      break;
    }
    oldrmsd=rmsd;
  }


  globalCoreRMSD = rmsd;
  globalCoreNumAligned = count;
  if(!converged) {
    printf("alignResidues did not converge\n");
  }


  for(i=0;i<=target->numPts;i++) {
    for(j=0;j<=query->numPts;j++) {
      free(pointarray[i][j]);
    }
    free(pointarray[i]);
    free(scorearray[i]);
  }
  free(pointarray);
  free(scorearray);

  globalLastAlignment = 1;
}


void doAllAlignments() {
  
  int oldt,num_aligned,index,order, first_alignment=1;
  OBJ tempObj0, tempObj1, *tempobj;
  int switched = 0;
  FILE *searchfile;
  FILE *points1f, *points2f;
  FILE *matrixfile;
  char myprot[50];
  float tempcentervector[3];
  float tempobjectmatrix[16];
  int i,k, done,j,temp,aligned,count;
  float rmsd, finalscore,maxscore;
  int wordlen;
  char line[200];
  char tempstr[300];
  int query_start,query_end,target_start,target_end;
  int SSEnum, query_alignment_start,query_alignment_end,target_alignment_start,target_alignment_end,SSEend,SSEstart,found_start,found_end;
  float percent_aligned;
  float overall_mu = 0.076;
  float overall_lambda = 24.33;
  float pvalue;


  if (strcmp(outputDir, "NONE") == 0) {
      searchfile = fopen("search.out", "w");
  }
  else {
    strcpy(tempstr, outputDir);
    strcat(tempstr, "/search.out");
    searchfile = fopen(tempstr, "w");
    if (searchfile == NULL) {
      printf("Trying to create directory %s\n", outputDir);
      sprintf(line, "mkdir %s", outputDir);
      system(line);
      searchfile = fopen(tempstr, "w");
    }
  }
  
  if (searchfile == NULL) {
    printf("Sorry.  Could not open output file search.out\n");
  }

  else {
    /*    start_time = time(&start_time);*/
    start_time = clock();
    /*Initialize query_points_to_save and target_points_to_save
    for(i=0;i<ObjList[0]->numPts;i++) {
      query_points_to_save[i][0] = NULLVALUE;
    }
    for(i=0;i<ObjList[1]->numPts;i++) {
      target_points_to_save[i][0] = NULLVALUE;
    }
    */

    if (globalSearchDatabase == 0) {
      printf("lock residues=%i, runstep1=%i\n",lock_residues,runStep1);
      if(!lock_residues) {
	if(runStep1) {
	  printf("runstep1=1\n");
	  /*step 1*/
	  if ((ObjList[0]->ssArrayLength > 0) && (ObjList[1]->ssArrayLength > 0)) {
	    initAlignUsingVectorsSingle(ObjList[0],ObjList[1],0);
	    alignUsingVectorsFull(ObjList[0],ObjList[1],0);
	  
	  }else {
	    printf("num aligned before checking for loops: %i\n",0);      	
	    num_aligned = 0;
	    for(i=0;i<ObjList[0]->numPts;i++) {
	      if(ObjList[0]->pairNums[i]!=-1 && ObjList[0]->sstype[i]!='U') {
		num_aligned++;
	      }
	    }
	    printf("Query num SSE residues aligned before loops: %i\n",num_aligned);
	    num_aligned = 0;
	    for(i=0;i<ObjList[1]->numPts;i++) {
	      if(ObjList[1]->pairNums[i]!=-1 && ObjList[1]->sstype[i]!='U') {
		num_aligned++;
	      }
	    }
	    printf("Target num SSE residues aligned before loops: %i\n",num_aligned);
	  }
      
	  if (globalNumAligned == 0) {
	    oldt = thresholdOFF;
	    thresholdOFF = 1;
	    /*structAlignSingle(ObjList[0],ObjList[1]);*/
	    thresholdOFF = oldt;
	  }
	




	  if(ObjList[0]->num_SSEs_aligned < 1) {
	    globalNumAligned = 0;
	    printOutput(ObjList[0],ObjList[1],0,-1);
	    exit(0);
	  }
	}else {
	  printf("initializing transformation\n");
	  init_pairs(); /*initialize transformation*/
	  for(i=0;i<3;i++) {
	    for(j=0;j<3;j++) {
	      ObjList[1]->objMat[i*4+j] = rotation[i][j];
	    }
	    ObjList[1]->objMat[i*4+4]=0.0;
	  }
	  for(i=0;i<3;i++) {
	    ObjList[1]->objMat[12+i] = translation[i];
	  }
	  ObjList[1]->objMat[15] = 1.0;
	  /*	printOutput(ObjList[0],ObjList[1],0,ObjList[0]->numPts-1);
	    exit(0);*/
	}
      }else{
	lockResidues(ObjList[0],ObjList[1]);
      }
    

      /*      structAlignFull(ObjList[0],ObjList[1]);*/

      
      /*take if statment out to test step2*/
 

      if (globalComputeCore) {
	oldt = distThreshold;
	/*	coreAlignFull(ObjList[0],ObjList[1]);
	structAlignFull(ObjList[0],ObjList[1]);
	coreAlignFull(ObjList[0],ObjList[1]);*/

	alignResidues(ObjList[0],ObjList[1]);
	globalLastAlignment=1;
      }



      printf("\nAligned Secondary Structure Elements after step 1: %i %i\n",ObjList[0]->num_SSEs_aligned,ObjList[0]->num_repeated_SSE_alignments+ObjList[1]->num_repeated_SSE_alignments);
      if(ObjList[0]->num_SSEs_aligned!=0) {
	for(i=0;i<ObjList[0]->ssArrayLength;i++) {
	  for(j=0;j<ObjList[1]->ssArrayLength;j++) {
	    if(ObjList[0]->ss_alignment[i][j]!=-1) {
	      query_start = ObjList[0]->ssArray[i*2];
	      query_end = ObjList[0]->ssArray[i*2+1]-1;
	      target_start = ObjList[1]->ssArray[ObjList[0]->ss_alignment[i][j]*2];
	      target_end = ObjList[1]->ssArray[ObjList[0]->ss_alignment[i][j]*2+1]-1;
	      printf("%c,%c\t%i%c%c,%i%c%c\t%i%c%c,%i%c%c\n",ObjList[0]->sstype[query_start],ObjList[1]->sstype[target_start],ObjList[0]->aaNums[query_start],ObjList[0]->insertionCodes[query_start],ObjList[0]->chains[query_start],ObjList[0]->aaNums[query_end],ObjList[0]->insertionCodes[query_end],ObjList[0]->chains[query_end],ObjList[1]->aaNums[target_start],ObjList[1]->insertionCodes[target_start],ObjList[1]->chains[target_start],ObjList[1]->aaNums[target_end],ObjList[1]->insertionCodes[target_end],ObjList[1]->chains[target_end]);
	    }
	  }
	}
      }
    
      if(ObjList[0]->num_SSEs_aligned!=0) {
	for(i=0;i<ObjList[0]->ssArrayLength;i++) {
	  for(k=0;k<ObjList[1]->ssArrayLength;k++) {
	    if(ObjList[0]->ss_alignment[i][k]!=-1) {
	      /*	      check to see if any points are still aligned*/
	      target_start =  ObjList[1]->ssArray[ObjList[0]->ss_alignment[i][k]*2];
	      target_end = ObjList[1]->ssArray[ObjList[0]->ss_alignment[i][k]*2+1]-1;
	      query_start = ObjList[0]->ssArray[i*2];
	      query_end = ObjList[0]->ssArray[i*2+1]-1;
	      aligned = 0;
	      for(j=ObjList[0]->ssArray[i*2];j<ObjList[0]->ssArray[i*2+1];j++) {
		if(ObjList[0]->pairNums[j] != -1) {
		  if(ObjList[0]->pairNums[j] >=target_start && ObjList[0]->pairNums[j]<=target_end) {
		    aligned++;
		  }
		}
	      }
	      if(query_end-query_start > target_end-target_start) {
		percent_aligned = ((float)(aligned))/((float)(target_end-target_start+1));
	      }else {
		percent_aligned = ((float)(aligned))/((float)(query_end-query_start+1));
	      }
	      
	      if(aligned<1) {
		printf("not aligned: query = %i, target = %i\n",i,ObjList[0]->ss_alignment[i][k]);
		for(j=0;j<ObjList[1]->ssArrayLength;j++) {
		  if(ObjList[1]->ss_alignment[ObjList[0]->ss_alignment[i][k]][j] == i) {
		    ObjList[1]->ss_alignment[ObjList[0]->ss_alignment[i][k]][j] = -1;
		    index = j;
		    break;
		  }
		}
		if(ObjList[1]->num_aligned[ObjList[0]->ss_alignment[i][k]]>1) {
		  ObjList[1]->num_repeated_SSE_alignments--;
		}
		ObjList[1]->num_aligned[ObjList[0]->ss_alignment[i][k]]--;
		ObjList[0]->ss_alignment[i][k] = -1;
		ObjList[0]->num_SSEs_aligned--;
		ObjList[1]->num_SSEs_aligned--;
		if(ObjList[0]->num_aligned[i]>1) {
		  ObjList[0]->num_repeated_SSE_alignments--;
		}
		ObjList[0]->num_aligned[i]--;
		  
		}
	      }
	    }
	  }
      }

      if(ObjList[0]->num_SSEs_aligned == 0) {
	printf("num aligned before checking for loops: %i\n",globalCoreNumAligned);      
	num_aligned = 0;
	for(i=0;i<ObjList[0]->numPts;i++) {
	  if(ObjList[0]->pairNums[i]!=-1 && ObjList[0]->sstype[i]!='U') {
	    num_aligned++;
	  }
	}
	printf("Query num SSE residues aligned before loops: %i\n",num_aligned);
	num_aligned = 0;
	for(i=0;i<ObjList[1]->numPts;i++) {
	  if(ObjList[1]->pairNums[i]!=-1 && ObjList[1]->sstype[i]!='U') {
	    num_aligned++;
	  }
	}
	printf("Target num SSE residues aligned before loops: %i\n",num_aligned);
	globalNumAligned = 0;
      }	  
      printf("num aligned before checking for loops: %i\n",globalCoreNumAligned);      
      num_aligned = 0;
      for(i=0;i<ObjList[1]->numPts;i++) {
	if(ObjList[1]->pairNums[i]!=-1 && ObjList[1]->sstype[i]!='U' && 
	   ObjList[0]->sstype[ObjList[1]->pairNums[i]]!='U') {
	  num_aligned++;
	}
      }
      printf("num SSE residues aligned before loops: %i\n",num_aligned);


      
  
      tempObj0.objSeq = ObjList[0]->objSeq; /*Array*/
      tempObj0.sstype = ObjList[0]->sstype; /*Array of SS types*/
      tempObj0.numPts = ObjList[0]->numPts; /*Array of points*/
      tempObj0.ssArrayLength = ObjList[0]->ssArrayLength;
      tempObj0.ssArray = ObjList[0]->ssArray; /*Array of #'s giving beginning and
					       *ends of ss elements*/
      tempObj0.num_SSEs_aligned = ObjList[0]->num_SSEs_aligned;
      tempObj0.ss_alignment = ObjList[0]->ss_alignment;
      applyMat(&tempObj0, ObjList[0], ObjList[0]->objMat);
      
      tempObj1.objSeq = ObjList[1]->objSeq;
      tempObj1.sstype = ObjList[1]->sstype;
      tempObj1.numPts = ObjList[1]->numPts;
      tempObj1.ssArrayLength = ObjList[1]->ssArrayLength;
      tempObj1.ssArray = ObjList[1]->ssArray;
      tempObj1.num_SSEs_aligned = ObjList[1]->num_SSEs_aligned;
      tempObj1.ss_alignment = ObjList[1]->ss_alignment;
      applyMat(&tempObj1, ObjList[1], ObjList[1]->objMat);
      if(!lock_residues) {
	initsequences_ss(&tempObj0,&tempObj1);
	finalscore = align_ss(1,ObjList[0],ObjList[1]);
      }

      if(ObjList[0]->ssArrayLength > ObjList[1]->ssArrayLength) {
	maxscore = 40*ObjList[0]->ssArrayLength-20.0;
      }else {
	maxscore = 40*ObjList[1]->ssArrayLength -20.0;
      }
      printOutput(ObjList[0],ObjList[1],0,ObjList[0]->numPts-1);

      pvalue = 1-exp(-exp(-overall_lambda*(finalscore/maxscore-overall_mu)));
      if (globalComputeCore) {
	fprintf(searchfile, "%s|%4.2f|%d|%d|%5.3f|%s", pdbname[1], finalscore/maxscore,ObjList[0]->num_SSEs_aligned,globalCoreNumAligned, globalCoreRMSD, ObjList[1]->header);

      }
      else {
	fprintf(searchfile, "%s|%4.2f|%d|%d|%5.3f|%s", pdbname[1], finalscore/maxscore,ObjList[0]->num_SSEs_aligned,globalNumAligned,globalRMSD, ObjList[1]->header);

      }

      if(!lock_residues) {
	/*now print SSE scores*/
	for(i=0;i<ObjList[0]->ssArrayLength;i++) {
	  if(ObjList[0]->ss_alignment[i][0] != -1) {
	    query_start = ObjList[0]->ssArray[i*2];
	    query_end = ObjList[0]->ssArray[i*2+1]-1;
	    target_start = ObjList[1]->ssArray[ObjList[0]->ss_alignment[i][0]*2];
	    target_end = ObjList[1]->ssArray[ObjList[0]->ss_alignment[i][0]*2+1]-1;
	    fprintf(searchfile,"|%i,%i,%i,%i,%8.3f,%c",ObjList[0]->aaNums[query_start],ObjList[1]->aaNums[target_start],ObjList[0]->aaNums[query_end],ObjList[1]->aaNums[target_end],ObjList[0]->ss_dpscores[i],ObjList[0]->sstype[query_start]);
	  }
	}
	fprintf(searchfile, "\n");
      }

      freeObj(ObjList[0]);
      freeObj(ObjList[1]);
      freeAlignments();

    }
  
    else if ((globalSearchDatabase == 1) && (globalDatabaseFile != NULL)) {
      
      rewind(globalDatabaseFile);
      
      for (i = 0; i < 3; i++) {
	tempcentervector[i] = ObjList[0]->center[i];
      }
      
      for (i = 0; i < 16; i++) {
	tempobjectmatrix[i] = ObjList[0]->objMat[i];
      }
      
      done = 0;
      
      while ((done == 0) && (fgets(line, 200, globalDatabaseFile))) {
	sscanf(line, "%s", myprot);
	
	if (strcmp(myprot, "LOCK_LIST") == 0) {	
	  done = 1;
	}
      }

         /*Print info to logfile: query SSEs*/
      printf("Query SSEs:");
      for(i=0;i<ObjList[0]->ssArrayLength;i++) {
	printf(" %c,%i,%i",ObjList[0]->sstype[ObjList[0]->ssArray[i*2]],ObjList[0]->aaNums[ObjList[0]->ssArray[i*2]],ObjList[0]->aaNums[ObjList[0]->ssArray[i*2+1]-1]);
      }
      printf("\n");
      while ((fscanf(globalDatabaseFile, "%s", myprot)) != EOF) {
	printf("running lock2 alignment for %s\n",myprot);
	fflush(stdout);
	for (i = 0; i < 3; i++) {
	  ObjList[0]->center[i] = tempcentervector[i];
	}
	
	for (i = 0; i < 16; i++) {
	  ObjList[0]->objMat[i] = tempobjectmatrix[i];
	}
	
	globalCoreNumAligned = 0;
	globalNumAligned = 0;
	globalCoreRMSD = 0.0;
	globalRMSD = 0.0;

	strcpy(username[1], myprot);
	chain[1] = ' ';

	if (inputObject(1)) {
	  start_time = clock();
	  /*need to create alignment arrays*/
	  if(!first_alignment) {
	    freeAlignments();
	  }
	  if(ObjList[0]->ss_alignment != NULL) {
	    for(i=0;i<=ObjList[0]->ssArrayLength;i++) {
	      free(ObjList[0]->ss_alignment[i]);
	    }
	    free(ObjList[0]->ss_alignment);
	  }
	  createAlignmentArray(ObjList[0],ObjList[1]);
	  if ((ObjList[0]->ssArrayLength > 0) && (ObjList[1]->ssArrayLength > 0)) {	    
	    
	    initAlignUsingVectorsSingle(ObjList[0],ObjList[1],0);
	    alignUsingVectorsFull(ObjList[0],ObjList[1],0);
	    fflush(stdout);
	    first_alignment = 0;
	    if (globalNumAligned == 0) {
	      oldt = thresholdOFF;
	      thresholdOFF = 1;
	      /*structAlignSingle(ObjList[0],ObjList[1]);*/
	      thresholdOFF = oldt;
	    }

	    if(ObjList[0]->num_SSEs_aligned >= 1) {	    
	      if (globalComputeCore) {
		alignResidues(ObjList[0],ObjList[1]);
		printf("globalRMSDs after align residues: %f, %f\n",globalRMSD,globalCoreRMSD);
	      }
	      
	      
	      tempObj0.objSeq = ObjList[0]->objSeq; /*Array*/
	      tempObj0.sstype = ObjList[0]->sstype; /*Array of SS types*/
	      tempObj0.numPts = ObjList[0]->numPts; /*Array of points*/
	      tempObj0.ssArrayLength = ObjList[0]->ssArrayLength;
	      tempObj0.ssArray = ObjList[0]->ssArray; /*Array of #'s giving beginning and
						       *ends of ss elements*/
	      tempObj0.num_SSEs_aligned = ObjList[0]->num_SSEs_aligned;
	      tempObj0.ss_alignment = ObjList[0]->ss_alignment;
	      applyMat(&tempObj0, ObjList[0], ObjList[0]->objMat);
	      
	      tempObj1.objSeq = ObjList[1]->objSeq;
	      tempObj1.sstype = ObjList[1]->sstype;
	      tempObj1.numPts = ObjList[1]->numPts;
	      tempObj1.ssArrayLength = ObjList[1]->ssArrayLength;
	      tempObj1.ssArray = ObjList[1]->ssArray;
	      tempObj1.num_SSEs_aligned = ObjList[1]->num_SSEs_aligned;
	      tempObj1.ss_alignment = ObjList[1]->ss_alignment;
	      applyMat(&tempObj1, ObjList[1], ObjList[1]->objMat);
	      finalscore = align_ss(1,ObjList[0],ObjList[1]);
	      if(ObjList[0]->ssArrayLength > ObjList[1]->ssArrayLength) {
		maxscore = 40*ObjList[0]->ssArrayLength-20.0;
	      }else {
		maxscore = 40*ObjList[1]->ssArrayLength -20.0;
	      }
	      printf("globalRMSDs before printing output: %f, %f\n",globalRMSD,globalCoreRMSD);
	      printOutput(ObjList[0],ObjList[1],0,ObjList[0]->numPts-1);
	  fflush(stdout);
	  pvalue = 1-exp(-exp(-overall_lambda*(finalscore-overall_mu)));
	  
	  pvalue = 1-exp(-exp(-overall_lambda*(finalscore/maxscore-overall_mu)));

	  if (globalComputeCore) {
	    fprintf(searchfile, "%s|%4.2f|%d|%d|%5.3f|%s", pdbname[1], finalscore/maxscore,ObjList[0]->num_SSEs_aligned,globalCoreNumAligned, globalCoreRMSD, ObjList[1]->header);
	  }
	  else {
	    fprintf(searchfile, "%s|%4.2f|%d|%d|%5.3f|%s", pdbname[1], finalscore/maxscore, ObjList[0]->num_SSEs_aligned,globalNumAligned, globalRMSD, ObjList[1]->header);		
	  }
	  fflush(searchfile);	  
	  /*now print SSE scores*/
	  for(i=0;i<ObjList[0]->ssArrayLength;i++) {
	    if(ObjList[0]->ss_alignment[i][0] != -1) {
	      query_start = ObjList[0]->ssArray[i*2];
	      query_end = ObjList[0]->ssArray[i*2+1]-1;
	      target_start = ObjList[1]->ssArray[ObjList[0]->ss_alignment[i][0]*2];
	      target_end = ObjList[1]->ssArray[ObjList[0]->ss_alignment[i][0]*2+1]-1;
	      fprintf(searchfile,"|%i,%i,%i,%i,%8.3f,%c",ObjList[0]->aaNums[query_start],ObjList[1]->aaNums[target_start],ObjList[0]->aaNums[query_end],ObjList[1]->aaNums[target_end],ObjList[0]->ss_dpscores[i],ObjList[0]->sstype[query_start]);
	


	    }
	  }
	  fprintf(searchfile, "\n");
	    }
	    fflush(searchfile);
      }

	}
      }
    }
    fclose(searchfile);
  }
  
}

void createAlignmentArray(OBJ* obj0, OBJ* obj1) {
  int i,j;

  obj0->ss_alignment = (int**)malloc(sizeof(int*)*(obj0->ssArrayLength+1));
  for(i=0;i<obj0->ssArrayLength+1;i++) {
    obj0->ss_alignment[i] = (int*)malloc(sizeof(int*)*(obj1->ssArrayLength+1));
    for(j=0;j<obj1->ssArrayLength;j++) {
      obj0->ss_alignment[i][j] = -1;
    }
  }

  obj1->ss_alignment = (int**)malloc(sizeof(int*)*(obj1->ssArrayLength+1));
  for(i=0;i<obj1->ssArrayLength+1;i++) {
    obj1->ss_alignment[i] = (int*)malloc(sizeof(int*)*(obj0->ssArrayLength+1));
    for(j=0;j<obj0->ssArrayLength;j++) {
      obj1->ss_alignment[i][j] = -1;
    }
  }
  obj0->num_SSEs_aligned = 0;
  obj1->num_SSEs_aligned = 0;

}

void main(int argc, char** argv) {
  char str[100];


  freopen("/dev/null", "w", stderr);

  mypid = getpid();

  lockdir = getenv("LOCKDIR");
  if (lockdir == NULL) {
    lockdir = (char*)(malloc((sizeof(char))*200));
    strcpy(lockdir, _LOCKDIR_);
  }

  dsspprogram = getenv("DSSPPROGRAM");
  if (dsspprogram == NULL) {
    dsspprogram = (char*)(malloc((sizeof(char))*200));
    strcpy(dsspprogram, _DSSPPROGRAM_);
  }

  scriptdir = getenv("SCRIPTDIR");
  if (scriptdir == NULL) {
    scriptdir = (char*)(malloc((sizeof(char))*200));
    strcpy(scriptdir, _SCRIPTDIR_);
  }

  checkopts(argc, argv);

#ifndef DISPLAYON
  globalStartDisplay = 0;
#endif

  if ((strcmp(username[0], "NONE") == 0) || (strcmp(username[1], "NONE") == 0)) {
    printf("\n\nPlease specify a query and target.\n");
    usageError();
  }

#ifdef DISPLAYON
  if (globalStartDisplay) {
    addReshapeFunc(myReshape);
    setBindings();
    
    InitPosition(0, 0, 500, 500);
    My_Tk_Init(1, argv, Tcl_AppInit);        
    setupGLWindow("LOCK");             
    init();

    strcpy(str, lockdir);
    strcat(str, "lock.tcl");
    runTclFile(str);
  }
#endif

  if ((inputObject(0)) && ((inputObject(1)) || (globalSearchDatabase))) {
    if(!globalSearchDatabase) {
      createAlignmentArray(ObjList[0],ObjList[1]);
    }
    if (globalStartDisplay) {      
#ifdef DISPLAYON
      jointModelNumber = glGenLists(1);
      Tk_MainLoop();       
#endif
    }    
    else {

      doAllAlignments();
    }

  }

} 








